/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/
#pragma warning( disable : 4518 )

#define BACKPLANE_DEFAULT_PORT 6743

// BackPanel.h
#include "BackPanel.h"

// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------

// STLPort STL
#include <sstream>
#include <algorithm>

// xerces 2.2
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLUni.hpp>
#include <xercesc/framework/XMLFormatter.hpp>
#include <xercesc/util/TranscodingException.hpp>

#include <xercesc/dom/deprecated/DOM_DOMException.hpp>
#include <xercesc/dom/deprecated/DOM_NamedNodeMap.hpp>
#include <xercesc/dom/deprecated/DOMParser.hpp>
#include <xercesc/dom/DOM.hpp>

XERCES_CPP_NAMESPACE_USE

#include "ws2tcpip.h"


// CORBA global variables
CORBA::ORB_var orb;
PortableServer::POA_var poa;
PortableServer::POAManager_var poa_mgr;
IORTable::Table_var iortable_adapter;

// Mapping elements for Pass2
map<long, OTIF_CompositionParentRole_i *> composition_parent_role_map; // map (uml_composition_id, compositionparentrole_servant)
map<long, OTIF_CompositionChildRole_i *> composition_child_role_map; // map (uml_composition_id, compositionchildrole_servant)
map<long, OTIF_Class_i *> class_servant_map; // map (uml_class_id, class_servant)
map<long, OTIF::Class_var> class_map; // map (uml_class_id, class_obj_var)
map<long, OTIF::TaggedValue_var> tagged_value_map; // map (uml_taggedvalue_id, taggedvalue_obj_var)
map<long, OTIF::AssociationRole_var> association_role_map; // map (uml_associationrole_id, associationrole_obj_var)


// Register remote metamodel
map<long, OTIF::Class_var> cls_servant_map; // Servant_ID
map<long, OTIF::Attrib_var> attrib_servant_map; // Servant_ID
map<long, OTIF::AssociationRole_var> assocrole_servant_map; // Servant_ID


// Servants repository
multimap<long, PortableServer::ObjectId_var> servants_id;	// MetaModel_ID, CORBA Object ID
multimap<long, PortableServer::Servant> servants_ref;		// MetaModel_ID, CORBA Servant Reference

// Forward declaration
void UML_pass2(Uml::Diagram &uml_diagram);
void UML_diagram_pass(Uml::Diagram &uml_diagram, long MetaModel_ID);
void backplane_load_config(DOM_Element root_element);
extern void printDOM(DOM_Node doc, const char *filename, XMLCh *encodingName = NULL);
void _unregisterAllTranslator();
void _unregisterAllParadigms();


// Global variables
long MetaModel_ID = 1;
long ToolAdaptor_Session_ID = 1;

long Servant_ID = 1;
long ManagerSession_ID = 1;
long Document_ID = 1;

Logger logger = Logger("Backplane");

OTIF_Backplane_i *backplane_servant;
int __translator_arguments_length;
string *__translator_arguments;

// XML config file
class STARTUP_PARAMETERS {
public:
	DOM_Element root_element;
};
STARTUP_PARAMETERS *startup_params;
bool isUMLregistered = false;

// HACK! because of a UdmDOM bug
bool UdmDOM_bug_exists = false;
// HACK!



// *******************************************************************************************************************************
// Function used by Backplane servant
// *******************************************************************************************************************************
string backplane_name("");
string backplane_version("1.0");
string manager_key("");
string meta_meta_uml_type_tooladaptor("");
long translator_keepalive_parameter = 0;
bool OnlyOneManager = true;
bool isActiveManager = false;
string log_level = "0";

// Mapping elements for MetaModels
map<long, OTIF::NameVersion> metamodel_repository; // map (MetaModel_ID, NameVersion of paradigm)
// MetaModel and root Diagram repository
map<long, OTIF::Diagram_var> metamodel_diagram_repository; // map (MetaModel_ID, Diagram_var servant)
// Lock information mapping
map<long, int> metamodel_lock_repository; // map (MetaModel_ID, locking info), 0 - unlocked; 1 - locked

// ManagerSession mapping
struct MANAGER_SESSION {
	OTIF_ManagerSession_i *servant_p;
	PortableServer::ObjectId_var servant;
};
map<long, MANAGER_SESSION> manager_session_rep; // map (ManagerSession_ID, Manager_Session)

// ToolAdaptorSession mapping
struct TOOLADAPTOR_SESSION {
	long ToolAdaptor_ID;							//	ToolAdaptor_ID (the type ID!)
	OTIF_ToolAdaptorSession_i *servant_p;
	PortableServer::ObjectId_var servant;
	OTIF::ToolAdaptorNotifyCallback_var callback;
};
map<long, TOOLADAPTOR_SESSION> tooladaptor_session_rep; // map (ToolAdaptor_Session_ID, ToolAdaptor_Session)

// ToolAdaptorSession, Document_ID mapping
multimap<long, long> datanetwork_tooladaptor_notification_rep; // tooladaptor notified repository
// Translator, Document_ID mapping
multimap<long, long> datanetwork_translator_notification_rep; // translator notified repository

// Mapping elements for DOCUMENT
struct DOCUMENT {
	long sender_id;									//	ToolAdaptor_ID (the type ID!) or Translator_ID
	OTIF::Document *document;

//	long paradigm_id;								// paradigm ID
	multiset<long> paradigm_ids;					// paradigm ID set

	OTIF::NameVersion document_name_version;
	char *remarks;
	long keepAlive;
	long creationTime;

	bool multidocument;				// true if the document is multi
	long source_docId;				// the source of this translated document, if not the value is -1
	list<long> waiting_translators; // the list of translators_ID which are waiting to translate the document
};
map<long, DOCUMENT> document_repository; // map (Document_ID, DOCUMENT struct)

// for updated documents
struct UPDATED_DOCUMENT {
	long destination;				// destination ID
	bool notified;
};
multimap<long, UPDATED_DOCUMENT> updated_document_repository;	// Document_ID, UPDATED_DOCUMENT struct

// Mapping elements for Translators
struct TRANSLATOR {
	int type; // 0 - EXE, 1 - XSLT, 2 - JAVA
	OTIF::TranslatorVersion translator;
	OTIF_TranslatorFeed_i *feed_servant_p;
	OTIF_TranslatorSink_i *sink_servant_p;
	PortableServer::ObjectId_var feed_servant;
	PortableServer::ObjectId_var sink_servant;
	string command;
	string feed_str;
	string sink_str;
	OTIF::TranslatorNotifyCallback_var callback;
	bool in_use; // false - not in use, true - in use
	long notified; // shows how much notification was sent to the translator
	bool callbacking; // false - callbacking finished, true - callbacking in progress
};
map<long, TRANSLATOR> translator_rep; // map (Translator_ID, TRANSLATOR struct)


// Mapping elements for ToolAdaptor
struct PARADIGMS {
	list<long> paradigms;
};

struct TOOLADAPTOR {
	string type;
	list<long> inputParadigms;
	list<PARADIGMS> outputParadigms;
};
map<long, TOOLADAPTOR> tooladaptor_rep; // map (ToolAdaptor_ID, TOOLADAPTOR struct)

// Mapping element for workflow
struct WORKFLOW {
	long destination;							//	ToolAdaptor_ID (the type ID!) or Translator_ID
	long outputParadigm;
};
multimap<long, WORKFLOW> workflow_rep; // map (source, WORKFLOW)


long __now() {
	struct _timeb tstruct;
	_ftime( &tstruct );
	return tstruct.time;
}

int _getBackplaneName(OTIF::NameVersion_var &name) {
	name->name = CORBA::string_dup(backplane_name.c_str());
	name->version = CORBA::string_dup(backplane_version.c_str());
	return 0;
}


void __insertParadigmIfNotExists(string name, string version, multimap<string, string> &temp_paradigms) {
	for(multimap<string, string>::iterator i=temp_paradigms.begin(); i!=temp_paradigms.end(); i++) {
		if (!name.compare((*i).first) && !version.compare((*i).second)) return;
	}
	temp_paradigms.insert(pair<const string, string>(name, version));
}


void __fillinToolAdaptorInfo(list<long> inputParadigms, list<PARADIGMS> outP, OTIF::ToolAdaptorInfo_var &ta_info) {

	try {
		// fetch paradigms
		multimap<string, string> fetch_paradigms;
		for(list<long>::iterator input_p=inputParadigms.begin(); input_p!=inputParadigms.end(); input_p++) {
			map<long, OTIF::NameVersion>::iterator ip=metamodel_repository.find(*input_p);
			if (ip!=metamodel_repository.end()) {
				__insertParadigmIfNotExists((string)(*ip).second.name, (string)(*ip).second.version, fetch_paradigms);
			}
		}

		CORBA::ULong offset = 0;
		ta_info->fetched.length(fetch_paradigms.size());

		for(multimap<string, string>::iterator i=fetch_paradigms.begin(); i!=fetch_paradigms.end(); i++, offset++) {

			OTIF::NameVersion_S_var paradigms = new OTIF::NameVersion_S();
			paradigms->length(1);
			paradigms[(CORBA::ULong)0].name = CORBA::string_dup((*i).first.c_str());
			paradigms[(CORBA::ULong)0].version = CORBA::string_dup((*i).second.c_str());

			ta_info->fetched[offset].paradigmName(paradigms);
		}
		logger.log(3, "get tooladaptorinfo - fetch paradigms size:", (long)offset);

		
		// publish paradigms
		offset = 0;
		ta_info->published.length(outP.size());

		for(list<PARADIGMS>::iterator out_i=outP.begin(); out_i!=outP.end(); out_i++, offset++) {
			list<long> outputParadigms = (*out_i).paradigms;
		
			multimap<string, string> publish_paradigms;
			for(list<long>::iterator output_p=outputParadigms.begin(); output_p!=outputParadigms.end(); output_p++) {
				map<long, OTIF::NameVersion>::iterator op=metamodel_repository.find(*output_p);
				if (op!=metamodel_repository.end()) {
					__insertParadigmIfNotExists((string)(*op).second.name, (string)(*op).second.version, publish_paradigms);
				}
			}

			OTIF::NameVersion_S_var paradigms = new OTIF::NameVersion_S();
			paradigms->length(publish_paradigms.size());

			CORBA::ULong offset2 = 0;
			for(multimap<string, string>::iterator i=publish_paradigms.begin(); i!=publish_paradigms.end(); i++, offset2++) {
				paradigms[offset2].name = CORBA::string_dup((*i).first.c_str());
				paradigms[offset2].version = CORBA::string_dup((*i).second.c_str());
			}
		
			ta_info->published[offset].paradigmName(paradigms);
		}

		logger.log(3, "get tooladaptorinfo - publish paradigms size:", (long)offset);
	} catch (...) {
		logger.log(1, "Error: _getToolAdaptorInfo fatal error!");
	}
}

int _getToolAdaptorInfoByType(const char *taType, OTIF::ToolAdaptorInfo_var &ta_info) {

	for(map<long, TOOLADAPTOR>::iterator ta=tooladaptor_rep.begin(); ta!=tooladaptor_rep.end(); ta++) {
		if (!((*ta).second.type).compare(taType)) {
			__fillinToolAdaptorInfo((*ta).second.inputParadigms, (*ta).second.outputParadigms, ta_info);
			return 0;
			}
	}

	return 1;
}

int _getToolAdaptorInfo(CORBA::Long session_id, OTIF::ToolAdaptorInfo_var &ta_info) {

	map<long, TOOLADAPTOR_SESSION>::iterator i=tooladaptor_session_rep.find(session_id);
	if (i==tooladaptor_session_rep.end())  return 1;

	map<long, TOOLADAPTOR>::iterator ta=tooladaptor_rep.find((*i).second.ToolAdaptor_ID);
	if (ta==tooladaptor_rep.end()) return 1;

	__fillinToolAdaptorInfo((*ta).second.inputParadigms, (*ta).second.outputParadigms, ta_info);

	return 0;
}

int _getToolAdaptorType(const OTIF::NameVersion &paradigm, OTIF::taMode mode, string &type) {

	bool found = false;
	for(map<long, OTIF::NameVersion>::iterator mmi=metamodel_repository.begin(); mmi!=metamodel_repository.end(); mmi++) {
		if (!strcmp((*mmi).second.name, paradigm.name) && !strcmp((*mmi).second.version, paradigm.version)) found = true;
	}
	if (!found) return 1;

	if (!strcmp("UML", paradigm.name) && !strcmp("1.0", paradigm.version)) type = meta_meta_uml_type_tooladaptor;
	
	return 0;
}

int _getParadigms(OTIF::NameVersion_S_var &names) {
	try {
		names->length( metamodel_repository.size() );

		CORBA::ULong offset = 0;
		for(map<long, OTIF::NameVersion>::iterator mmi=metamodel_repository.begin(); mmi!=metamodel_repository.end(); mmi++) {
	      (*names)[offset].name = (*mmi).second.name;
		  (*names)[offset].version = (*mmi).second.version;
	      offset++;
		}
		logger.log(3, "get paradigms - size:", (long)offset);

	} catch (...) {
		logger.log(1, "Error: _getParadigms fatal error!");
		return 1;
	}
	return 0;
}

int _getTranslators(OTIF::TranslatorVersion_S_var &translators) {
	try {
		translators->length( translator_rep.size() );

		CORBA::ULong offset = 0;
		for(map<long, TRANSLATOR>::iterator i=translator_rep.begin(); i!=translator_rep.end(); i++) {
			(*translators)[offset].name = (*i).second.translator.name;
			(*translators)[offset].from = (*i).second.translator.from;
			(*translators)[offset].to = (*i).second.translator.to;
			offset++;
		}
		logger.log(3, "get translators - size:", (long)offset);

	} catch (...) {
		logger.log(1, "Error: _getTranslators fatal error!");
		return 1;
	}
	return 0;
}

int _getDocuments(OTIF::DocumentInfo_S_var &datanetworks) {
	try {
		datanetworks->length( document_repository.size() );

		CORBA::ULong offset = 0;
		for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++, offset++) {
			(*datanetworks)[offset].docId = (*i).first;

			OTIF::NameVersion_S_var paradigms = new OTIF::NameVersion_S();

			CORBA::ULong _length = 0;
			for(CORBA::ULong j=0; j<(*i).second.document->documents.length(); j++) {
				if ((*i).second.document->documents[j]._d() == OTIF::structElem) _length++;
			}

			paradigms->length( _length );
			for(j=0; j<(*i).second.document->documents.length(); j++) {
				if ((*i).second.document->documents[j]._d() == OTIF::structElem) {
					(*paradigms)[j].name = (*i).second.document->documents[j].sElem().paradigm.name;
					(*paradigms)[j].version = (*i).second.document->documents[j].sElem().paradigm.version;
				}
			}
			(*datanetworks)[offset].paradigms = paradigms;
			
			(*datanetworks)[offset].documentName.name = (*i).second.document_name_version.name;
			(*datanetworks)[offset].documentName.version = (*i).second.document_name_version.version;
			(*datanetworks)[offset].remarks = CORBA::string_dup((*i).second.remarks);

			long diff = (*i).second.keepAlive - (__now() - (*i).second.creationTime);
			if (diff<0) diff=0;
			(*datanetworks)[offset].keepAlive = diff;
		}

		logger.log(3, "get documents - size:", (long)offset);

	} catch (...) {
		logger.log(1, "Error: _getDocuments fatal error!");
		return 1;
	}
	return 0;
}

int _getToolAdaptorTypes(OTIF::Name_S_var &types) {
	try {
		types->length( tooladaptor_rep.size()-1 );

		CORBA::ULong offset = 0;
		for(map<long, TOOLADAPTOR>::iterator i=tooladaptor_rep.begin(); i!=tooladaptor_rep.end(); i++) {
			if ((*i).first!=0) (*types)[offset++] = CORBA::string_dup( (*i).second.type.c_str() );
		}
		logger.log(3, "get tooladaptortypes - size:", (long)offset);

	} catch (...) {
		logger.log(1, "Error: _getToolAdaptorTypes fatal error!");
		return 1;
	}
	return 0;
}

int _lockParadigm(const OTIF::NameVersion &paradigm) {
	for(map<long, OTIF::NameVersion>::iterator mmi=metamodel_repository.begin(); mmi!=metamodel_repository.end(); mmi++) {
		if (!strcmp((*mmi).second.name, paradigm.name) && !strcmp((*mmi).second.version, paradigm.version)) {
			long MetaModel_ID = (*mmi).first;
			if (!metamodel_lock_repository[MetaModel_ID]) {
				metamodel_lock_repository[MetaModel_ID] = 1;
				logger.log(3, "lock paradigm:", MetaModel_ID);
				return 0;
			} else {
				logger.log(3, "paradigm is locked already");
				return 2;
			}
		}
	}
	logger.log(3, "lock paradigm: can't find paradigm");
	return 1;
}

int _fetchParadigm(const OTIF::NameVersion &paradigm, OTIF::Diagram_var &diagram) {
	for(map<long, OTIF::NameVersion>::iterator mmi=metamodel_repository.begin(); mmi!=metamodel_repository.end(); mmi++) {
		if (!strcmp((*mmi).second.name, paradigm.name) && !strcmp((*mmi).second.version, paradigm.version)) {
			long MetaModel_ID = (*mmi).first;
			diagram = metamodel_diagram_repository[MetaModel_ID];
			logger.log(3, "fetch paradigm:", MetaModel_ID);
			return 0;
		}
	}
	logger.log(3, "fetch paradigm: paradigm does not exist");
	diagram = OTIF::Diagram::_nil();
	return 1;
}

int _unlockParadigm(const OTIF::NameVersion &paradigm) {
	for(map<long, OTIF::NameVersion>::iterator mmi=metamodel_repository.begin(); mmi!=metamodel_repository.end(); mmi++) {
		if (!strcmp((*mmi).second.name, paradigm.name) && !strcmp((*mmi).second.version, paradigm.version)) {
			long MetaModel_ID = (*mmi).first;
			if (metamodel_lock_repository[MetaModel_ID]) {
				metamodel_lock_repository[MetaModel_ID] = 0;
				logger.log(3, "unlock paradigm:", MetaModel_ID);
				return 0;
			} else {
				logger.log(3, "paradigm is unlocked already");
				return 2;
			}
		}
	}
	logger.log(3, "unlock paradigm: can't find paradigm");
	return 1;
}

int _loginManager(const char *key, OTIF::ManagerSession_var &manager, CORBA::Long_out session_id) {
	if (OnlyOneManager && isActiveManager) return 2;
	if (strcmp(manager_key.c_str(), key)) return 1;

	OTIF_ManagerSession_i *ms_servant = new OTIF_ManagerSession_i();
	try {
		PortableServer::ObjectId_var oid = poa->activate_object( ms_servant );
	    CORBA::Object_var ms_obj = poa->id_to_reference( oid.in() );
		manager = OTIF::ManagerSession::_narrow(ms_obj.in());

		session_id = ManagerSession_ID;
		MANAGER_SESSION ms;
		ms.servant = oid;
		ms.servant_p = ms_servant;
		manager_session_rep[ManagerSession_ID++] = ms;
		isActiveManager = true;
	
		logger.log(3, "manager login:", (ManagerSession_ID-1));
	} catch (const CORBA::Exception &) {
		logger.log(1, "login manager: Caught a CORBA exception");
	}
	return 0;
}

int _logoutManager(CORBA::Long session_id) {
	map<long, MANAGER_SESSION>::iterator i = manager_session_rep.find(session_id);
	if (i!=manager_session_rep.end()) {
		poa->deactivate_object( (*i).second.servant );
		delete (*i).second.servant_p;
		isActiveManager = false;
		manager_session_rep.erase(session_id);
		logger.log(3, "manager logout:", session_id);
	} else {
		logger.log(3, "logout manager: can't find manager session");
		return 1;
	}
	return 0;
}

int _loginToolAdaptor(const char *ToolAdaptorType, OTIF::ToolAdaptorSession_var &tooladaptor, CORBA::Long_out session_id) {
	long ToolAdaptor_ID = -1;
	for(map<long, TOOLADAPTOR>::iterator i=tooladaptor_rep.begin(); i!=tooladaptor_rep.end(); i++) {
		if (!strcmp((*i).second.type.c_str(), ToolAdaptorType)) {
			ToolAdaptor_ID = (*i).first;
			break;
		}
	}
	if (ToolAdaptor_ID==-1) {
		logger.log(3, "login tooladaptor: can't find ToolAdaptor type");
		return 1;
	}
	
	OTIF_ToolAdaptorSession_i *tas_servant = new OTIF_ToolAdaptorSession_i();
	try {
		PortableServer::ObjectId_var oid = poa->activate_object( tas_servant );
	    CORBA::Object_var tas_obj = poa->id_to_reference( oid.in() );
		tooladaptor = OTIF::ToolAdaptorSession::_narrow(tas_obj.in());

		session_id = ToolAdaptor_Session_ID;
		TOOLADAPTOR_SESSION ts;
		ts.ToolAdaptor_ID = ToolAdaptor_ID;
		ts.servant = oid;
		ts.servant_p = tas_servant;
		tooladaptor_session_rep[ToolAdaptor_Session_ID++] = ts;

		logger.log(3, "tooladaptor login:", session_id);
	} catch (const CORBA::Exception &) {
		logger.log(1, "login tooladaptor: Caught a CORBA exception");
	}
	return 0;
}

/*
int _getParadigmsOfToolAdaptorType(CORBA::Long session_id, OTIF::NameVersion_S_var &paradigms) {
	map<long, TOOLADAPTOR_SESSION>::iterator i=tooladaptor_session_rep.find(session_id);
	if (i==tooladaptor_session_rep.end())  return 1;

	map<long, TOOLADAPTOR>::iterator ta=tooladaptor_rep.find((*i).second.ToolAdaptor_ID);
	if (ta==tooladaptor_rep.end()) return 1;

	multimap<string, string> temp_paradigms;
	
	list<long> inputParadigms = (*ta).second.inputParadigms;
	for(list<long>::iterator input_p=inputParadigms.begin(); input_p!=inputParadigms.end(); input_p++) {
		map<long, OTIF::NameVersion>::iterator ip=metamodel_repository.find(*input_p);
		if (ip!=metamodel_repository.end()) {
			__insertParadigmIfNotExists((string)(*ip).second.name, (string)(*ip).second.version, temp_paradigms);
		}
	}

	list<PARADIGMS> outP = (*ta).second.outputParadigms;
	for(list<PARADIGMS>::iterator out_i=outP.begin(); out_i!=outP.end(); out_i++) {
		list<long> outputParadigms = (*out_i).paradigms;
		for(list<long>::iterator output_p=outputParadigms.begin(); output_p!=outputParadigms.end(); output_p++) {
			map<long, OTIF::NameVersion>::iterator op=metamodel_repository.find(*output_p);
			if (op!=metamodel_repository.end()) {
				__insertParadigmIfNotExists((string)(*op).second.name, (string)(*op).second.version, temp_paradigms);
			}
		}
	}
	
	try {
		paradigms->length( temp_paradigms.size() );

		CORBA::ULong offset = 0;
		for(map<string, string>::iterator i=temp_paradigms.begin(); i!=temp_paradigms.end(); i++) {
			(*paradigms)[offset].name = CORBA::string_dup( (*i).first.c_str() );
			(*paradigms)[offset].version = CORBA::string_dup( (*i).second.c_str() );
			offset++;
		}
		logger.log(3, "get paradigmsoftooladaptortypes - size:", (long)offset);

	} catch (...) {
		logger.log(1, "Error: _getParadigmsOfToolAdaptorType fatal error!");
		return 1;
	}

	return 0;
}
*/

int _logoutToolAdaptor(CORBA::Long session_id) {
	map<long, TOOLADAPTOR_SESSION>::iterator i = tooladaptor_session_rep.find(session_id);
	if (i!=tooladaptor_session_rep.end()) {
		poa->deactivate_object( (*i).second.servant );
		delete (*i).second.servant_p;
		tooladaptor_session_rep.erase(session_id);
		datanetwork_tooladaptor_notification_rep.erase(session_id);
		logger.log(3, "tooladaptor logout:", session_id);
	} else {
		return 1;
		logger.log(3, "logout tooladaptor: can't find tooladaptor:", session_id);
	}

	return 0;
}

void _now(OTIF::timeStamp_out stamp) {
	struct _timeb tstruct;
	_ftime( &tstruct );

	stamp.second = tstruct.time;
	stamp.microsecond = tstruct.millitm;
}

int _uploadStartup(const char *xml) {
	ofstream config_file;
	config_file.open("BackPlane_config.xml");
	config_file << xml;
	config_file.close();
	config_file.clear();

	// everything must be unregistered before parsing the new config file
	_unregisterAllTranslator();
	_unregisterAllParadigms();
	tooladaptor_rep.clear();
	workflow_rep.clear();

	// we have to destroy the available datanetworks
	for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++) {
		delete (*i).second.document;
		document_repository.erase( (*i).first );
		updated_document_repository.erase( (*i).first );
	}

	// if the xml is invalid, then return 1
	DOMParser parser;
	parser.setDoNamespaces(false);
	parser.setIncludeIgnorableWhitespace(false);
	try	{
		parser.parse("./BackPlane_config.xml");
		startup_params->root_element = parser.getDocument().getDocumentElement();
		if (startup_params->root_element!=NULL) backplane_load_config(startup_params->root_element);
		else {
			cerr << "Error: The Backplane config file is corrupt." << endl;
			return 1;
		}

	} catch (const XMLException &e) {
		cerr << "Error: Error during parsing the config file! XMLException: " << e.getMessage() << endl;;
		return 1;
	} catch (DOM_DOMException e) {
		cerr << "Error: Error during parsing the config file! DOMException: " << e.msg.transcode() << endl;
		return 1;
	} catch (...) {
		cerr << "Error: Error during parsing the config file!";
		return 1;
	}

	cout << "Configuration changed: new configuration was processed in the Backplane." << endl;
	
	return 0;
}


// -------------------------------------------------------------------------------------------------------------------------------
//  CORBADataNetwork Unwrapping
// -------------------------------------------------------------------------------------------------------------------------------
	map<long, Udm::ObjectImpl*> objectimpl_rep;
	typedef map<long, Udm::Object> copy_assoc_map;

	void __extractUserData(CORBA::Any userData, long &subtype, long &realarchetype, long &archetype) {
		char * __userData;

		if (userData >>= __userData) {
			string s_archetype = __userData;
			string s_subtype = s_archetype.substr(0, 1);
			string s_realarchetype = s_archetype.substr(1, 1);
			s_archetype.erase(s_archetype.begin());
			s_archetype.erase(s_archetype.begin());

//			cout << s_subtype << ":" << s_realarchetype << ":" << s_archetype << endl;

			subtype = atol(s_subtype.c_str());
			realarchetype = atol(s_realarchetype.c_str());
			archetype = atol(s_archetype.c_str());
		}
	}

	Uml::CompositionChildRole getCompositionChildRoleMeta(const Uml::Class &class_meta, const Uml::Class &parent_meta) {
		set<Uml::Class> cancs = Uml::AncestorClasses(class_meta);
		for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++) {
			
			set<Uml::CompositionChildRole> uml_childroles = (*q).childRoles();
			for(set<Uml::CompositionChildRole>::iterator i=uml_childroles.begin(); i!=uml_childroles.end(); i++) {
				Uml::Class target = ((Uml::CompositionParentRole) ((Uml::Composition) (*i).parent()).parentRole()).target();
				if (target==parent_meta) return *i;
			}
		}
		return NULL;
	}

// Build up the Attributes for an ObjectImpl
	void ObjectImplAttrsPass(Udm::ObjectImpl *obj, const OTIF::Obj &udm_obj) {
		for(CORBA::ULong i=0; i< udm_obj.attrs.length(); i++) {

			map<long, OTIF::Attrib_var>::iterator iter = attrib_servant_map.find(udm_obj.attrs[i].tId);
			OTIF::Attrib_var attr_servant = (*iter).second;
			OTIF::AttrValueHolder attr_value = udm_obj.attrs[i].value;

			const Uml::Class &obj_meta = obj->type();
			Uml::Attribute attrib_meta;

			string meta_name = attr_servant->name();
			set<Uml::Class> cancs = Uml::AncestorClasses(obj_meta);
			for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++) {

				set<Uml::Attribute> uml_attrs = (*q).attributes();
				for(set<Uml::Attribute>::iterator i=uml_attrs.begin(); i!=uml_attrs.end(); i++) {
					if (!meta_name.compare( (*i).name() )) attrib_meta = *i;
				}
			}
			if (!attrib_meta) throw udm_exception("CORBADataNetwork is corrupt => A meta of an attribute is not found!");

			try {
				if (string(attr_servant->typeName())=="String") obj->setStringAttr(attrib_meta, attr_value.stringValue());
				if (string(attr_servant->typeName())=="Integer") obj->setIntegerAttr(attrib_meta, attr_value.intValue());
 				if (string(attr_servant->typeName())=="Boolean") obj->setBooleanAttr(attrib_meta, ((attr_value.boolValue()==0) ? false : true));
				if (string(attr_servant->typeName())=="Real") obj->setRealAttr(attrib_meta, attr_value.realValue());
			} catch (udm_exception&) {
//				cerr << "Warning: an attribute could not be set!" << endl;
				continue;
			}
		}
	}

	bool reqIsObjectCopied(const OTIF::Obj_S &udm_objects, const OTIF::Obj &srcDerived, const copy_assoc_map &cam) {
		
		for(CORBA::ULong i=1; i<udm_objects.length(); i++) {
			if (udm_objects[i].parent.parentId.id==srcDerived.id.id) {

				long __srcArchetypeID;
				long __realarchetype;
				long __subtype;
				__extractUserData(udm_objects[i].userData, __subtype, __realarchetype, __srcArchetypeID);

//				long srcArchetypeID = udm_objects[i].archetype;

//				if (srcArchetypeID!=-1 && udm_objects[i].realarchetype==0) {		// it is not a real archetype
				if (__srcArchetypeID!=-1 && __realarchetype==0) {		// it is not a real archetype

					copy_assoc_map::const_iterator cam_i = cam.find(__srcArchetypeID);
					if (cam_i != cam.end()) {
						if (!reqIsObjectCopied(udm_objects, udm_objects[i], cam)) return false;
					} else {
						return false;
					}
				}
			}
		}
				
		return true;
	}

	void reqCreateMappingforDerivedObject(const OTIF::Obj_S &udm_objects, const unsigned long udm_obj_id, const Udm::Object &dstDerived, copy_assoc_map &cam) {

		set<Udm::Object> dst_children = dstDerived.GetChildObjects();

		for(CORBA::ULong i=1; i<udm_objects.length(); i++) {
			if (udm_objects[i].parent.parentId.id==udm_obj_id) {

				long __srcArchetypeID;
				long __realarchetype;
				long __subtype;
				__extractUserData(udm_objects[i].userData, __subtype, __realarchetype, __srcArchetypeID);

//				long srcArchetypeID = udm_objects[i].archetype;

//				if (srcArchetypeID!=-1 && udm_objects[i].realarchetype==0) {		// it is not a real archetype
				if (__srcArchetypeID!=-1 && __realarchetype==0) {		// it is not a real archetype

					// src must be in the map
					copy_assoc_map::iterator cam_i=cam.find(__srcArchetypeID);
					if (cam_i!=cam.end()) {

						Udm::Object dst_child_archetype = (*cam_i).second;
						set<Udm::Object>::iterator dst_i = dst_children.begin();
						bool found = false;
						while (dst_i != dst_children.end() && !found) {

							Udm::Object dst_child = *dst_i++;
							if ((dst_child.archetype() == dst_child_archetype) && !dst_child.hasRealArchetype()) {

								copy_assoc_map::value_type map_item(udm_objects[i].id.id, dst_child);
								cam.insert(map_item);
								found=true;

								reqCreateMappingforDerivedObject(udm_objects, udm_objects[i].id.id, dst_child, cam);
							}
						}
					} else {
		
						throw udm_exception("Internal error");
					}

				} else {
					// this is not an error
					// it might be that udm_obj_id is a subtype and in adition has other objects
				}
			}
		}
	}

	void reqCopyObjectHierarchy(Udm::DataNetwork* p_dstBackend, const Uml::Diagram &root_meta, const OTIF::Obj_S &udm_objects, Udm::ObjectImpl *parent, const unsigned long parent_id, bool &finished, copy_assoc_map &cam) {

		for(CORBA::ULong i=1; i<udm_objects.length(); i++) {
			if (udm_objects[i].parent.parentId.id==parent_id) {

				map<long, OTIF::Class_var>::iterator iter = cls_servant_map.find(udm_objects[i].tId);
				string meta_name = (*iter).second->name();
				
				Uml::Class meta;
				set<Uml::Class> uml_classes = root_meta.classes();
				for(set<Uml::Class>::iterator j=uml_classes.begin(); j!=uml_classes.end(); j++) {
					if (!meta_name.compare( (*j).name() )) {
						meta = (*j);
						break;
					}
				}
				if (!meta) throw udm_exception("CORBADataNetwork is corrupt => A meta of a class is not found!");

				Uml::Class parent_meta = parent->type();
				Uml::CompositionChildRole childrole = getCompositionChildRoleMeta(meta, parent_meta);


				Udm::ObjectImpl* p_dstChild = &Udm::_null;
				Udm::Object dstChild;

				// check whether the object was copied first
				copy_assoc_map::iterator cam_i=cam.find( udm_objects[i].id.id );
				if (cam_i != cam.end()) {
					dstChild = (*cam_i).second;
					p_dstChild = dstChild.__impl()->clone();
				}
				
				long __srcArchetypeID;
				long __realarchetype;
				long __subtype;
				__extractUserData(udm_objects[i].userData, __subtype, __realarchetype, __srcArchetypeID);

				if (p_dstChild==&Udm::_null) {					// object has not created yet
//					long srcArchetypeID = udm_objects[i].archetype;
					Udm::ObjectImpl* p_dstChildArc = &Udm::_null;
					Udm::Object dstChildArc;

					if (__srcArchetypeID!=-1) {					// check if the source object has archetype

						//if it's not a real archetype, we skip this child
						//(it's there because the parent has an archetype)
						if (__realarchetype==0)	continue;

						// check if it was copied or not
						copy_assoc_map::iterator cam_i=cam.find(__srcArchetypeID);
						if (cam_i!=cam.end()) {

							// if it was already copied, assign the copied archetype to variables
							dstChildArc = (*cam_i).second;
							p_dstChildArc = dstChildArc.__impl();
						}

						if (p_dstChildArc==&Udm::_null) {

							// archetype is not copied, we need to walk the tree again
							finished = false;
							continue;

						} else {

							//check if archetype's all children were copied OK
							//(recursively)
							if ( !reqIsObjectCopied(udm_objects, udm_objects[i], cam) ) {
								finished = false;
								continue;
							}

							// archetype is ready, so we create the new object
							bool subtype=false;
//							if (udm_objects[i].subtype==1) subtype=true;
							if (__subtype==1) subtype=true;

							if (p_dstBackend->IsTypeSafe())
								p_dstChild = parent->createChild(childrole, meta, p_dstChildArc, subtype);
							else {
								const Uml::Class &safe_meta = Uml::SafeTypeContainer::GetSafeType(meta);
								p_dstChild = parent->createChild(childrole, safe_meta, p_dstChildArc, subtype);
							}

							// create the attributes
							ObjectImplAttrsPass(p_dstChild, udm_objects[i]); 
						}

						// do the mapping for the derived block as well
						reqCreateMappingforDerivedObject(udm_objects, udm_objects[i].id.id, p_dstChild->clone(), cam);

					} else {

						// it has not archetype, so just create the default way
						if (p_dstBackend->IsTypeSafe())
							p_dstChild = parent->createChild(childrole, meta);
						else {
							const Uml::Class &safe_meta = Uml::SafeTypeContainer::GetSafeType(meta);
							p_dstChild = parent->createChild(childrole, safe_meta);
						}

						// create the attributes
						ObjectImplAttrsPass(p_dstChild, udm_objects[i]); 
					}

					// create association for the map
					copy_assoc_map::iterator cam_i=cam.find(udm_objects[i].id.id);
					if (cam_i!=cam.end()) {
						(*cam_i).second = p_dstChild->clone();
					} else {
						copy_assoc_map::value_type map_item(udm_objects[i].id.id, p_dstChild->clone());
						cam.insert(map_item);
					}

				} else {
				}

				reqCopyObjectHierarchy(p_dstBackend, root_meta, udm_objects, p_dstChild, udm_objects[i].id.id, finished, cam);

				p_dstChild->release();
			}
		}
	}

	Uml::AssociationRole getAssocRoleMeta(Uml::Class &class_meta, const long assocrole_servant, bool &assoc_class) {
		map<long, OTIF::AssociationRole_var>::iterator iter = assocrole_servant_map.find(assocrole_servant);
		string meta_name = (*iter).second->name();

		set<Uml::Class> cancs = Uml::AncestorClasses(class_meta);
		for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++) {
			set<Uml::AssociationRole> uml_assocs = (*q).associationRoles();

			for(set<Uml::AssociationRole>::iterator i=uml_assocs.begin(); i!=uml_assocs.end(); i++) {
				Uml::AssociationRole target = Uml::theOther(*i);
				Uml::Class assocClass = Uml::Association(i->parent()).assocClass();
				if (!meta_name.compare( target.name() )) {
					if (assocClass) assoc_class = true;
					return *i;
				}
			}
		}
		return NULL;
	}

	// Build up the Assocations
	void ObjectImplAssocPass(const OTIF::Obj_S &udm_objects, copy_assoc_map &cam) {

		for(CORBA::ULong i=0; i<udm_objects.length(); i++) {

			copy_assoc_map::iterator rep_find = cam.find( udm_objects[i].id.id );
			if (rep_find==cam.end()) throw udm_exception("CORBADataNetwork is corrupt => Maybe an object is not created!");

			Udm::ObjectImpl *obj_impl = (*rep_find).second.__impl()->clone();
			Uml::Class obj_meta = obj_impl->type();

			for(CORBA::ULong w=0; w<udm_objects[i].assocs.length(); w++) {
				bool assoc_class = false;
		
				Uml::AssociationRole assocrole_meta = getAssocRoleMeta(obj_meta, udm_objects[i].assocs[w].tId, assoc_class);
				if (!assocrole_meta) throw udm_exception("CORBADataNetwork is corrupt => A meta of an AssocRole is not found!");
				Uml::AssociationRole target_meta = Uml::theOther(assocrole_meta);

				vector<Udm::ObjectImpl*> impl_vec;
				for(CORBA::ULong j=0; j<udm_objects[i].assocs[w].objectIds.length(); j++) {
					copy_assoc_map::iterator rep = cam.find( udm_objects[i].assocs[w].objectIds[j].id );
					if (rep!=cam.end()) {
						impl_vec.push_back( (*rep).second.__impl()->clone() );
					} else {
						throw udm_exception("CORBADataNetwork is corrupt!");
					}
 				}

				if (!assoc_class) {
					try {
						obj_impl->setAssociation(target_meta, impl_vec, Udm::TARGETFROMPEER, false);
					} catch (udm_exception e) {
						cout << e.what() << endl;
					}
				} else {
					try {
						obj_impl->setAssociation(target_meta, impl_vec, Udm::CLASSFROMTARGET, false);
					} catch (udm_exception e) {
						cout << e.what() << endl;
					}
				}

				for(vector<Udm::ObjectImpl*>::iterator qq=impl_vec.begin(); qq!=impl_vec.end(); qq++) (*qq)->release();
			} // for

			obj_impl->release();		
		}
	}

	// Build up the ObjectImpls
	void ObjectImplsPass(Udm::DataNetwork* p_dstBackend, const Uml::Diagram &root_meta, const OTIF::Obj_S &udm_objects, long parent_id, Udm::ObjectImpl *parent) {

		copy_assoc_map cam;

		copy_assoc_map::value_type map_item(parent_id, parent->clone());
		cam.insert(map_item);

		bool finished = true;
		bool first = true;

		while(!finished || first) {
			first = false;
			finished = true;

			reqCopyObjectHierarchy(p_dstBackend, root_meta, udm_objects, parent, parent_id, finished, cam);
		}

		ObjectImplAssocPass(udm_objects, cam);
	}

// *******************************************************************************************************************************
// Function used by ManagerSession servant
// *******************************************************************************************************************************
int _registerMetaModelFile(const OTIF::NameVersion &mm_info, const char *filename, 
						   OTIF::ManagerSession::RegistrationAction action, OTIF::Diagram_var &diagram, long ID) {

	char mm_name[1024];
	sprintf(mm_name, "name:%s version:%s file:%s", mm_info.name, mm_info.version, filename);
	logger.log(3, "obtain a request to register a paradigm: ", mm_name);

	for(map<long, OTIF::NameVersion>::iterator mmi=metamodel_repository.begin(); mmi!=metamodel_repository.end(); mmi++) {
	    if (!strcmp((*mmi).second.name, mm_info.name) && !strcmp((*mmi).second.version, mm_info.version)) {
			long MetaModel_ID = (*mmi).first;
			if (metamodel_lock_repository[MetaModel_ID]) return 2;

			logger.log(3, "paradigm exists: ", mm_name);
			return 1;
		}
	}

	try {

		if (ID!=-1) MetaModel_ID = ID;
		else MetaModel_ID++;

		string mm_loc = "paradigms/";
		mm_loc.append(filename);

		// HACK! because of UdmDOM bug
		string fn = filename;
		if (fn.length()<4) return 4;
		string suffix = fn.substr(fn.length()-4, fn.length());
		if (!suffix.compare(".xml")) UdmDOM_bug_exists = true;
		else UdmDOM_bug_exists = false;
		// HACK!

		Udm::SmartDataNetwork uml_net(Uml::diagram);
		uml_net.OpenExisting(mm_loc, "", Udm::CHANGES_PERSIST_ALWAYS);
		
		Udm::Object root = uml_net.GetRootObject();
		Uml::Diagram uml_diagram = Uml::Diagram::Cast(root);

		UML_diagram_pass(uml_diagram, MetaModel_ID);
		UML_pass2(uml_diagram);

		metamodel_repository[MetaModel_ID] = mm_info;
		diagram = metamodel_diagram_repository[MetaModel_ID];

		uml_net.CloseNoUpdate();
		logger.log(3, "register paradigm: ", fn);

	} catch (const udm_exception &e) {
		logger.log(1, "Error register paradigm:", e.what());
		return 4;
	}

	return 0;
}


int _registerMetaModel(const OTIF::NameVersion &mm_info, const OTIF::DataNetwork &metaData, 
					   OTIF::ManagerSession::RegistrationAction action, OTIF::Diagram_var &diagram) {


	try {
		string filename = (string)mm_info.name + "_v" + (string)mm_info.version + ".xml";
		string abs_file = "paradigms/" + filename;

		Udm::SmartDataNetwork uml_net(Uml::diagram);
		uml_net.CreateNew(abs_file.c_str(), "uml", Uml::Diagram::meta, Udm::CHANGES_PERSIST_ALWAYS);
		Uml::Diagram root_meta = uml_net.GetRootMeta();
		set<Uml::Class> uml_classes = root_meta.classes();

		// Search the root object
		OTIF::Obj_var root = new OTIF::Obj();
		for(CORBA::ULong i=0; i<metaData.objects.length(); i++) {
			if (metaData.objects[i].id.id==metaData.root.id) {
				*root=metaData.objects[i];
				break;
			}
		}

		Udm::Object root_obj = uml_net.GetRootObject();
		Udm::ObjectImpl *so = root_obj.__impl()->clone();
		objectimpl_rep[root->id.id] = so;
//		_setDNObjectParameters(root_obj.uniqueId(), root->id);

		// Build up the attributes for the rootobject
		ObjectImplAttrsPass(so, root); 
		// Build up the children of the rootobject
		ObjectImplsPass(&uml_net, root_meta, metaData.objects, root->id.id, so);

		uml_net.SaveAs(abs_file.c_str());
		uml_net.CloseNoUpdate();

		for(map<long, Udm::ObjectImpl*>::iterator qq=objectimpl_rep.begin(); qq!=objectimpl_rep.end(); qq++) delete (*qq).second;
		objectimpl_rep.clear();

		_registerMetaModelFile(mm_info, filename.c_str(), action, diagram);
		logger.log(3, "register paradigm: ", filename);

	} catch (udm_exception &e) {
		logger.log(1, "Error register paradigm:", e.what());
	} catch (...) {
		logger.log(1, "Error registering remote paradigm!");
	}

	return 0;
}

bool _isParadigmInUse (const char *name, const char *version) {

	for(map<long, TRANSLATOR>::iterator mmi=translator_rep.begin(); mmi!=translator_rep.end(); mmi++) {
		if ( (!strcmp((*mmi).second.translator.from.name, name) && !strcmp((*mmi).second.translator.from.version, version)) || 
			 (!strcmp((*mmi).second.translator.to.name, name) && !strcmp((*mmi).second.translator.to.version, version)) ) {
			return true;
		}
	}
	return false;
}

int _unregisterMetaModel(const OTIF::NameVersion &mm_info) {
	char mm_name[1024];
	sprintf(mm_name, "name:%s version:%s", mm_info.name, mm_info.version);
	logger.log(3, "obtain a request to unregister a MetaModel: ", mm_name);

	for(map<long, OTIF::NameVersion>::iterator mmi=metamodel_repository.begin(); mmi!=metamodel_repository.end(); mmi++) {
		if (!strcmp((*mmi).second.name, mm_info.name) && !strcmp((*mmi).second.version, mm_info.version)) {

			if (_isParadigmInUse(mm_info.name, mm_info.version)) {
				logger.log(3, "unregister paradigm: paradigm is in use");
				return 1; 
			}
			
			long MetaModel_ID = (*mmi).first;
			if (metamodel_lock_repository[MetaModel_ID]) {
				logger.log(3, "unregister paradigm: paradigm is locked");
				return 2;
			}
      
			for(multimap<long, PortableServer::ObjectId_var>::iterator id=servants_id.begin(); id!=servants_id.end(); id++) {
				try {
					if ((*id).first == MetaModel_ID) poa->deactivate_object( (*id).second );
				} catch (...) {
					logger.log(1, "servant deactivating exception occured");
				}
			}
			servants_id.erase(MetaModel_ID);

			for(multimap<long, PortableServer::Servant>::iterator ref=servants_ref.begin(); ref!=servants_ref.end(); ref++) {
				try {
					if ((*ref).first == MetaModel_ID) delete (*ref).second;
				} catch (...) {
					logger.log(1, "servant deleting exception occured");
				}
			}
			servants_ref.erase(MetaModel_ID);
			metamodel_diagram_repository.erase(MetaModel_ID);
			metamodel_lock_repository.erase(MetaModel_ID);

			// NOT CORRECT!
/*
			composition_parent_role_map.erase(MetaModel_ID);
			composition_child_role_map.erase(MetaModel_ID);
			class_servant_map.erase(MetaModel_ID);
			class_map.erase(MetaModel_ID);
			tagged_value_map.erase(MetaModel_ID);
			association_role_map.erase(MetaModel_ID);
*/

			metamodel_repository.erase(mmi);

			logger.log(3, "paradigm unregistered:", MetaModel_ID);
			return 0;
		}
	}

	logger.log(3, "paradigm could not be found: ", mm_name);
	return 3;
}


void _unregisterAllParadigms() {
	for(multimap<long, PortableServer::ObjectId_var>::iterator id=servants_id.begin(); id!=servants_id.end(); id++) {
		try {
			poa->deactivate_object( (*id).second );
		} catch (...) {
			logger.log(1, "servant deactivating exception occured");
		}
	}
	servants_id.clear();

	for(multimap<long, PortableServer::Servant>::iterator ref=servants_ref.begin(); ref!=servants_ref.end(); ref++) {
		try {
			delete (*ref).second;
		} catch (...) {
			logger.log(1, "servant deleting exception occured");
		}
	}
	servants_ref.clear();
	metamodel_diagram_repository.clear();
	metamodel_lock_repository.clear();

	composition_parent_role_map.clear();
	composition_child_role_map.clear();
	class_servant_map.clear();
	class_map.clear();
	tagged_value_map.clear();
	association_role_map.clear();

	metamodel_repository.clear();
}


bool __equalTranslators(const OTIF::TranslatorVersion &tr1, const OTIF::TranslatorVersion &tr2) {
	if (!strcmp(tr1.name.name, tr2.name.name) &&
		!strcmp(tr1.name.version, tr2.name.version) &&
		!strcmp(tr1.from.name, tr2.from.name) &&
		!strcmp(tr1.from.version, tr2.from.version) &&
		!strcmp(tr1.to.name, tr2.to.name) &&
		!strcmp(tr1.to.version, tr2.to.version)) return true;
	else return false;
}

int _registerTranslator(const OTIF::TranslatorVersion &translator, const char* activation, 
						OTIF::ManagerSession::RegistrationAction action, long ID) {

	char mm_name[1024];
	sprintf(mm_name, "name:%s version:%s activation:%s", translator.name.name, translator.name.version, activation);
	logger.log(3, "obtain a request to register a Translator: ", mm_name);

	// if exists then return 1
	for(map<long, TRANSLATOR>::iterator i=translator_rep.begin(); i!=translator_rep.end(); i++) {
		if (__equalTranslators((*i).second.translator, translator)) {
			logger.log(3, "register translator: translator exists");
			return 1;
		}
	}

	// check if InvalidParadigm
	bool invalid_fromParadigm = true;
	bool invalid_toParadigm = true;
	for(map<long, OTIF::NameVersion>::iterator j=metamodel_repository.begin(); j!=metamodel_repository.end(); j++) {
		if (!strcmp(translator.from.name, (*j).second.name) && !strcmp(translator.from.version, (*j).second.version)) invalid_fromParadigm=false;
		if (!strcmp(translator.to.name, (*j).second.name) && !strcmp(translator.to.version, (*j).second.version)) invalid_toParadigm=false;
	}
	if (invalid_fromParadigm || invalid_toParadigm) {
		logger.log(3, "register translator: invalid paradigm");
		return 4;
	}

	// check if translator valid
	string run_command = activation;
	if (run_command.length()<5) {
		logger.log(3, "register translator: invalid translator");
		return 3;
	}

	// if Manager application tries to register a translator, it should not allow!
	if (ID==-1) return 3;
	

	TRANSLATOR tr;
	string com_suffix = run_command.substr(run_command.length()-4, run_command.length());
	string com_suffix_6 = run_command.substr(run_command.length()-6, run_command.length());
	if (!com_suffix.compare(".xsl")) {
		// XSLT translator
		tr.type = 1;
	} else if (!com_suffix.compare(".exe")) {
		// EXE translator
		tr.type = 0;
	} else if (!com_suffix_6.compare(".class")) {
		// JAVA translator
		tr.type = 2;
		run_command = run_command.substr(0, run_command.length()-6);
	} else {
		return 3;
	}

	char suffix[64];
	ltoa(ID, suffix, 10);
	string feed_str = "translator_feed_";
	feed_str.append(suffix);
	string sink_str = "translator_sink_";
	sink_str.append(suffix);

	// register TranslatorFeed object
	OTIF_TranslatorFeed_i *translatorfeed_servant =  new OTIF_TranslatorFeed_i();
	translatorfeed_servant->session_id = ID;

	PortableServer::ObjectId_var feed_oid = poa->activate_object( translatorfeed_servant );
	CORBA::String_var ior = orb->object_to_string( translatorfeed_servant->_this() );
	iortable_adapter->rebind( ("OTIF/" + feed_str).c_str(), ior.in() );
	
	// register TranslatorSink object
	OTIF_TranslatorSink_i *translatorsink_servant =  new OTIF_TranslatorSink_i();
	translatorsink_servant->session_id = ID;

	PortableServer::ObjectId_var sink_oid = poa->activate_object( translatorsink_servant );
	ior = orb->object_to_string( translatorsink_servant->_this() );
	iortable_adapter->rebind( ("OTIF/" + sink_str).c_str(), ior.in() );

	tr.translator = translator;
	tr.feed_servant_p = translatorfeed_servant;
	tr.sink_servant_p = translatorsink_servant;
	tr.feed_servant = feed_oid;
	tr.sink_servant = sink_oid;
	tr.command = run_command;
	tr.feed_str = feed_str;
	tr.sink_str = sink_str;
	tr.in_use = false;
	tr.notified = 0;
	tr.callbacking = false;

	translator_rep.insert( pair<const long, TRANSLATOR>( ID, tr) );

	logger.log(3, "translator registered:", (ID));
	return 0;
}

int _unregisterTranslator(const OTIF::TranslatorVersion &translator) {
	for(map<long, TRANSLATOR>::iterator i=translator_rep.begin(); i!=translator_rep.end(); i++) {
		if (__equalTranslators((*i).second.translator, translator)) {

			try {
				poa->deactivate_object( (*i).second.feed_servant );
				poa->deactivate_object( (*i).second.sink_servant );
				delete (*i).second.feed_servant_p;
				delete (*i).second.sink_servant_p;

				iortable_adapter->unbind( ("OTIF/" + (*i).second.feed_str).c_str() );
				iortable_adapter->unbind( ("OTIF/" + (*i).second.sink_str).c_str() );

				translator_rep.erase( (*i).first );
				datanetwork_translator_notification_rep.erase( (*i).first );

				logger.log(3, "translator unregistered");
			} catch (...) {
				logger.log(1, "Error: In translator deactivating");
			}
			return 0;
		}
	}

	logger.log(3, "can't find translator");
	return 2;
}

void _unregisterAllTranslator() {
	for(map<long, TRANSLATOR>::iterator i=translator_rep.begin(); i!=translator_rep.end(); i++) {
		try {
			poa->deactivate_object( (*i).second.feed_servant );
			poa->deactivate_object( (*i).second.sink_servant );
			delete (*i).second.feed_servant_p;
			delete (*i).second.sink_servant_p;
				
			iortable_adapter->unbind( ("OTIF/" + (*i).second.feed_str).c_str() );
			iortable_adapter->unbind( ("OTIF/" + (*i).second.sink_str).c_str() );

			translator_rep.erase( (*i).first );
			datanetwork_translator_notification_rep.erase( (*i).first );

			logger.log(3, "translator unregistered");
		} catch (...) {
			logger.log(1, "Error: In translator deactivating");
		}
	}
}

int _change_password(const char *old_password, const char *new_password) {

	if (!manager_key.compare(old_password)) {
		try {

			if (startup_params->root_element!=NULL) {
				DOM_NodeList nodes = startup_params->root_element.getChildNodes();
				for(DOM_Node node = startup_params->root_element.getFirstChild(); node != NULL; node = node.getNextSibling()) {
					if( node.getNodeType() == DOM_Node::ELEMENT_NODE ) {
						DOM_Element e = static_cast<DOM_Element&>(node);
						string node_name(node.getNodeName().transcode());
						if (!node_name.compare("options")) e.setAttribute(DOMString("password"), DOMString(new_password));
					}
				}

				DOM_Document doc = startup_params->root_element.getOwnerDocument();
				printDOM(doc, "BackPlane_config.xml");
			}

			manager_key = new_password;

		} catch (XMLException &e) {
			logger.log(1, "Error change password:", *(e.getMessage()));
		}
	} else {
		logger.log(3, "wrong password");
		return 1;
	}

	logger.log(3, "password changed");
	return 0;
}

void _shutdown() {

	try {
		orb->shutdown(0);
	} catch (const CORBA::Exception &) {
		logger.log(1, "Error: Caught a CORBA exception");
	}

	logger.log(3, "ORB shutdowned");
}


void __notifyChangedDocuments(long docId, long keepAlive) {
	map<long, DOCUMENT>::iterator i=document_repository.find(docId);
	if (i==document_repository.end()) return;

	pair<multimap<long, WORKFLOW>::iterator, multimap<long, WORKFLOW>::iterator> wf=workflow_rep.equal_range((*i).second.sender_id);
	while (wf.first != wf.second)  {

		// notify policy for ToolAdaptors
		for(map<long, TOOLADAPTOR_SESSION>::iterator j=tooladaptor_session_rep.begin(); j!=tooladaptor_session_rep.end(); j++) {

			// no callback object, tooladaptor did not subscribe
			if ( CORBA::is_nil((*j).second.callback) ) continue;

			// if link destination is equal to the type of the tooladaptor session, then notify (callback)
			if ((*(wf.first)).second.destination == (*j).second.ToolAdaptor_ID) {
				OTIF::ToolAdaptorNotifyCallback_var callback = (*j).second.callback;
				logger.log(3, "ToolAdaptor callback:", (long)(OTIF::ToolAdaptorNotifyCallback_ptr)callback);
				try {

					callback->changedExistingDocument((*i).second.document_name_version, (*i).second.remarks, (*i).first, keepAlive, (*i).second.paradigm_ids.size());

				} catch (const CORBA::Exception&) {
					_logoutToolAdaptor((*j).first);
					logger.log(3, "The callback connection with client session_id is corrupted. Client session is destroyed:", (*i).first);
				}
			}
		}
		
		++wf.first;
	}
}

int _destroyDocument(OTIF::DocumentID docId) {
	bool found = false;
	for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++) {
		if ((*i).first == docId) {
			found = true;

			__notifyChangedDocuments(docId, -1);

			delete (*i).second.document;
			document_repository.erase( (*i).first );
			updated_document_repository.erase( (*i).first );
			logger.log(3, "Destroy document:", (*i).first);
		}
	}
	if (!found) return 1;
	return 0;
}

int _modifyDocumentName(OTIF::DocumentID docId, const char *name) {
	bool found = false;
	for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++) {
		if ((*i).first == docId) {
			found = true;
			(*i).second.document_name_version.name = CORBA::string_dup(name);
			logger.log(3, "Change document name:", (*i).first);
		}
	}
	if (!found) return 1;
	return 0;
}

int _modifyDocumentVersion(OTIF::DocumentID docId, const char *version) {
	bool found = false;
	for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++) {
		if ((*i).first == docId) {
			found = true;
			(*i).second.document_name_version.version = CORBA::string_dup(version);
			logger.log(3, "Change document version:", (*i).first);
		}
	}
	if (!found) return 1;
	return 0;
}

int _modifyDocumentRemarks(OTIF::DocumentID docId, const char *remarks) {
	bool found = false;
	for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++) {
		if ((*i).first == docId) {
			found = true;
			(*i).second.remarks = CORBA::string_dup(remarks);
			logger.log(3, "Change document remarks:", (*i).first);
		}
	}
	if (!found) return 1;
	return 0;
}

int _modifyDocumentKeepAlive(OTIF::DocumentID docId, CORBA::Long keepAlive) {
	bool found = false;
	for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++) {
		if ((*i).first == docId) {
			found = true;
			(*i).second.keepAlive = keepAlive;
			logger.log(3, "Change document keepAlive:", (*i).first);
		}
	}
	if (!found) return 1;
	return 0;
}


// *******************************************************************************************************************************
// Function used by ToolAdaptorSession servant
// *******************************************************************************************************************************

bool __isSetsEqual(const multiset<long> &a, const multiset<long> &b) {
	if (a.size()!=b.size()) return false;
	
	multiset<long> c;
	set_difference(a.begin(), a.end(), b.begin(), b.end(), inserter(c, c.begin()));

	if (c.size()==0) return true;
	else return false;
}

list<long> __getTheWaitingTranslatorsList(const long source) {
	list<long> return_value;

	pair<multimap<long, WORKFLOW>::iterator, multimap<long, WORKFLOW>::iterator> wf=workflow_rep.equal_range(source);
	while (wf.first != wf.second)  {
		map<long, TRANSLATOR>::iterator i=translator_rep.find((*(wf.first)).second.destination);
		if (i!=translator_rep.end()) return_value.push_back((*(wf.first)).second.destination);
		++wf.first;
	}

	return return_value;
}

list<long> __getTheWaitingMultipleTranslatorList(const long source, const multiset<long> &paradigm_ids) {
	list<long> return_value;

	for(map<long, TRANSLATOR>::iterator k=translator_rep.begin(); k!=translator_rep.end(); k++) {

		multiset<long> par_ids;
		pair<multimap<long, WORKFLOW>::iterator, multimap<long, WORKFLOW>::iterator> wf=workflow_rep.equal_range(source);
		while (wf.first != wf.second)  {
			if ((*(wf.first)).second.destination==(*k).first) par_ids.insert((*(wf.first)).second.outputParadigm);
			++wf.first;
		}

		if (__isSetsEqual(paradigm_ids, par_ids)) return_value.push_back((*k).first);
	}

	return return_value;
}

int __startEXETranslator(TRANSLATOR tr) {
	// start the EXE translator
	logger.log(3, "Run EXE translator:", tr.command + " " + tr.feed_str + " " + tr.sink_str);
	cout << "Run EXE translator: " << tr.command << " " << log_level << " " << tr.feed_str << " " << tr.sink_str << endl;

	int argc = __translator_arguments_length + 4;
	char **arguments = new char*[argc+1];

	arguments[0] = new char[tr.command.length()+1];
	strcpy(arguments[0], tr.command.c_str());
	arguments[1] = new char[log_level.length()+1];
	strcpy(arguments[1], log_level.c_str());
	arguments[2] = new char[tr.feed_str.length()+1];
	strcpy(arguments[2], tr.feed_str.c_str());
	arguments[3] = new char[tr.sink_str.length()+1];
	strcpy(arguments[3], tr.sink_str.c_str());

	for(int i=4; i<argc; i++) {
		arguments[i] = new char[__translator_arguments[i-4].size()];
		strcpy(arguments[i], __translator_arguments[i-4].c_str());
	}
	arguments[argc] = NULL;

	string run = "./translators/";
	run.append(tr.command);
	int handle = _spawnv(_P_NOWAIT, run.c_str(), arguments);
	if (handle==-1) {
		logger.log(1, "Error: Could not start the translator!");
		cout << "Error: Could not start the translator!" << endl;
	}

	delete[] arguments;
	logger.log(3, "translator handle:", handle);
	return handle;
}

int __startXSLTranslator(TRANSLATOR tr) {
	// start the XSLT translator
	logger.log(3, "Run XSL translator:", tr.command);
	cout << "Run XSL translator: " << log_level << " " << tr.feed_str << " " << tr.sink_str << " " << (string)tr.translator.from.name << " " << (string)tr.translator.from.version << " " << (string)tr.translator.to.name << " " << (string)tr.translator.to.version << " " << tr.command << endl;

	int argc = __translator_arguments_length + 9;
	char **arguments = new char*[argc+1];

	arguments[0] = "Generic_XSLT_Translator.exe";
	arguments[1] = new char[log_level.length()+1];
	strcpy(arguments[1], log_level.c_str());
	arguments[2] = new char[tr.feed_str.length()+1];
	strcpy(arguments[2], tr.feed_str.c_str());
	arguments[3] = new char[tr.sink_str.length()+1];
	strcpy(arguments[3], tr.sink_str.c_str());
	arguments[4] = new char[((string)tr.translator.from.name).length()+1];
	strcpy(arguments[4], ((string)tr.translator.from.name).c_str());
	arguments[5] = new char[((string)tr.translator.from.version).length()+1];
	strcpy(arguments[5], ((string)tr.translator.from.version).c_str());
	arguments[6] = new char[((string)tr.translator.to.name).length()+1];
	strcpy(arguments[6], ((string)tr.translator.to.name).c_str());
	arguments[7] = new char[((string)tr.translator.to.version).length()+1];
	strcpy(arguments[7], ((string)tr.translator.to.version).c_str());
	arguments[8] = new char[((string)"xsl_translators/" + tr.command).length()+1];
	strcpy(arguments[8], ((string)"xsl_translators/" + tr.command).c_str());

	for(int z=9; z<argc; z++) {
		arguments[z] = new char[__translator_arguments[z-9].size()];
		strcpy(arguments[z], __translator_arguments[z-9].c_str());
	}
	arguments[argc] = NULL;

	int handle = _spawnv(_P_NOWAIT, "./xsl_translators/Generic_XSLT_Translator.exe", arguments);
	if (handle==-1) {
		logger.log(1, "Error: Could not start the translator!");
		cout << "Error: Could not start the translator!" << endl;
	}

	delete[] arguments;
	logger.log(3, "translator handle:", handle);
	return handle;
}

int __startJAVATranslator(TRANSLATOR tr) {
	// start the JAVA translator
	logger.log(3, "Run JAVA translator:", tr.command + " " + tr.feed_str + " " + tr.sink_str);
	cout << "Run JAVA translator: " << tr.command << " " << log_level << " " << tr.feed_str << " " << tr.sink_str << endl;

	int argc = __translator_arguments_length + 5;
	char **arguments = new char*[argc+1];

	arguments[0] = "java.exe";
	arguments[1] = new char[tr.command.length()+1];
	strcpy(arguments[1], tr.command.c_str());
	arguments[2] = new char[log_level.length()+1];
	strcpy(arguments[2], log_level.c_str());
	arguments[3] = new char[tr.feed_str.length()+1];
	strcpy(arguments[3], tr.feed_str.c_str());
	arguments[4] = new char[tr.sink_str.length()+1];
	strcpy(arguments[4], tr.sink_str.c_str());

	for(int i=5; i<argc; i++) {
		arguments[i] = new char[__translator_arguments[i-5].size()];
		strcpy(arguments[i], __translator_arguments[i-5].c_str());
	}
	arguments[argc] = NULL;

	int handle = _spawnvp(_P_NOWAIT, "java.exe", arguments);
	if (handle==-1) {
		logger.log(1, "Error: Could not start the translator!");
		cout << "Error: Could not start the translator!" << endl;
	}

	delete[] arguments;
	logger.log(3, "translator handle:", handle);
	return handle;
}

void manageDocumentToolAdaptor(const long document_id, const DOCUMENT &document) {
	for(map<long, TOOLADAPTOR_SESSION>::iterator tas=tooladaptor_session_rep.begin(); tas!=tooladaptor_session_rep.end(); tas++) {
		
		OTIF::ToolAdaptorNotifyCallback_var callback = (*tas).second.callback;
		// no callback object, tooladaptor did not subscribe, so skip it!
		if ( CORBA::is_nil(callback) ) continue;

		// check if notified
		bool notified=false;

		pair<multimap<long, long>::iterator, multimap<long, long>::iterator> q=datanetwork_tooladaptor_notification_rep.equal_range( (*tas).first );
		while (q.first != q.second)  {
			if ( (*(q.first)).second == document_id ) notified=true;
			++q.first;
		}

		// check if updated notified
		bool updated_doc = false;
		pair<multimap<long, UPDATED_DOCUMENT>::iterator, multimap<long, UPDATED_DOCUMENT>::iterator> uq=updated_document_repository.equal_range( document_id );
		while (uq.first != uq.second)  {
			if ( (*(uq.first)).second.destination==(*tas).second.ToolAdaptor_ID && (*(uq.first)).second.notified==false) {
				notified=false;
				updated_doc=true;
			}
			++uq.first;
		}

		if (!notified) {
	
			pair<multimap<long, WORKFLOW>::iterator, multimap<long, WORKFLOW>::iterator> wf=workflow_rep.equal_range(document.sender_id);
			while (wf.first != wf.second)  {

				map<long, TOOLADAPTOR>::iterator ta=tooladaptor_rep.find((*tas).second.ToolAdaptor_ID);
				if (ta!=tooladaptor_rep.end()) {

					if ((*(wf.first)).second.destination==(*tas).second.ToolAdaptor_ID &&
						document.paradigm_ids.find( (*(wf.first)).second.outputParadigm ) != document.paradigm_ids.end()) {
//						document.paradigm_id==(*(wf.first)).second.outputParadigm) {

						logger.log(3, "ToolAdaptor callback, document id:", document_id);
						try {
							long diff = __now() - document.creationTime;

							if (!updated_doc) {

								callback->notify(document.document_name_version, document.remarks, document_id, (document.keepAlive - diff), document.paradigm_ids.size());
								datanetwork_tooladaptor_notification_rep.insert( pair<const long, long>( (*tas).first, document_id) );
								logger.log(3, "Notify tooladaptor about new document:", document_id);

							} else {

								callback->changedExistingDocument(document.document_name_version, document.remarks, document_id, (document.keepAlive - diff), document.paradigm_ids.size());
								pair<multimap<long, UPDATED_DOCUMENT>::iterator, multimap<long, UPDATED_DOCUMENT>::iterator> uq=updated_document_repository.equal_range( document_id );
								while (uq.first != uq.second)  {
									if ( (*(uq.first)).second.destination==(*tas).second.ToolAdaptor_ID) (*(uq.first)).second.notified=true;
									++uq.first;
								}
								logger.log(3, "Notify tooladaptor about updated document:", document_id);

							}
						
						} catch (const CORBA::Exception&) {
							_logoutToolAdaptor((*tas).first);
							logger.log(3, "The callback connection with client session_id is corrupted. Client session is destroyed:", (*tas).first);
						}
					}
				}
				++wf.first;
			}
		}
	}
}

void manageDocumentTranslator(const long document_id, const DOCUMENT &document)
{
	for(map<long, TRANSLATOR>::iterator k=translator_rep.begin(); k!=translator_rep.end(); k++)
  {

		// check if notified
		bool notified=false;
		pair<multimap<long, long>::iterator, multimap<long, long>::iterator> q=datanetwork_translator_notification_rep.equal_range( (*k).first );
		while (q.first != q.second) 
    {
			if ( (*(q.first)).second == document_id ) notified=true;
			++q.first;
		}

		// check if updated notified
		bool updated_doc = false;
		pair<multimap<long, UPDATED_DOCUMENT>::iterator, multimap<long, UPDATED_DOCUMENT>::iterator> uq=updated_document_repository.equal_range( document_id );
		while (uq.first != uq.second) 
    {
			if ( (*(uq.first)).second.destination==(*k).first && (*(uq.first)).second.notified==false)
      {
				notified=false;
				updated_doc=true;
			}
			++uq.first;
		}

		if (!notified)
    {
			TRANSLATOR tr = (*k).second;

			pair<multimap<long, WORKFLOW>::iterator, multimap<long, WORKFLOW>::iterator> wf=workflow_rep.equal_range(document.sender_id);
			while (wf.first != wf.second) 
      {
				if ((*(wf.first)).second.destination==(*k).first &&
					*(document.paradigm_ids.begin())==(*(wf.first)).second.outputParadigm)
        {
		
					if (!tr.in_use)
          {
						// only one of a kind translator can be running at once!!!
						(*k).second.in_use = true;
						
						if (tr.type == 0) { // EXE translator
							if (__startEXETranslator(tr)==-1) (*k).second.in_use = false; 
						} else if (tr.type == 1) { // XSLT translator
							if (__startXSLTranslator(tr)==-1) (*k).second.in_use = false; 
						} else if (tr.type == 2) { // JAVA translator
							if (__startJAVATranslator(tr)==-1) (*k).second.in_use = false; 
						}

					}
          else 
          {
						// Translator should be running!!

						// The translator was started, but maybe it terminated before it could subscribe to anything
						// or has not subscribed yet, so we just break, it will be notifed when it subscribes
						if (CORBA::is_nil(tr.callback)) break;

						OTIF::TranslatorNotifyCallback_var callback = tr.callback;
						logger.log(3, "Translator callback:", (long)(OTIF::TranslatorNotifyCallback_ptr)callback);
						try {

							callback->notify(document.document_name_version, document_id, document.keepAlive, 1);

							if (!updated_doc) {
								datanetwork_translator_notification_rep.insert( pair<const long, long>( (*k).first, document_id) );
							} else {
								pair<multimap<long, UPDATED_DOCUMENT>::iterator, multimap<long, UPDATED_DOCUMENT>::iterator> uq=updated_document_repository.equal_range( document_id );
								while (uq.first != uq.second)  {
									if ( (*(uq.first)).second.destination==(*k).first) (*(uq.first)).second.notified=true;
									++uq.first;
								}
							}

							(*k).second.notified = (*k).second.notified + 1;

						}
            
            catch (const CORBA::Exception& ex) 
            {
              
              std::stringstream str;
              str << ex;

              logger.log(3, str.str().c_str(), document_id);

							logger.log(3, "The callback connection with translator session_id is corrupted. Translator restarted:", document_id);


							// translator may be terminated, should be restarted
							(*k).second.callback = OTIF::TranslatorNotifyCallback::_nil();
							(*k).second.in_use = true;
							(*k).second.notified = 0;
							(*k).second.callbacking = false;

							if (tr.type == 0) { // EXE translator
								if (__startEXETranslator(tr)==-1) (*k).second.in_use = false; 
							} else if (tr.type == 1) { // XSLT translator
								if (__startXSLTranslator(tr)==-1) (*k).second.in_use = false; 
							} else if (tr.type == 2) { // JAVA translator
								if (__startJAVATranslator(tr)==-1) (*k).second.in_use = false; 
							}
						}

					} // if (!tr.in_use)
				}
				++wf.first;
			}

		} // if (!notified)
	}
}


void manageMultiDocumentTranslator(const long document_id, const DOCUMENT &document) {
	for(map<long, TRANSLATOR>::iterator k=translator_rep.begin(); k!=translator_rep.end(); k++) {

		// check if notified
		bool notified=false;
		pair<multimap<long, long>::iterator, multimap<long, long>::iterator> q=datanetwork_translator_notification_rep.equal_range( (*k).first );
		while (q.first != q.second)  {
			if ( (*(q.first)).second == document_id ) notified=true;
			++q.first;
		}

		// check if updated notified
		bool updated_doc = false;
		pair<multimap<long, UPDATED_DOCUMENT>::iterator, multimap<long, UPDATED_DOCUMENT>::iterator> uq=updated_document_repository.equal_range( document_id );
		while (uq.first != uq.second)  {
			if ( (*(uq.first)).second.destination==(*k).first && (*(uq.first)).second.notified==false) {
				notified=false;
				updated_doc=true;
			}
			++uq.first;
		}

		if (!notified) {
			TRANSLATOR tr = (*k).second;

			multiset<long> par_ids;
			pair<multimap<long, WORKFLOW>::iterator, multimap<long, WORKFLOW>::iterator> wf=workflow_rep.equal_range(document.sender_id);
			while (wf.first != wf.second)  {
				if ((*(wf.first)).second.destination==(*k).first) par_ids.insert((*(wf.first)).second.outputParadigm);
				++wf.first;
			}
			
			if (__isSetsEqual(document.paradigm_ids, par_ids)) {

				if (!tr.in_use) {
					// only one of a kind translator can be running at once!!!
					(*k).second.in_use = true;
						
					if (tr.type == 0) { // EXE translator
						if (__startEXETranslator(tr)==-1) (*k).second.in_use = false; 
					} else if (tr.type == 2) { // JAVA translator
						if (__startJAVATranslator(tr)==-1) (*k).second.in_use = false; 
					}

				} else {
					// Translator should be running!!

					// The translator was started, but maybe it terminated before it could subscribe to anything
					// or has not subscribed yet, so we just break, it will be notifed when it subscribes
					if (CORBA::is_nil(tr.callback)) break;

					OTIF::TranslatorNotifyCallback_var callback = tr.callback;
					logger.log(3, "Translator callback:", (long)(OTIF::TranslatorNotifyCallback_ptr)callback);
					try {

						callback->notify(document.document_name_version, document_id, document.keepAlive, document.paradigm_ids.size());

						if (!updated_doc) {
							datanetwork_translator_notification_rep.insert( pair<const long, long>( (*k).first, document_id) );
						} else {
							pair<multimap<long, UPDATED_DOCUMENT>::iterator, multimap<long, UPDATED_DOCUMENT>::iterator> uq=updated_document_repository.equal_range( document_id );
							while (uq.first != uq.second)  {
								if ( (*(uq.first)).second.destination==(*k).first) (*(uq.first)).second.notified=true;
								++uq.first;
							}
						}

						(*k).second.notified = (*k).second.notified + 1;

					} catch (const CORBA::Exception&) {
						logger.log(3, "The callback connection with translator session_id is corrupted. Translator restarted:", document_id);

						// translator may be terminated, should be restarted
						(*k).second.callback = OTIF::TranslatorNotifyCallback::_nil();
						(*k).second.in_use = true;
						(*k).second.notified = 0;
						(*k).second.callbacking = false;

						if (tr.type == 0) { // EXE translator
							if (__startEXETranslator(tr)==-1) (*k).second.in_use = false; 
						} else if (tr.type == 2) { // JAVA translator
							if (__startJAVATranslator(tr)==-1) (*k).second.in_use = false; 
						}
					}
				} // if (!tr.in_use)
			} // if (__isSetsEqual...
		} // if (!notified)
	}
}


void manageDocuments() {
	for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++) {

		// keepAlive policy
		if ( ((*i).second.creationTime + (*i).second.keepAlive) < __now() && (*i).second.waiting_translators.empty()) {
			delete (*i).second.document;
			document_repository.erase( (*i).first );
			updated_document_repository.erase( (*i).first );

			logger.log(3, "Destroy document:", (*i).first);
			continue;
		}

		// notify policy for ToolAdaptors
		manageDocumentToolAdaptor((*i).first, (*i).second);

		// notify policy for Translators
//		manageDocumentTranslator((*i).first, (*i).second);


		if ((*i).second.multidocument) {
			manageMultiDocumentTranslator((*i).first, (*i).second);
		} else {
			manageDocumentTranslator((*i).first, (*i).second);
		}


	}
}


int _publishDocument(const long session_id, const OTIF::NameVersion &document, const char *remarks, CORBA::Long keepAlive, 
					 const OTIF::Document &docData, OTIF::DocumentID previous_docId, OTIF::DocumentID &document_id) {

	map<long, TOOLADAPTOR_SESSION>::iterator i=tooladaptor_session_rep.find(session_id);
	if (i==tooladaptor_session_rep.end()) return 2; // invalid session

	long ToolAdaptor_ID = (*i).second.ToolAdaptor_ID;

	bool multidocument = false;
	long struct_length = 0;

	multiset<long> par_ids;
	for(CORBA::ULong j=0; j<docData.documents.length(); j++) {
		if (docData.documents[j]._d() == OTIF::structElem) {
			struct_length++;

			string par_name = docData.documents[j].sElem().paradigm.name;
			string par_version = docData.documents[j].sElem().paradigm.version;

			bool found = false;
			for(map<long, OTIF::NameVersion>::iterator q=metamodel_repository.begin(); q!=metamodel_repository.end(); q++) {
				if ( !par_name.compare((*q).second.name) && !par_version.compare((*q).second.version) ) {
					par_ids.insert((*q).first);
					found = true;
				}
			}
			if (!found) return 1; // invalid paradigm
		}
	}
	
	if (struct_length>1) multidocument = true;

	// Update the previous document
	if (previous_docId != -1) {
		map<long, DOCUMENT>::iterator di = document_repository.find(previous_docId);
		if (di != document_repository.end()) {
			
			// Need a deep copy!!!
			OTIF::Document *t_dn = new OTIF::Document(docData);
			// -------------------

			(*di).second.sender_id = ToolAdaptor_ID;
			(*di).second.document = t_dn;
			(*di).second.document_name_version = document;
			(*di).second.keepAlive = keepAlive;
			(*di).second.remarks = CORBA::string_dup(remarks);
			(*di).second.creationTime = __now();

			(*di).second.paradigm_ids = par_ids;
			(*di).second.multidocument = multidocument;

			(*di).second.source_docId = -1;

			document_id = previous_docId;

			// store the update
			for(multimap<long, WORKFLOW>::iterator wi=workflow_rep.begin(); wi!=workflow_rep.end(); wi++) {
				if ((*wi).first == ToolAdaptor_ID) {
					UPDATED_DOCUMENT ud;
					ud.destination = (*wi).second.destination;
					ud.notified = false;
					updated_document_repository.insert(pair<const long, UPDATED_DOCUMENT>(previous_docId, ud));
				}
			}

			logger.log(3, "tooladaptor updated document:", (string)document.name + ":" + (string)document.version + ":" + (string)remarks);

		} else return 3; // invalid document


	// New document
	} else {

		// Need a deep copy!!!
		OTIF::Document *t_dn = new OTIF::Document(docData);
		// -------------------

		document_repository[Document_ID].sender_id = ToolAdaptor_ID;
		document_repository[Document_ID].document = t_dn;
		document_repository[Document_ID].document_name_version = document;
		document_repository[Document_ID].keepAlive = keepAlive;
		document_repository[Document_ID].remarks = CORBA::string_dup(remarks);
		document_repository[Document_ID].creationTime = __now();

		document_repository[Document_ID].paradigm_ids = par_ids;
		document_repository[Document_ID].multidocument = multidocument;

		document_repository[Document_ID].source_docId = -1;

		document_id = Document_ID;

		if (!multidocument) {
			document_repository[Document_ID].waiting_translators = __getTheWaitingTranslatorsList(ToolAdaptor_ID);
		} else {
			document_repository[Document_ID].waiting_translators = __getTheWaitingMultipleTranslatorList(ToolAdaptor_ID, par_ids);
		}

		Document_ID++;

		logger.log(3, "tooladaptor published document:", (string)document.name + ":" + (string)document.version + ":" + (string)remarks);
	}

	manageDocuments();
	return 0;
}


int _subscribe (const long session_id, OTIF::ToolAdaptorNotifyCallback_ptr callback) {
	map<long, TOOLADAPTOR_SESSION>::iterator i = tooladaptor_session_rep.find(session_id);
	if (i!=tooladaptor_session_rep.end()) {
		tooladaptor_session_rep[session_id].callback = OTIF::ToolAdaptorNotifyCallback::_duplicate(callback);
		logger.log(3, "Subscribe tooladaptor callback:", (long)(OTIF::ToolAdaptorNotifyCallback_ptr)callback);
		logger.log(3, "Subscribe tooladaptor session:", session_id);
	} else return 2;

	manageDocuments();
	return 0;
}

int _fetchDocument(const OTIF::DocumentID docId, OTIF::Document_out docData) {

	map<long, DOCUMENT>::iterator i = document_repository.find(docId);
	if (i==document_repository.end()) {
		logger.log(3, "Did not find the corresponding document docId:", docId);
		return 1;
	}

	// Have to create a copy with deep copy, because _retn() owns the pointer!!!
	OTIF::Document_var t_dn = new OTIF::Document();
	*t_dn = *((*i).second.document);
	docData = t_dn._retn();
	// -----------------------------------------------------------------------

	logger.log(3, "Fetched document:", docId);
	return 0;
}

int _getToolAdaptorDocuments(const long session_id, OTIF::DocumentInfo_S_var &datanetworks) {
	try {

		long size = 0;
		CORBA::ULong offset = 0;
		for(map<long, DOCUMENT>::iterator i=document_repository.begin(); i!=document_repository.end(); i++, offset++) {
			long diff = (*i).second.keepAlive - (__now() - (*i).second.creationTime);
			if (diff>0) size++;
		}
		
		datanetworks->length( size );

		offset = 0;
		for(i=document_repository.begin(); i!=document_repository.end(); i++, offset++) {
			long diff = (*i).second.keepAlive - (__now() - (*i).second.creationTime);
			if (diff<=0) continue;

			(*datanetworks)[offset].docId = (*i).first;

			OTIF::NameVersion_S_var paradigms = new OTIF::NameVersion_S();

			CORBA::ULong _length = 0;
			for(CORBA::ULong j=0; j<(*i).second.document->documents.length(); j++) {
				if ((*i).second.document->documents[j]._d() == OTIF::structElem) _length++;
			}

			paradigms->length( _length );
			for(j=0; j<(*i).second.document->documents.length(); j++) {
				if ((*i).second.document->documents[j]._d() == OTIF::structElem) {
					(*paradigms)[j].name = (*i).second.document->documents[j].sElem().paradigm.name;
					(*paradigms)[j].version = (*i).second.document->documents[j].sElem().paradigm.version;
				}
			}
			(*datanetworks)[offset].paradigms = paradigms;
			
			(*datanetworks)[offset].documentName.name = (*i).second.document_name_version.name;
			(*datanetworks)[offset].documentName.version = (*i).second.document_name_version.version;
			(*datanetworks)[offset].remarks = CORBA::string_dup((*i).second.remarks);
			(*datanetworks)[offset].keepAlive = diff;
		}

		logger.log(3, "get tooladaptor documents - size:", (long)offset);

	} catch (...) {
		logger.log(1, "Error: _getToolAdaptorDocuments fatal error!");
		return 1;
	}
	return 0;
}

int _getDocumentRootObjectMeta(const OTIF::DocumentID docId, CORBA::String_out rootobjectmeta) {

	map<long, DOCUMENT>::iterator i = document_repository.find(docId);
	if (i==document_repository.end()) {
		logger.log(3, "Did not find the corresponding document docId:", docId);
		return 1;
	}

//	What happens with the multidocuments?
//	if ((*i).second.multidocument) return 1;


	unsigned long ro_tid;
	unsigned long rootobject_id = (*i).second.document->documents[0].sElem().data.root.id;
	for(CORBA::ULong ww=0; ww<(*i).second.document->documents[0].sElem().data.objects.length(); ww++) {
		if ((*i).second.document->documents[0].sElem().data.objects[ww].id.id == rootobject_id)
			ro_tid = (*i).second.document->documents[0].sElem().data.objects[ww].tId;
	}

	for(map<long, OTIF_Class_i *>::iterator q=class_servant_map.begin(); q!=class_servant_map.end(); q++) {
		if ((unsigned)(*q).second->get_id() == ro_tid) rootobjectmeta = CORBA::string_dup((*q).second->get_name());
	}
	
	logger.log(3, "Get rootobject's meta of the document:", docId);
	return 0;
}

int _getDocumentParadigm(const OTIF::DocumentID docId, OTIF::NameVersion *paradigm) {

	map<long, DOCUMENT>::iterator i = document_repository.find(docId);
	if (i==document_repository.end()) {
		logger.log(3, "Did not find the corresponding document docId:", docId);
		return 1;
	}

//	What happens with the multidocuments?
//	if ((*i).second.multidocument) return 1;

	paradigm->name = CORBA::string_dup((*i).second.document->documents[0].sElem().paradigm.name);
	paradigm->version = CORBA::string_dup((*i).second.document->documents[0].sElem().paradigm.version);

	logger.log(3, "Get paradigm of the document:", docId);
	return 0;
}


// *******************************************************************************************************************************
// Functions used by TranslatorFeed & TranslatorSink servants
// *******************************************************************************************************************************
int _tf_subscribe(long session_id, OTIF::TranslatorNotifyCallback_ptr callback) {
	map<long, TRANSLATOR>::iterator i = translator_rep.find(session_id);
	if (i!=translator_rep.end()) {
		(*i).second.callback = OTIF::TranslatorNotifyCallback::_duplicate(callback);
		logger.log(3, "Subscribe translator callback:", (long)(OTIF::TranslatorNotifyCallback_ptr)callback);
		logger.log(3, "Subscribe translator session:", session_id);
	} else return 1;

	manageDocuments();
	return 0;
}

int _ts_publishDocument(long session_id, const OTIF::NameVersion & document, const char * remarks, CORBA::Long keepAlive, 
						const OTIF::Document & docData, OTIF::DocumentID source_docId) {

	map<long, TRANSLATOR>::iterator i = translator_rep.find(session_id);
	if (i==translator_rep.end()) return 1; // invalid session

	bool multidocument = false;
	long struct_length = 0;

	multiset<long> par_ids;
	for(CORBA::ULong j=0; j<docData.documents.length(); j++) {
		if (docData.documents[j]._d() == OTIF::structElem) {
			struct_length++;

			string par_name = docData.documents[j].sElem().paradigm.name;
			string par_version = docData.documents[j].sElem().paradigm.version;

			bool found = false;
			for(map<long, OTIF::NameVersion>::iterator q=metamodel_repository.begin(); q!=metamodel_repository.end(); q++) {
				if ( !par_name.compare((*q).second.name) && !par_version.compare((*q).second.version) ) {
					par_ids.insert((*q).first);
					found = true;
				}
			}
			if (!found) return 1; // invalid paradigm
		}
	}

	if (struct_length>1) multidocument = true;


	// decide new or updated doc
	bool updated_doc=false;
	pair<multimap<long, UPDATED_DOCUMENT>::iterator, multimap<long, UPDATED_DOCUMENT>::iterator> uq=updated_document_repository.equal_range(source_docId);
	while (uq.first != uq.second)  {
		if ( (*(uq.first)).second.destination==(*i).first && (*(uq.first)).second.notified==true) {
			updated_doc=true;
		}
		++uq.first;
	}

	// updated document by translator
	if (updated_doc) {

		for(map<long, DOCUMENT>::iterator doc=document_repository.begin(); doc!=document_repository.end(); doc++) {
			if ((*doc).second.source_docId == source_docId && (*doc).second.paradigm_ids == par_ids ) {
				
				// Need a deep copy!!!
				OTIF::Document *t_dn = new OTIF::Document(docData);
				// -------------------

				(*doc).second.sender_id = session_id;
				(*doc).second.document = t_dn;
				(*doc).second.document_name_version = document;
				(*doc).second.keepAlive = (translator_keepalive_parameter > keepAlive) ? translator_keepalive_parameter : keepAlive;
				(*doc).second.remarks = CORBA::string_dup(remarks);
				(*doc).second.creationTime = __now();

				(*doc).second.paradigm_ids = par_ids;
				(*doc).second.multidocument = multidocument;

				(*doc).second.source_docId = source_docId;

				if (!multidocument) {
					(*doc).second.waiting_translators = __getTheWaitingTranslatorsList(session_id);
				} else {
					(*doc).second.waiting_translators = __getTheWaitingMultipleTranslatorList(session_id, par_ids);
				}
			}
		}

	// new document by translator
	} else {

		// Need a deep copy!!!
		OTIF::Document *t_dn = new OTIF::Document(docData);
		// -------------------

		document_repository[Document_ID].sender_id = session_id;
		document_repository[Document_ID].document = t_dn;
		document_repository[Document_ID].document_name_version = document;
		document_repository[Document_ID].keepAlive = (translator_keepalive_parameter > keepAlive) ? translator_keepalive_parameter : keepAlive;
		document_repository[Document_ID].remarks = CORBA::string_dup(remarks);
		document_repository[Document_ID].creationTime = __now();

		document_repository[Document_ID].paradigm_ids = par_ids;
		document_repository[Document_ID].multidocument = multidocument;

		document_repository[Document_ID].source_docId = source_docId;

		if (!multidocument) {
			document_repository[Document_ID].waiting_translators = __getTheWaitingTranslatorsList(session_id);
		} else {
			document_repository[Document_ID].waiting_translators = __getTheWaitingMultipleTranslatorList(session_id, par_ids);
		}

		Document_ID++;
	}

	(*i).second.notified = (*i).second.notified - 1;
	// translator is ready, may be reseted
	if ((*i).second.notified == 0) {
		(*i).second.in_use = false;
		(*i).second.callback = OTIF::TranslatorNotifyCallback::_nil();
		logger.log(3, "translator is ready, may be reseted");
	}

	// document's translator waiting list management
	map<long, DOCUMENT>::iterator doc = document_repository.find(source_docId);
	if (doc != document_repository.end()) doc->second.waiting_translators.remove(session_id);

	logger.log(3, "translator published document:", (string)document.name + ":" + (string)document.version + ":" + (string)remarks);

	manageDocuments();
	return 0;
}


// *******************************************************************************************************************************
// *******************************************************************************************************************************
// *******************************************************************************************************************************
// Proces a MetaModel
// *******************************************************************************************************************************
// *******************************************************************************************************************************
// *******************************************************************************************************************************
OTIF::Attrib_var UML_attribute_pass(const Uml::Attribute &uml_attribute, OTIF::Class_var &parent_class, long MetaModel_ID) {
  OTIF_Attrib_i *attribute_servant = new OTIF_Attrib_i();
  OTIF::Attrib_var attribute_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( attribute_servant );
    CORBA::Object_var attribute_obj = poa->id_to_reference( oid.in() );
    attribute_obj_var = OTIF::Attrib::_narrow(attribute_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) attribute_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  if (MetaModel_ID==0) attrib_servant_map[Servant_ID] = attribute_obj_var;

  (*attribute_servant).set_id( Servant_ID++ );
  (*attribute_servant).set_name(((string)uml_attribute.name()).c_str());
  (*attribute_servant).set_type(((string)uml_attribute.type()).c_str());
  (*attribute_servant).set_minCard((long)uml_attribute.min());
  (*attribute_servant).set_maxCard((long)uml_attribute.max());
  (*attribute_servant).set_nonpersistent(uml_attribute.nonpersistent());
  (*attribute_servant).set_ordered(uml_attribute.ordered());
  (*attribute_servant).set_visibility(((string)uml_attribute.visibility()).c_str());
  (*attribute_servant).set_registry(uml_attribute.registry());
  (*attribute_servant).set_parent(parent_class);

  vector<string> defvls = uml_attribute.defvalue();
  OTIF::string_S  *defvalues = new OTIF::string_S(defvls.size());
  defvalues->length(defvls.size());

  CORBA::ULong offset = 0;
  for(vector<string>::iterator k = defvls.begin(); k!= defvls.end(); k++, offset++) {
	CORBA::String_var defvalue = CORBA::string_dup((*k).c_str());
    (*defvalues)[offset] = defvalue;
	logger.log(5, "    -Attribute defvalue: ", (string)defvalue);
  }
  (*attribute_servant).set_defvalue(defvalues);

  logger.log(5, "    -Attribute name: ", uml_attribute.name());

  return attribute_obj_var;
}

OTIF::CompositionParentRole_var UML_compositionparent_role_pass(const Uml::CompositionParentRole &uml_composition_parent_role, OTIF::Class_var &target_class, long MetaModel_ID) {
  OTIF_CompositionParentRole_i *compositionparent_role_servant = new OTIF_CompositionParentRole_i();
  OTIF::CompositionParentRole_var compositionparent_role_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( compositionparent_role_servant );
    CORBA::Object_var compositionparent_role_obj = poa->id_to_reference( oid.in() );
    compositionparent_role_obj_var = OTIF::CompositionParentRole::_narrow(compositionparent_role_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) compositionparent_role_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  (*compositionparent_role_servant).set_id( Servant_ID++ );
  string p_name = Uml::MakeRoleName(uml_composition_parent_role);
  (*compositionparent_role_servant).set_name(p_name.c_str());
  (*compositionparent_role_servant).set_isNavigable(uml_composition_parent_role.isNavigable());
  (*compositionparent_role_servant).set_target(target_class);
  //  (*compositionparent_role_servant).set_parent();

  Uml::Composition uml_composition = uml_composition_parent_role.parent();
  long parent_id = uml_composition.uniqueId();
  composition_parent_role_map[parent_id] = compositionparent_role_servant;

  logger.log(5, "    -CompositionParentRole name: ", p_name);

  return compositionparent_role_obj_var;
}

OTIF::CompositionChildRole_var UML_compositionchild_role_pass(const Uml::CompositionChildRole &uml_composition_child_role, OTIF::Class_var &target_class, long MetaModel_ID) {
  OTIF_CompositionChildRole_i *compositionchild_role_servant = new OTIF_CompositionChildRole_i();
  OTIF::CompositionChildRole_var compositionchild_role_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( compositionchild_role_servant );
    CORBA::Object_var compositionchild_role_obj = poa->id_to_reference( oid.in() );
    compositionchild_role_obj_var = OTIF::CompositionChildRole::_narrow(compositionchild_role_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) compositionchild_role_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  (*compositionchild_role_servant).set_id( Servant_ID++ );
  string c_name = Uml::MakeShortRoleName(uml_composition_child_role);
  (*compositionchild_role_servant).set_name(c_name.c_str());
  (*compositionchild_role_servant).set_isNavigable(uml_composition_child_role.isNavigable());
  (*compositionchild_role_servant).set_target(target_class);
  (*compositionchild_role_servant).set_minCard((long)uml_composition_child_role.min());
  (*compositionchild_role_servant).set_maxCard((long)uml_composition_child_role.max());

  //  (*compositionchild_role_servant).set_parent();

  Uml::Composition uml_composition = uml_composition_child_role.parent();
  long parent_id = uml_composition.uniqueId();
  composition_child_role_map[parent_id] = compositionchild_role_servant;

  logger.log(5, "    -CompositionChildRole name: ", c_name);

  return compositionchild_role_obj_var;
}

OTIF::Class_var UML_class_pass(const Uml::Class &uml_class, OTIF::Diagram_var &parent_diagram, long MetaModel_ID) {
  OTIF_Class_i *class_servant = new OTIF_Class_i();
  OTIF::Class_var class_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( class_servant );
    CORBA::Object_var class_obj = poa->id_to_reference( oid.in() );
    class_obj_var = OTIF::Class::_narrow(class_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) class_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  if (MetaModel_ID==0) cls_servant_map[Servant_ID] = class_obj_var;
// DONT FORGET TO CLEAR THE ARRAY!!!
//  cls_servant_map[Servant_ID] = class_obj_var;

  (*class_servant).set_id( Servant_ID++ );
  (*class_servant).set_name(((string)uml_class.name()).c_str());
  (*class_servant).set_parent(parent_diagram);
  (*class_servant).set_stereotype(((string)uml_class.stereotype()).c_str());
  (*class_servant).set_isAbstract(uml_class.isAbstract());

  logger.log(5, "  -Class name: ", uml_class.name());

  long class_id = uml_class.uniqueId();
  class_servant_map[class_id] = class_servant;
  class_map[class_id] = class_obj_var;

  // Get the composition parent roles from the class
  set<Uml::CompositionParentRole> uml_compositionparent_roles = uml_class.parentRoles();
  OTIF::CompositionParentRole_S *compositionparent_roles = new OTIF::CompositionParentRole_S(uml_compositionparent_roles.size());
  compositionparent_roles->length(uml_compositionparent_roles.size());

  CORBA::ULong offset = 0;
  for(set<Uml::CompositionParentRole>::iterator i = uml_compositionparent_roles.begin(); i!= uml_compositionparent_roles.end(); i++, offset++) {
    OTIF::CompositionParentRole_var parent_role_obj_var = UML_compositionparent_role_pass(*i, class_obj_var, MetaModel_ID);
    (*compositionparent_roles)[offset] = parent_role_obj_var;
  }
  (*class_servant).set_parentRoles(compositionparent_roles);
  //  delete compositionparent_roles;

  // Get the composition child roles from the class
  set<Uml::CompositionChildRole> uml_compositionchild_roles = uml_class.childRoles();
  OTIF::CompositionChildRole_S *compositionchild_roles = new OTIF::CompositionChildRole_S(uml_compositionchild_roles.size());
  compositionchild_roles->length(uml_compositionchild_roles.size());

  offset = 0;
  for(set<Uml::CompositionChildRole>::iterator j = uml_compositionchild_roles.begin(); j!= uml_compositionchild_roles.end(); j++, offset++) {
	
	// HACK!!!!
	if (UdmDOM_bug_exists) {
		Uml::CompositionParentRole _temp = Uml::theOther( (*j) );
		Uml::CompositionChildRole j_ok = Uml::theOther( _temp );
		OTIF::CompositionChildRole_var child_role_obj_var = UML_compositionchild_role_pass(j_ok, class_obj_var, MetaModel_ID);
	    (*compositionchild_roles)[offset] = child_role_obj_var;
		// HACK!!!!

	} else {
		OTIF::CompositionChildRole_var child_role_obj_var = UML_compositionchild_role_pass(*j, class_obj_var, MetaModel_ID);
	    (*compositionchild_roles)[offset] = child_role_obj_var;
	}
  }
  (*class_servant).set_childRoles(compositionchild_roles);
  //  delete compositionchild_roles;

  // Get the attributes from the class
  set<Uml::Attribute> uml_attributes = uml_class.attributes();
  OTIF::Attrib_S  *attributes = new OTIF::Attrib_S(uml_attributes.size());
  attributes->length(uml_attributes.size());

  offset = 0;
  for(set<Uml::Attribute>::iterator k = uml_attributes.begin(); k!= uml_attributes.end(); k++, offset++) {
    OTIF::Attrib_var attribute_obj_var = UML_attribute_pass(*k, class_obj_var, MetaModel_ID);
    (*attributes)[offset] = attribute_obj_var;
  }
  (*class_servant).set_attributes(attributes);
  //  delete attributes;

  return class_obj_var;
}

OTIF::Composition_var UML_composition_pass(const Uml::Composition &uml_composition, OTIF::Diagram_var &parent_diagram, long MetaModel_ID) {
  OTIF_Composition_i *composition_servant = new OTIF_Composition_i();
  OTIF::Composition_var composition_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( composition_servant );
    CORBA::Object_var composition_obj = poa->id_to_reference( oid.in() );
    composition_obj_var = OTIF::Composition::_narrow(composition_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) composition_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  (*composition_servant).set_id( Servant_ID++ );
  (*composition_servant).set_name(((string)uml_composition.name()).c_str());
  (*composition_servant).set_parent(parent_diagram);
  (*composition_servant).set_nonpersistent(uml_composition.nonpersistent());

  long parent_id = uml_composition.uniqueId();
  if (composition_parent_role_map.count(parent_id) == 1) {
    OTIF_CompositionParentRole_i *compositionparent_role_servant = composition_parent_role_map[parent_id];
    OTIF::CompositionParentRole_var compositionparent_role_obj_var;

    try {
      PortableServer::ObjectId_var oid = poa->servant_to_id( compositionparent_role_servant );
      CORBA::Object_var compositionparent_role_obj = poa->id_to_reference( oid.in() );
      compositionparent_role_obj_var = OTIF::CompositionParentRole::_narrow(compositionparent_role_obj.in());
    } catch (const CORBA::Exception &) {
	  logger.log(1, "Error: Caught a CORBA exception");
    }

    (*compositionparent_role_servant).set_parent(composition_obj_var);
    (*composition_servant).set_parentRole(compositionparent_role_obj_var);
  }
  if (composition_child_role_map.count(parent_id) == 1) {
    OTIF_CompositionChildRole_i *compositionchild_role_servant = composition_child_role_map[parent_id];
    OTIF::CompositionChildRole_var compositionchild_role_obj_var;

    try {
      PortableServer::ObjectId_var oid = poa->servant_to_id( compositionchild_role_servant );
      CORBA::Object_var compositionchild_role_obj = poa->id_to_reference( oid.in() );
      compositionchild_role_obj_var = OTIF::CompositionChildRole::_narrow(compositionchild_role_obj.in());
    } catch (const CORBA::Exception &) {
      logger.log(1, "Error: Caught a CORBA exception");
    }

    (*compositionchild_role_servant).set_parent(composition_obj_var);
    (*composition_servant).set_childRole(compositionchild_role_obj_var);
  }

  logger.log(5, "  -Composition name: ", uml_composition.name());

  return composition_obj_var;
}

OTIF::AssociationRole_var UML_association_role_pass(const Uml::AssociationRole &uml_association_role, OTIF::Association_var &parent_association, long MetaModel_ID) {
  OTIF_AssociationRole_i *association_role_servant = new OTIF_AssociationRole_i();
  OTIF::AssociationRole_var association_role_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( association_role_servant );
    CORBA::Object_var association_role_obj = poa->id_to_reference( oid.in() );
    association_role_obj_var = OTIF::AssociationRole::_narrow(association_role_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) association_role_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  if (MetaModel_ID==0) assocrole_servant_map[Servant_ID] = association_role_obj_var;

  (*association_role_servant).set_id( Servant_ID++ );
  (*association_role_servant).set_name(((string)uml_association_role.name()).c_str());
  (*association_role_servant).set_isNavigable(uml_association_role.isNavigable());
  (*association_role_servant).set_parent(parent_association);
  (*association_role_servant).set_minCard((long)uml_association_role.min());
  (*association_role_servant).set_maxCard((long)uml_association_role.max());
  (*association_role_servant).set_isPrimary(uml_association_role.isPrimary());

  // Create class-taggedvalue one directional reference
  Uml::Class uml_class = uml_association_role.target();
  long class_id = uml_class.uniqueId();
  if (class_servant_map.count(class_id) == 1) {
    OTIF_Class_i *class_servant = class_servant_map[class_id];
    OTIF::Class_var class_obj_var;

    try {
      PortableServer::ObjectId_var oid = poa->servant_to_id(class_servant);
      CORBA::Object_var class_obj = poa->id_to_reference(oid.in());
      class_obj_var = OTIF::Class::_narrow(class_obj.in());
    } catch (const CORBA::Exception &) {
	  logger.log(1, "Error: Caught a CORBA exception");
    }

    (*association_role_servant).set_target(class_obj_var);
  }

  long ar_id = uml_association_role.uniqueId();
  association_role_map[ar_id] = association_role_obj_var;

  logger.log(5, "    -AssociationRole name: ", uml_association_role.name());

  return association_role_obj_var;
}

OTIF::TaggedValue_var UML_tagged_value_pass(const Uml::TaggedValue &uml_tagged_value, OTIF::Association_var &parent_association, long MetaModel_ID) {
  OTIF_TaggedValue_i *tagged_value_servant = new OTIF_TaggedValue_i();
  OTIF::TaggedValue_var tagged_value_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( tagged_value_servant );
    CORBA::Object_var tagged_value_obj = poa->id_to_reference( oid.in() );
    tagged_value_obj_var = OTIF::TaggedValue::_narrow(tagged_value_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) tagged_value_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  (*tagged_value_servant).set_id(Servant_ID++);
  (*tagged_value_servant).set_name(((string)uml_tagged_value.name()).c_str());
  (*tagged_value_servant).set_value(((string)uml_tagged_value.value()).c_str());
  (*tagged_value_servant).set_assocParent(parent_association);

  // Create class-taggedvalue one directional reference
  Uml::Class uml_class = uml_tagged_value.classParent();
  long class_id = uml_class.uniqueId();
  if (class_servant_map.count(class_id) == 1) {
    OTIF_Class_i *class_servant = class_servant_map[class_id];
    OTIF::Class_var class_obj_var;

    try {
      PortableServer::ObjectId_var oid = poa->servant_to_id(class_servant);
      CORBA::Object_var class_obj = poa->id_to_reference(oid.in());
      class_obj_var = OTIF::Class::_narrow(class_obj.in());
    } catch (const CORBA::Exception &) {
      logger.log(1, "Error: Caught a CORBA exception");
    }

    (*tagged_value_servant).set_classParent(class_obj_var);
  }

  long tv_id = uml_tagged_value.uniqueId();
  tagged_value_map[tv_id] = tagged_value_obj_var;

  logger.log(5, "    -TaggedValue name: ", uml_tagged_value.name());

  return tagged_value_obj_var;
}

OTIF::Association_var UML_association_pass(const Uml::Association &uml_association, OTIF::Diagram_var &parent_diagram, long MetaModel_ID) {
  OTIF_Association_i *association_servant = new OTIF_Association_i();
  OTIF::Association_var association_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( association_servant );
    CORBA::Object_var association_obj = poa->id_to_reference( oid.in() );
    association_obj_var = OTIF::Association::_narrow(association_obj.in());

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) association_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  (*association_servant).set_id(Servant_ID++);
  (*association_servant).set_name(((string)uml_association.name()).c_str());
  (*association_servant).set_parent(parent_diagram);
  (*association_servant).set_nonpersistent(uml_association.nonpersistent());

  // Create class-association bidirectional reference
  Uml::Class uml_class = uml_association.assocClass();
  long class_id = uml_class.uniqueId();
  if (class_servant_map.count(class_id) == 1) {
    OTIF_Class_i *class_servant = class_servant_map[class_id];
    OTIF::Class_var class_obj_var;

    try {
      PortableServer::ObjectId_var oid = poa->servant_to_id(class_servant);
      CORBA::Object_var class_obj = poa->id_to_reference(oid.in());
      class_obj_var = OTIF::Class::_narrow(class_obj.in());
    } catch (const CORBA::Exception &) {
      logger.log(1, "Error: Caught a CORBA exception");
    }

    (*class_servant).set_assoc(association_obj_var);
    (*association_servant).set_assocClass(class_obj_var);
  }
  
  // Get the asociation roles from the diagram
  set<Uml::AssociationRole> uml_association_role = uml_association.roles();
  OTIF::AssociationRole_S *association_roles = new OTIF::AssociationRole_S(uml_association_role.size());
  association_roles->length(uml_association_role.size());

  CORBA::ULong offset = 0;
  for(set<Uml::AssociationRole>::iterator i = uml_association_role.begin(); i!= uml_association_role.end(); i++, offset++) {
    OTIF::AssociationRole_var association_role_obj_var = UML_association_role_pass(*i, association_obj_var, MetaModel_ID);
    (*association_roles)[offset] = association_role_obj_var;
  }
  (*association_servant).set_roles(association_roles);
//  delete association_roles;

  // Get the tagged values from the diagram
  set<Uml::TaggedValue> uml_tagged_value = uml_association.tagValues();
  OTIF::TaggedValue_S *tagged_values = new OTIF::TaggedValue_S(uml_tagged_value.size());
  tagged_values->length(uml_tagged_value.size());

  offset = 0;
  for(set<Uml::TaggedValue>::iterator j = uml_tagged_value.begin(); j!= uml_tagged_value.end(); j++, offset++) {
    OTIF::TaggedValue_var tagged_value_obj_var = UML_tagged_value_pass(*j, association_obj_var, MetaModel_ID);
    (*tagged_values)[offset] = tagged_value_obj_var;
  }
  (*association_servant).set_tagValues(tagged_values);
//  delete tagged_values;

  logger.log(5, "  -Association name: ", uml_association.name());

  return association_obj_var;
}

void UML_diagram_pass(Uml::Diagram &uml_diagram, long MetaModel_ID) {
  OTIF_Diagram_i *diagram_servant = new OTIF_Diagram_i();
  OTIF::Diagram_var diagram_obj_var;

  try {
    PortableServer::ObjectId_var oid = poa->activate_object( diagram_servant );
    CORBA::Object_var diagram_obj = poa->id_to_reference( oid.in() );
    diagram_obj_var = OTIF::Diagram::_narrow(diagram_obj.in());

    metamodel_diagram_repository[MetaModel_ID] = diagram_obj_var;

    servants_id.insert(pair<long const, PortableServer::ObjectId_var>(MetaModel_ID, oid));
    servants_ref.insert(pair<long const, PortableServer::Servant>(MetaModel_ID, (PortableServer::Servant) diagram_servant));

  } catch (const CORBA::Exception &) {
	logger.log(1, "Error: Caught a CORBA exception");
  }

  (*diagram_servant).set_id(Servant_ID++);
  (*diagram_servant).set_name(((string)uml_diagram.name()).c_str());
  (*diagram_servant).set_version(((string)uml_diagram.version()).c_str());

  logger.log(5, "-Diagram name: ", uml_diagram.name());
  logger.log(5, "-Diagram version: ", uml_diagram.version());

  // Get the classes from the diagram
  set<Uml::Class> uml_classes = uml_diagram.classes();
  OTIF::Class_S  *classes = new OTIF::Class_S(uml_classes.size());
  classes->length(uml_classes.size());

  CORBA::ULong offset = 0;
  for(set<Uml::Class>::iterator i = uml_classes.begin(); i!= uml_classes.end(); i++, offset++) {
    OTIF::Class_var class_obj_var = UML_class_pass(*i, diagram_obj_var, MetaModel_ID);
    (*classes)[offset] = class_obj_var;
  }
  (*diagram_servant).set_classes(classes);
  //  delete classes;

  // Get the compositions from the diagram
  set<Uml::Composition> uml_compositions = uml_diagram.compositions();
  OTIF::Composition_S  *compositions = new OTIF::Composition_S(uml_compositions.size());
  compositions->length(uml_compositions.size());

  offset = 0;
  for(set<Uml::Composition>::iterator j = uml_compositions.begin(); j!= uml_compositions.end(); j++, offset++) {
    OTIF::Composition_var composition_obj_var = UML_composition_pass(*j, diagram_obj_var, MetaModel_ID);
    (*compositions)[offset] = composition_obj_var;
  }
  (*diagram_servant).set_compositions(compositions);
  //  delete compositions;

  // Get the associations from the diagram
  set<Uml::Association> uml_associations = uml_diagram.associations();
  OTIF::Association_S  *associations = new OTIF::Association_S(uml_associations.size());
  associations->length(uml_associations.size());

  offset = 0;
  for(set<Uml::Association>::iterator k = uml_associations.begin(); k!= uml_associations.end(); k++, offset++) {
    OTIF::Association_var association_obj_var = UML_association_pass(*k, diagram_obj_var, MetaModel_ID);
    (*associations)[offset] = association_obj_var;
  }
  (*diagram_servant).set_associations(associations);
  //  delete associations;

}

void UML_pass2(Uml::Diagram &uml_diagram) {
  logger.log(5, "UML_pass2 activated");

  set<Uml::Class> uml_main_classes = uml_diagram.classes();
  for(set<Uml::Class>::iterator i=uml_main_classes.begin(); i!=uml_main_classes.end(); i++) {

    long class_id = (*i).uniqueId();
    if (class_servant_map.count(class_id) == 1) {
      OTIF_Class_i *class_servant = class_servant_map[class_id];

      // Set TaggedValue_S for Class
      set<Uml::TaggedValue> uml_tagged_values = (*i).tagValues();
      OTIF::TaggedValue_S *tagged_values = new OTIF::TaggedValue_S(uml_tagged_values.size());
      tagged_values->length(uml_tagged_values.size());
      
      CORBA::ULong offset = 0;
      for(set<Uml::TaggedValue>::iterator j=uml_tagged_values.begin(); j!=uml_tagged_values.end(); j++, offset++) {
		long tv_id = (*j).uniqueId();
		(*tagged_values)[offset] = tagged_value_map[tv_id];
      }
      (*class_servant).set_tagValues(tagged_values);

      // Set AssociationRole_S for Class
      set<Uml::AssociationRole> uml_association_roles = (*i).associationRoles();
      OTIF::AssociationRole_S *association_roles = new OTIF::AssociationRole_S(uml_association_roles.size());
      association_roles->length(uml_association_roles.size());
      
      offset = 0;
      for(set<Uml::AssociationRole>::iterator k=uml_association_roles.begin(); k!=uml_association_roles.end(); k++, offset++) {
		long ar_id = (*k).uniqueId();
		(*association_roles)[offset] = association_role_map[ar_id];
      }
      (*class_servant).set_associationRoles(association_roles);

      // Set baseTypes for Class
      set<Uml::Class> uml_classes = (*i).baseTypes();
      OTIF::Class_S *base_types = new OTIF::Class_S(uml_classes.size());
      base_types->length(uml_classes.size());
      
      offset = 0;
      for(set<Uml::Class>::iterator l=uml_classes.begin(); l!=uml_classes.end(); l++, offset++) {
		long bt_id = (*l).uniqueId();
		(*base_types)[offset] = class_map[bt_id];
      }
      (*class_servant).set_baseTypes(base_types);

      // Set subTypes for Class
      uml_classes = (*i).subTypes();
      OTIF::Class_S *sub_types = new OTIF::Class_S(uml_classes.size());
      sub_types->length(uml_classes.size());
      
      offset = 0;
      for(set<Uml::Class>::iterator m=uml_classes.begin(); m!=uml_classes.end(); m++, offset++) {
		long st_id = (*m).uniqueId();
		(*sub_types)[offset] = class_map[st_id];
      }
      (*class_servant).set_subTypes(sub_types);
    }
  }
}


// *******************************************************************************************************************************
// Startup function
// *******************************************************************************************************************************
struct PARADIGM {
	string name;
	string version;
	string filename;
};

bool __registerUMLParadigm(string filename) {
	bool uml_registered = false;

	OTIF::NameVersion_var name_version = new OTIF::NameVersion();
	name_version->name = CORBA::string_dup("UML");
	name_version->version = CORBA::string_dup("1.0");
	OTIF::Diagram_var diagram;
	int result = _registerMetaModelFile(name_version, filename.c_str(), OTIF::ManagerSession::RegistrationAction::add, diagram, 0);
	if (result!=0) {
		logger.log(1, "Error: could not register the UML v1.0 paradigm!");
		cerr << "Error: could not register the UML v1.0 paradigm!" << endl;

	} else {
		uml_registered = true;
		if (_lockParadigm(name_version) != 0) {
			logger.log(1, "Error: Could not lock the paradigm!");
			cerr << "Error: Could not lock the paradigm!" << endl;
		}
	}

	return uml_registered;
}

void backplane_load_config(DOM_Element root_element) {
	
	map<long, PARADIGM> paradigms_rep;

	// ------------------------------------------------------------------------------------------
	// get the backplane name parameter
	// ------------------------------------------------------------------------------------------
	DOM_Attr as = root_element.getAttributeNode("name");
	if (as != NULL && as.getSpecified()) {
		backplane_name = as.getValue().transcode();
//		cout << "Backplane name: " << backplane_name << endl;
	}
	// ------------------------------------------------------------------------------------------
	// get the backplane version parameter
	// ------------------------------------------------------------------------------------------
	as = root_element.getAttributeNode("version");
	if (as != NULL && as.getSpecified()) {
		backplane_version = as.getValue().transcode();
//		cout << "Backplane version: " << backplane_version << endl;
	}

	// ------------------------------------------------------------------------------------------
	// get the options, startup parameters
	// ------------------------------------------------------------------------------------------
	DOM_NodeList nodes = root_element.getChildNodes();
	DOM_Node startup_node;
	bool uml_registered = false;

	for(DOM_Node node = root_element.getFirstChild(); node != NULL; node = node.getNextSibling()) {

		if( node.getNodeType() == DOM_Node::ELEMENT_NODE ) {

			DOM_Element e = static_cast<DOM_Element&>(node);
			string node_name(node.getNodeName().transcode());

			if (!node_name.compare("options")) {
				// ------------------------------------------------------------------------------------------
				// get the manager key parameter
				// ------------------------------------------------------------------------------------------
				DOM_Attr as = e.getAttributeNode("password");
				if (as != NULL && as.getSpecified()) {
					manager_key = as.getValue().transcode();
//					cout << "Manager key: " << manager_key << endl;
				}
				// ------------------------------------------------------------------------------------------
				// get the onlyOneManager parameter
				// ------------------------------------------------------------------------------------------
				as = e.getAttributeNode("onlyOneManager");
				if (as != NULL && as.getSpecified()) {
					string manager = as.getValue().transcode();
					OnlyOneManager = (!manager.compare("true")) ? true : false;
//					cout << "Only one manager: " << OnlyOneManager << endl;
				}
				// ------------------------------------------------------------------------------------------
				// get the keepalive parameter
				// ------------------------------------------------------------------------------------------
				as = e.getAttributeNode("translator_documents");
				if (as != NULL && as.getSpecified()) {
					string keepalive = as.getValue().transcode();
					translator_keepalive_parameter = atol(keepalive.c_str());
//					cout << "Translator published documents keepalive: " << translator_keepalive_parameter << "sec" << endl;
				}
				// ------------------------------------------------------------------------------------------
				// get the log_level parameter
				// ------------------------------------------------------------------------------------------
				as = e.getAttributeNode("log_level");
				if (as != NULL && as.getSpecified()) {
					log_level = as.getValue().transcode();
					long log_ll = atol(log_level.c_str());
					logger.setLevel(log_ll);
				}
				// ------------------------------------------------------------------------------------------
				// register the uml_paradigm
				// ------------------------------------------------------------------------------------------
				as = e.getAttributeNode("uml_paradigm");
//				if (as != NULL && as.getSpecified() && !isUMLregistered) {
				if (as != NULL && as.getSpecified()) {
					string filename = as.getValue().transcode();
					isUMLregistered = __registerUMLParadigm(filename);
				}

				meta_meta_uml_type_tooladaptor="META_META_UML_PUBLISHER_RECEIVER";

			} else if (!node_name.compare("startup")) {
				startup_node = node;
			}
		}
	}

	if (!isUMLregistered) {
		cout << "Error: UML paradigm could not be registered. Backplane exits." << endl;
		exit(0);
	}


	if (!startup_node.isNull()) {
		DOM_Node paradigms_node;
		DOM_Node translators_node;
		DOM_Node tooladaptors_node;
		DOM_Node workflow_node;
		long paradigm_max_id = 0;
		
		for(DOM_Node node = startup_node.getFirstChild(); node != NULL; node = node.getNextSibling()) {
			if( node.getNodeType() == DOM_Node::ELEMENT_NODE ) {
				string node_name(node.getNodeName().transcode());
				if (!node_name.compare("paradigms")) {
					paradigms_node = node;
				} else if (!node_name.compare("translators")) {
					translators_node = node;
				} else if (!node_name.compare("tooladaptors")) {
					tooladaptors_node = node;
				} else if (!node_name.compare("workflow")) {
					workflow_node = node;
				}
			}
		}			

		// ------------------------------------------------------------------------------------------
		// register the paradigms
		// ------------------------------------------------------------------------------------------
		if (!paradigms_node.isNull()) {
			for(DOM_Node node = paradigms_node.getFirstChild(); node != NULL; node = node.getNextSibling()) {

				if( node.getNodeType() == DOM_Node::ELEMENT_NODE ) {
					DOM_Element e = static_cast<DOM_Element&>(node);
					string node_name(node.getNodeName().transcode());
		
					if (node_name.compare("paradigm")) continue;

					long id = atol(e.getAttribute("id").transcode());
					if (paradigm_max_id<id) paradigm_max_id = id;
					PARADIGM par;
					par.name = e.getAttribute("name").transcode();
					par.version = e.getAttribute("version").transcode();
					par.filename = e.getAttribute("filename").transcode();
					paradigms_rep[id] = par;					

					OTIF::NameVersion_var name_version = new OTIF::NameVersion();
					name_version->name = CORBA::string_dup(par.name.c_str());
					name_version->version = CORBA::string_dup(par.version.c_str());
					OTIF::Diagram_var diagram;
					int result = _registerMetaModelFile(name_version, par.filename.c_str(), OTIF::ManagerSession::RegistrationAction::add, diagram, id);
					if (result!=0) {
						logger.log(1, "Error: could not register the paradigm:", par.name + " v" + par.version);
						cerr << "Error: could not register the " << par.name << " v" << par.version << " paradigm!" << endl;
					}

					if ( !((string)e.getAttribute("locked").transcode()).compare("true") ) {
						if (_lockParadigm(name_version) != 0) {
							logger.log(1, "Error: Could not lock the paradigm!");
							cerr << "Error: Could not lock the paradigm!" << endl;
						}
					}
					
				}
			}
		}
		MetaModel_ID = paradigm_max_id;

		// ------------------------------------------------------------------------------------------
		// register the translators
		// ------------------------------------------------------------------------------------------
		if (!translators_node.isNull()) {
			for(DOM_Node node = translators_node.getFirstChild(); node != NULL; node = node.getNextSibling()) {

				if( node.getNodeType() == DOM_Node::ELEMENT_NODE ) {
					DOM_Element e = static_cast<DOM_Element&>(node);
					string node_name(node.getNodeName().transcode());

					if (node_name.compare("translator")) continue;

					long id = atol(e.getAttribute("id").transcode());
					string name = e.getAttribute("name").transcode();
					string version = e.getAttribute("version").transcode();
					long input = atol(e.getAttribute("inputParadigms").transcode());
					long output = atol(e.getAttribute("outputParadigms").transcode());
					string command = e.getAttribute("command").transcode();
					
					map<long, PARADIGM>::iterator i=paradigms_rep.find(input);
					if (i==paradigms_rep.end()) continue;
					i=paradigms_rep.find(output);
					if (i==paradigms_rep.end()) continue;

					OTIF::NameVersion_var name_version = new OTIF::NameVersion();
					name_version->name = CORBA::string_dup(name.c_str());
					name_version->version = CORBA::string_dup(version.c_str());
					OTIF::TranslatorVersion_var tran = new OTIF::TranslatorVersion();
					tran->name = name_version;
					tran->from.name = CORBA::string_dup(paradigms_rep[input].name.c_str());
					tran->from.version = CORBA::string_dup(paradigms_rep[input].version.c_str());
					tran->to.name = CORBA::string_dup(paradigms_rep[output].name.c_str());
					tran->to.version = CORBA::string_dup(paradigms_rep[output].version.c_str());
					int result = _registerTranslator(tran, command.c_str(), OTIF::ManagerSession::RegistrationAction::add, id);
					if (result!=0) {
						logger.log(1, "Error: could not register the translator:", name + " v" + version);
						cerr << "Error: could not register the " << name << " v" << version << " translator!" << endl;
					}

//					cout << "Translator: " << id << ":" << name << ":" << version << ":" << input << ":" << output << ":" << command << endl;
				}
			}
		}

		// ------------------------------------------------------------------------------------------
		// register the tooladaptors
		// ------------------------------------------------------------------------------------------
		TOOLADAPTOR ta;
		ta.type = meta_meta_uml_type_tooladaptor;
		ta.inputParadigms.push_back(0);
		PARADIGMS par;
		par.paradigms.push_back(0);
		ta.outputParadigms.push_back(par);
		tooladaptor_rep.insert(pair<const long, TOOLADAPTOR>(0, ta));
//		cout << "Tooladaptor: 0" << ":" << ta.type << ":" << ta.inputParadigms.size() << ":" << ta.outputParadigms.size() << endl;
		
		if (!tooladaptors_node.isNull()) {
			for(DOM_Node node = tooladaptors_node.getFirstChild(); node != NULL; node = node.getNextSibling()) {

				if( node.getNodeType() == DOM_Node::ELEMENT_NODE ) {
					DOM_Element e = static_cast<DOM_Element&>(node);
					string node_name(node.getNodeName().transcode());

					if (node_name.compare("tooladaptor")) continue;

					long id = atol(e.getAttribute("id").transcode());
					string temp = "";
					TOOLADAPTOR ta;
					ta.type = e.getAttribute("type").transcode();

					string input_str = e.getAttribute("inputParadigms").transcode();
					if (input_str.empty()) ta.inputParadigms.clear();
					else {
						for(unsigned int i=0; i<=input_str.size(); i++) {
							if (input_str.c_str()[i]!=',' && input_str.c_str()[i]!='\0') temp += input_str.c_str()[i];
							else {
								ta.inputParadigms.push_back(atol(temp.c_str()));
								temp = "";
							}
						}
					}

					list<long> temp_pars;
					string output_str = e.getAttribute("outputParadigms").transcode();
					if (output_str.empty()) ta.outputParadigms.clear();
					else {
						for(unsigned int i=0; i<=output_str.size(); i++) {
							if (output_str.c_str()[i]!=',' && output_str.c_str()[i]!='\0' 
								&& output_str.c_str()[i]!='|') {
								temp += output_str.c_str()[i];
							} else if (output_str.c_str()[i]=='|') {
								temp_pars.push_back(atol(temp.c_str()));
								temp = "";
							} else if (output_str.c_str()[i]==',' || output_str.c_str()[i]=='\0') {
								PARADIGMS par;
								temp_pars.push_back(atol(temp.c_str()));
								par.paradigms = temp_pars;
								ta.outputParadigms.push_back(par);
								temp_pars.clear();
								temp = "";
							}
						}
					}

					tooladaptor_rep.insert(pair<const long, TOOLADAPTOR>(id, ta));
//					cout << "Tooladaptor: " << id << ":" << ta.type << ":" << ta.inputParadigms.size() << ":" << ta.outputParadigms.size() << endl;
				}
			}
		}

		// ------------------------------------------------------------------------------------------
		// register the workflow
		// ------------------------------------------------------------------------------------------
		if (!workflow_node.isNull()) {
			for(DOM_Node node = workflow_node.getFirstChild(); node != NULL; node = node.getNextSibling()) {

				if( node.getNodeType() == DOM_Node::ELEMENT_NODE ) {
					DOM_Element e = static_cast<DOM_Element&>(node);
					string node_name(node.getNodeName().transcode());

					if (node_name.compare("link")) continue;

					long source = atol(e.getAttribute("source").transcode());
					long destination = atol(e.getAttribute("destination").transcode());
					long outputParadigm = atol(e.getAttribute("outputParadigm").transcode());

					WORKFLOW wf;
					wf.destination = destination;
					wf.outputParadigm = outputParadigm;

					workflow_rep.insert(pair<const long, WORKFLOW>(source, wf));
					
//					cout << "Workflow: " << source << ":" << destination << ":" << outputParadigm << endl;
				}
			}
		}


	}
}

//===================================
void network_init(string &IP_ADDRESS) 
{

    WSAData wsaData;
    if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0)
    {
      throw udm_exception("Could not get IP address!");
    }

    char ac[80];
    if (gethostname(ac, sizeof(ac)) == SOCKET_ERROR)
    {
      throw udm_exception("Could not get IP address!");   
    }
    cout << "Host name is " << ac << "." << endl;

    struct hostent *phe = gethostbyname(ac);
    if (phe == 0)
    {
        throw udm_exception("Could not get IP address!");   
    }

    for (int i = 0; phe->h_addr_list[i] != 0; ++i)
    {
        struct in_addr addr;
        memcpy(&addr, phe->h_addr_list[i], sizeof(struct in_addr));
        string add(inet_ntoa(addr));
        if (add.find(".0.") == std::string::npos)
          IP_ADDRESS = add;

    }


    WSACleanup();

}

//=====================================================
/*
void network_init(string &IP_ADDRESS) {
	WORD wVersionRequested;
	WSADATA wsaData;
	char name[255];
	string ip;
	PHOSTENT hostinfo;
	wVersionRequested = MAKEWORD( 2, 0 );

	if ( WSAStartup( wVersionRequested, &wsaData ) == 0 ) {
		if( gethostname ( name, sizeof(name)) == 0) {
			if((hostinfo = gethostbyname(name)) != NULL) 
      {
				IP_ADDRESS = inet_ntoa (*(struct in_addr *)*hostinfo->h_addr_list);
			}
		}
		WSACleanup();
	}
}
*/

int ip_port_init(string &config_ip_address) {

	int port = 0;
	string temp_port = "";

	// Get the URL config
	ifstream url_config("backplane_server_url.config");
	if (url_config.fail()) {
		port = BACKPLANE_DEFAULT_PORT;
//		cout << "Note: it happens when you first run the framework or delete the backplane_server_url.config file."<< endl;
	} else {
		int n = 0;
		bool isPort = false;
		for (;; ++n) {
			char c = url_config.get();
			if (url_config.eof()) break;
			if (c=='\n') break;
			if (c==':') {
				isPort = true;
				continue;
			}
			if (isPort) temp_port += c;
			if (!isPort) config_ip_address += c;
		}

		port = atoi(temp_port.c_str());		
	}
	url_config.close();

	return port;
}

void update_config_file(string &ip_address, int port)
{
  cout << "backplane_server_url.config has been updated" << endl;
	ofstream url_config("backplane_server_url.config", ios::trunc);
//	char t_port[10];
//	itoa(port, t_port, 10);
	url_config << ip_address << ":" << port;
	url_config.close();
}


bool try_to_open_port(int argc, char **argv, string &ip_address, int port) {

	cout << "Backplane tries to use the port:" << port << endl;

	// Set arguments
	int _argc = argc+1;
	char **_argv = new char*[_argc];

	for(int i=0; i<argc; i++) {
		string _temp = argv[i];
		_argv[i] = new char[_temp.length()+1];
		strcpy(_argv[i], _temp.c_str());
	}

	char t_port[10];
	itoa(port, t_port, 10);
	string _temp = "-ORBEndPoint iiop://" + ip_address + ":" + t_port;
	_argv[argc] = new char[_temp.length()+1];
	strcpy(_argv[argc], _temp.c_str());

	// CORBA initialization
	try {
		// Obtain ORB
		if (!CORBA::is_nil(orb)) orb->destroy();
		orb = CORBA::ORB_init(_argc, _argv);

		// Obtain POA and POA manager
		CORBA::Object_var obj = orb->resolve_initial_references("RootPOA");
		poa = PortableServer::POA::_narrow(obj.in());
		if (CORBA::is_nil(poa.in())) {
			cerr << "Failed to narrow root POA." << endl;
			exit(1);
		}
		poa_mgr = poa->the_POAManager();
		if (CORBA::is_nil(poa_mgr.in())) {
			cerr << "Failed to narrow root POA manager." << endl;
			exit(1);
		}
		logger.log(0, "ORB, POA and POA manager initialized");

	} catch (const CORBA::Exception &) {
		cout << "Port could not be opened." << endl;
		return false;
	}
	
	// Register backplane
	try {
		// Obtain IORTable reference (Note: this is not CORBA compliant)
		CORBA::Object_var obj = orb->resolve_initial_references("IORTable");
		iortable_adapter = IORTable::Table::_narrow(obj.in());
		if (CORBA::is_nil(iortable_adapter.in())) {
			cerr << "Failed to narrow IORTable." << endl;
			exit(1);
		}
		logger.log(0, "IORTable initialized");
		// ------------------------------------------------------------------------------------------

		// We should register the backplane object in the NameService
		backplane_servant =  new OTIF_Backplane_i();
		poa->activate_object(backplane_servant);
//		OTIF::Backplane_var _backplane_obj = backplane_servant->_this();

		CORBA::String_var ior = orb->object_to_string(backplane_servant->_this());
		iortable_adapter->bind("OTIF/BACKPLANE", ior.in());


		poa_mgr->activate();

	} catch (const CORBA::Exception &ex) {
		logger.log(1, "Error: Caught a CORBA exception");
    cerr << "Caught a CORBA exception: " << ex << endl;
		exit(-1);
	}

	return true;
}
	
// *******************************************************************************************************************************
// MAIN
// *******************************************************************************************************************************
int main(int argc, char** argv) {
	logger.log(0, "BackPlane started");

	// The translators should be started with the same arguments as the backplane server!
	__translator_arguments_length = argc-1;
	__translator_arguments = new string[argc-1];
	for(int i=1; i<argc; i++) {
		__translator_arguments[i-1] = argv[i];
	}

	// get IP address
	string ip_address;
	network_init(ip_address);
	string config_ip_address;
  ip_port_init(config_ip_address);

	int port = BACKPLANE_DEFAULT_PORT;
  // Set the URL config
	if (config_ip_address.compare(ip_address)) 
    update_config_file(ip_address, port);

	while (!try_to_open_port(argc, argv, ip_address, port)) {
		port++;
		if (port>65000) {
			cout << "Error: Could not open port for Backplane server." << endl;
			exit(-1);
		}
	}

	
	cout << "Listening on: " << ip_address << ":" << port << endl;

	// ------------------------------------------------------------------------------------------
	// startup
	// ------------------------------------------------------------------------------------------
	startup_params = new STARTUP_PARAMETERS();
	
	DOMParser parser;
	parser.setDoNamespaces(false);
	parser.setIncludeIgnorableWhitespace(false);
	try	{
		parser.parse("./BackPlane_config.xml");
		startup_params->root_element = parser.getDocument().getDocumentElement();
		if (startup_params->root_element!=NULL) backplane_load_config(startup_params->root_element);
		else {
			cout << "Warning: There is no Backplane config file. Trying to use the default values." << endl;

			isUMLregistered = __registerUMLParadigm("uml.xml");
			if (!isUMLregistered) {
				cout << "Error: UML paradigm could not be registered. Backplane exits." << endl;
				exit(0);
			}
		}

	} catch (const XMLException &e) {
		cerr << "Error: Error during parsing the config file! XMLException: " << e.getMessage() << endl;;
		exit(-1);
	} catch (DOM_DOMException e) {
		cerr << "Error: Error during parsing the config file! DOMException: " << e.msg.transcode() << endl;
		exit(-1);
	} catch (...) {
		cerr << "Error: Error during parsing the config file!" << endl;
		exit(-1);
	}

	cout << "Configuration is loaded." << endl;
	
	// ------------------------------------------------------------------------------------------
	// ORB running
	// ------------------------------------------------------------------------------------------
	try {
		
		cout << "Waiting for requests..." << endl;
		orb->run();

	// ------------------------------------------------------------------------------------------
	// ORB stopped, quit
	// ------------------------------------------------------------------------------------------
		poa->destroy(1, 1);
		
		orb->destroy();

		cout << "Backplane exits." << endl;
	
	} catch (const CORBA::Exception &) {
		logger.log(1, "Error: Caught a CORBA exception");
	}

	delete startup_params;
	logger.log(0, "ORB destroyed");

	return 0;
}
