/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

// *******************************************************************************************************************************
// CORBA specific headers
// *******************************************************************************************************************************
//#include <ace/streams.h>
#include "tao/IORTable/IORTable.h"
#include "otifdata_impl.h"
#include "otif_impl.h"


// *******************************************************************************************************************************
// C specific headers
// *******************************************************************************************************************************
#include <time.h>

// *******************************************************************************************************************************
// UDM specific headers
// *******************************************************************************************************************************
#include "UdmBase.h"
#include "UdmDom.h"
#include "UmlExt.h"
#include "UdmCORBA_Logger.h"

// *******************************************************************************************************************************
// UML meta-metamodel specific header
// *******************************************************************************************************************************
//#include "uml.h"

// *******************************************************************************************************************************
// Functions used by BackPlane servant
// *******************************************************************************************************************************
int _getBackplaneName(OTIF::NameVersion_var &name);
int _lockParadigm(const OTIF::NameVersion &paradigm);
int _fetchParadigm(const OTIF::NameVersion &paradigm, OTIF::Diagram_var &diagram);
int _unlockParadigm(const OTIF::NameVersion &paradigm);
int _getParadigms(OTIF::NameVersion_S_var &names);
int _getTranslators(OTIF::TranslatorVersion_S_var &translators);
int _getDocuments(OTIF::DocumentInfo_S_var &datanetworks);
int _getToolAdaptorTypes(OTIF::Name_S_var &types);

int _getToolAdaptorInfoByType(const char *taType, OTIF::ToolAdaptorInfo_var &ta_info);
int _getToolAdaptorInfo(CORBA::Long session_id, OTIF::ToolAdaptorInfo_var &ta_info);
int _getToolAdaptorType(const OTIF::NameVersion &paradigm, OTIF::taMode mode, string &type);

int _loginManager(const char *key, OTIF::ManagerSession_var &manager, CORBA::Long_out session_id);
int _logoutManager(CORBA::Long session_id);

int _loginToolAdaptor(const char * ToolAdaptorType, OTIF::ToolAdaptorSession_var &tooladaptor, CORBA::Long_out session_id);
//int _getParadigmsOfToolAdaptorType(CORBA::Long session_id, OTIF::NameVersion_S_var &paradigms);
int _logoutToolAdaptor(CORBA::Long session_id);

void _now(OTIF::timeStamp_out stamp);
int _uploadStartup(const char *xml);


// *******************************************************************************************************************************
// Functions used by ManagerSession servant
// *******************************************************************************************************************************
int _registerMetaModel(const OTIF::NameVersion &mm_info, const OTIF::DataNetwork &metaData, 
					   OTIF::ManagerSession::RegistrationAction action, OTIF::Diagram_var &diagram);
int _registerMetaModelFile(const OTIF::NameVersion &mm_info, const char *filename, 
						   OTIF::ManagerSession::RegistrationAction action, OTIF::Diagram_var &diagram, long ID=-1);
int _unregisterMetaModel(const OTIF::NameVersion &mm_info);

int _registerTranslator(const OTIF::TranslatorVersion &translator, const char* activation, 
						OTIF::ManagerSession::RegistrationAction action, long ID=-1);
int _unregisterTranslator(const OTIF::TranslatorVersion &translator);

int _change_password (const char *old_password, const char *new_password);
void _shutdown();

int _destroyDocument(OTIF::DocumentID docId);
int _modifyDocumentName(OTIF::DocumentID docId, const char *name);
int _modifyDocumentVersion(OTIF::DocumentID docId, const char *version);
int _modifyDocumentRemarks(OTIF::DocumentID docId, const char *remarks);
int _modifyDocumentKeepAlive(OTIF::DocumentID docId, CORBA::Long keepAlive);

// *******************************************************************************************************************************
// Functions used by ToolAdapterSession servant
// *******************************************************************************************************************************
int _publishDocument(const long session_id, const OTIF::NameVersion &document, const char *remarks, CORBA::Long keepAlive, 
					 const OTIF::Document &docData, OTIF::DocumentID previous_docId, OTIF::DocumentID &document_id);
int _subscribe(const long session_id, OTIF::ToolAdaptorNotifyCallback_ptr callback);
int _fetchDocument(const OTIF::DocumentID docId, OTIF::Document_out docData);

int _getToolAdaptorDocuments(const long session_id, OTIF::DocumentInfo_S_var &datanetworks);

int _getDocumentRootObjectMeta(const OTIF::DocumentID docId, CORBA::String_out rootobjectmeta);
int _getDocumentParadigm(const OTIF::DocumentID docId, OTIF::NameVersion *paradigm);

// *******************************************************************************************************************************
// Functions used by TranslatorFeed & TranslatorSink servants
// *******************************************************************************************************************************
int _tf_subscribe(long session_id, OTIF::TranslatorNotifyCallback_ptr callback);
int _ts_publishDocument(long session_id, const OTIF::NameVersion & document, const char * remarks, CORBA::Long keepAlive, 
						const OTIF::Document & docData, OTIF::DocumentID source_docId);
