package org.eclipse.otif.generictooladaptor;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.ole.win32.OLE;
import org.eclipse.swt.ole.win32.OleAutomation;
import org.eclipse.swt.ole.win32.OleControlSite;
import org.eclipse.swt.ole.win32.OleEvent;
import org.eclipse.swt.ole.win32.OleFrame;
import org.eclipse.swt.ole.win32.OleListener;
import org.eclipse.swt.ole.win32.Variant;
import org.eclipse.swt.widgets.TableItem;

/**
 * @author lang
 *
 * To change this generated comment edit the template variable "typecomment":
 * Window>Preferences>Java>Templates.
 * To enable and disable the creation of type comments go to
 * Window>Preferences>Java>Code Generation.
 */
public class Backplane {

	private OleAutomation activex;
	private Layout layout;

	public Backplane(OleFrame frame, Layout layout) {
		this.layout = layout;

		// Get the Generic ToolAdaptor ActiveX component into activex variable
		try {
			OleControlSite controlSite = new OleControlSite(frame, SWT.NONE, "GENERICTOOLADAPTORACTIVEX.GenericToolAdaptorActiveXCtrl.1");
			controlSite.doVerb(OLE.OLEIVERB_UIACTIVATE);
			activex = new OleAutomation(controlSite);		

			// Get document arrived notifications
			controlSite.addEventListener(activex, 1, new OleListener() {
				public void handleEvent(OleEvent event) {
					Variant[] param = event.arguments;

 					TableItem tableItem = new TableItem(Layout.datanetworks_table, SWT.NONE);
					Document doc = new Document(param[0].getString(), param[1].getString(),
						param[2].getString(), param[3].getInt(), 
						param[4].getInt(), param[5].getInt(), 0);

					long diff = doc.keepAlive - System.currentTimeMillis()/1000;
					long sec = diff % 60;
					long d_min = diff / 60;
					long min = d_min % 60;
					long hour = d_min / 60;
					String strs[];
					if (hour>0) strs = new String[]{doc.name, doc.version, hour + "h " + min + "min " + sec + "sec", doc.remarks};
					else if (min>0) strs = new String[]{doc.name, doc.version, min + "min " + sec + "sec", doc.remarks};
					else strs = new String[]{doc.name, doc.version, sec + "sec", doc.remarks};
					tableItem.setText(strs);

					doc.key = tableItem.hashCode();
					Document.putDocument(doc);

					if (Layout.getNotificationCheck()) {
						MessageDialog.openInformation(Layout.getShell(), 
							"Document is available", 
							"Name: " + doc.name + "\n" +
							"Version: " + doc.version + "\n" +
							"Remarks: " + doc.remarks + "\n" +
							"keepAlive: " + doc.keepAlive + " sec\n");
					}
				}
			});

			// Get document changed notifications
			controlSite.addEventListener(activex, 2, new OleListener() {
				public void handleEvent(OleEvent event) {
					Variant[] param = event.arguments;

					// Document was destoryed (keepAlive is -1!)
					if (param[3].getInt() == -1) {
						TableItem[] items = Layout.datanetworks_table.getItems();
						for(int i=0; i<items.length; i++) {
							Document doc = Document.find_key(items[i].hashCode());
							if (param[4].getInt() == doc.docId) {
								Layout.datanetworks_table.remove(i);
								Document.removeDocument(doc);
							}
						}

					// Document was updated!
					} else {
						TableItem[] items = Layout.datanetworks_table.getItems();
						for(int i=0; i<items.length; i++) {
							Document doc = Document.find_key(items[i].hashCode());
							if (param[4].getInt() == doc.docId) {
								Layout.datanetworks_table.remove(i);
								Document.removeDocument(doc);

								TableItem tableItem = new TableItem(Layout.datanetworks_table, SWT.NONE);
								Document up_doc = new Document(param[0].getString(), param[1].getString(),
									param[2].getString(), param[3].getInt(), 
									param[4].getInt(), param[5].getInt(), 0);

								long diff = up_doc.keepAlive - System.currentTimeMillis()/1000;
								long sec = diff % 60;
								long d_min = diff / 60;
								long min = d_min % 60;
								long hour = d_min / 60;
								String strs[];
								if (hour>0) strs = new String[]{up_doc.name, up_doc.version, hour + "h " + min + "min " + sec + "sec", up_doc.remarks};
								else if (min>0) strs = new String[]{up_doc.name, up_doc.version, min + "min " + sec + "sec", up_doc.remarks};
								else strs = new String[]{up_doc.name, up_doc.version, sec + "sec", up_doc.remarks};
								tableItem.setText(strs);

								up_doc.key = tableItem.hashCode();
								Document.putDocument(up_doc);
							}
						}
					}
				}
			});


		} catch (Exception e) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			e.printStackTrace();
		}
	}

	public int initialize(String[] args) {
		// Initialize the ActiveX component		
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return -1;
		}

		Layout.datanetworks_table.removeAll();
		Layout.paradigms_table.removeAll();
		Layout.ta_types_table.removeAll();
		Document.removeAll();
		Paradigm.removeAll();
		Type.removeAll();

		String param = "";
//		if (args.length != 0) {
//			if (args.length > 0) {
//				if (args[0].length() != 0) param = "-ORBInitRef NameService=corbaloc:iiop:" + args[0] + "/NameService";
//			}
//		}

		try {
			int[] rgdispid = activex.getIDsOfNames(new String[]{"initialize"});
			int dispIdMember = rgdispid[0];
			Variant[] rgvarg = new Variant[1];
			rgvarg[0] = new Variant(param);

			Variant pVarResult = activex.invoke(dispIdMember, rgvarg);
			if (pVarResult.getInt()==5 || pVarResult.getInt()==6 || pVarResult.getInt()==7) {
				layout.status_text.setText("Could not find the Backplane. Please check that the Backplane is running!");
				return -1;
			} else if (pVarResult.getInt()!=0) {
				layout.status_text.setText("Could not connect to the Backplane!");
				return -1;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

		if (getRemoteTypes()==0) layout.status_text.setText("connected to the Backplane.");
		else layout.status_text.setText("Could not get ToolAdaptor types from the Backplane.");

		try {
			int[] rgdispid = activex.getIDsOfNames(new String[]{"setXSLTScriptPath"});
			int dispIdMember = rgdispid[0];
			Variant[] rgvarg = new Variant[1];
			rgvarg[0] = new Variant(Layout.xslt_script_path);

			Variant pVarResult = activex.invoke(dispIdMember, rgvarg);
		} catch (Exception e) {
			e.printStackTrace();
		}

		return 0;
	}		

/*
	public void default_subscribe(String default_subscribed_paradigms) {
		// Subscribe to the default paradigms
		char[] result = default_subscribed_paradigms.toCharArray();

		int offset = 0;
		String name = "";
		String version = "";
		String temp = "";
		for(int i=0; i<result.length-1; i++) {
			if (result[i] == '\255') {
				if (result[i+1] == '\255') {
					version = temp;
					Paradigm par = Paradigm.find_paradigm(name, version);
					if (par == null) return;
					subscribeParadigm(par);
					name = "";
					version = "";
				}
				else name = temp;
				temp = "";
			} else {
				temp += result[i];
			}
		}	
	}
*/

	public int loginToolAdaptor(String type) {

		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return -1;
		}

		Layout.datanetworks_table.removeAll();
		Layout.paradigms_table.removeAll();
		Document.removeAll();
		Paradigm.removeAll();

		int[] rgdispid = activex.getIDsOfNames(new String[]{"loginToolAdaptor"});
		int dispIdMember = rgdispid[0];
		Variant[] rgvarg = new Variant[1];
		rgvarg[0] = new Variant(type);
		try {
			Variant pVarResult = activex.invoke(dispIdMember, rgvarg);
			if (pVarResult.getInt()!=0) {
				MessageDialog.openError(Layout.getShell(), "Error in loginToolAdaptor invocation", "Error");
			}
		} catch (Exception ex) {
			ex.printStackTrace();	
		}
				
		if (getRemoteParadigms()==0) layout.status_text.setText("connected to the Backplane. Type: " + type);
		else layout.status_text.setText("Could not login to the Backplane with this type.");

		return 0;		
	}

	public int getRemoteTypes() {
		// Get the remote tooladaptor types
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return -1;
		}

		Type.ta_types = new Type[0];
		try {
			int[] rgdispid = activex.getIDsOfNames(new String[]{"remoteTypes"});
			int dispIdMember = rgdispid[0];
			Variant pVarResult = activex.invoke(dispIdMember);

			if (pVarResult.getString()==null) return 0;
			if (pVarResult.getString().compareTo("-1")==0) return -2;				

			char[] result = pVarResult.getString().toCharArray();

			int length = 0;
			for(int i=0; i<result.length; i++) {
				if (result[i]=='\n') length++;
			}

			Type.ta_types = new Type[length];
			int offset = 0;
			String temp = "";
			for(int i=0; i<result.length; i++) {
				if (result[i] == '\n') {
					Type.ta_types[offset++] = new Type(temp);
					temp = "";
				} else {
					temp += result[i];
				}
			}	
		} catch (Exception e) {
			e.printStackTrace();
		}

		if (Type.ta_types.length!=0) Type.fillTable();
		return 0;
	}		

	public int getRemoteParadigms() {
		// Get the remote paradigms
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return -1;
		}

		try {
			// get the working paradigms
			int[] rgdispid = activex.getIDsOfNames(new String[]{"remoteParadigms"});
			int dispIdMember = rgdispid[0];
			Variant pVarResult = activex.invoke(dispIdMember);

			if (pVarResult.getString()==null) {
				layout.status_text.setText("Could not find the Backplane. Please check that the Backplane is running!");
				return -2;
			}
			
			char[] result = pVarResult.getString().toCharArray();

			int length = 0;
			for(int i=0; i<result.length-1; i++) {
				if (result[i]=='\n')
					if (result[i+1]=='\n') length++;
			}

			Paradigm.remote_paradigms = new Paradigm[length];
			int offset = 0;
			String temp = "";
			for(int i=0; i<result.length-1; i++) {
				if (result[i] == '\n') {
					if (result[i+1] == '\n') Paradigm.remote_paradigms[offset++].version = temp;
					else Paradigm.remote_paradigms[offset] = new Paradigm(temp);
					temp = "";
				} else {
					temp += result[i];
				}
			}	

			// set the subscribable paradigms
			rgdispid = activex.getIDsOfNames(new String[]{"getSubscribable"});
			dispIdMember = rgdispid[0];
			pVarResult = activex.invoke(dispIdMember);

			if (pVarResult.getString().compareTo("null")==0) {
				layout.status_text.setText("Could not find the Backplane. Please check that the Backplane is running!");
				return -2;
			}

			result = pVarResult.getString().toCharArray();

			length = 0;
			for(int i=0; i<result.length-1; i++) {
				if (result[i]=='\n')
					if (result[i+1]=='\n') length++;
			}

			offset = 0;
			temp = "";
			String name = "";
			for(int i=0; i<result.length-1; i++) {
				if (result[i] == '\n') {
					if (result[i+1] == '\n') {
						Paradigm par = Paradigm.find_paradigm(name, temp);
						par.subscribable = true;
						Main.backplane.subscribeParadigm(par);
					} else name = temp;
					temp = "";
				} else {
					temp += result[i];
				}
			}	

		} catch (Exception e) {
			e.printStackTrace();
		}

		layout.buildParadigmMenu();
		return 0;
	}		

	public void unsubscribeParadigm(String name, String version) {
		// invoke unsubscribeParadigm	
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return;
		}

		int[] rgdispid = activex.getIDsOfNames(new String[]{"unsubscribeParadigm"});
		int dispIdMember = rgdispid[0];
		Variant[] rgvarg = new Variant[2];
		rgvarg[0] = new Variant(name);
		rgvarg[1] = new Variant(version);
		try {
			Variant pVarResult = activex.invoke(dispIdMember, rgvarg);
			if (pVarResult.getInt()!=0) {
				MessageDialog.openError(Layout.getShell(), "Error in unsubscribeParadigm invocation", "Error");
			}
		} catch (Exception ex) {
			ex.printStackTrace();	
		}
	}

	public void subscribeParadigm(Paradigm par) {
		// invoke subscribeParadigm	
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return;
		}

		int[] rgdispid = activex.getIDsOfNames(new String[]{"subscribeParadigm"});
		int dispIdMember = rgdispid[0];
		Variant[] rgvarg = new Variant[2];
		rgvarg[0] = new Variant(par.name);
		rgvarg[1] = new Variant(par.version);
		try {
			Variant pVarResult = activex.invoke(dispIdMember, rgvarg);
			if (pVarResult.getInt()!=0) {
				MessageDialog.openError(Layout.getShell(), "Error", "Error in subscribeParadigm invocation");
			} else {

				// Put out the item in the widget
				TableItem tableItem = new TableItem(Layout.paradigms_table, SWT.NONE);
				String strs[] = new String[]{par.name, par.version};
				tableItem.setText(strs);
				par.putTableKey(tableItem.hashCode());
			}

		} catch (Exception ex) {
			ex.printStackTrace();	
		}
	}

	public void fetchDocument(long subscriptionId, long docId, String working_dir, String filename) {
		// invoke fetchDocument	
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return;
		}

		int[] rgdispid = activex.getIDsOfNames(new String[]{"fetchDocument"});
		int dispIdMember = rgdispid[0];
		Variant[] rgvarg = new Variant[4];
		rgvarg[0] = new Variant(subscriptionId);
		rgvarg[1] = new Variant(docId);
		rgvarg[2] = new Variant(working_dir);
		rgvarg[3] = new Variant(filename);
		try {
			Variant pVarResult = activex.invoke(dispIdMember, rgvarg);
			if (pVarResult.getInt()!=0) {
				MessageDialog.openError(Layout.getShell(), "Error in fetchDocument invocation", "Error");
			}
		} catch (Exception ex) {
			ex.printStackTrace();	
		}

		try {
			rgdispid = activex.getIDsOfNames(new String[]{"getXSLTScriptPath"});
			dispIdMember = rgdispid[0];
			Variant pVarResult = activex.invoke(dispIdMember);
			if (pVarResult.getString()==null) return;
			Layout.xslt_script_path = pVarResult.getString();
		} catch (Exception ex) {
			ex.printStackTrace();	
		}
	}

	public void publishDocument(String filename) {
		// publish this dragged and dropped file
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return;
		}

		int[] rgdispid = activex.getIDsOfNames(new String[]{"publishDocument"});
		int dispIdMember = rgdispid[0];
		Variant[] rgvarg = new Variant[1];
		rgvarg[0] = new Variant(filename);
		try {
			Variant pVarResult = activex.invoke(dispIdMember, rgvarg);
			if (pVarResult.getInt()==-1)
				MessageDialog.openError(Layout.getShell(), "Error", "Error in publishDocument invocation");
			if (pVarResult.getInt()==-2)
				MessageDialog.openError(Layout.getShell(), "Error", "There is no available paradigms. Please login as one of the ToolAdaptor type.");
			if (pVarResult.getInt()==-3)
				MessageDialog.openError(Layout.getShell(), "Error", "More types of publishable multiple and single documents is not supported yet.");
			if (pVarResult.getInt()==-4)
				MessageDialog.openError(Layout.getShell(), "Error", "Only one document can be published with this type of ToolAdaptor.");
			if (pVarResult.getInt()==-5)
				MessageDialog.openError(Layout.getShell(), "Error", "Only multiple documents can be published with this type of ToolAdaptor.");
		} catch (Exception ex) {
			ex.printStackTrace();	
		}
	}

	public void aboutBox() {
		if (activex == null) {
			layout.status_text.setText("Probably the GenericTooladaptor ActiveX component is not registered. The framework won't function properly!");
			return;
		}

		int[] rgdispid = activex.getIDsOfNames(new String[]{"AboutBox"});
		int dispIdMember = rgdispid[0];
		activex.invoke(dispIdMember);
	}

}
