/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

// Generic_ToolAdaptorDlg.cpp : implementation file
//
#include "TA_XSLT_Support.h"
#include "UdmHeader.h"
#include "stdafx.h"

#include "Generic_ToolAdaptor.h"
#include "Generic_ToolAdaptorDlg.h"

#include "Publish_document.h"
#include "Fetch_document.h"
#include "Subscribe_dlg.h"
#include "Unsubscribe_dlg.h"
#include "mga_paradigm_DLG.h"
#include "ta_types_dlg.h"
#include "Progress_DLG.h"

#include <fstream>
#include <time.h>

using namespace UdmCORBA;
void dummy(void) {};

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// *************************************************************************************************************************
// global variables
// *************************************************************************************************************************
extern CGeneric_ToolAdaptorDlg *main_dlg;

map<long, PARADIGM_STRUCT> remote_paradigms;
list<CONFIG> config;
list<CONFIG> subscribable;
bool publishing_multiple_documents;
vector<DOCUMENT_STRUCT> existing_documents;

map<long, SUBSCRIPTION> subscriptions;
long subscriptions_id=0;

map<string, string> attachments; // filename, directory

struct THREAD_PARAM {
	PARADIGM_STRUCT paradigm;
	string type;
};

struct REGISTERED_PARADIGM {
	string name;
	string version;
	Udm::UdmDiagram udmDataDiagram;
	Uml::Diagram paradigm_diagram;
};

list<REGISTERED_PARADIGM> registered_paradigms;

Udm::UdmDiagram udmDataDiagram;
Uml::Diagram paradigm_diagram;

CString xsl_file_for_fetch = "";

// *************************************************************************************************************************
//	my_UML_ToolAdaptor
// *************************************************************************************************************************
class my_UML_ToolAdaptor : public ToolAdaptor {
public:

	my_UML_ToolAdaptor(int argc, char **argv) :
		ToolAdaptor(argc, argv) {};

	void notify(const char *paradigm_name, const char *paradigm_version, \
		const char *name, const char *version, const char *remarks, long docId, long keepAlive) {};
};

my_UML_ToolAdaptor *uml_ta;

// *************************************************************************************************************************
// my_ToolAdaptor
// *************************************************************************************************************************
class my_ToolAdaptor : public ToolAdaptor {

private:
	Uml::Diagram meta_diagram;

public:
	Progress_DLG progress_dlg;

	my_ToolAdaptor(const char *tooladaptor_type, int argc, char **argv) :
		ToolAdaptor(tooladaptor_type, argc, argv) {};

	my_ToolAdaptor(const char *tooladaptor_type, int argc, char **argv, Uml::Diagram meta_diagram) :
		ToolAdaptor(tooladaptor_type, argc, argv), meta_diagram(meta_diagram) {};

	void notify(const char *paradigm_name, const char *paradigm_version, \
		const char *name, const char *version, const char *remarks, long docId, long keepAlive) {
		
		try {
			CString filename;
			CString filetitle;
			CString temp_filename;
			CString xsl_file;
			
			Fetch_document fetch_dlg;
			fetch_dlg.m_name = name;
			fetch_dlg.m_version = version;
			fetch_dlg.m_remarks = remarks;
			char keep[32];
			ltoa(keepAlive, keep, 10);
			string t_keep = keep;
			t_keep += "sec";
			fetch_dlg.m_keepalive = t_keep.c_str();
			fetch_dlg.m_paradigm = (dn->getParadigmName() + " v" + dn->getParadigmVersion()).c_str();
			fetch_dlg.m_xslt_file = xsl_file_for_fetch;

			int result = fetch_dlg.DoModal();
			if (result==IDCANCEL) return;

			if (fetch_dlg.m_xslt_translation) xsl_file = fetch_dlg.m_xslt_file;

			while (filename.GetLength()<5 || !(filename.Right(4)==".xml" || filename.Right(4)==".mem" || filename.Right(4)==".mga")) {
				CFileDialog save_dlg(FALSE, "", "");
				result = save_dlg.DoModal();
				filename = save_dlg.GetPathName();
				filetitle = save_dlg.GetFileName();
				if (result==IDCANCEL) return;

				if (filename.GetLength()<5 || !(filename.Right(4)==".xml" || filename.Right(4)==".mem" || filename.Right(4)==".mga"))
					main_dlg->MessageBox("Invalid filename! Filename must end with .mem or .xml or .mga", "Error", MB_OK + MB_ICONEXCLAMATION + MB_TOPMOST);
			}

			string ff_name = "";
			string xsd_file = "";

			// Get the rootobject
			Uml::Class rootobject = getDocumentRootobjectMeta(docId, meta_diagram);
			if (!rootobject) {
				main_dlg->MessageBox("Could not find the meta of the rootobject of the document! Could not save it.", "Error", MB_OK + MB_ICONEXCLAMATION);
				return;
			}

			if (filename.Right(4)==".xml") {
				// xml layer specific metalocator
				ff_name = dn->getParadigmName() + "_v" + dn->getParadigmVersion();

				// Create temporary xsd for the paradigm			
				xsd_file = ff_name + ".xsd";
				OTIF_XSLT::GenerateXMLSchemaFile(meta_diagram, xsd_file.c_str());

				if (xsl_file.Compare("")) {
					temp_filename = filename;
					filename = "__temp_" + filetitle + "_xslt.xml";
				}

			} else if (filename.Right(4)==".mga") {
				// mga layer specific metalocator
				mga_paradigm_DLG mga_dlg;
				int result = mga_dlg.DoModal();
				if (result == IDOK) ff_name = mga_dlg.m_paradigm;
				else return;
			}

			// Create the new datanetwork
			dn->CreateNew((LPCSTR)filename, ff_name.c_str(), rootobject, Udm::CHANGES_PERSIST_ALWAYS);
//			main_dlg->m_status.AddString(("Fetching " + (string)name + " v" + (string)version + " document...").c_str());


			// Display the progress bar for the fetch process
			progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
			progress_dlg.SetWindowText("Fetching document:");
			progress_dlg.ShowWindow(SW_SHOW);

			dn->fetchDocument(docId);
			dn->SaveAs((LPCSTR)filename);
			dn->CloseNoUpdate();
			dn->release();
			
			progress_dlg.DestroyWindow();
			
			// XSLT translation
			if (xsl_file.Compare("") && filename.Right(4)==".xml") {

				// Display the progress bar for the XSLT post-process process
				progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
				progress_dlg.SetWindowText("XSLT post-processing:");
				progress_dlg.ShowWindow(SW_SHOW);

				// do the transform
				int result;
				try {
					OTIF_XSLT::XSLTransformer xslt;
					result = xslt.executeTransform((LPCSTR)filename, (LPCSTR)xsl_file, (LPCSTR)temp_filename, false, "");
			
				} catch (OTIF_XSLT::xslt_exception &e) {
					main_dlg->MessageBox(e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
					progress_dlg.DestroyWindow();
					return;
				}
			
				if (result!=0) {
					main_dlg->MessageBox("Could not transform the document with the XSLT script", "Error", MB_OK + MB_ICONEXCLAMATION);
					progress_dlg.DestroyWindow();
					return;
				}

				progress_dlg.m_progress.SetPos(100);

				DeleteFile(filename);
				DeleteFile(xsd_file.c_str());

				progress_dlg.DestroyWindow();
				filename = temp_filename;
			}

			main_dlg->m_status.AddString( ((string)name + " v" + (string)version + " document saved in " + (string)((LPCSTR)filename) + " file").c_str());
			main_dlg->m_status.AddString("");

		} catch (udm_exception &e) {
			main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		} catch (udmcorba_exception &e) {
			main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		} catch (...) {
			main_dlg->MessageBox("Fatal error!", "Error", MB_OK + MB_ICONEXCLAMATION);
		}

	};

	void changedExistingDocument(const char *name, const char *version, const char *remarks, long docId, long keepAlive) {
		notify("", "", name, version, remarks, docId, keepAlive);
	}
	
	void getFetchStatus(long percent) {
		progress_dlg.m_progress.SetPos(percent);
	}

	void getPublishStatus(long percent) {
		progress_dlg.m_progress.SetPos(percent);
	}
};

my_ToolAdaptor *ta = NULL;

// *************************************************************************************************************************
// fetching Thread
// *************************************************************************************************************************
CWinThread *fetching_thread;

UINT MyFetchingThread( LPVOID pParam ) {

	::CoInitialize(NULL);

	try {

		THREAD_PARAM *par = (THREAD_PARAM*) pParam;
		string name = (string) par->paradigm.name;
		string version = (string) par->paradigm.version;
		string type = par->type;

		Uml::Diagram paradigm_diagram = uml_ta->remoteFetchMyParadigm(name.c_str(), version.c_str());

		Udm::UdmDiagram udmDataDiagram;
		udmDataDiagram.dgr = &paradigm_diagram;
		udmDataDiagram.init = dummy;

		my_ToolAdaptor *ta = new my_ToolAdaptor(type.c_str(), __argc, __argv, paradigm_diagram);

		SUBSCRIPTION subscr;
		subscr.paradigm.name = name;
		subscr.paradigm.version = version;
		subscr.fetch_ta = ta;
		subscriptions.insert(pair<const long, SUBSCRIPTION>(subscriptions_id++, subscr));

		delete par;
		
		ta->subscribe(udmDataDiagram, name.c_str(), version.c_str());
		ta->run();

		ta->logout();
		delete ta;

	} catch (udm_exception &e) {
		main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		return -1;
	} catch (udmcorba_exception &e) {
		main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		return -1;
	}

	return 0;
}


/////////////////////////////////////////////////////////////////////////////
// CGeneric_ToolAdaptorDlg dialog

CGeneric_ToolAdaptorDlg::CGeneric_ToolAdaptorDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CGeneric_ToolAdaptorDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CGeneric_ToolAdaptorDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

void CGeneric_ToolAdaptorDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CGeneric_ToolAdaptorDlg)
	DDX_Control(pDX, IDC_LIST1, m_status);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CGeneric_ToolAdaptorDlg, CDialog)
	//{{AFX_MSG_MAP(CGeneric_ToolAdaptorDlg)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_BN_CLICKED(IDC_BUTTON1, OnPublish)
	ON_BN_CLICKED(IDC_BUTTON2, OnSubscribe)
	ON_BN_CLICKED(IDC_BUTTON3, OnButton3)
	ON_BN_CLICKED(IDC_BUTTON4, OnButton4)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void getDocumentsInfo() {
	existing_documents.clear();
	existing_documents = ta->getDocuments();
}


/////////////////////////////////////////////////////////////////////////////
// CGeneric_ToolAdaptorDlg message handlers

BOOL CGeneric_ToolAdaptorDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	
	// TODO: Add extra initialization here
	m_status.AddString("Generic ToolAdaptor started...");
	m_status.AddString("");

	if (ta) delete ta;
	if (uml_ta) delete uml_ta;

	try {
		CString type = "";
		if (__argc>1) {
			CString temp = __argv[1];
			if (temp.GetAt(0)!='-') type = __argv[1];
		}
		
		if (!type.Compare("")) {
			ta_types_dlg types_dlg;
			int nResponse = types_dlg.DoModal();
			if (nResponse == IDCANCEL) exit(0);
			type = types_dlg.m_ta_type;
		}

		this->SetWindowText("Generic ToolAdaptor - " + type);
		
		ta = new my_ToolAdaptor(type, __argc, __argv);
		uml_ta = new my_UML_ToolAdaptor(__argc, __argv);
		uml_ta->getDatanetwork(Uml::diagram);

		remote_paradigms.clear();
		ta->getWorkingParadigms(remote_paradigms);

		config.clear();
		ta->getTAConfig(config);

		subscribable.clear();
		ta->getTASubscribable(subscribable);

/*
		for(list<CONFIG>::iterator i=subscribable.begin(); i!=subscribable.end(); i++) {
			list<PARADIGM_STRUCT> pars = (*i).paradigms;
			for(list<PARADIGM_STRUCT>::iterator j=pars.begin(); j!=pars.end(); j++) {
				MessageBox(((*j).name + (*j).version).c_str(), "info", MB_OK);
			}
		}
*/

	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		exit(-1);
	} catch (udm_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		exit(-1);
	}

	return TRUE;  // return TRUE  unless you set the focus to a control
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CGeneric_ToolAdaptorDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CGeneric_ToolAdaptorDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

// *************************************************************************************************************************
//	Publish a document
// *************************************************************************************************************************
void CGeneric_ToolAdaptorDlg::OnPublish() 
{
	// TODO: Add your control notification handler code here
	::CoInitialize(NULL);
	
	UpdateData(true);

	// decide what to publish
	list<PARADIGM_STRUCT> paradigms;
	publishing_multiple_documents = false;
	for(list<CONFIG>::iterator i=config.begin(); i!=config.end(); i++) {
		paradigms = (*i).paradigms;		
		if (paradigms.size()>1 && config.size()>1) {
			MessageBox("More types of publishable multiple and single documents is not supported yet!", "Error", MB_OK + MB_ICONEXCLAMATION);
			return;
		}

		if (paradigms.size()!=1) publishing_multiple_documents = true;
	}
	
	Publish_document publish_dlg;
	publish_dlg.m_updated_document_id = -1;

	try {
		CString filename;
		int result = publish_dlg.DoModal();
		if (result == IDOK) {
			CString name = publish_dlg.m_name;
			CString version = publish_dlg.m_version;
			CString remarks = publish_dlg.m_remarks;
			BOOL xslt_support = publish_dlg.m_xslt_support;
			long updated_document_id = publish_dlg.m_updated_document_id;

			long keepAlive = atol(publish_dlg.m_keepalive);
			switch (publish_dlg.m_time) {
			case 1 : keepAlive *= 60; break;
			case 2 : keepAlive *= 3600; break;
			}


			// Publishing simple document
			if (!publishing_multiple_documents) {

				CFileDialog open_dlg(TRUE, "", "");
				open_dlg.m_ofn.lpstrTitle = "Select the document to publish";
				result = open_dlg.DoModal();
				if (result==IDCANCEL) return;
				filename = open_dlg.GetPathName();
				CString filetitle = open_dlg.GetFileTitle();
				CString fileext = open_dlg.GetFileExt();
				CString xsl_file = "";

				if (xslt_support && !fileext.Compare("xml")) {
					CFileDialog open_xsl_dlg(TRUE, "xsl", "");
					open_xsl_dlg.m_ofn.lpstrTitle = "Open the XSLT file for preprocessing";
					int result = open_xsl_dlg.DoModal();
					if (result==IDCANCEL) return;
					xsl_file = open_xsl_dlg.GetPathName();
				}
				
				map<long, PARADIGM_STRUCT>::iterator i = remote_paradigms.find(publish_dlg.m_paradigm);
				if (i==remote_paradigms.end())  return;

				string par_name = (*i).second.name;
				string par_version = (*i).second.version;
				
				// if it is fetched remotely we will use that Diagram and paradigm
				for(list<REGISTERED_PARADIGM>::iterator j=registered_paradigms.begin(); j!=registered_paradigms.end(); j++) {
					if (!par_name.compare((*j).name) && !par_version.compare((*j).version)) {
						udmDataDiagram = (*j).udmDataDiagram;
						paradigm_diagram = (*j).paradigm_diagram;
					}
				}
				
				// if it is not fetched yet, we fetch it and put the paradigm into our store to use it now and later
				if (!ta->isRegisteredDatanetwork((*i).second.name.c_str(), (*i).second.version.c_str())) {
					paradigm_diagram = uml_ta->remoteFetchMyParadigm((*i).second.name.c_str(), (*i).second.version.c_str());
					udmDataDiagram.dgr = &paradigm_diagram;
					udmDataDiagram.init = dummy;

					REGISTERED_PARADIGM temp_reg_par;
					temp_reg_par.name = par_name;
					temp_reg_par.version = par_version;
					temp_reg_par.paradigm_diagram = paradigm_diagram;
					temp_reg_par.udmDataDiagram = udmDataDiagram;
					registered_paradigms.push_front(temp_reg_par);
				}
			

				// XSLT translation
				if (xsl_file.Compare("") && !fileext.Compare("xml")) {

					// Display the progress bar for the XSLT pre-process process
					ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
					ta->progress_dlg.SetWindowText("XSLT pre-processing:");
					ta->progress_dlg.ShowWindow(SW_SHOW);

					// do the trasnform
					int result;
					try {
						OTIF_XSLT::XSLTransformer xslt;
						result = xslt.executeTransform((LPCSTR)filename, (LPCSTR)xsl_file, \
							(LPCSTR)("__temp_" + filetitle + "_xslt." + fileext), false, "");

					} catch (OTIF_XSLT::xslt_exception &e) {
						MessageBox(e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
						ta->progress_dlg.DestroyWindow();
						return;
					}

					if (result!=0) {
						MessageBox("Could not transform the document with the XSLT script", "Error", MB_OK + MB_ICONEXCLAMATION);
						ta->progress_dlg.DestroyWindow();
						return;
					}

					ta->progress_dlg.m_progress.SetPos(100);

					filename = "__temp_" + filetitle + "_xslt." + fileext;

					ta->progress_dlg.DestroyWindow();
				}

				// Display the progress bar for the publishing process
				ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
				ta->progress_dlg.SetWindowText("Publishing document:");
				ta->progress_dlg.ShowWindow(SW_SHOW);

				UdmCORBA::CORBADataNetwork *dn;
				try {
					// create DN with the remote paradigm
					dn = ta->getDatanetwork(udmDataDiagram, (*i).second.name.c_str(), (*i).second.version.c_str());
					// open the publishable document
					dn->OpenExisting((LPCSTR)filename, "", Udm::CHANGES_LOST_DEFAULT);
				} catch (udm_exception &e) {
					string error = e.what();
					if (!error.compare("Not found")) MessageBox("Invalid input file!", "Error", MB_OK + MB_ICONEXCLAMATION);
					else MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);

					ta->progress_dlg.DestroyWindow();
					return;
				}


				// attachment files
				for(map<string, string>::iterator addfile = attachments.begin(); addfile!=attachments.end(); addfile++) {
					dn->addAdditionalFile( (*addfile).first.c_str(), (*addfile).second.c_str() );
				}

				// publish the document
				dn->publishDocument((LPCSTR)name, (LPCSTR)version, (LPCSTR)remarks, keepAlive, updated_document_id);
				// close the document
				dn->CloseNoUpdate();

				ta->progress_dlg.DestroyWindow();

				// delete temporary docs after XSLT transform
				if (xsl_file.Compare("")) DeleteFile((LPCSTR)filename);

				m_status.AddString("Document is successfully sent.");
				m_status.AddString("");


			// Publishing multiple document
			} else { 

				list<UdmCORBA::CORBADataNetwork*> docs;
				list<CString> temp_xsl_files;
				
				for(list<PARADIGM_STRUCT>::iterator i=paradigms.begin(); i!=paradigms.end(); i++) {
					CFileDialog open_dlg(TRUE, "", "");
					open_dlg.m_ofn.lpstrTitle = "Open the " + (CString)((*i).name.c_str()) + " v" + (CString)((*i).version.c_str()) + " document";
					result = open_dlg.DoModal();
					if (result==IDCANCEL) return;
					filename = open_dlg.GetPathName();
					CString filetitle = open_dlg.GetFileTitle();
					CString fileext = open_dlg.GetFileExt();
					CString xsl_file = "";

					if (xslt_support) {
						CFileDialog open_xsl_dlg(TRUE, "xsl", "");
						open_xsl_dlg.m_ofn.lpstrTitle = "Open the XSLT file for the " + (CString)((*i).name.c_str()) + " v" + (CString)((*i).version.c_str()) + " document";
						int result = open_xsl_dlg.DoModal();
						if (result==IDOK) xsl_file = open_xsl_dlg.GetPathName();
					}

					Uml::Diagram paradigm_diagram = uml_ta->remoteFetchMyParadigm((*i).name.c_str(), (*i).version.c_str());
					Udm::UdmDiagram udmDataDiagram;
					udmDataDiagram.dgr = &paradigm_diagram;
					udmDataDiagram.init = dummy;

					// XSLT translation
					if (xsl_file.Compare("") && !fileext.Compare("xml")) {

						// Display the progress bar for the XSLT pre-process process
						ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
						ta->progress_dlg.SetWindowText("XSLT pre-processing:");
						ta->progress_dlg.ShowWindow(SW_SHOW);

						// do the trasnform
						int result = 0;
						try {
							OTIF_XSLT::XSLTransformer xslt;
							int result = xslt.executeTransform((LPCSTR)filename, (LPCSTR)xsl_file, \
								(LPCSTR)("__temp_" + filetitle + "_xslt." + fileext), false, "");

						} catch (OTIF_XSLT::xslt_exception &e) {
							MessageBox(e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
							ta->progress_dlg.DestroyWindow();
							return;
						}

						if (result!=0) {
							MessageBox("Could not transform the document with the XSLT script", "Error", MB_OK + MB_ICONEXCLAMATION);
							ta->progress_dlg.DestroyWindow();
							return;
						}

						ta->progress_dlg.m_progress.SetPos(100);

						filename = "__temp_" + filetitle + "_xslt." + fileext;
						temp_xsl_files.push_back(filename);

						ta->progress_dlg.DestroyWindow();
					}

					UdmCORBA::CORBADataNetwork *dn;
					try {
						// create DN with the remote paradigm
						dn = ta->getDatanetwork(udmDataDiagram, (*i).name.c_str(), (*i).version.c_str());
						// open the publishable document
						dn->OpenExisting((LPCSTR)filename, "", Udm::CHANGES_LOST_DEFAULT);
					} catch (udm_exception &e) {
						string error = e.what();
						if (!error.compare("Not found")) MessageBox("Invalid input file!", "Error", MB_OK + MB_ICONEXCLAMATION);
						else MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);

						ta->progress_dlg.DestroyWindow();
						return;
					}

					docs.push_back(dn);
				}

				// Display the progress bar for the publishing process
				ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
				ta->progress_dlg.SetWindowText("Publishing document:");
				ta->progress_dlg.ShowWindow(SW_SHOW);

				
				// attachment files
				for(map<string, string>::iterator addfile = attachments.begin(); addfile!=attachments.end(); addfile++) {
					(*docs.begin())->addAdditionalFile( (*addfile).first.c_str(), (*addfile).second.c_str() );
				}

				
				ta->publishMultipleDocuments(docs, (LPCSTR)name, (LPCSTR)version, (LPCSTR)remarks, keepAlive, updated_document_id);

				for(list<UdmCORBA::CORBADataNetwork*>::iterator q=docs.begin(); q!=docs.end(); q++) (*q)->CloseNoUpdate();

				ta->progress_dlg.DestroyWindow();

				// delete temporary docs after XSLT transform
				for(list<CString>::iterator txf=temp_xsl_files.begin(); txf!=temp_xsl_files.end(); txf++) DeleteFile((LPCSTR)(*txf));
				
				m_status.AddString("Document is successfully sent.");
				m_status.AddString("");
			} // publishing_multiple_documents
		
		} // if IDOK

	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
	} catch (udm_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
	}
}

// *************************************************************************************************************************
//	Subscribe to a paradigm
// *************************************************************************************************************************
void CGeneric_ToolAdaptorDlg::OnSubscribe() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
	Subscribe_dlg subscribe_dlg;

	int result = subscribe_dlg.DoModal();
	if (result == IDOK) {
		map<long, PARADIGM_STRUCT>::iterator i = remote_paradigms.find(subscribe_dlg.m_paradigm);
		if (i!=remote_paradigms.end()) {
			THREAD_PARAM *param = new THREAD_PARAM();
			param->paradigm.name = (*i).second.name;
			param->paradigm.version = (*i).second.version;
			param->type = ta->getToolAdaptorType();
			
			fetching_thread = AfxBeginThread(MyFetchingThread, param);
			m_status.AddString(("Subscribe to " + (*i).second.name + " v" + (*i).second.version + " paradigm").c_str());
			m_status.AddString("");
		}
	}
}


// *************************************************************************************************************************
// Unsubscribe a paradigm
// *************************************************************************************************************************
void CGeneric_ToolAdaptorDlg::OnButton3() 
{
	// TODO: Add your control notification handler code here
	Unsubscribe_dlg unsub_dlg;
	int result = unsub_dlg.DoModal();
	if (result==IDOK) {
		if (unsub_dlg.m_paradigms>-1) {
			for(map<long, SUBSCRIPTION>::iterator j=subscriptions.begin(); j!=subscriptions.end(); j++) {
				if (j->second.list_offset==unsub_dlg.m_paradigms) {
					j->second.fetch_ta->terminate();
					m_status.AddString(("Unsubscribed the " + j->second.paradigm.name + " v" 
						+ j->second.paradigm.version + " paradigm").c_str());
					m_status.AddString("");
					subscriptions.erase(j);
				}
			}
		}
	}
}

// *************************************************************************************************************************
// Clear status listbox
// *************************************************************************************************************************
void CGeneric_ToolAdaptorDlg::OnButton4() 
{
	// TODO: Add your control notification handler code here
	m_status.ResetContent();
}
