// ASP.h : Declaration of the CASP

#ifndef __ASP_H_
#define __ASP_H_

#include "resource.h"       // main symbols
#include <asptlb.h>         // Active Server Pages Definitions

#include <time.h>
#include <direct.h>
#include "mgainterfaces.h"

using namespace UdmCORBA;

struct DOCUMENTS {
	string name;
	string version;
	long keepalive;
	string remarks;
	long subscription_id;
	long notified_time;

	string filename;
	string xsd_filename;
};

class CASP;

// *************************************************************************************************************************
//	my_UML_ToolAdaptor
// *************************************************************************************************************************
class ASP_my_UML_ToolAdaptor : public ToolAdaptor {
public:

	ASP_my_UML_ToolAdaptor(int argc, char **argv) :
		ToolAdaptor(argc, argv) {};

	void notify(const char *paradigm_name, const char *paradigm_version, \
		const char *name, const char *version, const char *remarks, long docId, long keepAlive) {};
};

// *************************************************************************************************************************
// my_ToolAdaptor
// *************************************************************************************************************************
class ASP_my_ToolAdaptor : public ToolAdaptor {

friend CASP;

private:
	Uml::Diagram meta_diagram;
	long subscription_id;
	map<long, DOCUMENTS> documents;

public:
	ASP_my_ToolAdaptor(const char *tooladaptor_type, int argc, char **argv) :
		ToolAdaptor(tooladaptor_type, argc, argv) {};

	ASP_my_ToolAdaptor(const char *tooladaptor_type, int argc, char **argv, Uml::Diagram meta_diagram, long subscription_id) 
		: ToolAdaptor(tooladaptor_type, argc, argv), meta_diagram(meta_diagram), subscription_id(subscription_id) {};

	void changedExistingDocument(const char *name, const char *version, const char *remarks, long docId, long keepAlive) {
	
		map<long, DOCUMENTS>::iterator j = this->documents.find(docId);
		if (j == this->documents.end()) return;

		(*j).second.name = name;
		(*j).second.version = version;
		(*j).second.keepalive = keepAlive;
		(*j).second.remarks = remarks;
		(*j).second.subscription_id = this->subscription_id;
		(*j).second.filename = "";
		(*j).second.xsd_filename = "";
	};
	
	void notify(const char *paradigm_name, const char *paradigm_version, const char *name, \
		const char *version, const char *remarks, long docId, long keepAlive) {
	
		time_t ltime;
		time( &ltime );

		DOCUMENTS doc;
		doc.name = name;
		doc.version = version;
		doc.keepalive = keepAlive;
		doc.remarks = remarks;
		doc.subscription_id = this->subscription_id;
		doc.notified_time = ltime - 2; // 2 is a theoretical value :-)
		documents.insert( pair<long, DOCUMENTS>(docId, doc) );			
	};
	
};

struct ASP_SUBSCRIPTION {
	long list_offset;
	ASP_my_ToolAdaptor *fetch_ta;
	UdmCORBA::PARADIGM_STRUCT paradigm;
};

/////////////////////////////////////////////////////////////////////////////
// CASP
class ATL_NO_VTABLE CASP : 
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CASP, &CLSID_ASP>,
	public ISupportErrorInfo,
	public IDispatchImpl<IASP, &IID_IASP, &LIBID_GENERIC_TOOLADAPTOR_ACTIVEXLib>
{
public:
	CASP()
	{ 
		this->ta = NULL;
		this->uml_ta = NULL;
	
		subscriptions_id = 0;
		m_bOnStartPageCalled = FALSE;
	}


	// Destructor
	~CASP() {

		if (this->ta) delete this->ta;
		if (this->uml_ta) delete this->uml_ta;

		for(map<long, ASP_SUBSCRIPTION>::iterator j=this->subscriptions.begin(); j!=this->subscriptions.end(); j++) {
			for(map<long, DOCUMENTS>::iterator i=(*j).second.fetch_ta->documents.begin(); i!=(*j).second.fetch_ta->documents.end(); i++) {
				if (!(*i).second.filename.empty()) DeleteFile( (*i).second.filename.c_str() );
				if (!(*i).second.xsd_filename.empty()) DeleteFile( (*i).second.xsd_filename.c_str() );
			}
			j->second.fetch_ta->terminate();
		}
		
		this->remote_paradigms.clear();
		this->config.clear();
		this->subscribable.clear();
		this->registered_paradigms.clear();
		this->subscriptions.clear();
		this->subscribed.clear();
	}

public:

DECLARE_REGISTRY_RESOURCEID(IDR_ASP)

DECLARE_PROTECT_FINAL_CONSTRUCT()
DECLARE_NOT_AGGREGATABLE(CASP)

BEGIN_COM_MAP(CASP)
	COM_INTERFACE_ENTRY(IASP)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(ISupportErrorInfo)
END_COM_MAP()

// ISupportsErrorInfo
	STDMETHOD(InterfaceSupportsErrorInfo)(REFIID riid);

// IASP
public:
	STDMETHOD(initialize)(/*[in]*/ BSTR parameters, /*[out,retval]*/ BSTR *err);
	STDMETHOD(remoteTypes)(/*[out,retval]*/ BSTR *ret);
	STDMETHOD(remoteParadigms)(/*[out,retval]*/ BSTR *ret);
  STDMETHOD(getActions)(/*[out,retval]*/ BSTR *ret);
	STDMETHOD(getSubscribable)(/*[out,retval]*/ BSTR *ret);
	STDMETHOD(getPublishable)(/*[out,retval]*/ BSTR *ret);
	STDMETHOD(loginToolAdaptor)(/*[in]*/ BSTR type, /*[out,retval]*/ BSTR *err);
	STDMETHOD(publishDocument)(/*[in]*/ BSTR name, /*[in]*/ BSTR version, /*[in]*/ long keepAlive, /*[in]*/ BSTR remarks, 
		/*[in]*/ BSTR filename, /*[in]*/ BSTR xslt_filename, /*[in]*/ BSTR attachments, /*[in]*/ long update_id, /*[out,retval]*/ BSTR *err);
	STDMETHOD(subscribeParadigm)(/*[in]*/ BSTR par_name, /*[in]*/ BSTR par_version, /*[out,retval]*/ BSTR *err);
	STDMETHOD(availableDocuments)(/*[out,retval]*/ BSTR *ret);
	STDMETHOD(requestDocument)(/*[in]*/ BSTR extension, /*[in]*/ BSTR mga_par, /*[in]*/ BSTR directory, /*[in]*/ long doc_id, 
		/*[in]*/ BSTR xslt_filename, /*[out,retval]*/ BSTR *err);
	STDMETHOD(availableGMEParadigms)(/*[out,retval]*/ BSTR *ret);
	STDMETHOD(logout)(/*[out,retval]*/ BSTR *ret);
	STDMETHOD(unsubscribeParadigm)(/*[in]*/ BSTR par_name, /*[in]*/ BSTR par_version, /*[out,retval]*/ BSTR *err);
	STDMETHOD(getExistingDocuments)(/*[out,retval]*/ BSTR *ret);

	//Active Server Pages Methods
	STDMETHOD(OnStartPage)(IUnknown* IUnk);
	STDMETHOD(OnEndPage)();

private:
	CComPtr<IRequest> m_piRequest;					//Request Object
	CComPtr<IResponse> m_piResponse;				//Response Object
	CComPtr<ISessionObject> m_piSession;			//Session Object
	CComPtr<IServer> m_piServer;					//Server Object
	CComPtr<IApplicationObject> m_piApplication;	//Application Object
	BOOL m_bOnStartPageCalled;						//OnStartPage successful?


	// ----------------------------------------------------------------------
public:
	ASP_my_UML_ToolAdaptor *uml_ta;

private:	
	ASP_my_ToolAdaptor *ta;

	map<long, PARADIGM_STRUCT> remote_paradigms;
	list<CONFIG> config;
	list<CONFIG> subscribable;

	list<REGISTERED_PARADIGM> registered_paradigms;

	map<long, bool> subscribed;

public:
	Udm::UdmDiagram udmDataDiagram;
	Uml::Diagram paradigm_diagram;

public:

	long subscriptions_id;
	map<long, ASP_SUBSCRIPTION> subscriptions;
	vector<DOCUMENT_STRUCT> existing_documents;
};

#endif //__ASP_H_
