/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

// MainDialog.cpp : implementation file
//
#include "UdmCORBA.h"
#include "TA_XSLT_Support.h"
#include "stdafx.h"

#include "Generic_ToolAdaptor_ActiveX.h"
#include "Generic_ToolAdaptor_ActiveXCtl.h"
#include "MainDialog.h"

#include "Publish_DLG.h"
#include "Subscribe_DLG.h"
#include "Unsubscribe_DLG.h"
#include "Fetch_A_Document.h"
#include "Fetch_DLG.h"
#include "Mga_DLG.h"
#include "ta_types_dlg.h"
#include "Choose_paradigm_dlg.h"
#include "Progress_DLG.h"

#include <fstream>
#include <time.h>
#include <direct.h>

using namespace UdmCORBA;
void dummy(void) {};

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// *************************************************************************************************************************
// global variables
// *************************************************************************************************************************
extern CMainDialog *main_dlg;
static HWND w_parent;

map<long, PARADIGM_STRUCT> remote_paradigms;
list<CONFIG> config;
list<CONFIG> subscribable;
bool publishing_multiple_documents;
vector<DOCUMENT_STRUCT> existing_documents;

map<long, SUBSCRIPTION> subscriptions;
long subscriptions_id=0;

map<string, string> attachments; // filename, directory

struct ARG_LIST {
	HWND parent;
	string name;
	string version;
	string type;
};

list<REGISTERED_PARADIGM> registered_paradigms;

Udm::UdmDiagram udmDataDiagram;
Uml::Diagram paradigm_diagram;

CString xsl_file_for_fetch = "";

// *************************************************************************************************************************
//	my_UML_ToolAdaptor
// *************************************************************************************************************************
class my_UML_ToolAdaptor : public ToolAdaptor {
public:

	my_UML_ToolAdaptor(int argc, char **argv) :
		ToolAdaptor(argc, argv) {};

	void notify(const char *paradigm_name, const char *paradigm_version, \
		const char *name, const char *version, const char *remarks, long docId, long keepAlive) {};
};

my_UML_ToolAdaptor *uml_ta;

// *************************************************************************************************************************
// my_ToolAdaptor
// *************************************************************************************************************************
class my_ToolAdaptor : public ToolAdaptor {

private:
	Uml::Diagram meta_diagram;
	long subscription_id;

public:
	Progress_DLG progress_dlg;

	my_ToolAdaptor(const char *tooladaptor_type, int argc, char **argv) :
		ToolAdaptor(tooladaptor_type, argc, argv) {};

//	my_ToolAdaptor(const char *tooladaptor_type, int argc, char **argv, Uml::Diagram meta_diagram) :
//		ToolAdaptor(tooladaptor_type, argc, argv), meta_diagram(meta_diagram) {};

	my_ToolAdaptor(const char *tooladaptor_type, int argc, char **argv, Uml::Diagram meta_diagram, long subscription_id) 
		: ToolAdaptor(tooladaptor_type, argc, argv), meta_diagram(meta_diagram), subscription_id(subscription_id) {};


	void changedExistingDocument(const char *name, const char *version, const char *remarks, long docId, long keepAlive) {

		message_param.name = name;
		message_param.version = version;
		message_param.remarks = remarks;
		message_param.keepAlive = keepAlive;
		message_param.docId = docId;
		message_param.subscriptionId = subscription_id;
		::SendMessage(w_parent, wm_Message, 1, reinterpret_cast<long>(&message_param));
	};
	

	void notify(const char *paradigm_name, const char *paradigm_version, const char *name, \
		const char *version, const char *remarks, long docId, long keepAlive) {
		
		message_param.name = name;
		message_param.version = version;
		message_param.remarks = remarks;
		message_param.keepAlive = keepAlive;
		message_param.docId = docId;
		message_param.subscriptionId = subscription_id;
		::SendMessage(w_parent, wm_Message, 0, reinterpret_cast<long>(&message_param));
	};
	
	long __fetchDocument(long docId, CString directory, CString filename) {
		try {
			// Get the rootobject
			Uml::Class rootobject = getDocumentRootobjectMeta(docId, meta_diagram);
			if (!rootobject) {
				if (main_dlg) main_dlg->MessageBox("Could not find the meta of the rootobject of the document! Could not save it.", "Error", MB_OK + MB_ICONEXCLAMATION);
				return -1;
			}
			string ff_name = "";
			string xsd_file = "";

			CString filetitle;
			CString temp_filename;
			CString xsl_file = "";

			if (!directory.Compare("")) {

				do {
					CFileDialog save_dlg(FALSE, "", filename, OFN_NOCHANGEDIR);
					int result = save_dlg.DoModal();
					filename = save_dlg.GetPathName();
					filetitle = save_dlg.GetFileName();
					directory = filename.Left(filename.GetLength() - save_dlg.GetFileName().GetLength());
					if (result==IDCANCEL) return 0;

					if (filename.GetLength()<5 || !(filename.Right(4)==".xml" || filename.Right(4)==".mem" || filename.Right(4)==".mga"))
						main_dlg->MessageBox("Invalid filename! Filename must end with .mem or .xml or .mga", "Error", MB_OK + MB_ICONEXCLAMATION + MB_TOPMOST);
				} while (filename.GetLength()<5 || !(filename.Right(4)==".xml" || filename.Right(4)==".mem" || filename.Right(4)==".mga"));

				_chdir((LPCSTR)directory);
				
				if (filename.Right(4)==".xml") {
					// xml layer specific metalocator
					ff_name = dn->getParadigmName() + "_v" + dn->getParadigmVersion();

					// Create temporary xsd for the paradigm			
					xsd_file = ff_name + ".xsd";
					OTIF_XSLT::GenerateXMLSchemaFile(meta_diagram, xsd_file.c_str());

					if (xsl_file.Compare("")) {
						temp_filename = filename;
						filename = "__temp_" + filetitle + "_xslt.xml";
					}

				} else if (filename.Right(4)==".mga") {
					// mga layer specific metalocator
					CMga_DLG mga_dlg;
					int result = mga_dlg.DoModal();
					if (result == IDOK) ff_name = mga_dlg.m_paradigm;
					else return 0;;
				}

			} else {
				
				Fetch_A_Document fetch_doc_dlg;
				fetch_doc_dlg.m_working_dir = directory;
				fetch_doc_dlg.m_filename = filename;
				fetch_doc_dlg.m_xslt_file = xsl_file_for_fetch;
				int result = fetch_doc_dlg.DoModal();
				if (result==IDCANCEL) return 0;

				CString extension = fetch_doc_dlg.m_file_extension;
				filename = fetch_doc_dlg.m_filename + "." + extension;

				if (fetch_doc_dlg.m_xslt_translation) xsl_file = fetch_doc_dlg.m_xslt_file;
				
				_chdir((LPCSTR)directory);

				if (!extension.Compare("mga")) {
					ff_name = fetch_doc_dlg.m_gme_paradigm;

				} else if (!extension.Compare("xml")) {
					ff_name = dn->getParadigmName() + "_v" + dn->getParadigmVersion();

					// Create temporary xsd for the paradigm			
					xsd_file = ff_name + ".xsd";
					OTIF_XSLT::GenerateXMLSchemaFile(meta_diagram, xsd_file.c_str());

					if (xsl_file.Compare("")) {
						temp_filename = filename;
						filename = "__temp_" + filetitle + "_xslt.xml";
					}
				}
			}
			
			// Create the new datanetwork
			dn->CreateNew((LPCSTR)(filename), ff_name.c_str(), rootobject, Udm::CHANGES_PERSIST_ALWAYS);

			// Display the progress bar for the fetch process
			progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
			progress_dlg.SetWindowText("Fetching document:");
			progress_dlg.ShowWindow(SW_SHOW);

			dn->fetchDocument(docId);
			dn->SaveAs((LPCSTR)(filename));
			dn->CloseNoUpdate();
			dn->release();

			progress_dlg.DestroyWindow();

			// XSLT translation
			if (xsl_file.Compare("") && filename.Right(4)==".xml") {

				// Display the progress bar for the XSLT post-process process
				progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
				progress_dlg.SetWindowText("XSLT post-processing:");
				progress_dlg.ShowWindow(SW_SHOW);

				// do the transform
				int result = 0;
				try {
					OTIF_XSLT::XSLTransformer xslt;
					result = xslt.executeTransform((LPCSTR)filename, (LPCSTR)xsl_file, (LPCSTR)temp_filename, false, "");

				} catch (OTIF_XSLT::xslt_exception &e) {
					if (main_dlg) main_dlg->MessageBox(e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
					else MessageBox(NULL, e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
					progress_dlg.DestroyWindow();
					return -2;
				}

				if (result!=0) {
					if (main_dlg) main_dlg->MessageBox("Could not transform the document with the XSLT script", "Error", MB_OK + MB_ICONEXCLAMATION);
					progress_dlg.DestroyWindow();
					return -2;
				}

				progress_dlg.m_progress.SetPos(100);

				DeleteFile(filename);
				DeleteFile(xsd_file.c_str());

				progress_dlg.DestroyWindow();
			}

		} catch (udm_exception &e) {
			if (main_dlg) main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
			else MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		} catch (udmcorba_exception &e) {
			if (main_dlg) main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
			else MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		} catch (...) {
			if (main_dlg) main_dlg->MessageBox("Fatal error!", "Error", MB_OK + MB_ICONEXCLAMATION);
			else MessageBox(NULL, "Fatal error!", "Error", MB_OK + MB_ICONEXCLAMATION);
		}

		return 0;
	};
	

	void getFetchStatus(long percent) {
		progress_dlg.m_progress.SetPos(percent);
	}

	void getPublishStatus(long percent) {
		progress_dlg.m_progress.SetPos(percent);
	}
};

my_ToolAdaptor *ta = NULL;


long _fetchDocument(long subscriptionId, long docId, CString directory, CString filename) {

	map<long, SUBSCRIPTION>::iterator i = subscriptions.find(subscriptionId);
	if (i!=subscriptions.end()) {
		(*i).second.fetch_ta->__fetchDocument(docId, directory, filename);
	} else {
		if (main_dlg) main_dlg->MessageBox("Could not find the corresponding subscription!", "Error", MB_OK + MB_ICONEXCLAMATION);
		else MessageBox(NULL, "Could not find the corresponding subscription!", "Error", MB_OK + MB_ICONEXCLAMATION);
	}

	return 0;
}


// *************************************************************************************************************************
// login ToolAdaptor
// *************************************************************************************************************************
long _loginToolAdaptor(CString type) {
	try {
		if (ta) delete ta;
		if (uml_ta) delete uml_ta;
		
		if (!type.Compare("")) {
			ta_types_dlg types_dlg;
			int nResponse = types_dlg.DoModal();
			if (nResponse == IDCANCEL) return 1;

			ta = new my_ToolAdaptor((LPCSTR)types_dlg.m_ta_type, __argc, __argv);
		} else {
			ta = new my_ToolAdaptor((LPCSTR)type, 0, NULL);
		}

		uml_ta = new my_UML_ToolAdaptor(0, NULL);
		uml_ta->getDatanetwork(Uml::diagram);

		remote_paradigms.clear();
		ta->getWorkingParadigms(remote_paradigms);

		config.clear();
		ta->getTAConfig(config);

		subscribable.clear();
		ta->getTASubscribable(subscribable);

	} catch (udmcorba_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		return -1;
	} catch (udm_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		return -1;
	}

	return 0;
}

// *************************************************************************************************************************
// fetching Thread
// *************************************************************************************************************************
CWinThread *fetching_thread;

UINT MyFetchingThread( LPVOID pParam ) {

	::CoInitialize(NULL);

	try {

		ARG_LIST *arg = (ARG_LIST*) pParam;
		string name = (string) arg->name;
		string version = (string) arg->version;
		w_parent = (HWND) arg->parent;
		string type = arg->type;

		Uml::Diagram paradigm_diagram = uml_ta->remoteFetchMyParadigm(name.c_str(), version.c_str());

		Udm::UdmDiagram udmDataDiagram;
		udmDataDiagram.dgr = &paradigm_diagram;
		udmDataDiagram.init = dummy;

		my_ToolAdaptor *ta = new my_ToolAdaptor(type.c_str(), __argc, __argv, paradigm_diagram, subscriptions_id);

		SUBSCRIPTION subscr;
		subscr.paradigm.name = name;
		subscr.paradigm.version = version;
		subscr.fetch_ta = ta;
		subscriptions.insert(pair<const long, SUBSCRIPTION>(subscriptions_id++, subscr));

		delete arg;
		
		ta->subscribe(udmDataDiagram, name.c_str(), version.c_str());
		ta->run();

		ta->logout();
		delete ta;

	} catch (udm_exception &e) {
		if (main_dlg) main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		return -1;
	} catch (udmcorba_exception &e) {
		if (main_dlg) main_dlg->MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		return -1;
	}

	return 0;
}


/////////////////////////////////////////////////////////////////////////////
// CMainDialog dialog


CMainDialog::CMainDialog(CWnd* pParent /*=NULL*/)
	: CDialog(CMainDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(CMainDialog)
	//}}AFX_DATA_INIT
}


void CMainDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CMainDialog)
	DDX_Control(pDX, IDC_LIST1, m_status);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CMainDialog, CDialog)
	//{{AFX_MSG_MAP(CMainDialog)
	ON_BN_CLICKED(IDC_CLEAR, OnClear)
	ON_BN_CLICKED(IDC_PUBLISH, OnPublish)
	ON_BN_CLICKED(IDC_SUBSCRIBE, OnSubscribe)
	ON_BN_CLICKED(IDC_UNSUBSCRIBE, OnUnsubscribe)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

void getDocumentsInfo() {
	existing_documents.clear();
	existing_documents = ta->getDocuments();
}


/////////////////////////////////////////////////////////////////////////////
// CMainDialog message handlers

// *************************************************************************************************************************
//	Publish a document
// *************************************************************************************************************************
long _publishDocument(CString filename) {

	::CoInitialize(NULL);

	if (remote_paradigms.size()==0) {
		if (main_dlg) main_dlg->MessageBox("There is no available paradigms.", "Error", MB_OK + MB_ICONEXCLAMATION);
		return -2;
	}
	
	// get the publishable files into a list
	list<CString> files;
	int offset = filename.Find("|");
	while (offset!=-1) {
		CString temp = filename.Left(offset);
		files.push_back(temp);
		filename = filename.Mid(offset+1, filename.GetLength());
		offset = filename.Find("|");
	}

	// decide what to publish (single/multiple)
	list<PARADIGM_STRUCT> paradigms;
	publishing_multiple_documents = false;
	for(list<CONFIG>::iterator i=config.begin(); i!=config.end(); i++) {
		paradigms = (*i).paradigms;		
		if (paradigms.size()>1 && config.size()>1) {
			if (main_dlg) main_dlg->MessageBox("More types of publishable multiple and single documents is not supported yet!", "Error", MB_OK + MB_ICONEXCLAMATION);
			return -3;
		}

		if (paradigms.size()!=1) publishing_multiple_documents = true;
	}
	

	CPublish_DLG publish_dlg;
	publish_dlg.m_updated_document_id = -1;

	try {
		CString filename;
		CString filetitle;
		CString fileext;
		int result = publish_dlg.DoModal();
		if (result == IDOK) {
			CString name = publish_dlg.m_name;
			CString version = publish_dlg.m_version;
			CString remarks = publish_dlg.m_remarks;
			BOOL xslt_support = publish_dlg.m_xslt_support;
			long updated_document_id = publish_dlg.m_updated_document_id;

			long keepAlive = atol(publish_dlg.m_keepalive);
			switch (publish_dlg.m_time) {
			case 1 : keepAlive *= 60; break;
			case 2 : keepAlive *= 3600; break;
			}

			// Publishing simple document
			if (!publishing_multiple_documents) {
			
				if (files.size()!=1) return -4;
				CString filename = *files.begin();

				int offset = filename.ReverseFind('\\');
				CString _file = filename.Mid(offset+1);
				filetitle = _file.Left(_file.GetLength()-4);
				fileext = _file.Right(3);

				CString xsl_file = "";

				if (xslt_support && !fileext.Compare("xml")) {
					CFileDialog open_xsl_dlg(TRUE, "xsl", "");
					open_xsl_dlg.m_ofn.lpstrTitle = "Open the XSLT file for the document";
					int result = open_xsl_dlg.DoModal();
					if (result==IDCANCEL) return 0;
					xsl_file = open_xsl_dlg.GetPathName();
				}

				map<long, PARADIGM_STRUCT>::iterator i = remote_paradigms.find(publish_dlg.m_paradigm);
				if (i==remote_paradigms.end()) return 0;

				string par_name = (*i).second.name;
				string par_version = (*i).second.version;
				
				// if it is fetched remotely we will use that Diagram and paradigm
				for(list<REGISTERED_PARADIGM>::iterator j=registered_paradigms.begin(); j!=registered_paradigms.end(); j++) {
					if (!par_name.compare((*j).name) && !par_version.compare((*j).version)) {
						udmDataDiagram = (*j).udmDataDiagram;
						paradigm_diagram = (*j).paradigm_diagram;
					}
				}

				// if it is not fetched yet, we fetch it and put the paradigm into our store to use it now and later
				if (!ta->isRegisteredDatanetwork((*i).second.name.c_str(), (*i).second.version.c_str())) {
					paradigm_diagram = uml_ta->remoteFetchMyParadigm((*i).second.name.c_str(), (*i).second.version.c_str());
					udmDataDiagram.dgr = &paradigm_diagram;
					udmDataDiagram.init = dummy;

					REGISTERED_PARADIGM temp_reg_par;
					temp_reg_par.name = par_name;
					temp_reg_par.version = par_version;
					temp_reg_par.paradigm_diagram = paradigm_diagram;
					temp_reg_par.udmDataDiagram = udmDataDiagram;
					registered_paradigms.push_front(temp_reg_par);
				}


				// XSLT translation
				if (xsl_file.Compare("") && !fileext.Compare("xml")) {

					// Display the progress bar for the XSLT pre-process process
					ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
					ta->progress_dlg.SetWindowText("XSLT pre-processing:");
					ta->progress_dlg.ShowWindow(SW_SHOW);

					// do the trasnform
					int result = 0;
					try {
						OTIF_XSLT::XSLTransformer xslt;
						result = xslt.executeTransform((LPCSTR)filename, (LPCSTR)xsl_file, \
							(LPCSTR)("__temp_" + filetitle + "_xslt." + fileext), false, "");

					} catch (OTIF_XSLT::xslt_exception &e) {
						if (main_dlg) main_dlg->MessageBox(e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
						else MessageBox(NULL, e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
						ta->progress_dlg.DestroyWindow();
						return -6;
					}

					if (result!=0) {
						if (main_dlg) main_dlg->MessageBox("Could not transform the document with the XSLT script", "Error", MB_OK + MB_ICONEXCLAMATION);
						ta->progress_dlg.DestroyWindow();
						return -6;
					}

					ta->progress_dlg.m_progress.SetPos(100);

					filename = "__temp_" + filetitle + "_xslt." + fileext;

					ta->progress_dlg.DestroyWindow();
				}

				// Display the progress bar for the publishing process
				ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
				ta->progress_dlg.SetWindowText("Publishing document:");
				ta->progress_dlg.ShowWindow(SW_SHOW);

				UdmCORBA::CORBADataNetwork *dn;
				try {
					// create DN with the remote paradigm
					dn = ta->getDatanetwork(udmDataDiagram, (*i).second.name.c_str(), (*i).second.version.c_str());
					// open the publishable document
					dn->OpenExisting((LPCSTR)filename, "", Udm::CHANGES_LOST_DEFAULT);
				} catch (udm_exception &e) {
					string error = e.what();
					if (!error.compare("Not found")) MessageBox(NULL, "Invalid input file!", "Error", MB_OK + MB_ICONEXCLAMATION);
					else MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);

					ta->progress_dlg.DestroyWindow();
					return -100;
				}

				// attachment files
				for(map<string, string>::iterator addfile = attachments.begin(); addfile!=attachments.end(); addfile++) {
					dn->addAdditionalFile( (*addfile).first.c_str(), (*addfile).second.c_str() );
				}

				if (main_dlg) main_dlg->m_status.AddString("Start sending the document...");
				// publish the document
				dn->publishDocument((LPCSTR)name, (LPCSTR)version, (LPCSTR)remarks, keepAlive, updated_document_id);
				// close the document
				dn->CloseNoUpdate();

				ta->progress_dlg.DestroyWindow();

				// delete temporary docs after XSLT transform
				if (xsl_file.Compare("")) DeleteFile((LPCSTR)filename);

				if (main_dlg) main_dlg->m_status.AddString("Document is successfully sent.");
				if (main_dlg) main_dlg->m_status.AddString("");


				
			// Publishing multiple document
			} else { 

				if (files.size()<2) return -5;

				list<UdmCORBA::CORBADataNetwork*> docs;
				list<CString> temp_xsl_files;
				
				for(list<PARADIGM_STRUCT>::iterator i=paradigms.begin(); i!=paradigms.end(); i++) {
					CString filename = "";

					Choose_paradigm_DLG choose_dlg;
					choose_dlg.window_title = "For the " + (CString)((*i).name.c_str()) + " v" + (CString)((*i).version.c_str()) + " paradigm";
					for(list<CString>::iterator ff=files.begin(); ff!=files.end(); ff++) choose_dlg.files.push_back(*ff);
					int result = choose_dlg.DoModal();
					if (result==IDOK) filename = choose_dlg.m_file;
					else return 0;					
					
					int offset = filename.ReverseFind('\\');
					CString _file = filename.Mid(offset+1);
					filetitle = _file.Left(_file.GetLength()-4);
					fileext = _file.Right(3);
				
					CString xsl_file = "";

					if (xslt_support) {

						CFileDialog open_xsl_dlg(TRUE, "xsl", "");
						open_xsl_dlg.m_ofn.lpstrTitle = "Open the XSLT file for the " + (CString)((*i).name.c_str()) + " v" + (CString)((*i).version.c_str()) + " document";
						int result = open_xsl_dlg.DoModal();
						if (result==IDOK) xsl_file = open_xsl_dlg.GetPathName();
					}

					Uml::Diagram paradigm_diagram = uml_ta->remoteFetchMyParadigm((*i).name.c_str(), (*i).version.c_str());
					Udm::UdmDiagram udmDataDiagram;
					udmDataDiagram.dgr = &paradigm_diagram;
					udmDataDiagram.init = dummy;

					// XSLT translation
					if (xsl_file.Compare("") && !fileext.Compare("xml")) {

						// Display the progress bar for the XSLT pre-process process
						ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
						ta->progress_dlg.SetWindowText("XSLT pre-processing:");
						ta->progress_dlg.ShowWindow(SW_SHOW);

						// do the trasnform
						int result = 0;
						try {
							OTIF_XSLT::XSLTransformer xslt;
							result = xslt.executeTransform((LPCSTR)filename, (LPCSTR)xsl_file, \
								(LPCSTR)("__temp_" + filetitle + "_xslt." + fileext), false, "");

						} catch (OTIF_XSLT::xslt_exception &e) {
							if (main_dlg) main_dlg->MessageBox(e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
							else MessageBox(NULL, e.what(), "XSLT Error", MB_OK + MB_ICONEXCLAMATION);
							ta->progress_dlg.DestroyWindow();
							return -6;
						}

						if (result!=0) {
							if (main_dlg) main_dlg->MessageBox("Could not transform the document with the XSLT script", "Error", MB_OK + MB_ICONEXCLAMATION);
							ta->progress_dlg.DestroyWindow();
							return -6;
						}

						ta->progress_dlg.m_progress.SetPos(100);

						filename = "__temp_" + filetitle + "_xslt." + fileext;
						temp_xsl_files.push_back(filename);

						ta->progress_dlg.DestroyWindow();
					}

					UdmCORBA::CORBADataNetwork *dn;
					try {
						// create DN with the remote paradigm
						dn = ta->getDatanetwork(udmDataDiagram, (*i).name.c_str(), (*i).version.c_str());
						// open the publishable document
						dn->OpenExisting((LPCSTR)filename, "", Udm::CHANGES_LOST_DEFAULT);
					} catch (udm_exception &e) {
						string error = e.what();
						if (!error.compare("Not found")) MessageBox(NULL, "Invalid input file!", "Error", MB_OK + MB_ICONEXCLAMATION);
						else MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);

						ta->progress_dlg.DestroyWindow();
						return -100;
					}

					docs.push_back(dn);
				}

				// Display the progress bar for the publishing process
				ta->progress_dlg.Create(IDD_PROGRESS_BAR, main_dlg);
				ta->progress_dlg.SetWindowText("Publishing document:");
				ta->progress_dlg.ShowWindow(SW_SHOW);


				// attachment files
				for(map<string, string>::iterator addfile = attachments.begin(); addfile!=attachments.end(); addfile++) {
					(*docs.begin())->addAdditionalFile( (*addfile).first.c_str(), (*addfile).second.c_str() );
				}


				ta->publishMultipleDocuments(docs, (LPCSTR)name, (LPCSTR)version, (LPCSTR)remarks, keepAlive, updated_document_id);

				for(list<UdmCORBA::CORBADataNetwork*>::iterator q=docs.begin(); q!=docs.end(); q++) (*q)->CloseNoUpdate();

				ta->progress_dlg.DestroyWindow();

				// delete temporary docs after XSLT transform
				for(list<CString>::iterator txf=temp_xsl_files.begin(); txf!=temp_xsl_files.end(); txf++) DeleteFile((LPCSTR)(*txf));
				
				if (main_dlg) main_dlg->m_status.AddString("Document is successfully sent.");
				if (main_dlg) main_dlg->m_status.AddString("");
			} // publishing_multiple_documents
		
		} // if IDOK

	} catch (udmcorba_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
	} catch (udm_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
	}

	return 0;
}


void CMainDialog::OnPublish() 
{
	// TODO: Add your control notification handler code here
	_publishDocument("");
}


// *************************************************************************************************************************
//	Subscribe to a paradigm
// *************************************************************************************************************************

long _getParadigmKey(CString name, CString version) {
	for(map<long, PARADIGM_STRUCT>::iterator i=remote_paradigms.begin(); i!=remote_paradigms.end(); i++) {
		if (!name.Compare((*i).second.name.c_str()) && !version.Compare((*i).second.version.c_str())) 
			return (*i).first;
	}
	return -1;
}

long _subscribeParadigm(CString name, CString version, HWND parent) {

	try {
		long key;

		if (!name.Compare("") && !version.Compare("")) {
			CSubscribe_DLG subscribe_dlg;
			int result = subscribe_dlg.DoModal();
			if (result == IDCANCEL) return 0;
			key = subscribe_dlg.m_paradigm;
		} else {
			key = _getParadigmKey(name, version);
			if (key == -1) return -1;
		}

		map<long, PARADIGM_STRUCT>::iterator i = remote_paradigms.find(key);
		if (i!=remote_paradigms.end()) {

			ARG_LIST *arg = new ARG_LIST();
			arg->parent = parent;
			arg->name = (*i).second.name;
			arg->version = (*i).second.version;		
			arg->type = ta->getToolAdaptorType();

			fetching_thread = AfxBeginThread(MyFetchingThread, arg);
				
			if (main_dlg) {
				main_dlg->m_status.AddString(("Subscribe to " + (*i).second.name + " v" + (*i).second.version + " paradigm").c_str());
				main_dlg->m_status.AddString("");
			}
		}

	} catch (udmcorba_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
	} catch (udm_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
	}

	return 0;
}


void CMainDialog::OnSubscribe() 
{
	// TODO: Add your control notification handler code here
	_subscribeParadigm("", "", m_hWnd);
}

// *************************************************************************************************************************
// Unsubscribe a paradigm
// *************************************************************************************************************************

long _unsubscribeParadigm(CString name, CString version) {

	if (!name.Compare("") && !version.Compare("")) {
		CUnsubscribe_DLG unsub_dlg;
		int result = unsub_dlg.DoModal();
		if (result==IDCANCEL || unsub_dlg.m_paradigms==-1) return 0;

		for(map<long, SUBSCRIPTION>::iterator j=subscriptions.begin(); j!=subscriptions.end(); j++) {
			if (j->second.list_offset==unsub_dlg.m_paradigms) {
				j->second.fetch_ta->terminate();

				if (main_dlg) {
					main_dlg->m_status.AddString(("Unsubscribed the " + j->second.paradigm.name + " v" 
						+ j->second.paradigm.version + " paradigm").c_str());
					main_dlg->m_status.AddString("");
				}
				subscriptions.erase(j);
			}
		}
	
	} else {
		for(map<long, SUBSCRIPTION>::iterator j=subscriptions.begin(); j!=subscriptions.end(); j++) {
			if (!name.Compare(j->second.paradigm.name.c_str()) && !version.Compare(j->second.paradigm.version.c_str())) {
				j->second.fetch_ta->terminate();
				subscriptions.erase(j);
				return 0;
			}
		}
		return -1;
	}

	return 0;
}

void CMainDialog::OnUnsubscribe() 
{
	// TODO: Add your control notification handler code here
	_unsubscribeParadigm("", "");
}

// *************************************************************************************************************************
// Clear status listbox
// *************************************************************************************************************************
void CMainDialog::OnClear() 
{
	// TODO: Add your control notification handler code here
	m_status.ResetContent();
}
