/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

// Publish_DLG.cpp : implementation file
//
#include "UdmCORBA.h"
#include "stdafx.h"
#include "Generic_ToolAdaptor_ActiveX.h"
#include "Publish_DLG.h"
#include "update_document_dlg.h"

using namespace UdmCORBA;

extern map<long, PARADIGM_STRUCT> remote_paradigms;
extern bool publishing_multiple_documents;
extern map<string, string> attachments; // filename, directory

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CPublish_DLG dialog


CPublish_DLG::CPublish_DLG(CWnd* pParent /*=NULL*/)
	: CDialog(CPublish_DLG::IDD, pParent)
{
	//{{AFX_DATA_INIT(CPublish_DLG)
	m_name = _T("");
	m_keepalive = _T("");
	m_remarks = _T("");
	m_version = _T("");
	m_paradigm = -1;
	m_time = -1;
	m_xslt_support = FALSE;
	m_update_doc = FALSE;
	//}}AFX_DATA_INIT
}


void CPublish_DLG::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CPublish_DLG)
	DDX_Control(pDX, IDC_ATTACHMENTS, m_attachments);
	DDX_Control(pDX, IDC_TIME, m_time_control);
	DDX_Control(pDX, IDC_PARADIGM, m_paradigm_control);
	DDX_Text(pDX, IDC_EDIT1, m_name);
	DDX_Text(pDX, IDC_KEEPALIVE, m_keepalive);
	DDX_Text(pDX, IDC_REMARKS, m_remarks);
	DDX_Text(pDX, IDC_VERSION, m_version);
	DDX_CBIndex(pDX, IDC_PARADIGM, m_paradigm);
	DDX_CBIndex(pDX, IDC_TIME, m_time);
	DDX_Check(pDX, IDC_CHECK1, m_xslt_support);
	DDX_Check(pDX, IDC_CHECK2, m_update_doc);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CPublish_DLG, CDialog)
	//{{AFX_MSG_MAP(CPublish_DLG)
	ON_CBN_SELCHANGE(IDC_PARADIGM, OnSelchangeParadigm)
	ON_CBN_SELCHANGE(IDC_TIME, OnSelchangeTime)
	ON_BN_CLICKED(IDC_ADD, OnAdd)
	ON_BN_CLICKED(IDC_DELETE, OnDelete)
	ON_BN_CLICKED(IDC_CHECK2, OnCheck2)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPublish_DLG message handlers

void CPublish_DLG::OnOK() 
{
	// TODO: Add extra validation here
	UpdateData(true);

	for(int i=0; i<m_version.GetLength(); i++) {
		char c = m_version.GetAt(i);
		long l_c = atol(&c);
		if (c!='.' && c!='0' && l_c==0) {
			m_version.Remove(c); 
			i--;
		}
	}

	for(i=0; i<m_keepalive.GetLength(); i++) {
		char c = m_keepalive.GetAt(i);
		long l_c = atol(&c);
		if (c!='0' && l_c==0) {
			m_keepalive.Remove(c); 
			i--;
		}
	}

	UpdateData(false);

	if (m_name.GetLength()==0) {
		MessageBox("You did NOT specify the name of the datanetwork!", "Warning", MB_OK + MB_ICONWARNING);
		return;
	}
	if (m_version.GetLength()==0) {
		MessageBox("You did NOT specify the version of the datanetwork!", "Warning", MB_OK + MB_ICONWARNING);
		return;
	}
	
	long l = atol((LPCSTR)m_keepalive);

/*	
	if (l<10 && m_time_control.GetCurSel()==0) { 
		int result = MessageBox("The keepAlive parameter is too small, so maybe the datanetwork will be destroyed before it can be processed.\n\nContinue?", "Warning", MB_OKCANCEL + MB_ICONQUESTION);
		if (result==IDOK) CDialog::OnOK();
	} else {
		CDialog::OnOK();
	}
*/

	CDialog::OnOK();
}

BOOL CPublish_DLG::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// TODO: Add extra initialization here
	m_time_control.AddString("sec");
	m_time_control.AddString("min");
	m_time_control.AddString("hour");

	if (publishing_multiple_documents) {
		m_paradigm_control.EnableWindow(false);
		SetWindowText("Publishing multiple documents");
	}

	// default values
	m_version = "1.0";
	m_keepalive = "1";
	UpdateData(false);

	m_time_control.SetCurSel(1);

	for(map<long, PARADIGM_STRUCT>::iterator i=remote_paradigms.begin(); i!=remote_paradigms.end(); i++) {
		CString item = (CString)(*i).second.name.c_str() + " v" + (CString)(*i).second.version.c_str();
		m_paradigm_control.AddString(item);
	}
	if (!publishing_multiple_documents) m_paradigm_control.SetCurSel(0);

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

void CPublish_DLG::OnSelchangeParadigm() 
{
	// TODO: Add your control notification handler code here
	m_paradigm = m_paradigm_control.GetCurSel();
}

void CPublish_DLG::OnSelchangeTime() 
{
	// TODO: Add your control notification handler code here
	m_time = m_time_control.GetCurSel();
}

/*
void CPublish_DLG::OnXsltSelection() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);

	CFileDialog open_dlg(TRUE, "xsl", "");
	open_dlg.m_ofn.lpstrTitle = "Open an XSLT file...";
	int result = open_dlg.DoModal();
	if (result==IDOK) m_xslt_url = open_dlg.GetPathName();
	else m_xslt_url = "";

	UpdateData(false);
	
}
*/
void CPublish_DLG::OnAdd() 
{
	// TODO: Add your control notification handler code here
	CFileDialog open_dlg(TRUE, "", "", OFN_ALLOWMULTISELECT);
	open_dlg.m_ofn.lpstrTitle = "Select the attachment files";

	// need larger filename buffer to support multiselect
	TCHAR filenameBuffer[16384] = { 0 } ;
	open_dlg.m_ofn.lpstrFile = filenameBuffer;
	open_dlg.m_ofn.nMaxFile = sizeof(filenameBuffer) / sizeof(filenameBuffer[0]);

	int result = open_dlg.DoModal();
	if (result==IDOK) {
		POSITION pos = open_dlg.GetStartPosition();
		while (pos != NULL) {
			CString attachment_file = open_dlg.GetNextPathName(pos);
			int offset = attachment_file.ReverseFind('\\');
			attachments.insert( pair<string, string>( (string)attachment_file.Right(attachment_file.GetLength()-(offset+1)), (string)attachment_file.Left(offset+1) ) );
			m_attachments.AddString(attachment_file.Right(attachment_file.GetLength()-(offset+1)));
		}
	}
}

void CPublish_DLG::OnDelete() 
{
	// TODO: Add your control notification handler code here
	int selected = m_attachments.GetCurSel();
	if (selected != LB_ERR) {
		CString text;
		m_attachments.GetText(selected, text);
		attachments.erase((string)text);
		m_attachments.DeleteString(selected);
	}
}

void CPublish_DLG::OnCheck2() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
	if (!m_update_doc) {
		m_updated_document_id = -1;
		return;
	}

	Update_Document_dlg update_dlg;

	int result = update_dlg.DoModal();
	if (result==IDOK) {
		m_updated_document_id = update_dlg.m_updated_document_id;
	} else {
		m_update_doc = false;
	}

	UpdateData(false);
}
