#include "Uml.h"
#include "UdmGme.h"
#include "UdmDom.h"
#include "UdmCORBA.h"

using namespace UdmCORBA;

void dummy(void) {};
Udm::UdmDiagram udmDataDiagram;
Uml::Diagram paradigm_diagram;

string file_to_save;
string paradigm_locator;

// *************************************************************************************************************************
//	my_UML_ToolAdaptor
// *************************************************************************************************************************
class my_UML_ToolAdaptor : public ToolAdaptor {
public:

	my_UML_ToolAdaptor(int argc, char **argv) :
		ToolAdaptor(argc, argv) {};

	void notify(const char *paradigm_name, const char *paradigm_version, \
		const char *name, const char *version, const char *remarks, long docId, long keepAlive) {};
};
my_UML_ToolAdaptor *uml_ta = NULL;

// *************************************************************************************************************************
// my_ToolAdaptor
// *************************************************************************************************************************
class my_ToolAdaptor : public UdmCORBA::ToolAdaptor {

public:
	my_ToolAdaptor(const char *type, int argc, char **argv) 
		: ToolAdaptor(type, argc, argv) {};

	void notify(const char *par_name, const char *par_version, const char *name, const char *version, const char *remarks, long docId, long keepAlive) {
	
		// Get the rootobject
		Uml::Class rootobject = getDocumentRootobjectMeta(docId, paradigm_diagram);
		if (!rootobject) {
			cerr << "Could not find the meta of the rootobject of the document!" << endl;
			return;
		}

		cout << endl << "***Restore CORBADataNetwork***" << endl;

		dn->CreateNew(file_to_save.c_str(), paradigm_locator.c_str(), rootobject, Udm::CHANGES_PERSIST_ALWAYS);
		dn->fetchDocument(docId);

		cout << endl << "DataNetwork pulled..." << endl;

		dn->SaveAs(file_to_save.c_str());
		dn->CloseNoUpdate();
		dn->release();

		terminate();
	};

	void getPublishStatus(long percent) {
		cout << "#";
	}

	void getFetchStatus(long percent) {
		cout << "#";
	}
};
my_ToolAdaptor *ta = NULL;


// *************************************************************************************************************************
// main
// *************************************************************************************************************************
int main(int argc, char** argv) {

	if (argc<6) {
		cout << "Usage: GTA <tooladaptor type> <paradigm_locator> <paradigm_name> <paradigm_version> <filename> [<keepAlive in second>] [<additional files list>]" << endl;
		exit(1);
	}


	try {
		uml_ta = new my_UML_ToolAdaptor(argc, argv);
		uml_ta->getDatanetwork(Uml::diagram);

		ta = new my_ToolAdaptor(argv[1], argc, argv);
	
		paradigm_diagram = uml_ta->remoteFetchMyParadigm(argv[3], argv[4]);
		udmDataDiagram.dgr = &paradigm_diagram;
		udmDataDiagram.init = dummy;


		// Fetch a document
		if (argc==6) {

			paradigm_locator = argv[2];
			file_to_save = argv[5];

			ta->subscribe(udmDataDiagram, argv[3], argv[4]);
			ta->run();


		// Publish a document
		} else if (argc>6) {

			// create datanetwork
			UdmCORBA::CORBADataNetwork *dn = ta->getDatanetwork(udmDataDiagram, argv[3], argv[4]);

			// open the document
			dn->OpenExisting(argv[5], "", Udm::CHANGES_LOST_DEFAULT);


			// add additional files
			for(int i=7; i<argc; i++) {
				dn->addAdditionalFile(argv[i]);
			}
			
			// publish the document
			dn->publishDocument(argv[5], "1.0", "GTA published document", atol(argv[6]));

			// close the document
			dn->CloseNoUpdate();

			cout << endl << "DataNetwork pushed..." << endl;
		}
		
	} catch (udm_exception &e) {
		cerr << e.what() << endl;
	} catch (udmcorba_exception &e) {
		if ( e.code()<11 ) {
			cerr << endl << "CORBA exception: " << e.what() << endl;
		} else if (e.code()<21) {
			cerr << endl << "Paradigm exception: " << e.what() << endl;
		} else if (e.code()<41) {
			cerr << endl << "UdmCORBA exception: " << e.what() << endl;
		} else if (e.code()<91) {
			cerr << endl << "ToolAdaptor exception: " << e.what() << endl;
		}
	}

	return 0;
}
