/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

#ifndef __ToolAdaptor_XSLT_SUPPORT__H
#define __ToolAdaptor_XSLT_SUPPORT__H

#include "Uml.h"
#include "UmlExt.h"

#include <time.h>
#include <string>
#include <map>
#include <exception>

using namespace Uml;


namespace OTIF_XSLT {

	// XSLT transformer class
	class XSLTransformer {

	public:
		XSLTransformer() {};
		~XSLTransformer() {};

		int executeTransform(const char *input, const char *xslt, const char *output, const bool generate_xsd, const char *rootclass);

	};

	//generates complete DTD file
	void GenerateDTDFile(const Uml::Diagram &diagram, const char *dtd_file);

	//generates complete XML Schema file
	void GenerateXMLSchemaFile(const Uml::Diagram &diagram, const char *xsd_file);


	// OTIF XSLT exception handling
	class xslt_exception : public exception {

	public:
		xslt_exception() throw();
		xslt_exception(const xslt_exception &e) throw() : description(e.description), exception_code(e.exception_code) {}
		xslt_exception(const string &d, const int c) throw() : description(d), exception_code(c) {}
		xslt_exception(const char *d, const int c) throw() : description(d), exception_code(c) {}

		const xslt_exception &operator =(const xslt_exception &e) throw() { 
			description = e.description; 
			exception_code = e.exception_code;
		}

		virtual ~xslt_exception() {}

		virtual const char *what() const throw() { 
			return description.c_str(); 
		}

		virtual const int code() const throw() {
			return exception_code;
		}

	protected:
		int exception_code;
		string description;
	};

};

#endif