/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

// ManagerDlg.cpp : implementation file
//
#include "uml.h"
#include "UdmCORBA.h"



#include "stdafx.h"
#include "Manager.h"
#include "ManagerDlg.h"
#include "ChangePassword_DLG.h"
#include "about_dlg.h"

#include <map>


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
using namespace UdmCORBA;
// *************************************************************************************************************************
// Metamodel & Translator & Datanetwork repository
// *************************************************************************************************************************
map<long, PARADIGM_STRUCT> metamodel_rep;
map<long, TRANSLATOR_STRUCT> translator_rep;
map<long, DOCUMENT_STRUCT> datanetwork_rep;

ManagerObject *manager;

/////////////////////////////////////////////////////////////////////////////
// CManagerDlg dialog

CManagerDlg::CManagerDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CManagerDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CManagerDlg)
	m_name = _T("");
	m_version = _T("");
	m_url = _T("");
	m_locked = FALSE;
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

void CManagerDlg::OnOK() {};

// *************************************************************************************************************************
// logout
// *************************************************************************************************************************
void CManagerDlg::OnCancel() {
	try {
		if (manager) manager->logout();

    } catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error" , MB_OK + MB_ICONEXCLAMATION);
	} catch (...) {
		MessageBox("Error: could not communicate with the server!", "Error", MB_OK + MB_ICONEXCLAMATION);
	}

	CDialog::OnCancel();
};

void CManagerDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CManagerDlg)
	DDX_Control(pDX, IDC_DATANETWORKS, m_datanetwork_listbox);
	DDX_Control(pDX, IDC_LIST1, m_translator_listbox);
	DDX_Control(pDX, IDC_LIST3, m_listbox);
	DDX_Text(pDX, IDC_EDIT1, m_name);
	DDX_Text(pDX, IDC_EDIT2, m_version);
	DDX_Text(pDX, IDC_EDIT4, m_url);
	DDX_Check(pDX, IDC_CHECK1, m_locked);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CManagerDlg, CDialog)
	//{{AFX_MSG_MAP(CManagerDlg)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_BN_CLICKED(IDC_BUTTON2, OnButton2)
	ON_LBN_SELCHANGE(IDC_LIST3, OnSelchangeList3)
	ON_BN_CLICKED(IDC_CHECK1, OnCheck1)
	ON_BN_CLICKED(IDC_BUTTON7, OnButton7)
	ON_BN_CLICKED(IDC_BUTTON8, OnButton8)
	ON_BN_CLICKED(IDC_BUTTON9, OnButton9)
	ON_BN_CLICKED(IDC_BUTTON10, OnButton10)
	ON_BN_CLICKED(IDC_BUTTON11, OnButton11)
	ON_BN_CLICKED(IDC_BUTTON1, OnAbout)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CManagerDlg message handlers

// *************************************************************************************************************************
// init
// *************************************************************************************************************************
BOOL CManagerDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	
	// TODO: Add extra initialization here
	m_version = "1.0";
	m_translator_version = "1.0";

	UpdateData(false);

	m_listbox.ResetContent();
	metamodel_rep.clear();
	char text[1024];

	try {
		vector<PARADIGM_STRUCT> ps = manager->backplane_paradigms();

		int j = 0;
		for(vector<PARADIGM_STRUCT>::iterator i=ps.begin(); i!=ps.end(); i++, j++) {
			sprintf(text, "%s v%s", (*i).name.c_str(), (*i).version.c_str());
			m_listbox.AddString(text);
			metamodel_rep[j] = (*i);
		}
	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		exit(-1);
	}
	
	m_translator_listbox.ResetContent();
	translator_rep.clear();
	try {
		vector<TRANSLATOR_STRUCT> ps = manager->backplane_translators();

		int j = 0;
		for(vector<TRANSLATOR_STRUCT>::iterator i=ps.begin(); i!=ps.end(); i++, j++) {
			sprintf(text, "%s v%s  Inputparadigm:%s v%s  Outputparadigm:%s v%s", 
				(*i).name.c_str(), (*i).version.c_str(), (*i).fromName.c_str(), (*i).fromVersion.c_str(), 
				(*i).toName.c_str(), (*i).toVersion.c_str());
			m_translator_listbox.AddString(text);
			translator_rep[j] = (*i);
		}
	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		exit(-1);
	}

	m_datanetwork_listbox.ResetContent();
	datanetwork_rep.clear();
	try {
		vector<DOCUMENT_STRUCT> ps = manager->backplane_documents();

		int j = 0;
		for(vector<DOCUMENT_STRUCT>::iterator i=ps.begin(); i!=ps.end(); i++, j++) {

			if ((*i).paradigms.size() > 1) {
				string temp_paradigms;
				for (list<PARADIGM_STRUCT>::iterator q=(*i).paradigms.begin(); q!=(*i).paradigms.end(); q++) {
					if (++q==(*i).paradigms.end()) {
						--q;
						temp_paradigms += (*q).name + " v" + (*q).version;
					} else {
						--q;
						temp_paradigms += (*q).name + " v" + (*q).version + ",";
					}
				}	
				sprintf(text, "Name:%s  Version:%s  Remarks:%s  keepAlive:%u sec  Paradigm:(%s)", 
				(*i).name.c_str(), (*i).version.c_str(), (*i).remarks.c_str(), (*i).keepAlive, temp_paradigms.c_str());
			} else {
				sprintf(text, "Name:%s  Version:%s  Remarks:%s  keepAlive:%u sec  Paradigm:%s %s", 
				(*i).name.c_str(), (*i).version.c_str(), (*i).remarks.c_str(), (*i).keepAlive, 
				(*(*i).paradigms.begin()).name.c_str(), (*(*i).paradigms.begin()).version.c_str());
			}
				
			m_datanetwork_listbox.AddString(text);
			datanetwork_rep[j] = (*i);
		}
	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		exit(-1);
	}

	return true;  // return TRUE  unless you set the focus to a control
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.
void CManagerDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CManagerDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}


// *************************************************************************************************************************
// Upload metamodel button
// *************************************************************************************************************************
void CManagerDlg::OnButton2() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);

	if (m_url.GetLength()>6 && m_url.Find("http://")==0) {
	// have to download the paradigm through http, and send to the backplane via datanetwork

		string filename = m_name + "_v" + m_version + ".xml";
		
		HRESULT hr = URLDownloadToFile(NULL, m_url, filename.c_str(), 0, NULL);
		if (SUCCEEDED(hr)) {
			
			MessageBox("Download succeded. Paradigm will be registered.", "Success", MB_OK + MB_ICONINFORMATION);

			// Setup the parameters of the Paradigm
			PARADIGM_STRUCT paradigm;
			paradigm.name = m_name;
			paradigm.version = m_version;

			try {
				manager->registerMetaModel(paradigm, filename.c_str());
			} catch (udm_exception &e) {
				MessageBox(e.what(), "Udm error", MB_OK);
			} catch (udmcorba_exception &e) {
				MessageBox(e.what(), "UdmCORBA error", MB_OK);
			}

			DeleteFile(filename.c_str());

		} else {
			MessageBox("Wrong URL! Paradigm could not be downloaded.", "Error", MB_OK + MB_ICONEXCLAMATION);
		}


	} else if (m_url.GetLength()>7 && m_url.Find("local://")==0) {
	// paradigm is a local file on the manager client, have to send to the backplane via datanetwork.

		// Setup the parameters of the Paradigm
		PARADIGM_STRUCT paradigm;
		paradigm.name = m_name;
		paradigm.version = m_version;

		string filename = m_url.Mid(8);

		// Because of remote mem type paradigm BUG in the Backplane
		if (filename.length()<5) {
			MessageBox("Wrong filename!", "Error", MB_OK + MB_ICONEXCLAMATION);
			return;
		}
		string suffix = filename.substr(filename.length()-4, filename.length());
		if (suffix.compare(".xml")) {
			MessageBox("File should be xml type!", "Error", MB_OK + MB_ICONEXCLAMATION);
			return;
		}
		// !Because of remote mem type paradigm BUG in the Backplane

		try {
			manager->registerMetaModel(paradigm, filename.c_str());
		} catch (udm_exception &e) {
			MessageBox(e.what(), "Udm error", MB_OK + MB_ICONEXCLAMATION);
		} catch (udmcorba_exception &e) {
			MessageBox(e.what(), "UdmCORBA error", MB_OK + MB_ICONEXCLAMATION);
		}

/*
	} else if (m_url.GetLength()>8 && m_url.Find("remote://")==0) {
	// paradigm is a local file on the server

		// Setup the parameters of the Paradigm
		PARADIGM_STRUCT paradigm;
		paradigm.name = m_name;
		paradigm.version = m_version;

		string filename = m_url.Mid(9);

		// Register a MetaModel
		try {
			manager->registerMetaModelFile(paradigm, filename.c_str());
		} catch (udm_exception &e) {
			MessageBox(e.what(), "Udm error", MB_OK + MB_ICONEXCLAMATION);
	    } catch (udmcorba_exception &e) {
			MessageBox(e.what(), "UdmCORBA error" , MB_OK + MB_ICONEXCLAMATION);
		}
*/

	} else {
		MessageBox("Invalid URL! URL has to be started with 'http://' or 'local://'.", "Error", MB_OK + MB_ICONEXCLAMATION);
		return;
	}

	// print out the paradigms
	m_listbox.ResetContent();
	metamodel_rep.clear();
	char text[1024];

	try {
		vector<PARADIGM_STRUCT> ps = manager->backplane_paradigms();

		int j = 0;
		for(vector<PARADIGM_STRUCT>::iterator i=ps.begin(); i!=ps.end(); i++, j++) {
			sprintf(text, "%s v%s", (*i).name.c_str(), (*i).version.c_str());
			m_listbox.AddString(text);
			metamodel_rep[j] = (*i);
		}
	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		exit(-1);
	}
}

// *************************************************************************************************************************
// selection in metamodel list
// *************************************************************************************************************************
void CManagerDlg::OnSelchangeList3() 
{
	UpdateData(true);
	// TODO: Add your control notification handler code here
	map<long, PARADIGM_STRUCT>::iterator i = metamodel_rep.find(m_listbox.GetCurSel());

	m_name = (*i).second.name.c_str();
	m_version = (*i).second.version.c_str();
	m_url = "";

	m_locked = ( manager->isLocked((*i).second) ) ? 1 : 0;

	UpdateData(false);
}

// *************************************************************************************************************************
// lock/unlock checkbox
// *************************************************************************************************************************
void CManagerDlg::OnCheck1() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);

	// Setup the parameters of the MetaModel
	PARADIGM_STRUCT ps;
	ps.name = m_name;
	ps.version = m_version;

	if (m_locked) {
		try {
			manager->lockParadigm(ps);
		} catch (udmcorba_exception &e) {
			MessageBox(e.what(), "Error" , MB_OK + MB_ICONEXCLAMATION);
		}
	} else {
		try {
			manager->unlockParadigm(ps);
		} catch (udmcorba_exception &e) {
			MessageBox(e.what(), "Error" , MB_OK + MB_ICONEXCLAMATION);
		}
	}
}

// *************************************************************************************************************************
// open file button
// *************************************************************************************************************************
void CManagerDlg::OnButton7() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);

	CString filename;
	CFileDialog open_dlg(TRUE, "xml", "");
	int result = open_dlg.DoModal();
	if (result==IDOK) {
		filename = open_dlg.GetPathName();
		m_url = "local://" + filename;
	}

	UpdateData(false);
}


// *************************************************************************************************************************
// backplane shutdown button
// *************************************************************************************************************************
void CManagerDlg::OnButton8() 
{
	// TODO: Add your control notification handler code here
	int result = MessageBox("Are you sure to shutdown the Backplane server?", "Warning", MB_YESNO + MB_ICONQUESTION);
	
	if (result==IDYES) {
		try {
			manager->backplane_shutdown();

			Sleep(100);
			exit(-1);

		} catch (udmcorba_exception &e) {
			MessageBox(e.what(), "Error" , MB_OK + MB_ICONEXCLAMATION);
		} catch (...) {
			MessageBox("Error: could not communicate with the server!", "Error", MB_OK + MB_ICONEXCLAMATION);
		}
	}

}

// *************************************************************************************************************************
// change password button
// *************************************************************************************************************************
void CManagerDlg::OnButton9() 
{
	// TODO: Add your control notification handler code here
	UpdateData(true);

	ChangePassword_DLG change_dlg;
	int result = change_dlg.DoModal();
	if (result==IDOK) {
	
		if (!change_dlg.m_newpassword1.Compare(change_dlg.m_newpassword2)) {

			try {
				manager->change_password((LPCSTR) change_dlg.m_oldpassword, (LPCSTR) change_dlg.m_newpassword1);
				MessageBox("Password has changed!", "Information", MB_OK + MB_ICONINFORMATION);

			} catch (udmcorba_exception &e) {
				MessageBox(e.what(), "Error" , MB_OK + MB_ICONEXCLAMATION);
			} catch (...) {
				MessageBox("Error: could not communicate with the server!", "Error", MB_OK + MB_ICONEXCLAMATION);
			}

		} else {
			MessageBox("The passwords are not the same! Password is not changed.", "Error", MB_OK + MB_ICONEXCLAMATION);
		}
	}
}


// *************************************************************************************************************************
// refresh datanetworks button
// *************************************************************************************************************************
void CManagerDlg::OnButton10() 
{
	// TODO: Add your control notification handler code here
	m_datanetwork_listbox.ResetContent();
	datanetwork_rep.clear();
	char text[1024];

	try {
		vector<DOCUMENT_STRUCT> ps = manager->backplane_documents();

		int j = 0;
		for(vector<DOCUMENT_STRUCT>::iterator i=ps.begin(); i!=ps.end(); i++, j++) {

			if ((*i).paradigms.size() > 1) {
				string temp_paradigms;
				for (list<PARADIGM_STRUCT>::iterator q=(*i).paradigms.begin(); q!=(*i).paradigms.end(); q++) {
					if (++q==(*i).paradigms.end()) {
						--q;
						temp_paradigms += (*q).name + " v" + (*q).version;
					} else {
						--q;
						temp_paradigms += (*q).name + " v" + (*q).version + ",";
					}
				}	
				sprintf(text, "Name:%s  Version:%s  Remarks:%s  keepAlive:%u sec  Paradigm:(%s)", 
				(*i).name.c_str(), (*i).version.c_str(), (*i).remarks.c_str(), (*i).keepAlive, temp_paradigms.c_str());
			} else {
				sprintf(text, "Name:%s  Version:%s  Remarks:%s  keepAlive:%u sec  Paradigm:%s %s", 
				(*i).name.c_str(), (*i).version.c_str(), (*i).remarks.c_str(), (*i).keepAlive, 
				(*(*i).paradigms.begin()).name.c_str(), (*(*i).paradigms.begin()).version.c_str());
			}

			m_datanetwork_listbox.AddString(text);
			datanetwork_rep[j] = (*i);
		}
	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error", MB_OK + MB_ICONEXCLAMATION);
		exit(-1);
	}
}

// *************************************************************************************************************************
// destroy datanetwork button
// *************************************************************************************************************************
void CManagerDlg::OnButton11() 
{
	// TODO: Add your control notification handler code here
	if (m_datanetwork_listbox.GetCurSel()==-1) {
		MessageBox("No datanetwork is selected!", "Warning", MB_OK + MB_ICONEXCLAMATION);
		return;
	}

	map<long, DOCUMENT_STRUCT>::iterator i = datanetwork_rep.find(m_datanetwork_listbox.GetCurSel());
	if (i==datanetwork_rep.end()) return;
	
	try {
		manager->destroyDocument((*i).second.docID);

		// refresh datanetworks
		OnButton10();

	} catch (udmcorba_exception &e) {
		MessageBox(e.what(), "Error" , MB_OK + MB_ICONEXCLAMATION);
	} catch (...) {
		MessageBox("Error: could not communicate with the server!", "Error", MB_OK + MB_ICONEXCLAMATION);
	}
}

void CManagerDlg::OnAbout() 
{
	// TODO: Add your control notification handler code here
	about_dlg dlg;
	dlg.DoModal();
}
