// Matlab_ToolAdaptor.cpp : Defines the class behaviors for the application.
//

#include "stdafx.h"
#include "Matlab_ToolAdaptor.h"


#include "TA_MFC_Support.h"

#include "Matlab.h"
#include "UdmDom.h"
#include "UdmCORBA.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


// CMatlab_ToolAdaptorApp

BEGIN_MESSAGE_MAP(CMatlab_ToolAdaptorApp, CWinApp)
	ON_COMMAND(ID_HELP, CWinApp::OnHelp)
END_MESSAGE_MAP()


// CMatlab_ToolAdaptorApp construction

CMatlab_ToolAdaptorApp::CMatlab_ToolAdaptorApp()
{
	// TODO: add construction code here,
	// Place all significant initialization in InitInstance
}

// *************************************************************************************************************************
// myToolAdaptor
// *************************************************************************************************************************
class myToolAdaptor : public UdmCORBA::ToolAdaptor {

public:
	myToolAdaptor(const char *type, int argc, char **argv) 
		: ToolAdaptor(type, argc, argv) {};

	void notify(const char *par_name, const char *par_version, const char *name, const char *version, const char *remarks, long docId, long keepAlive) {

		try {
			CString filename;
			
			while (filename.GetLength()<5 || filename.Right(4)!=".xml") {
				int result = showDatanetworkArrivedDlg(name, version, remarks, keepAlive, &filename, "xml");
				if (result==IDCANCEL) return;
				if (filename.GetLength()<5 || filename.Right(4)!=".xml") 
					MessageBox(NULL, "Invalid filename! Filename must end with .xml", "Error", MB_OK);
			}
		
			dn->CreateNew((LPCTSTR)filename, "matlab", matlab::Simulink::meta, Udm::CHANGES_PERSIST_ALWAYS);
			dn->fetchDocument(docId);
			addStringFetchDlg("Datanetwork fetched.");

			dn->SaveAs((LPCSTR)filename);
			addStringFetchDlg("Datanetwork saved as " + filename);

			dn->CloseNoUpdate();
			dn->release();

			addStringFetchDlg("");
			addStringFetchDlg("Waiting for Datanetworks...");

		} catch (udm_exception &e) {
			addStringFetchDlg(e.what());
		} catch (udmcorba_exception &e) {
			addStringFetchDlg(e.what());
		} catch (...) {
			addStringFetchDlg("Fatal Error!");
			return;
		}

	};
};

myToolAdaptor *ta;

// *************************************************************************************************************************
// working Thread
// *************************************************************************************************************************
CWinThread *working_thread;

UINT MyWorkingThread( LPVOID pParam ) {

	try {
		myToolAdaptor *ta = new myToolAdaptor("MATLAB_RECEIVER", __argc, __argv);

		addStringFetchDlg("Subscribe to Matlab paradigm v1.0");
		ta->subscribe(matlab::diagram, "MATLAB", "1.0");
		addStringFetchDlg("Waiting for Datanetworks...");
		ta->run();

		ta->logout();
		delete ta;
		
	} catch (udm_exception &e) {
		MessageBox(NULL, e.what(), "Error!", MB_OK);
		return -1;
	} catch (udmcorba_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK);
		return -1;
	}

	return 0;
}

// The one and only CMatlab_ToolAdaptorApp object

CMatlab_ToolAdaptorApp theApp;


// CMatlab_ToolAdaptorApp initialization

BOOL CMatlab_ToolAdaptorApp::InitInstance()
{
	// InitCommonControls() is required on Windows XP if an application
	// manifest specifies use of ComCtl32.dll version 6 or later to enable
	// visual styles.  Otherwise, any window creation will fail.
	InitCommonControls();

	CWinApp::InitInstance();

	AfxEnableControlContainer();

	try {
	
		// publish Document
		CString name, version, remarks, filename;
		long keepAlive;

		int result = showPublishDlg(&name, &version, &remarks, &keepAlive, &filename, "Matlab", "xml");

		if (result==IDOK) {
			ta = new myToolAdaptor("MATLAB_PUBLISHER", __argc, __argv);

			UdmCORBA::CORBADataNetwork *dn = ta->getDatanetwork(matlab::diagram, "MATLAB", "1.0");

			try {
				dn->OpenExisting((LPCSTR)filename, "Matlab", Udm::CHANGES_LOST_DEFAULT);
			} catch (udm_exception &e) {
				string error = e.what();
				if (!error.compare("Not found")) MessageBox(NULL, "Invalid input file!", "Error", MB_OK);
				else MessageBox(NULL, e.what(), "Error", MB_OK);
				return false;
			}

			dn->publishDocument((LPCSTR)name, (LPCSTR)version, (LPCSTR)remarks, keepAlive);
			
			ta->logout();
		}

		// fetch Document
		working_thread = AfxBeginThread(MyWorkingThread, NULL);
		showFetchDlg("Matlab ToolAdaptor");

	} catch (udm_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK);
	} catch (udmcorba_exception &e) {
		MessageBox(NULL, e.what(), "Error", MB_OK);
	}
	return FALSE;
}
