#include "matlab.h"
#include "UdmDom.h"
#include "UdmCORBA.h"

#include "UdmCORBA_Logger.h"
Logger logger("Pull_Matlab");

using namespace matlab;

class myToolAdaptor : public UdmCORBA::ToolAdaptor {

public:
	myToolAdaptor(const char *type, int argc, char **argv, Logger* logger) 
		: ToolAdaptor(type, argc, argv, logger) {};

	void notify(const char *par_name, const char *par_version, const char *name, const char *version, const char * remarks, long docId, long keepAlive) {
	
		cout << endl << "***Restore CORBADataNetwork***" << endl;
		dn->CreateNew("__pull_matlab.xml", "matlab", matlab::Simulink::meta, Udm::CHANGES_PERSIST_ALWAYS);
		dn->fetchDocument(docId);

		cout << endl << "DataNetwork pulled..." << endl;

		dn->SaveAs("__pull_matlab.xml");
		dn->CloseNoUpdate();

		terminate();
	};

	void getFetchStatus(long percent) {
		cout << "#";
//		cout << percent << "%" << endl;
	}
};

int main(int argc, char* argv[]) {

	if (argc!=2) {
		cout << "Usage: Pull_Matlab <tooladaptor type>" << endl;
		exit(1);
	}

	try {
		myToolAdaptor ta(argv[1], argc, argv, &logger);

		ta.subscribe(diagram);
		ta.run();
	
	} catch (udm_exception &e) {
		cerr << e.what() << endl;
	} catch (udmcorba_exception &e) {
		if ( e.code()<11 ) {
			cerr << endl << "CORBA exception: " << e.what() << endl;
		} else if (e.code()<21) {
			cerr << endl << "Paradigm exception: " << e.what() << endl;
		} else if (e.code()<41) {
			cerr << endl << "UdmCORBA exception: " << e.what() << endl;
		} else if (e.code()<91) {
			cerr << endl << "ToolAdaptor exception: " << e.what() << endl;
		}
	}

	return 0;
}
