#include "CORBAObject.h"

namespace UdmCORBA {
// *******************************************************************************************************************************	
// ===============================================================================================================================
//  CORBAObject
// ===============================================================================================================================
// *******************************************************************************************************************************	
		// ---------------------------------------------------------------------------------------------------------------------------
		//	constructor
		// ---------------------------------------------------------------------------------------------------------------------------
  CORBAObject::CORBAObject(const Uml::Class &meta, int ref) : m_type(meta), m_parent(NULL), refCount(ref), m_servant_id(-1) {
			m_servant._diagram = NULL;
		};

		// ---------------------------------------------------------------------------------------------------------------------------
		//	destructor
		// ---------------------------------------------------------------------------------------------------------------------------
		CORBAObject::~CORBAObject() {};

		// ---------------------------------------------------------------------------------------------------------------------------
		//	clone function
		// ---------------------------------------------------------------------------------------------------------------------------
	    ObjectImpl *CORBAObject::clone() {
		  ++refCount;
	      return this;
		}
 
		// ---------------------------------------------------------------------------------------------------------------------------
		//	release function
		// ---------------------------------------------------------------------------------------------------------------------------
		void CORBAObject::release() {
			if (--refCount==0) {
				if (m_servant._diagram) m_servant._diagram->_remove_ref();
				delete this;
			}
		}


		// ---------------------------------------------------------------------------------------------------------------------------
		//	__getdn function
		// ---------------------------------------------------------------------------------------------------------------------------
	    Udm::DataNetwork *CORBAObject::__getdn() { 
//			throw udmcorba_exception("CORBAObject's __getdn method not implemented", 30);
			return NULL;
		}

		// ---------------------------------------------------------------------------------------------------------------------------
		//	type function
		// ---------------------------------------------------------------------------------------------------------------------------
	    const Uml::Class &CORBAObject::type() const { 
			return m_type;
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	uniqueId function
		// ---------------------------------------------------------------------------------------------------------------------------
		long CORBAObject::uniqueId() const
    { 
			return reinterpret_cast<long>(this); 
		}

		// ---------------------------------------------------------------------------------------------------------------------------
		//	getStringAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		string CORBAObject::getStringAttr(const Uml::Attribute &meta) const 
    {


			try {
				CORBA::String_var obj_var;
				string obj = "";
				bool cached = _find_in_cache(uniqueId(), meta.uniqueId(), obj);
				if (cached) return obj;
//				cout << "String cache not used for: " << (long) uniqueId() << ":" << (long) meta.uniqueId() << endl;

				if (meta==Uml::Diagram::meta_name) { obj_var = m_servant._diagram->name(); obj = obj_var; }
				else if (meta==Uml::Diagram::meta_version) { obj_var = m_servant._diagram->version(); obj = obj_var; }
				else if (meta==Uml::Composition::meta_name) { obj_var = m_servant._composition->name(); obj = obj_var; }
				else if (meta==Uml::CompositionParentRole::meta_name) { obj_var = m_servant._parentrole->name(); obj = obj_var; }
				else if (meta==Uml::CompositionChildRole::meta_name) { obj_var = m_servant._childrole->name(); obj = obj_var; }
				else if (meta==Uml::Association::meta_name) { obj_var = m_servant._association->name(); obj = obj_var; }
				else 
        if (meta==Uml::Attribute::meta_name) 
        { 
          // temporary attribute, it has no servant, it's kinda hack
          // 
          if (m_servant_id == -1) 
          {
            obj = m_name;
          }
          else
          {
					  obj_var = m_servant._attrib->name(); obj = obj_var; 
          }
        }
				else if (meta==Uml::Attribute::meta_type) { obj_var = m_servant._attrib->typeName(); obj = obj_var; }
				else if (meta==Uml::Attribute::meta_visibility) { obj_var = m_servant._attrib->visibility(); obj = obj_var; }

				// defvalue is a vector<string>!!!
				else if (meta==Uml::Attribute::meta_defvalue) {
					OTIF::string_S_var defvls = m_servant._attrib->defvalue();
					for(CORBA::ULong i=0; i<defvls->length(); i++) {
						obj += defvls[i];
						if (i != defvls->length()-1) obj += ";";
					}
				}

				else if (meta==Uml::Class::meta_name) { obj_var = m_servant._class->name(); obj = obj_var; }
				else 
        if (meta==Uml::Class::meta_stereotype) 
        { obj_var = m_servant._class->stereotypeName(); obj = obj_var; }
				else 
				{
					throw udmcorba_corbalayer_exception("Fatal exception: Unknown Attribute", 1);
					return 					obj;
				}

				_put_in_attr_cache(uniqueId(), meta.uniqueId(), obj);
				return obj;

			} catch (const CORBA::Exception &) {
				throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
			} catch (udm_exception&) {
				throw;
			} catch (udmcorba_exception&) {
				throw;
			} catch (...) {
				throw udmcorba_exception("Fatal exception in the getStringAttr function.", 0);
			}
		}

		// ---------------------------------------------------------------------------------------------------------------------------
		//	getBooleanAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		bool CORBAObject::getBooleanAttr(const Uml::Attribute &meta) const {
			try {
				bool obj;
				bool cached = _find_in_cache(uniqueId(), meta.uniqueId(), obj);
				if (cached) return obj;
//				cout << "Boolean cache not used for: " << (long) uniqueId() << ":" << (long) meta.uniqueId() << endl;

				if (meta==Uml::Association::meta_nonpersistent) return (m_servant._association->nonpersistent()==0) ? false : true;
				else if (meta==Uml::Attribute::meta_nonpersistent) return (m_servant._attrib->nonpersistent()==0) ? false : true;
				else if (meta==Uml::Attribute::meta_ordered) return (m_servant._attrib->ordered()==0) ? false : true;
				else if (meta==Uml::Attribute::meta_registry) return (m_servant._attrib->registry()==0) ? false : true;
				else if (meta==Uml::Class::meta_isAbstract) return (m_servant._class->isAbstract()==0) ? false : true;
				else if (meta==Uml::GenericRole::meta_isNavigable && m_type==Uml::CompositionParentRole::meta) return (m_servant._parentrole->isNavigable()==0) ? false : true;
				else if (meta==Uml::GenericRole::meta_isNavigable && m_type==Uml::CompositionChildRole::meta) return (m_servant._childrole->isNavigable()==0) ? false : true;
				else if (meta==Uml::GenericRole::meta_isNavigable && m_type==Uml::AssociationRole::meta) return (m_servant._associationrole->isNavigable()==0) ? false : true;
				else return false;

				_put_in_attr_cache(uniqueId(), meta.uniqueId(), obj);
				return obj;

			} catch (const CORBA::Exception &) {
				throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
			} catch (udm_exception&) {
				throw;
			} catch (udmcorba_exception&) {
				throw;
			} catch (...) {
				throw udmcorba_exception("Fatal exception in the getBooleanAttr function.", 0);
			}
		}

    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getIntegerAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		__int64 CORBAObject::getIntegerAttr(const Uml::Attribute &meta) const 
    {
			try {
				long obj;
				bool cached = _find_in_cache(uniqueId(), meta.uniqueId(), obj);
				if (cached) return obj;
//				cout << "Long cache not used for: " << (long) uniqueId() << ":" << (long) meta.uniqueId() << endl;
		
				if (meta==Uml::CompositionChildRole::meta_min) return m_servant._childrole->minCard();
				else if (meta==Uml::CompositionChildRole::meta_max) return m_servant._childrole->maxCard();
				else if (meta==Uml::AssociationRole::meta_min) return m_servant._associationrole->minCard();
				else if (meta==Uml::AssociationRole::meta_max) return m_servant._associationrole->maxCard();
				else if (meta==Uml::Attribute::meta_min) return m_servant._attrib->minCard();
				else if (meta==Uml::Attribute::meta_max) return m_servant._attrib->maxCard();
				else return 0;
		
				_put_in_attr_cache(uniqueId(), meta.uniqueId(), obj);
				return obj;

			} catch (const CORBA::Exception &) {
				throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
			} catch (udm_exception&) {
				throw;
			} catch (udmcorba_exception&) {
				throw;
			} catch (...) {
				throw udmcorba_exception("Fatal exception in the getIntegerAttr function.", 0);
			}
		}
    
    
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getParentAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		ObjectImpl *CORBAObject::getParent(const Uml::CompositionParentRole &role) const 
    {
      // hack - the namspace should be introduced into OTIF
      if (role == ::Uml::Class::meta_parent_ns)
        return NULL;

			if (m_parent) ++(m_parent->refCount);
			return m_parent; 
		}


    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getChildren function
		// ---------------------------------------------------------------------------------------------------------------------------
	   vector<ObjectImpl*> CORBAObject::getChildren(const Uml::CompositionChildRole &meta, const Uml::Class &uml_class) const 
     {
			try {
				vector<ObjectImpl*> ret;

				bool cached = _find_in_cache(uniqueId(), meta.uniqueId(), ret);
				if (cached) return ret;
//				cout << "Children cache not used for: " << (long) uniqueId() << ":" << (long) meta.uniqueId() << endl;

				if (meta==Uml::Diagram::meta_classes || (m_type==Uml::Diagram::meta && !meta)) {
					OTIF::Class_S_var cls = m_servant._diagram->classes();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}	
					if (cls->length()==0) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
				}
				if (meta==Uml::Diagram::meta_compositions || (m_type==Uml::Diagram::meta && !meta)) {
					OTIF::Composition_S_var cls = m_servant._diagram->compositions();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}	
					if (cls->length()==0) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
				}
				if (meta==Uml::Diagram::meta_associations || (m_type==Uml::Diagram::meta && !meta)) {
					OTIF::Association_S_var cls = m_servant._diagram->associations();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}	
					if (cls->length()==0) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
				}
	
				if (meta==Uml::Composition::meta_parentRole || (m_type==Uml::Composition::meta && !meta)) {
					OTIF::CompositionParentRole_var cls = m_servant._composition->parentRole();
					long _id = cls->tId();
					map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
					if (j!=servant_rep.end()) {
//						ret.push_back( (*j).second );
						ret.push_back( (*j).second->clone() );
					}
					_put_in_cache(uniqueId(), meta.uniqueId(), _id);
				}
				if (meta==Uml::Composition::meta_childRole || (m_type==Uml::Composition::meta && !meta)) {
					OTIF::CompositionChildRole_var cls = m_servant._composition->childRole();
					long _id = cls->tId();
					map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
					if (j!=servant_rep.end()) {
//						ret.push_back( (*j).second );
						ret.push_back( (*j).second->clone() );
					}
					_put_in_cache(uniqueId(), meta.uniqueId(), _id);
				}
			
				if (meta==Uml::Association::meta_roles || (m_type==Uml::Association::meta && !meta)) {
					OTIF::AssociationRole_S_var cls = m_servant._association->roles();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}	
					if (cls->length()==0) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
				}
			
				if (meta==Uml::Class::meta_attributes || (m_type==Uml::Class::meta && !meta)) {
					long not_found = true;
					OTIF::Attrib_S_var cls = m_servant._class->attributes();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) { 
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						not_found = false;
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}	
					if (not_found) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
				}

				return ret;

			} catch (const CORBA::Exception &) {
				throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
			} catch (udm_exception&) {
				throw;
			} catch (udmcorba_exception&) {
				throw;
			} catch (...) {
				throw udmcorba_exception("Fatal exception in the getChildren function.", 0);
			}
		}

    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getAssociation function
		// ---------------------------------------------------------------------------------------------------------------------------
		vector<ObjectImpl*> CORBAObject::getAssociation(const Uml::AssociationRole &meta, int mode ) const 
    {
			try {
				vector<ObjectImpl*> ret;
	
				bool cached = _find_in_cache(uniqueId(), meta.uniqueId(), ret);
				if (cached) return ret;
//				cout << "Association cache not used for: " << (long) uniqueId() << ":" << (long) meta.uniqueId() << endl;

				if (meta==Uml::CompositionParentRole::meta_target) {
					OTIF::Class_var cls_var = m_servant._parentrole->target();
					long _id = cls_var->tId();
					map<long, CORBAObject*>::iterator i = servant_rep.find( _id );
					if (i!=servant_rep.end()) {
//						ret.push_back( (*i).second );
						ret.push_back( (*i).second->clone() );
					}
					_put_in_cache(uniqueId(), meta.uniqueId(), _id);
				}
				if (meta==Uml::CompositionChildRole::meta_target) {
					OTIF::Class_var cls_var = m_servant._childrole->target();
					long _id = cls_var->tId();
					map<long, CORBAObject*>::iterator i = servant_rep.find( _id );
					if (i!=servant_rep.end()) {
//						ret.push_back( (*i).second );
						ret.push_back( (*i).second->clone() );
					}
					_put_in_cache(uniqueId(), meta.uniqueId(), _id);
				}
				if (meta==Uml::AssociationRole::meta_target) {
					OTIF::Class_var cls_var = m_servant._associationrole->target();
					long _id = cls_var->tId();
					map<long, CORBAObject*>::iterator i = servant_rep.find( _id );
					if (i!=servant_rep.end()) {
//						ret.push_back( (*i).second );
						ret.push_back( (*i).second->clone() );
					}
					_put_in_cache(uniqueId(), meta.uniqueId(), _id);
				}
				if (meta==Uml::Association::meta_assocClass) {
					OTIF::Class_var _assocClass = m_servant._association->assocClass();
					if (!CORBA::is_nil(_assocClass)) {
						long _id = _assocClass->tId();
						map<long, CORBAObject*>::iterator i = servant_rep.find( _id );
						if (i!=servant_rep.end()) {
//							ret.push_back( (*i).second );
							ret.push_back( (*i).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					} else {
						_put_in_cache(uniqueId(), meta.uniqueId(), -1);
					}
				}

				if (meta==Uml::Class::meta_parentRoles) {
					bool not_found = true;
					OTIF::CompositionParentRole_S_var cls = m_servant._class->parentRoles();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						not_found = false;
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}
					if (not_found) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
		
				}
				if (meta==Uml::Class::meta_childRoles) {
					bool not_found = true;
					OTIF::CompositionChildRole_S_var cls = m_servant._class->childRoles();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						not_found = false;
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}
					if (not_found) _put_in_cache(uniqueId(), meta.uniqueId(), -1);

				}
				if (meta==Uml::Class::meta_associationRoles) {
					bool not_found = true;
					OTIF::AssociationRole_S_var cls = m_servant._class->associationRoles();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						not_found = false;
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}
					if (not_found) _put_in_cache(uniqueId(), meta.uniqueId(), -1);

				}

				if (meta==Uml::Class::meta_association) {
					OTIF::Association_var _assoc = m_servant._class->assoc();
					if (!CORBA::is_nil(_assoc)) {
						long _id = _assoc->tId();
						map<long, CORBAObject*>::iterator i = servant_rep.find( _id );
						if (i!=servant_rep.end()) {
//							ret.push_back( (*i).second );
							ret.push_back( (*i).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					} else {
						_put_in_cache(uniqueId(), meta.uniqueId(), -1);
					}
				}

				if (meta==Uml::Class::meta_subTypes) {
					OTIF::Class_S_var cls = m_servant._class->subTypes();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}	
					if (cls->length()==0) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
				}

				if (meta==Uml::Class::meta_baseTypes) {
					OTIF::Class_S_var cls = m_servant._class->baseTypes();
					for(CORBA::ULong i=0; i<cls->length(); i++) {
						long _id = cls[i]->tId();
						map<long, CORBAObject*>::iterator j = servant_rep.find( _id );
						if (j!=servant_rep.end()) {
//							ret.push_back( (*j).second );
							ret.push_back( (*j).second->clone() );
						}
						_put_in_cache(uniqueId(), meta.uniqueId(), _id);
					}	
					if (cls->length()==0) _put_in_cache(uniqueId(), meta.uniqueId(), -1);
				}
	
				return ret;

			} catch (const CORBA::Exception &) {
				throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
			} catch (udm_exception&) {
				throw;
			} catch (udmcorba_exception&) {
				throw;
			} catch (...) {
				throw udmcorba_exception("Fatal exception in the getAssociation function.", 0);
			}
		
		}

}//namespace UdmCORBA {
