#ifndef CORBAObject_H
#define CORBAObject_H

#include "otifdataC.h"
#include "otifC.h"
#include "otif_i.h"
#include <ace/streams.h>

#include <winsock2.h>

#ifndef MOBIES_UDMBASE_H
#include "UdmBase.h"
#endif

#ifndef MOBIES_UDMSTATIC_H
#include "UdmStatic.h"
#endif

#ifndef MOBIES_UDMCORBA_EXCEPTION_H
#include <UdmCORBA_Exception.h>
#include <UdmCORBA_Logger.h>
#endif

#include <UmlExt.h>

namespace UdmCORBA
{
	typedef map<long, Udm::Object> copy_assoc_map;

  class CORBAObject;
  extern map<long, CORBAObject*> servant_rep; 

  typedef Udm::ObjectImpl ObjectImpl;

	void _put_in_cache(long id, long meta_id, long servant_id);
	void _put_in_attr_cache(long id, long meta_id, string STRING);
	void _put_in_attr_cache(long id, long meta_id, bool BOOL);
	void _put_in_attr_cache(long id, long meta_id, long LONG);
	bool _find_in_cache(long id, long meta_id, vector<ObjectImpl*> &ret);
	bool _find_in_cache(long id, long meta_id, string &obj);
	bool _find_in_cache(long id, long meta_id, long &obj);
	bool _find_in_cache(long id, long meta_id, bool &obj);



// *******************************************************************************************************************************	
// ===============================================================================================================================
//  CORBAObject
// ===============================================================================================================================
// *******************************************************************************************************************************	
	class CORBAObject : public ObjectImpl {

	public:

		union {
			OTIF::Diagram_ptr _diagram;
			OTIF::Composition_ptr _composition;
			OTIF::CompositionChildRole_ptr _childrole;
			OTIF::CompositionParentRole_ptr _parentrole;
			OTIF::AssociationRole_ptr _associationrole;
			OTIF::Association_ptr _association;
			OTIF::TaggedValue_ptr _taggedvalue;
			OTIF::Attrib_ptr _attrib;
			OTIF::Class_ptr _class;
		} m_servant;
		
		long m_servant_id;
	    int refCount;

		// Optimalization
		string m_name;
		string m_attrib_type;
		// Optimalization
		
		const Uml::Class &m_type;

		CORBAObject *m_parent;	// not reference counted

		// ---------------------------------------------------------------------------------------------------------------------------
		//	constructor
		// ---------------------------------------------------------------------------------------------------------------------------
	    CORBAObject(const Uml::Class &meta, int ref);

		// ---------------------------------------------------------------------------------------------------------------------------
		//	destructor
		// ---------------------------------------------------------------------------------------------------------------------------
		~CORBAObject() ;

		// ---------------------------------------------------------------------------------------------------------------------------
		//	clone function
		// ---------------------------------------------------------------------------------------------------------------------------
	    virtual ObjectImpl *clone() ;
 
		// ---------------------------------------------------------------------------------------------------------------------------
		//	release function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void release();


		// ---------------------------------------------------------------------------------------------------------------------------
		//	__getdn function
		// ---------------------------------------------------------------------------------------------------------------------------
	    virtual Udm::DataNetwork *__getdn();

		// ---------------------------------------------------------------------------------------------------------------------------
		//	type function
		// ---------------------------------------------------------------------------------------------------------------------------
	    virtual const Uml::Class &type() const ;
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	uniqueId function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual long uniqueId() const ;

		// ---------------------------------------------------------------------------------------------------------------------------
		//	getStringAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual string getStringAttr(const Uml::Attribute &meta) const ;
		// ---------------------------------------------------------------------------------------------------------------------------
		//	setStringAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		void setStringAttr(const Uml::Attribute &meta, const string &a, const bool direct = true) {
			throw udmcorba_exception("CORBAObject's setStringAttr method not implemented", 21);
		}


		// ---------------------------------------------------------------------------------------------------------------------------
		//	getBooleanAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual bool getBooleanAttr(const Uml::Attribute &meta) const;
		// ---------------------------------------------------------------------------------------------------------------------------
		//	setBooleanAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void setBooleanAttr(const Uml::Attribute &meta, bool a, const bool direct = true) {
			throw udmcorba_exception("CORBAObject's setBooleanAttr method not implemented", 22);
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getIntegerAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual __int64 getIntegerAttr(const Uml::Attribute &meta) const ;
		// ---------------------------------------------------------------------------------------------------------------------------
		//	setIntegerAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void setIntegerAttr(const Uml::Attribute &meta, __int64 a, const bool direct = true) {
			throw udmcorba_exception("CORBAObject's setIntegerAttr method not implemented", 23);
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getRealAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
	    virtual double getRealAttr(const Uml::Attribute &meta) const {
			return 0.0;
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	setRealAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void setRealAttr(const Uml::Attribute &meta, double a, const bool direct = true) {
			throw udmcorba_exception("CORBAObject's setRealAttr method not implemented", 24);
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getParentAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual ObjectImpl *getParent(const Uml::CompositionParentRole &role) const;

		// ---------------------------------------------------------------------------------------------------------------------------
		//	setParentAttr function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void setParent(ObjectImpl *a, const Uml::CompositionParentRole &role, const bool direct = true) {
			throw udmcorba_exception("CORBAObject's setParent method not implemented", 25);
		}

		// ---------------------------------------------------------------------------------------------------------------------------
		//	detach function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void detach() { 
			throw udmcorba_exception("CORBAObject's detach method not implemented", 26);
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getChildren function
		// ---------------------------------------------------------------------------------------------------------------------------
	    virtual vector<ObjectImpl*> getChildren(const Uml::CompositionChildRole &meta, const Uml::Class &uml_class) const;
		// ---------------------------------------------------------------------------------------------------------------------------
		//	setChildren function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void setChildren(const Uml::CompositionChildRole &meta, const vector<ObjectImpl*> &a, const bool direct = true) {
			throw udmcorba_exception("CORBAObject's setChildren method not implemented", 27);
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	createChild function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual ObjectImpl *createChild(const Uml::CompositionChildRole &childrole, const Uml::Class &meta, const Udm::ObjectImpl* archetype = NULL, const bool subtype = false, const bool real_archetype = true, const bool need_safetype = false) {
			throw udmcorba_exception("CORBAObject's createChild method not implemented", 28);
			return NULL;
		}
    
		// ---------------------------------------------------------------------------------------------------------------------------
		//	getAssociation function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual vector<ObjectImpl*> getAssociation(const Uml::AssociationRole &meta, int mode = Udm::TARGETFROMPEER) const;  
		// ---------------------------------------------------------------------------------------------------------------------------
		//	setAssociation function
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual void setAssociation(const Uml::AssociationRole &meta, const vector<ObjectImpl*> &nvect, int mode = Udm::TARGETFROMPEER, const bool direct = true) {
			throw udmcorba_exception("CORBAObject's setAssociation method not implemented", 29);
		}

		// ---------------------------------------------------------------------------------------------------------------------------
		// acrhetype/derived/instances
		// ---------------------------------------------------------------------------------------------------------------------------
		virtual vector<ObjectImpl*> getDerived() const {
			throw udmcorba_exception("CORBAObject's getDerived method not implemented", 0);
		}

		virtual vector<ObjectImpl*> getInstances() const {
			throw udmcorba_exception("CORBAObject's getInstances method not implemented", 0);
		}

		virtual ObjectImpl* getArchetype() const {
			throw udmcorba_exception("CORBAObject's getArchetype method not implemented", 0);
		}
		
		virtual bool hasRealArchetype() const {
			throw udmcorba_exception("CORBAObject's hasRealArchetype method not implemented", 0);
		}

		virtual bool isSubtype() const {
			throw udmcorba_exception("CORBAObject's isSubtype method not implemented", 0);
		}

		virtual bool isInstance() const {
			throw udmcorba_exception("CORBAObject's isInstance method not implemented", 0);
		}
	};
}//namespace UdmCORBA 
#endif //CORBAObject_H