/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

#pragma warning( disable : 4786 )
#pragma warning( disable : 4518 )

#include "CORBAObject.h"

namespace UdmCORBA
{
long status_percent;
long fetch_document_offset;
long temp_fetch_document_offset;

typedef map<long, long> LongLongMap;

// *******************************************************************************************************************************	
// ===============================================================================================================================
//  CORBADataNetwork Unwrapping
// ===============================================================================================================================
// *******************************************************************************************************************************	
	void _setDNObjectParameters(long id, OTIF::objectId dn_id);
	map<long, Udm::ObjectImpl*> objectimpl_rep;
//========================================================================
	void __printTime(Logger *logger)
  {
		char time_buf[128];
		struct _timeb tstruct;
		_strtime(time_buf);
		_ftime( &tstruct );
		char milli[64];
		itoa(tstruct.millitm, milli, 10);
		logger->log(3, "Time:", (string)time_buf + "." + milli);
	}
//========================================================================	
	Uml::CompositionChildRole getCompositionChildRoleMeta(Uml::Class &class_meta, Uml::Class &parent_meta) 
  {
		set<Uml::Class> cancs = Uml::AncestorClasses(class_meta);
		for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++) {
			
			set<Uml::CompositionChildRole> uml_childroles = (*q).childRoles();
			for(set<Uml::CompositionChildRole>::iterator i=uml_childroles.begin(); i!=uml_childroles.end(); i++) {
				Uml::Class target = ((Uml::CompositionParentRole) ((Uml::Composition) (*i).parent()).parentRole()).target();
				if (target==parent_meta) return *i;
			}
		}
		return NULL;
	}
//========================================================================
  
// Build up the Attributes for an ObjectImpl
void ObjectImplAttrsPass(Udm::ObjectImpl *obj, const OTIF::Obj &udm_obj)
{
	set<string> isSetAttribute;	// stores the attributes name for an ObjectImpl
	for(CORBA::ULong i=0; i< udm_obj.attrs.length(); i++)
  {
		OTIF::AttrValueHolder attr_value = udm_obj.attrs[i].value;

		const Uml::Class &obj_meta = obj->type();
		Uml::Attribute attrib_meta;



		map<long, CORBAObject*>::iterator m_obj = servant_rep.find( udm_obj.attrs[i].tId );
		if (m_obj==servant_rep.end()) 
      return;

		// Optimalization
		string meta_name = m_obj->second->m_name;
//			CORBA::String_var meta_name_var = m_obj->second->m_servant._attrib->name();//udm_obj.attrs[i].type->name();
//			string meta_name = meta_name_var;

//  cout << "obj meta name " << (string)obj_meta.name() << std::endl;
    
        

    
  
		// Optimalization
		string attr_class_name = m_obj->second->m_parent->m_name;
//			Class_var parent_class_var = m_obj->second->m_servant._attrib->parent();//udm_obj.attrs[i].type->parent();
//			CORBA::String_var attr_class_name_var = parent_class_var->name();
//			string attr_class_name = attr_class_name_var;

    
		// Optimalization
		string meta_type = m_obj->second->m_attrib_type;
//			CORBA::String_var meta_type_var = m_obj->second->m_servant._attrib->type();//udm_obj.attrs[i].type->type();
//			string meta_type = meta_type_var;

		// *****************************************************************************************************
		// XML,GME backend BUG, if the string attribute value is "" and the attribute meta is a base class
		// we don't want to set it because the inherited class has the value already!!!
		// The problem is not a good solution, because if somebody has set both of the base and the inherited class
		// attribute, one of them will be overwritten!!!

    
    
		if (isSetAttribute.find(meta_name) != isSetAttribute.end())
    {
			if (meta_type=="String" && 
             attr_class_name.compare(obj_meta.name())) 
      {
				if (!((string)attr_value.stringValue()).compare("")) continue;
			}
			else
          if (meta_type=="Integer" &&
             attr_class_name.compare(obj_meta.name()))
          {
				if (attr_value.intValue()==0) continue;
			}
			else
          if (meta_type=="Boolean" &&
             attr_class_name.compare(obj_meta.name()))
          {
				if (attr_value.boolValue()==0) continue;
			}
			else
          if (meta_type=="Real" && 
             attr_class_name.compare(obj_meta.name()))
          {
				if (attr_value.realValue()==0.0) continue;
			}
		}
    
			// *****************************************************************************************************

		set<Uml::Class> cancs = Uml::AncestorClasses(obj_meta);
		for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++)
    {
			set<Uml::Attribute> uml_attrs = (*q).attributes();
//				set<Uml::Attribute> uml_attrs = (*cancs.begin()).attributes();
			for(set<Uml::Attribute>::iterator i=uml_attrs.begin(); i!=uml_attrs.end(); i++)
      {
				if (!meta_name.compare((*i).name()) 
            && !attr_class_name.compare((*q).name()))
        {
					attrib_meta = *i;
				}
			}
		}
		if (!attrib_meta) throw 
      udm_exception("CORBADataNetwork is corrupt => A meta of an attribute is not found!");

		try
    {
			if (meta_type=="String") 
      {
        obj->setStringAttr(attrib_meta, attr_value.stringValue(), false);
//        if (meta_name =="name")
 //       cout << "attr " << meta_name << " " << attr_value.stringValue() << std::endl;
      }
			else 
      if (meta_type=="Integer") 
        obj->setIntegerAttr(attrib_meta, attr_value.intValue(), false);
 			else 
      if (meta_type=="Boolean") 
        obj->setBooleanAttr(attrib_meta, ((attr_value.boolValue()==0) ? false : true), false);
			else 
      if (meta_type=="Real") 
        obj->setRealAttr(attrib_meta, attr_value.realValue(), false);

			isSetAttribute.insert(attrib_meta.name());
		}
    catch (udm_exception &e)
    {
			cerr << "Warning: Attribute (" << (string)obj_meta.name() << ":" << (string)attrib_meta.name() << ") couldn't be set: " << e.what() << endl;
			continue;
		}
	}
}

//========================================================================

	
// ---------------------------------------------------------------------------------------------------------------------

//========================================================================

	void __extractUserData(CORBA::Any userData, long &subtype, long &realarchetype, long &archetype) 
  {
		char * __userData;

		if (userData >>= __userData) {
			string s_archetype = __userData;
			string s_subtype = s_archetype.substr(0, 1);
			string s_realarchetype = s_archetype.substr(1, 1);
			s_archetype.erase(s_archetype.begin());
			s_archetype.erase(s_archetype.begin());

//			cout << s_subtype << ":" << s_realarchetype << ":" << s_archetype << endl;

			subtype = atol(s_subtype.c_str());
			realarchetype = atol(s_realarchetype.c_str());
			archetype = atol(s_archetype.c_str());
		}
	}
	
//========================================================================	
	bool reqIsObjectCopied(const OTIF::Obj_S &udm_objects, const OTIF::Obj &srcDerived, const copy_assoc_map &cam)
  {
		
		for(CORBA::ULong i=1; i<udm_objects.length(); i++) 
    {
			if (udm_objects[i].parent.parentId.id==srcDerived.id.id) 
      {

				long __srcArchetypeID;
				long __realarchetype;
				long __subtype;
				__extractUserData(udm_objects[i].userData, __subtype, __realarchetype, __srcArchetypeID);

//				long srcArchetypeID = udm_objects[i].archetype;

//				if (srcArchetypeID!=-1 && udm_objects[i].realarchetype==0) 		// it is not a real archetype
				if (__srcArchetypeID!=-1 && __realarchetype==0)
        {		// it is not a real archetype

					copy_assoc_map::const_iterator cam_i = cam.find(__srcArchetypeID);
					if (cam_i != cam.end())
          {
						if (!reqIsObjectCopied(udm_objects, udm_objects[i], cam)) return false;
					} else 
          {
						return false;
					}
				}
      }
		}
				
		return true;
	}
//========================================================================
	void reqCreateMappingforDerivedObject(const OTIF::Obj_S &udm_objects, const unsigned long udm_obj_id, const Udm::Object &dstDerived, copy_assoc_map &cam) 
  {

		set<Udm::Object> dst_children = dstDerived.GetChildObjects();

		for(CORBA::ULong i=1; i<udm_objects.length(); i++) {
			if (udm_objects[i].parent.parentId.id==udm_obj_id) {

				long __srcArchetypeID;
				long __realarchetype;
				long __subtype;
				__extractUserData(udm_objects[i].userData, __subtype, __realarchetype, __srcArchetypeID);

//				long srcArchetypeID = udm_objects[i].archetype;

//				if (srcArchetypeID!=-1 && udm_objects[i].realarchetype==0) 		// it is not a real archetype
				if (__srcArchetypeID!=-1 && __realarchetype==0) {		// it is not a real archetype

					// src must be in the map
					copy_assoc_map::iterator cam_i=cam.find(__srcArchetypeID);
					if (cam_i!=cam.end()) {

						Udm::Object dst_child_archetype = (*cam_i).second;
						set<Udm::Object>::iterator dst_i = dst_children.begin();
						bool found = false;
						while (dst_i != dst_children.end() && !found) {

							Udm::Object dst_child = *dst_i++;
							if ((dst_child.archetype() == dst_child_archetype) && !dst_child.hasRealArchetype()) {

								// record mapping
								copy_assoc_map::value_type map_item(udm_objects[i].id.id, dst_child);
								cam.insert(map_item);
								found=true;

								// recursive step, we do it for the whole subtree
								reqCreateMappingforDerivedObject(udm_objects, udm_objects[i].id.id, dst_child, cam);
							}
						}
					} else {
		
						throw udm_exception("Internal error");
					}

				} else {
					// this is not an error
					// it might be that udm_obj_id is a subtype and in adition has other objects
				}
			}
		}
	}

//========================================================================
void reqCopyObjectHierarchy(Udm::DataNetwork* p_dstBackend, 
                            const Uml::Diagram &root_meta, 
                            const OTIF::Obj_S &udm_objects, 
                            ObjectImpl *parent, 
                            const unsigned long parent_id, 
                            bool &finished, 
                            copy_assoc_map &cam, 
                            ToolAdaptor *tooladaptor) 
{

  Uml::Class parent_meta = parent->type();
//    cout << "-----------" << endl;
//    cout << "parent " << parent_meta.name() << endl;

	for(CORBA::ULong i=1; i<udm_objects.length(); i++)
  {
		if (udm_objects[i].parent.parentId.id==parent_id)
    {

			map<long, CORBAObject*>::iterator m_obj = servant_rep.find( udm_objects[i].tId );
			if (m_obj == servant_rep.end())
        throw udmcorba_exception("A type is not found among the remote types!", 0);

			// Optimalization
			string meta_name = m_obj->second->m_name;
//				CORBA::String_var meta_name_var = m_obj->second->m_servant._class->name();//(udm_objects[i].type)->name();
//				string meta_name = meta_name_var;

//				cout << "child " << meta_name << endl;
			
			Uml::Class meta;
			set<Uml::Class> uml_classes = root_meta.classes();
			for(set<Uml::Class>::iterator j=uml_classes.begin(); j!=uml_classes.end(); j++)
      {
				if (!meta_name.compare( (*j).name() ))
        {
					meta = (*j);
					break;
				}
			}
			if (!meta) 
        throw udm_exception("CORBADataNetwork is corrupt => A meta of a class is not found!");

			Uml::Class parent_meta = parent->type();
			Uml::CompositionChildRole childrole = getCompositionChildRoleMeta(meta, parent_meta);


			ObjectImpl* p_dstChild = &Udm::_null;
			Udm::Object dstChild;

			// check whether the object was copied first
			copy_assoc_map::iterator cam_i=cam.find( udm_objects[i].id.id );
			if (cam_i != cam.end())
      {
				dstChild = (*cam_i).second;
				p_dstChild = dstChild.__impl()->clone();
			}
			
			long __srcArchetypeID;
			long __realarchetype;
			long __subtype;
			__extractUserData(udm_objects[i].userData, __subtype, __realarchetype, __srcArchetypeID);
			
			if (p_dstChild==&Udm::_null)
      {					// object has not created yet
//        	cout << "object has not created yet" << endl;
//					long srcArchetypeID = udm_objects[i].archetype;
				ObjectImpl* p_dstChildArc = &Udm::_null;
				Udm::Object dstChildArc;

				if (__srcArchetypeID!=-1)
        {					// check if the source object has archetype

          	//cout << "source object has archetype" << endl;
					//if it's not a real archetype, we skip this child
					//(it's there because the parent has an archetype)
//						if (udm_objects[i].realarchetype==0)	continue;
					if (__realarchetype==0)	
          {
            //cout << " it's not a real archetype, we skip this child"  << endl;
            continue;
          }

					// check if it was copied or not
					copy_assoc_map::iterator cam_i=cam.find(__srcArchetypeID);
					if (cam_i!=cam.end())
          {

						// if it was already copied, assign the copied archetype to variables
						dstChildArc = (*cam_i).second;
						p_dstChildArc = dstChildArc.__impl();
					}

					if (p_dstChildArc==&Udm::_null)
          {

						//cout <<  "archetype is not copied, we need to walk the tree again" << endl;
						finished = false;
						continue;

					}//if (p_dstChildArc==&Udm::_null)
          else 
          {
             //cout << "archetype's all children were copied " << endl;
						//check if archetype's all children were copied OK
						//(recursively)
						if ( !reqIsObjectCopied(udm_objects, udm_objects[i], cam) )
            {
              //cout << "archetype's all children were copied not OK" << endl;
							finished = false;
							continue;
						}

						// archetype is ready, so we create the new object
						bool subtype=false;
//							if (udm_objects[i].subtype==1) subtype=true;
						if (__subtype==1) subtype=true;

							//cout << "subtype: " << subtype << endl;
//							cout << "Creating archetype: " << udm_obj->id.id << endl;

						if (p_dstBackend->IsTypeSafe())
							p_dstChild = parent->createChild(childrole, meta, p_dstChildArc, subtype);
						else
            {
							const Uml::Class &safe_meta = Uml::SafeTypeContainer::GetSafeType(meta);
							p_dstChild = parent->createChild(childrole, safe_meta, p_dstChildArc, subtype);
						}

						// create the attributes
            //cout << "set attr "<< endl;
						ObjectImplAttrsPass(p_dstChild, udm_objects[i]); 
					}//else (p_dstChildArc==&Udm::_null)

					
					// do the mapping for the derived block as well
					reqCreateMappingforDerivedObject(udm_objects, udm_objects[i].id.id, p_dstChild->clone(), cam);

				}//if (__srcArchetypeID!=-1)
        else
        {

					// it has not archetype, so just create the default way
						//cout << "Creating: " << endl;//udm_obj->id.id << endl;

					if (p_dstBackend->IsTypeSafe())
						p_dstChild = parent->createChild(childrole, meta);
					else {
						const Uml::Class &safe_meta = Uml::SafeTypeContainer::GetSafeType(meta);
						p_dstChild = parent->createChild(childrole, safe_meta);
					}

					// create the attributes
          //cout << "set attr "<< endl;
					ObjectImplAttrsPass(p_dstChild, udm_objects[i]); 
				}//else (__srcArchetypeID!=-1)


				// create association for the map
				copy_assoc_map::iterator cam_i=cam.find(udm_objects[i].id.id);
				if (cam_i!=cam.end())
        {
					(*cam_i).second = p_dstChild->clone();
				}
        else 
        {
					copy_assoc_map::value_type map_item(udm_objects[i].id.id, p_dstChild->clone());
					cam.insert(map_item);
				}

			}//if (p_dstChild==&Udm::_null)
      else
      {
        //cout << "already copied" << endl;
        // kalmar
        // derived objects may have different atribute settings
//  				if (__srcArchetypeID!=-1)
//         {					// check if the source object has archetype
// 
// 
// 					ObjectImplAttrsPass(p_dstChild, udm_objects[i]); 
//         }
				//in case of inherited children object
				//which where created by createChild() call
				//attributes must be copied.

				//getArhcetype() clone()-s it first
//					ObjectImpl * p_srcChildArc = p_srcChild->getArchetype();
//					if ( p_srcChildArc  && (p_srcChildArc != &Udm::_null))
//					{
//						if (!p_srcChild->hasRealArchetype())
//							p_dstChild->CopyAttributesFrom(p_srcChild);	
//						
//						p_srcChildArc->release();
//					}

			}


			// show the status of the fetch process
			if (tooladaptor)
      {

				if (temp_fetch_document_offset==0) {
					fetch_document_offset++;

					tooladaptor->getFetchStatus(fetch_document_offset, udm_objects.length());

					long percent = fetch_document_offset*100/udm_objects.length();
					if (percent != status_percent) {
						status_percent = percent;
						tooladaptor->getFetchStatus(percent);
					}
				
				} else {
					temp_fetch_document_offset--;
				}
			}

			reqCopyObjectHierarchy(p_dstBackend, root_meta, udm_objects,
        p_dstChild, udm_objects[i].id.id, finished, cam, tooladaptor);

      //cout << "====================" << endl;
			p_dstChild->release();

		}
	}
  //cout << "========================" << endl;
}
//==========================================

  void reqCopyObjectAttrs(Udm::DataNetwork* p_dstBackend, 
                            const Uml::Diagram &root_meta, 
                            const OTIF::Obj_S &udm_objects, 
                            ObjectImpl *parent, 
                            const unsigned long parent_id, 
                            bool &finished, 
                            copy_assoc_map &cam, 
                            ToolAdaptor *tooladaptor) 
{
    //cout << "=======ATTR================" << endl;

  Uml::Class parent_meta = parent->type();
//   cout << "-----------" << endl;
//   cout << "parent " << parent_meta.name() << endl;


  LongLongMap not_ready;

 	for(CORBA::ULong i=1; i<udm_objects.length(); i++)
  {
    long id = udm_objects[i].id.id;
    not_ready[id] = i;
  }  

  LongLongMap::iterator itr =  not_ready.begin();
  while (!not_ready.empty())
  {
	
    i = itr->second;

    long id = udm_objects[i].id.id;
  
		long __srcArchetypeID;
		long __realarchetype;
		long __subtype;


		__extractUserData(udm_objects[i].userData, __subtype, __realarchetype, __srcArchetypeID);

		ObjectImpl* p_dstChild = &Udm::_null;
		Udm::Object dstChild;

    LongLongMap::const_iterator it = not_ready.find(__srcArchetypeID);
    if (it != not_ready.end())
    {
      itr++;
      continue;
    }
    else
    {
      not_ready.erase(itr);
      itr = not_ready.begin();
    }

    

		// get the dst
		copy_assoc_map::iterator cam_i=cam.find( udm_objects[i].id.id );
		if (cam_i != cam.end())
    {
			dstChild = (*cam_i).second;
			p_dstChild = dstChild.__impl()->clone();
		}
		
    if (p_dstChild == 0)
      throw udm_exception("CORBADataNetwork is corrupt => A meta of a class is not found!");

    //cout << "set attr";
		
		ObjectImplAttrsPass(p_dstChild, udm_objects[i]); 
    

		// show the status of the fetch process
		if (tooladaptor)
    {

			if (temp_fetch_document_offset==0) {
				fetch_document_offset++;

				tooladaptor->getFetchStatus(fetch_document_offset, udm_objects.length());

				long percent = fetch_document_offset*100/udm_objects.length();
				if (percent != status_percent) {
					status_percent = percent;
					tooladaptor->getFetchStatus(percent);
				}
			
			} else {
				temp_fetch_document_offset--;
			}
		}

	    //cout << "====================" << endl;
		p_dstChild->release();

    

  }//while
  //cout << "========================" << endl;
}


}//namespace UdmCORBA