
// 11-28-04 fixed - derived attributes were not set  - kalmar
// 11-05-04 fixed - temporary attributes had no servant - kalmar

/*
Copyright (c) Vanderbilt University, 2000-2001
ALL RIGHTS RESERVED
Vanderbilt University disclaims all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall Vanderbilt University be liable for
any special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

#pragma warning( disable : 4786 )
#pragma warning( disable : 4518 )

// -------------------------------------------------------------------------------------------------------------------------------
// CORBA includes
// -------------------------------------------------------------------------------------------------------------------------------
#include "otifdataC.h"
#include "otifC.h"
#include "otif_i.h"
#include <ace/streams.h>

// -------------------------------------------------------------------------------------------------------------------------------
// Networking include
// -------------------------------------------------------------------------------------------------------------------------------
#include <winsock2.h>
#include <direct.h>

// -------------------------------------------------------------------------------------------------------------------------------
// CORBA global variables
// -------------------------------------------------------------------------------------------------------------------------------
#include "CORBAObject.h"
CORBA::ORB_var __orb;
PortableServer::POA_var __poa;
PortableServer::POAManager_var __poa_mgr;
string SERVER_IP_PORT;

OTIF::Backplane_var __backplane_obj;
OTIF::Diagram_var __diagram;
long __diagram_id;
OTIF::NameVersion __paradigm;

int __UdmCORBA_argc;
char **__UdmCORBA_argv;

// -------------------------------------------------------------------------------------------------------------------------------
// Global variables
// -------------------------------------------------------------------------------------------------------------------------------
string IP_ADDRESS;
unsigned long PROCESS_ID;
unsigned long THREAD_ID;

map<long, OTIF::ToolAdaptorSession_var> tooladaptor_rep; // Map for tooladaptor_session, tooladaptor servant


map<long, OTIF_ToolAdaptorNotifyCallback_i *> subscriptions_rep; // session_id, OTIF_ToolAdaptorNotifyCallback_i *

// Internal Error report stuff
const static char INTERNAL_ERROR_FILE[] = "_otif_internal_error.log";

// -------------------------------------------------------------------------------------------------------------------------------
//	UdmCORBA namespace
// -------------------------------------------------------------------------------------------------------------------------------
namespace UdmCORBA
{

  void reqCopyObjectAttrs(Udm::DataNetwork* p_dstBackend, 
                            const Uml::Diagram &root_meta, 
                            const OTIF::Obj_S &udm_objects, 
                            ObjectImpl *parent, 
                            const unsigned long parent_id, 
                            bool &finished, 
                            copy_assoc_map &cam, 
                            ToolAdaptor *tooladaptor) ;
  void reqCopyObjectHierarchy(Udm::DataNetwork* p_dstBackend, 
                            const Uml::Diagram &root_meta, 
                            const OTIF::Obj_S &udm_objects, 
                            ObjectImpl *parent, 
                            const unsigned long parent_id, 
                            bool &finished, 
                            copy_assoc_map &cam, 
                            ToolAdaptor *tooladaptor) ;
	void __printTime(Logger *logger);
  void ObjectImplAttrsPass(Udm::ObjectImpl *obj, const OTIF::Obj &udm_obj);
extern	map<long, Udm::ObjectImpl*> objectimpl_rep;

extern  long status_percent;
extern  long fetch_document_offset;
extern  long temp_fetch_document_offset;



	class CORBAObject;
	typedef Udm::ObjectImpl ObjectImpl;

	// Forward declaration
	void corba_shutdown();

	map<long, CORBAObject*> servant_rep; // map (servant unique ID, CORBAObject pointer)

	struct CACHE_BODY {
		long meta_id;
		long servant_id;
	};
	struct ATTR_CACHE_BODY {
		long meta_id;
		string STRING;
		bool BOOL;
		long LONG;
	};
	multimap<long, CACHE_BODY> cache; // map (CORBAObject uniqueId, CACHE_BODY)
	multimap<long, ATTR_CACHE_BODY> attr_cache; // map (CORBAObject uniqueId, CACHE_BODY)

  


// *******************************************************************************************************************************	
// ===============================================================================================================================
//  CORBADataNetwork Wrapping
// ===============================================================================================================================
// *******************************************************************************************************************************	
	CORBA::ULong datanetwork_object_offset = 0;
	long obtainID(long unique_id);


// Build up the AttrValue_S
	void AttrValueSeqPass(Udm::ObjectImpl *object, OTIF::AttrValue_S_var &udm_attrs)
   {
		Uml::Class meta = object->type();
		CORBA::ULong offset = 0;
		
		CORBA::ULong size = 0;
		set<Uml::Class> cancs = Uml::AncestorClasses(meta);
		for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++)
      {
			set<Uml::Attribute> attrs = (*q).attributes();
			size += attrs.size();
		}
		udm_attrs->length(size);

		for(q=cancs.begin(); q!=cancs.end(); q++) {
			set<Uml::Attribute> attrs = (*q).attributes();

			for(set<Uml::Attribute>::iterator i=attrs.begin(); i!=attrs.end(); i++)
         {
				OTIF::AttrValue_var udm_attr = new OTIF::AttrValue();
				OTIF::AttrValueHolder attr_value;

//  				string temp = i->name();
//  				if (!temp.compare("name")) 
//            cout << "attr val: " << object->getStringAttr(*i).c_str() << endl;

				try {
					if ( string(i->type())=="String" ) 
                  attr_value.stringValue( ((string) object->getStringAttr(*i)).c_str() );
					if ( string(i->type())=="Integer" ) 
                  attr_value.intValue( (long)object->getIntegerAttr(*i) );
					if ( string(i->type())=="Boolean" ) 
                  attr_value.boolValue( object->getBooleanAttr(*i) );
					if ( string(i->type())=="Real" ) 
                  attr_value.realValue( object->getRealAttr(*i) );
				} catch (udm_exception&) {
					continue;
				}

				CORBAObject *co = (CORBAObject*) (*i).__impl();
//				udm_attr->type = Attrib::_duplicate( co->m_servant._attrib );
				udm_attr->tId = co->m_servant_id;	

				udm_attr->value = attr_value;
				udm_attrs[offset++] = udm_attr;
			}
		}
	}

// Build up the ParentRole_S
	void ParentRoleSeqPass(Udm::ObjectImpl *object, OTIF::ParentRole_var &udm_parent) {
		Uml::Class meta = object->type();

		Udm::ObjectImpl *parent = object->getParent(NULL);
		if (parent==&Udm::_null || !parent) return;

		Uml::Composition comp = Uml::matchChildToParent(meta, parent->type());
		Uml::CompositionParentRole c_pr = comp.parentRole();

		CORBAObject *co = (CORBAObject*) c_pr.__impl();
//		udm_parent->type = CompositionParentRole::_duplicate( co->m_servant._parentrole );
		udm_parent->tId = co->m_servant_id;

		udm_parent->parentId.id = obtainID( parent->uniqueId() );

		parent->release();
	}

// Build up the AssocRole_S
	void AssocRoleSeqPass(Udm::ObjectImpl *object, OTIF::AssocRole_S_var &udm_assocs) {
		bool assoc_class = false;
		Uml::Class meta = object->type();

//		string class_meta = meta.name();
//		if (!class_meta.compare("Transition"))
//			cout << class_meta << endl;

		CORBA::ULong offset=0;
		set<Uml::Class> cancs = Uml::AncestorClasses(meta);
		for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++) {
			set<Uml::AssociationRole> assoc_role = q->associationRoles();

			for(set<Uml::AssociationRole>::iterator i=assoc_role.begin(); i!=assoc_role.end(); i++) {

				Uml::AssociationRole o_role = Uml::theOther(*i);
				Uml::Class assocClass = Uml::Association(i->parent()).assocClass();

				vector<Udm::ObjectImpl*> srcPeers = object->getAssociation( \
					Uml::theOther(*i), assocClass ? Udm::CLASSFROMTARGET : Udm::TARGETFROMPEER);
				
				// If there is no association, we don`t want to store anything :)
				if (srcPeers.size()==0) continue;

				OTIF::AssocRole_var udm_assoc_role = new OTIF::AssocRole();
				CORBAObject *co = (CORBAObject*) o_role.__impl();
//				udm_assoc_role->type = AssociationRole::_duplicate( co->m_servant._associationrole );
				udm_assoc_role->tId = co->m_servant_id;

				OTIF::objectId_S_var ids = new OTIF::objectId_S();
				ids->length(srcPeers.size());
				CORBA::ULong id_offset=0;

				for(vector<ObjectImpl*>::iterator j=srcPeers.begin(); j!=srcPeers.end(); j++) {
//					string name = (*j)->type().name();
//					long asd = obtainID( (*j)->uniqueId() );
				
					ids[id_offset++].id = obtainID( (*j)->uniqueId() );

					(*j)->release();
				}
				udm_assoc_role->objectIds = ids;

				udm_assocs->length(offset+1);
				udm_assocs[offset++] = udm_assoc_role;
			}
		}
	}


// Build up the Obj_S
	void ObjSeqPass(Udm::ObjectImpl *object, OTIF::Obj_S_var &udm_objects, ToolAdaptor *tooladaptor, long object_size) {

		OTIF::Obj_var udm_object = new OTIF::Obj();
		Uml::Class meta = object->type();
		string meta_name = meta.name();

		CORBAObject *co = (CORBAObject*) meta.__impl();
		if (!co->m_servant._class) throw udmcorba_exception("There is no remote type information to send the datanetwork.", 0);
//		udm_object->type = Class::_duplicate( co->m_servant._class );
		udm_object->tId = co->m_servant_id;

		udm_object->id.id = obtainID( object->uniqueId() );
		udm_object->id.host = inet_addr(IP_ADDRESS.c_str());
		udm_object->id.process = PROCESS_ID;
		udm_object->id.thread = THREAD_ID;
//		cout << "Object_meta:uniqueId => " << (string) meta.name() << ":" << udm_object->id.id << endl;

		// Build up the AttrValue_S
		OTIF::AttrValue_S_var udm_attrs = new OTIF::AttrValue_S();
		udm_attrs->length(0);
		AttrValueSeqPass(object, udm_attrs);
		udm_object->attrs = udm_attrs;

		// Build up the ParentRole_S
		OTIF::ParentRole_var udm_parent = new OTIF::ParentRole();
		ParentRoleSeqPass(object, udm_parent);
		udm_object->parent = udm_parent;

		// Build up the AssocRole_S
		OTIF::AssocRole_S_var udm_assocs = new OTIF::AssocRole_S();
		udm_assocs->length(0);
		AssocRoleSeqPass(object, udm_assocs);
		udm_object->assocs = udm_assocs;

		// archetype mapping
		// userData value:
		// ---------------
		//	boolean subtype;		// true - it is derived, false - it is an instance
		//	boolean realarchetype;	// true - it is real archetype, false - it is NOT real archetpye
		//	long archetype;			// if archetype an object it points to the ID of its archetype object, if not it is -1

		string userData = "";
		ObjectImpl *p_srcChildArc = object->getArchetype();
		Udm::Object srcChildArc(p_srcChildArc);
		if (p_srcChildArc && p_srcChildArc!=&Udm::_null) 
      {

			long _archetype_id = obtainID(p_srcChildArc->uniqueId());

			// is this a real archetype
			if (object->hasRealArchetype()) {
				bool subtype = 0;

				// get derived for subtype
				Udm::Object srcChild = object->clone();
				set<Udm::Object> deriveds = srcChildArc.derived();
				if (deriveds.find(srcChild) != deriveds.end()) subtype = 1;
				
				userData += (subtype==1) ? "1" : "0";
				userData += "1";
				char _temp[16];
				ltoa(_archetype_id, _temp, 10);
				userData += _temp;
				
//				udm_object->archetype = obtainID(p_srcChildArc->uniqueId());
//				udm_object->subtype = subtype;
//				udm_object->realarchetype = 1;

			} else {
				userData += "0";
				userData += "0";
				char _temp[16];
				ltoa(_archetype_id, _temp, 10);
				userData += _temp;

//				udm_object->archetype = obtainID(p_srcChildArc->uniqueId());
//				udm_object->subtype = 0;
//				udm_object->realarchetype = 0;
			}

		}
      else
      {
			userData += "0";
			userData += "0";
			userData += "-1";

//			udm_object->archetype = -1;
		}

		udm_object->userData <<= userData.c_str();

//		cout << obtainID(object->uniqueId()) << " : " << udm_object->archetype << " : " << udm_object->realarchetype << endl;
		
		udm_objects[datanetwork_object_offset++] = udm_object;

		// show the status of the publish process
		if (tooladaptor) {
			tooladaptor->getPublishStatus(datanetwork_object_offset, object_size);

			long percent = datanetwork_object_offset*100/object_size;
			if (percent != status_percent) {
				status_percent = percent;
				tooladaptor->getPublishStatus(percent);
			}
		}

		// Get the Children classes reqursive
		vector<ObjectImpl*> o = object->getChildren(NULL, NULL);

		for(vector<ObjectImpl*>::iterator j = o.begin(); j != o.end(); j++) {
			ObjSeqPass((*j), udm_objects, tooladaptor, object_size);

			(*j)->release();
		}
	}



///=====================================================
	Uml::AssociationRole getAssocRoleMeta(Uml::Class &class_meta, const long assocrole_servant, bool &assoc_class) {
		map<long, CORBAObject*>::iterator m_obj = servant_rep.find( assocrole_servant );

		// Optimalization
		string meta_name = m_obj->second->m_name;
//		CORBA::String_var meta_name_var = m_obj->second->m_servant._associationrole->name();//assocrole_servant->name();
//		string meta_name = meta_name_var;

		set<Uml::Class> cancs = Uml::AncestorClasses(class_meta);
		for(set<Uml::Class>::iterator q=cancs.begin(); q!=cancs.end(); q++) {
			set<Uml::AssociationRole> uml_assocs = (*q).associationRoles();

			for(set<Uml::AssociationRole>::iterator i=uml_assocs.begin(); i!=uml_assocs.end(); i++) {
				Uml::AssociationRole target = Uml::theOther(*i);
				Uml::Class assocClass = Uml::Association(i->parent()).assocClass();
				if (!meta_name.compare( target.name() )) {
					if (assocClass) assoc_class = true;
					return *i;
				}
			}
		}
		return NULL;
	}

///=====================================================
// Build up the Assocations
	void ObjectImplAssocPass(const OTIF::Obj_S &udm_objects, copy_assoc_map &cam) {

		for(CORBA::ULong i=0; i<udm_objects.length(); i++) {
			// if it is an archetype we don't have to create links, backend does! i was wrong, backend did not do ;-))
			//			if (udm_obj->archetype!=-1) continue;

			copy_assoc_map::iterator rep_find = cam.find( udm_objects[i].id.id );
			if (rep_find==cam.end()) throw udm_exception("CORBADataNetwork is corrupt => Maybe an object is not created!");

			ObjectImpl *obj_impl = (*rep_find).second.__impl()->clone();
			Uml::Class obj_meta = obj_impl->type();

//			cout << "Class type: " << (string)obj_meta.name() << endl;

			for(CORBA::ULong w=0; w<udm_objects[i].assocs.length(); w++) {
				bool assoc_class = false;
		
				Uml::AssociationRole assocrole_meta = getAssocRoleMeta(obj_meta, udm_objects[i].assocs[w].tId, assoc_class);
				if (!assocrole_meta) throw udm_exception("CORBADataNetwork is corrupt => A meta of an AssocRole is not found!");
				Uml::AssociationRole target_meta = Uml::theOther(assocrole_meta);

//				cout << (string)assocrole_meta.name() << ":" << (string)target_meta.name() << endl;

				vector<ObjectImpl*> impl_vec;
				for(CORBA::ULong j=0; j<udm_objects[i].assocs[w].objectIds.length(); j++) {
					copy_assoc_map::iterator rep = cam.find( udm_objects[i].assocs[w].objectIds[j].id );
					if (rep!=cam.end()) {
//						cout << (string) (*rep).second->type().name() << endl;
						impl_vec.push_back( (*rep).second.__impl()->clone() );
					} else {
						throw udm_exception("CORBADataNetwork is corrupt!");
					}
 				}

				if (!assoc_class) {
					try {
//						cout << "Normal setAssociation: " << impl_vec.size() << endl;
						obj_impl->setAssociation(target_meta, impl_vec, Udm::TARGETFROMPEER, false);
					} catch (udm_exception e) {
						cout << e.what() << endl;
					}
				} else {
					try {
//						cout << "AssocClass setAssociation: " << impl_vec.size() << endl;
						obj_impl->setAssociation(target_meta, impl_vec, Udm::CLASSFROMTARGET, false);
					} catch (udm_exception e) {
						cout << e.what() << endl;
					}
				}

				for(vector<ObjectImpl*>::iterator qq=impl_vec.begin(); qq!=impl_vec.end(); qq++) (*qq)->release();
			} // for

			obj_impl->release();		
		}
	}


// Build up the ObjectImpls
void ObjectImplsPass(Udm::DataNetwork* p_dstBackend, const Uml::Diagram &root_meta, const OTIF::Obj_S &udm_objects, Udm::ObjectImpl *parent, long parent_id, ToolAdaptor *tooladaptor) {

		fetch_document_offset = 1;
		temp_fetch_document_offset = 0;

		copy_assoc_map cam;

		copy_assoc_map::value_type map_item(parent_id, parent->clone());
		cam.insert(map_item);

		bool finished = true;
		bool first = true;

		while(!finished || first) {
			first = false;
			finished = true;

			reqCopyObjectHierarchy(p_dstBackend, root_meta, udm_objects, parent, parent_id, finished, cam, tooladaptor);

			temp_fetch_document_offset = fetch_document_offset-1;
		}
 
 			reqCopyObjectAttrs(p_dstBackend, root_meta, udm_objects, parent, parent_id, finished, cam, tooladaptor);

    

		ObjectImplAssocPass(udm_objects, cam);
	}




// *******************************************************************************************************************************	
// ===============================================================================================================================
//  Public functions
// ===============================================================================================================================
// *******************************************************************************************************************************	
	Object CreateObject(const Uml::Class &meta) {
		return new CORBAObject(meta, 1);
	}

	typedef map<string, OTIF::Class_var> CLASS_SERVANTS;
	CLASS_SERVANTS class_servants;

	struct ATTRIB_SERVANT {
		string name;
		OTIF::Attrib_var pointer;
	};
	typedef multimap<long, ATTRIB_SERVANT> ATTRIB_SERVANTS;
	ATTRIB_SERVANTS attrib_servants;

	struct COMPOSITION_SERVANT {
		string parentTarget_name;
		string parentRole_name;
		string childTarget_name;
		string childRole_name;
		OTIF::Composition_var pointer;
	};
	typedef multimap<string, COMPOSITION_SERVANT> COMP_SERVANTS;
	COMP_SERVANTS comp_servants;

	struct ASSOCIATION_SERVANT {
		string role1_name;
		string role2_name;
		string target1_name;
		string target2_name;
		OTIF::Association_var pointer;
		OTIF::AssociationRole_var role1_pointer;
		OTIF::AssociationRole_var role2_pointer;
	};
	typedef multimap<string, ASSOCIATION_SERVANT> ASSOC_SERVANTS;
	ASSOC_SERVANTS assoc_servants;
 
	// -------------------------------------------------------------------------------------------------------------------------------
	//  InitDiagram function
	// -------------------------------------------------------------------------------------------------------------------------------
	void InitDiagram(const Uml::Diagram &obj) {
		try {
			vector<string> null_vector;
			null_vector.push_back("");
			if (CORBA::is_nil(__backplane_obj)) corba_startup(null_vector);

			// Check if it was initialized and the remote paradigm is specified
			if (!((string)__paradigm.name).compare("") && !((string)__paradigm.version).compare(""))
				throw udmcorba_exception("UdmCORBA was not initialized. Please use UdmCORBA::Initialize function before creating UdmCORBA datanetwork.", 0);

			// Fetch paradigm
			__diagram = __backplane_obj->fetchParadigm(__paradigm);
			
			CORBAObject *d = (CORBAObject*) obj.__impl();
			d->m_servant._diagram = OTIF::Diagram::_duplicate(__diagram);
			d->m_servant_id = d->m_servant._diagram->tId();
			servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );

			__diagram_id = d->m_servant_id;

//			cout << "Paradigm name: " << __diagram->name() << endl;
//			cout << "Paradigm version: " << __diagram->version() << endl;

			class_servants.clear();
			attrib_servants.clear();
			comp_servants.clear();
			assoc_servants.clear();
	
			// Get the classes, attributes and cache them
			OTIF::Class_S_var classes = d->m_servant._diagram->classes();
			for(CORBA::ULong i=0; i<classes->length(); i++) {

				CORBA::String_var name_var = classes[i]->name();
				string name = name_var;
				class_servants[name] = OTIF::Class::_duplicate(classes[i]);

				long _id = classes[i]->tId();
				OTIF::Attrib_S_var attribs = classes[i]->attributes();
				for(CORBA::ULong j=0; j<attribs->length(); j++) {

					ATTRIB_SERVANT as;
					CORBA::String_var name_var = attribs[j]->name();
					as.name = name_var;
					as.pointer = OTIF::Attrib::_duplicate(attribs[j]);

					attrib_servants.insert( pair<const long, ATTRIB_SERVANT>(_id, as) );
				}
			}
	
			// Get the compositions and cache them
			OTIF::Composition_S_var comps = d->m_servant._diagram->compositions();
			for(i=0; i<comps->length(); i++) {
				CORBA::String_var name_var = comps[i]->name();
				string name = name_var;

				COMPOSITION_SERVANT cs;
		
				OTIF::CompositionParentRole_var cparent_var = comps[i]->parentRole();
				OTIF::Class_var class_var = cparent_var->target();
				CORBA::String_var parentTarget_name_var = class_var->name();
				cs.parentTarget_name = parentTarget_name_var;

				CORBA::String_var parentRole_name_var = cparent_var->name();
				cs.parentRole_name = parentRole_name_var; 

				OTIF::CompositionChildRole_var cchild_var = comps[i]->childRole();
				OTIF::Class_var class2_var = cchild_var->target();
				CORBA::String_var childTarget_name_var = class2_var->name();
				cs.childTarget_name = childTarget_name_var;

				CORBA::String_var childRole_name_var = cchild_var->name();
				cs.childRole_name = childRole_name_var;

				cs.pointer = OTIF::Composition::_duplicate(comps[i]);
		
				comp_servants.insert( pair<const string, COMPOSITION_SERVANT>(name, cs) );
			}

			// Get the associations and cache them
			OTIF::Association_S_var assocs = d->m_servant._diagram->associations();
			for(i=0; i<assocs->length(); i++) {
				CORBA::String_var name_var = assocs[i]->name();
				string name = name_var;

				OTIF::AssociationRole_S_var ass_roles = assocs[i]->roles();
				if (ass_roles->length()!=2) throw udm_exception("UdmCORBA ERROR: An association without two endpoints is invalid!");

				ASSOCIATION_SERVANT as;
				CORBA::String_var role1_name_var = ass_roles[(CORBA::ULong) 0]->name();
				as.role1_name = role1_name_var;

				CORBA::String_var role2_name_var = ass_roles[(CORBA::ULong) 1]->name();
				as.role2_name = role2_name_var;

				OTIF::Class_var cls1_var = ass_roles[(CORBA::ULong) 0]->target();
				CORBA::String_var target1_name_var = cls1_var->name();
				as.target1_name = target1_name_var;

				OTIF::Class_var cls2_var = ass_roles[(CORBA::ULong) 1]->target();
				CORBA::String_var target2_name_var = cls2_var->name();
				as.target2_name = target2_name_var;

				as.pointer = OTIF::Association::_duplicate(assocs[i]);
				as.role1_pointer = OTIF::AssociationRole::_duplicate(ass_roles[(CORBA::ULong) 0]);
				as.role2_pointer = OTIF::AssociationRole::_duplicate(ass_roles[(CORBA::ULong) 1]);
		
				assoc_servants.insert( pair<const string, ASSOCIATION_SERVANT>(name, as) );
			}

		} catch(OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is not found!", 11);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the InitDiagram function.", 0);
		}
	}

	// -------------------------------------------------------------------------------------------------------------------------------
	//  InitClass function
	// -------------------------------------------------------------------------------------------------------------------------------
	void InitClass(const Uml::Class &obj, const Uml::Diagram &parent, const char *name) {
		try {
			CLASS_SERVANTS::iterator i = class_servants.find(name);
			if (i==class_servants.end()) throw udmcorba_exception("A class was not found in the paradigm. Maybe the paradigms are different.", 19);

			CORBAObject *par = (CORBAObject*) parent.__impl();
			CORBAObject *d = (CORBAObject*) obj.__impl();
			d->m_servant._class = OTIF::Class::_duplicate((*i).second);
			d->m_servant_id = d->m_servant._class->tId();
			d->m_parent = par;

			// Optimalization
			d->m_name = (*i).first;

			servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the InitClass function.", 0);
		}
	}
  
	// -------------------------------------------------------------------------------------------------------------------------------
	//  InitAttribute function
	// -------------------------------------------------------------------------------------------------------------------------------
	void InitAttribute(const Uml::Attribute &obj, const Uml::Class &parent, const char *name) {

		try {
		    CORBAObject *par = (CORBAObject*) parent.__impl();

			pair<ATTRIB_SERVANTS::iterator, ATTRIB_SERVANTS::iterator> i = attrib_servants.equal_range( par->m_servant._class->tId() );

      bool found = false;
			while (i.first != i.second)
      {
				if (!strcmp((*(i.first)).second.name.c_str(), name))
        {

					CORBAObject *d = (CORBAObject*) obj.__impl();
					d->m_servant._attrib = OTIF::Attrib::_duplicate((*(i.first)).second.pointer);
					d->m_servant_id = d->m_servant._attrib->tId();
					d->m_parent = par;

					// Optimalization
					d->m_name = (*(i.first)).second.name;
					d->m_attrib_type = d->m_servant._attrib->typeName();

					servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );
          found = true;
        }
				++i.first;
			}
    //it is a temporary attribute, there is no servat for this, we fill out some necessay members
 		if (!found)
		{
			CORBAObject *d = (CORBAObject*) obj.__impl();

      d->m_parent = (CORBAObject*) parent.__impl();
      d->m_name = std::string(name);
		}

      
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the InitAttribute function.", 0);
		}
	}

	// -------------------------------------------------------------------------------------------------------------------------------
	//  InitComposition function
	// -------------------------------------------------------------------------------------------------------------------------------
	void InitComposition(const Uml::Composition &obj, const Uml::Diagram &parent, const char *name, 
		const Uml::CompositionParentRole &prole, const char *pname, const Uml::Class &ptarget, 
		const Uml::CompositionChildRole &crole, const char *cname, const Uml::Class &ctarget) {

		try {
			CORBAObject *par = (CORBAObject*) parent.__impl();

			CORBAObject *par2 = (CORBAObject*) ptarget.__impl();
			CORBA::String_var parent_target_name_var = par2->m_servant._class->name();
			string parent_target_name = parent_target_name_var;

			CORBAObject *par3 = (CORBAObject*) ctarget.__impl();
			CORBA::String_var child_target_name_var = par3->m_servant._class->name();
			string child_target_name = child_target_name_var;

			bool error = true;
			for(COMP_SERVANTS::iterator i=comp_servants.begin(); i!=comp_servants.end(); i++) {
				if (!strcmp((*i).second.parentTarget_name.c_str(), parent_target_name.c_str()) && 
					!strcmp((*i).second.parentRole_name.c_str(), pname) &&
					!strcmp((*i).second.childTarget_name.c_str(), child_target_name.c_str()) &&	
					!strcmp((*i).second.childRole_name.c_str(), cname)
					) {

					// has to be cloned, udm generated api does not create it global...
					CORBAObject *d = (CORBAObject*) obj.__impl()->clone();
					d->m_servant._composition = OTIF::Composition::_duplicate((*i).second.pointer);
					d->m_servant_id = d->m_servant._composition->tId();
					d->m_parent = par;
					servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );

					CORBAObject *d2 = (CORBAObject*) prole.__impl();
					d2->m_servant._parentrole = OTIF::CompositionParentRole::_duplicate((*i).second.pointer->parentRole());
					d2->m_servant_id = d2->m_servant._parentrole->tId();
					d2->m_parent = d;
					servant_rep.insert( pair<long const, CORBAObject*>(d2->m_servant_id, (CORBAObject*) d2) );

					CORBAObject *d3 = (CORBAObject*) crole.__impl();
					d3->m_servant._childrole = OTIF::CompositionChildRole::_duplicate((*i).second.pointer->childRole());
					d3->m_servant_id = d3->m_servant._childrole->tId();
					d3->m_parent = d;
					servant_rep.insert( pair<long const, CORBAObject*>(d3->m_servant_id, (CORBAObject*) d3) );

					error = false;
					break;
				}
			}
			if (error) throw udmcorba_paradigm_exception("A composition was not found in the paradigm", 12);

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the InitComposition function.", 0);
		}
	}

	// -------------------------------------------------------------------------------------------------------------------------------
	//  InitAssociation function
	// -------------------------------------------------------------------------------------------------------------------------------
	void InitAssociation(const Uml::Association &obj, const Uml::Diagram &parent, const char *name, 
		const Uml::AssociationRole &ass1_role, const char *ass1_name, const Uml::Class &ass1_target, 
		const Uml::AssociationRole &ass2_role, const char *ass2_name, const Uml::Class &ass2_target) {

		try {
		
			CORBAObject *par = (CORBAObject*) parent.__impl();

			CORBAObject *par3 = (CORBAObject*) ass1_target.__impl();
			CORBA::String_var target1_class_name_var = par3->m_servant._class->name();
			string target1_class_name = target1_class_name_var;

			CORBAObject *par4 = (CORBAObject*) ass2_target.__impl();
			CORBA::String_var target2_class_name_var = par4->m_servant._class->name();
			string target2_class_name = target2_class_name_var;

			bool error = true;
			for(ASSOC_SERVANTS::iterator i=assoc_servants.begin(); i!=assoc_servants.end(); i++) {
				if (!strcmp((*i).second.role1_name.c_str(), ass1_name) && 
					!strcmp((*i).second.role2_name.c_str(), ass2_name) &&
					!strcmp((*i).second.target1_name.c_str(), target1_class_name.c_str()) && 
					!strcmp((*i).second.target2_name.c_str(), target2_class_name.c_str())
					) {

					// has to be cloned, udm generated api does not create it global...
					CORBAObject *d = (CORBAObject*) obj.__impl()->clone();
					d->m_servant._association = OTIF::Association::_duplicate((*i).second.pointer);
					d->m_servant_id = d->m_servant._association->tId();
					d->m_parent = par;
					servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );

					CORBAObject *d2 = (CORBAObject*) ass1_role.__impl();
					d2->m_servant._associationrole = OTIF::AssociationRole::_duplicate((*i).second.role1_pointer);
					d2->m_servant_id = d2->m_servant._associationrole->tId();

					// Optimalization
					d2->m_name = d2->m_servant._associationrole->name();

					d2->m_parent = d;
					servant_rep.insert( pair<long const, CORBAObject*>(d2->m_servant_id, (CORBAObject*) d2) );

					CORBAObject *d3 = (CORBAObject*) ass2_role.__impl();
					d3->m_servant._associationrole = OTIF::AssociationRole::_duplicate((*i).second.role2_pointer);
					d3->m_servant_id = d3->m_servant._associationrole->tId();

					// Optimalization
					d3->m_name = d3->m_servant._associationrole->name();

					d3->m_parent = d;
					servant_rep.insert( pair<long const, CORBAObject*>(d3->m_servant_id, (CORBAObject*) d3) );
			
					error = false;
					break;
				}
				if (!strcmp((*i).second.role2_name.c_str(), ass1_name) && 
					!strcmp((*i).second.role1_name.c_str(), ass2_name) &&
					!strcmp((*i).second.target2_name.c_str(), target1_class_name.c_str()) && 
					!strcmp((*i).second.target1_name.c_str(), target2_class_name.c_str())
					) {

					// has to be cloned, udm generated api does not create it global...
					CORBAObject *d = (CORBAObject*) obj.__impl()->clone();
					d->m_servant._association = OTIF::Association::_duplicate((*i).second.pointer);
					d->m_servant_id = d->m_servant._association->tId();
					d->m_parent = par;
					servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );

					CORBAObject *d2 = (CORBAObject*) ass1_role.__impl();
					d2->m_servant._associationrole = OTIF::AssociationRole::_duplicate((*i).second.role2_pointer);
					d2->m_servant_id = d2->m_servant._associationrole->tId();

					// Optimalization
					d2->m_name = d2->m_servant._associationrole->name();

					d2->m_parent = d;
					servant_rep.insert( pair<long const, CORBAObject*>(d2->m_servant_id, (CORBAObject*) d2) );

					CORBAObject *d3 = (CORBAObject*) ass2_role.__impl();
					d3->m_servant._associationrole = OTIF::AssociationRole::_duplicate((*i).second.role1_pointer);
					d3->m_servant_id = d3->m_servant._associationrole->tId();

					// Optimalization
					d3->m_name = d3->m_servant._associationrole->name();

					d3->m_parent = d;
					servant_rep.insert( pair<long const, CORBAObject*>(d3->m_servant_id, (CORBAObject*) d3) );
			
					error = false;
					break;
				}
			}
			if (error) throw udmcorba_paradigm_exception("An association was not found in the paradigm", 13);

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the InitAssociation function.", 0);
		}
	}


// -------------------------------------------------------------------------------------------------------------------------------
//  CACHING stuffs
// -------------------------------------------------------------------------------------------------------------------------------
	void _put_in_cache(long id, long meta_id, long servant_id) {
		CACHE_BODY cb;
		cb.meta_id = meta_id;
		cb.servant_id = servant_id;
		cache.insert(pair<long const, CACHE_BODY>(id, cb));
	}

	void _put_in_attr_cache(long id, long meta_id, string STRING) {
		ATTR_CACHE_BODY cb;
		cb.meta_id = meta_id;
		cb.STRING = STRING;
		attr_cache.insert(pair<long const, ATTR_CACHE_BODY>(id, cb));
	}
	void _put_in_attr_cache(long id, long meta_id, bool BOOL) {
		ATTR_CACHE_BODY cb;
		cb.meta_id = meta_id;
		cb.BOOL = BOOL;
		attr_cache.insert(pair<long const, ATTR_CACHE_BODY>(id, cb));
	}
	void _put_in_attr_cache(long id, long meta_id, long LONG) {
		ATTR_CACHE_BODY cb;
		cb.meta_id = meta_id;
		cb.LONG = LONG;
		attr_cache.insert(pair<long const, ATTR_CACHE_BODY>(id, cb));
	}

	bool _find_in_cache(long id, long meta_id, vector<ObjectImpl*> &ret) {
		bool cached = false;

    
		pair< typename multimap<long, CACHE_BODY>::iterator, multimap<long, CACHE_BODY>::iterator > t = cache.equal_range(id);
		while(t.first != t.second) {
			if ((*t.first).second.meta_id == meta_id) {
				cached = true;
				map<long, CORBAObject*>::iterator i = servant_rep.find( (*t.first).second.servant_id );
				if (i!=servant_rep.end()) {
					ret.push_back( (*i).second->clone() );
				}
			}
			++t.first;
		}
		return cached;
	}

	bool _find_in_cache(long id, long meta_id, string &obj) {
		pair<multimap<long, ATTR_CACHE_BODY>::iterator, multimap<long, ATTR_CACHE_BODY>::iterator> t = attr_cache.equal_range(id);
		while(t.first != t.second) {
			if ((*t.first).second.meta_id == meta_id) {
				obj = (*t.first).second.STRING;
				return true;
			}
			++t.first;
		}
		return false;
	}

	bool _find_in_cache(long id, long meta_id, long &obj) {
		pair<multimap<long, ATTR_CACHE_BODY>::iterator, multimap<long, ATTR_CACHE_BODY>::iterator> t = attr_cache.equal_range(id);
		while(t.first != t.second) {
			if ((*t.first).second.meta_id == meta_id) {
				obj = (*t.first).second.LONG;
				return true;
			}
			++t.first;
		}
		return false;
	}

	bool _find_in_cache(long id, long meta_id, bool &obj) {
		pair<multimap<long, ATTR_CACHE_BODY>::iterator, multimap<long, ATTR_CACHE_BODY>::iterator> t = attr_cache.equal_range(id);
		while(t.first != t.second) {
			if ((*t.first).second.meta_id == meta_id) {
				obj = (*t.first).second.BOOL;
				return true;
			}
			++t.first;
		}
		return false;
	}


// -------------------------------------------------------------------------------------------------------------------------------
//  Network stuffs
// -------------------------------------------------------------------------------------------------------------------------------
  /*
	void network_init() {
		WORD wVersionRequested;
		WSADATA wsaData;
		char name[255];
		string ip;
		PHOSTENT hostinfo;
		wVersionRequested = MAKEWORD( 2, 0 );

		if ( WSAStartup( wVersionRequested, &wsaData ) == 0 ) {
			if( gethostname ( name, sizeof(name)) == 0) {
				if((hostinfo = gethostbyname(name)) != NULL) {
					IP_ADDRESS = inet_ntoa (*(struct in_addr *)*hostinfo->h_addr_list);
				}
			}
			WSACleanup();
		}
		PROCESS_ID = GetCurrentProcessId();
		THREAD_ID = GetCurrentThreadId();
	}
*/
void network_init() 
{

    WSAData wsaData;
    if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0)
    {
      throw udm_exception("Could not get IP address!");
    }

    char ac[80];
    if (gethostname(ac, sizeof(ac)) == SOCKET_ERROR)
    {
      throw udm_exception("Could not get IP address!");   
    }
 
    struct hostent *phe = gethostbyname(ac);
    if (phe == 0)
    {
        throw udm_exception("Could not get IP address!");   
    }

    for (int i = 0; phe->h_addr_list[i] != 0; ++i)
    {
        struct in_addr addr;
        memcpy(&addr, phe->h_addr_list[i], sizeof(struct in_addr));
        string add(inet_ntoa(addr));
        if (add.find(".0.") == std::string::npos)
          IP_ADDRESS = add;

    }

    WSACleanup();

 		PROCESS_ID = GetCurrentProcessId();
		THREAD_ID = GetCurrentThreadId();
}

// -------------------------------------------------------------------------------------------------------------------------------
//  CORBA stuffs
// -------------------------------------------------------------------------------------------------------------------------------
	bool getConfig_from_dir(string directory) {

    string aa (directory + string("/backplane_server_url.config"));
   char buffer[200];
	_getcwd(buffer, 200);

		ifstream url_config( (directory + "/backplane_server_url.config").c_str() );
		if (!url_config.fail()) {

			SERVER_IP_PORT = "";
			int n = 0;
			for (;; ++n) {
				char c = url_config.get();
				if (url_config.eof()) break;
				if (c=='\n') break;
				SERVER_IP_PORT += c;
			} 

			url_config.close();
			return true;
		}

		url_config.close();
		return false;
	}
	
	bool getConfig_from_environmentvariable() {
		char *envvar;
		envvar = getenv("OTIF_CONFIG_FILE_PATH");
		if (envvar == NULL) return false;
			
		bool result = getConfig_from_dir(envvar);

		return result;
	}
	
	void corba_startup(vector<string> parameters) {
		try {
			network_init();

			
			// Get the arguments and parameters to initialize ORB
			int argc = __UdmCORBA_argc + parameters.size();
			char **arguments = new char*[argc];
			for(int i=0; i<__UdmCORBA_argc; i++) parameters.push_back(__UdmCORBA_argv[i]);

			int q = 0;
			for(vector<string>::iterator j=parameters.begin(); j!= parameters.end(); j++) {
				arguments[q] = new char[j->length()+1];
				strcpy(arguments[q], j->c_str());
				q++;
			}

			if (__orb) __orb->destroy();
			__orb = CORBA::ORB_init(argc, arguments);

			for(int qq=0; qq<argc; qq++) delete arguments[qq];
			delete[] arguments;

		    // Obtain POA and POA manager
			CORBA::Object_var obj = __orb->resolve_initial_references("RootPOA");
			__poa = PortableServer::POA::_narrow(obj.in());
			if (CORBA::is_nil(__poa.in())) {
				throw udmcorba_corbalayer_exception("Failed to narrow root POA", 2);
			}
			__poa_mgr = __poa->the_POAManager();
			if (CORBA::is_nil(__poa_mgr.in())) {
				throw udmcorba_corbalayer_exception("Failed to narrow root POA manager", 3);
			}
			__poa_mgr->activate();

			// Get backplane obj

			// try to open the corresponding file
			bool result = false;
			
			result = getConfig_from_dir(".");
			if (!result) result = getConfig_from_dir("..");
			if (!result) result = getConfig_from_environmentvariable();
			if (!result) throw udmcorba_corbalayer_exception("Could not find the backplane_server_url.config file.", 0);

			obj = __orb->string_to_object( ((string)"corbaloc::" + SERVER_IP_PORT + "/OTIF/BACKPLANE").c_str() );

			// Narrow to Backplane object
			__backplane_obj = OTIF::Backplane::_narrow(obj.in());
			if (CORBA::is_nil(__backplane_obj.in())) {
				throw udmcorba_corbalayer_exception("Failed to narrow to Backplane object", 5);
			}

		} catch (const CORBA::Exception &) {
//			ofstream internal_error_file(INTERNAL_ERROR_FILE, ios::trunc);
//			internal_error_file << e << endl;
//			internal_error_file.close();
			throw udmcorba_corbalayer_exception("Probably there is no active Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the corba_startup function.", 0);
		}
	}

	void corba_shutdown() {
		try {
			__orb->destroy();

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the corba_shutdown function.", 0);
		}
	}



// *******************************************************************************************************************************	
// ===============================================================================================================================
//  CORBA DataNetwork functions
// ===============================================================================================================================
// *******************************************************************************************************************************	
	void CORBADataNetwork::Initialize(const char *paradigm_name, const char *paradigm_version) {
		__paradigm.name = CORBA::string_dup(paradigm_name);
		__paradigm.version = CORBA::string_dup(paradigm_version);
	};

	CORBADataNetwork::CORBADataNetwork(const Udm::UdmDiagram &metainfo) 
		: SmartDataNetwork(metainfo), tooladaptor(NULL), translator(NULL) {
		this->paradigm_name = __paradigm.name;
		this->paradigm_version = __paradigm.version;

		_additional_files.clear();
		_attached_documents.clear();
	};

	CORBADataNetwork::CORBADataNetwork(const Udm::UdmDiagram &metainfo, ToolAdaptor *ta) 
		: SmartDataNetwork(metainfo), tooladaptor(ta), translator(NULL) {
		this->paradigm_name = __paradigm.name;
		this->paradigm_version = __paradigm.version;

		_additional_files.clear();
		_attached_documents.clear();
	};
	CORBADataNetwork::CORBADataNetwork(const Udm::UdmDiagram &metainfo, Translator *tr) 
		: SmartDataNetwork(metainfo), translator(tr), tooladaptor(NULL) {
		this->paradigm_name = __paradigm.name;
		this->paradigm_version = __paradigm.version;

		_additional_files.clear();
		_attached_documents.clear();
	};
	
	long offset = 0;
	map<long, long> datanetwork_id_rep; // map unique_id, dn_id
	map<long, long> datanetwork_host_rep;
	map<long, long> datanetwork_process_rep;
	map<long, long> datanetwork_thread_rep;

	long _getIDfromMAP(long unique_id) {
		map<long, long>::iterator i = datanetwork_id_rep.find(unique_id);
		if (i != datanetwork_id_rep.end()) return (*i).second;
		return 0;
	}

	long _getDNIDfromMAP(long dn_id) {
		for(map<long, long>::iterator i=datanetwork_id_rep.begin(); i!=datanetwork_id_rep.end(); i++)
			if ((*i).second==dn_id) return (*i).first;
		return 0;
	}

	void _setIDforMAP(long unique_id, long dn_id) {
		if (dn_id==0) throw udm_exception("UdmCORBA ERROR: 0 is not allowed for ID!");
		for(map<long, long>::iterator i=datanetwork_id_rep.begin(); i!=datanetwork_id_rep.end(); i++)
			if ((*i).second==dn_id) throw udm_exception("UdmCORBA ERROR: Not allowed to use the same ID for multiple Objects!");
		datanetwork_id_rep[unique_id] = dn_id;
	}
  
	bool _existsIDinMAP(long id) {
		bool ret = false;
		for(map<long, long>::iterator i=datanetwork_id_rep.begin(); i!=datanetwork_id_rep.end(); i++)
			if ((*i).second==id) { ret=true; break; }
		return ret;
	}

	long obtainID(long unique_id) {
		long ret = _getIDfromMAP(unique_id);

		// id not exists in the MAP, let`s create one :)
		if (ret==0) {
			while ( _existsIDinMAP(++offset) );
			ret = offset;
			datanetwork_id_rep[unique_id] = ret;
		}

		return ret;
	}

	void _setDNObjectParameters(long id, OTIF::objectId dn_id) {
		_setIDforMAP(id, dn_id.id);
		datanetwork_host_rep[id] = dn_id.host;
		datanetwork_process_rep[id] = dn_id.process;
		datanetwork_thread_rep[id] = dn_id.thread;
	}
	
	// Get the datanetwork_ID functions
	long CORBADataNetwork::GetDatanetworkObjectID(long unique_id) {
		return _getIDfromMAP( unique_id );
	}

	long CORBADataNetwork::GetDatanetworkObjectID(Udm::Object &obj) {
		return _getIDfromMAP( obj.uniqueId() );
	}

	// Set the datanetwork_ID functions
	void CORBADataNetwork::SetDatanetworkObjectID(long unique_id, long dn_id) {
		_setIDforMAP(unique_id, dn_id);
	}

	void CORBADataNetwork::SetDatanetworkObjectID(Udm::Object &obj, long dn_id) {
		_setIDforMAP(obj.uniqueId(), dn_id);
	}

	// Get Object By ID
	Udm::Object CORBADataNetwork::GetObjectById(long dn_id) {
		long unique_id = _getDNIDfromMAP( dn_id );
		return testdn()->ObjectById(unique_id);
	}

	long CORBADataNetwork::GetObjectHostInfo(Udm::Object &obj) {
		map<long, long>::iterator i = datanetwork_host_rep.find(obj.uniqueId());
		if (i != datanetwork_host_rep.end()) return (*i).second;
		return 0;
	}

	long CORBADataNetwork::GetObjectProcessInfo(Udm::Object &obj) {
		map<long, long>::iterator i = datanetwork_process_rep.find(obj.uniqueId());
		if (i != datanetwork_process_rep.end()) return (*i).second;
		return 0;
	}

	long CORBADataNetwork::GetObjectThreadInfo(Udm::Object &obj) {
		map<long, long>::iterator i = datanetwork_thread_rep.find(obj.uniqueId());
		if (i != datanetwork_thread_rep.end()) return (*i).second;
		return 0;
	}

	void _getNumberOfObjectsInDatanetwork(Udm::ObjectImpl *object, unsigned int &number) {
		vector<ObjectImpl*> o = object->getChildren(NULL, NULL);
		for(vector<ObjectImpl*>::iterator j = o.begin(); j != o.end(); j++) {
			_getNumberOfObjectsInDatanetwork(*j, ++number);

			(*j)->release();
		}
	}


	string CORBADataNetwork::getParadigmName() {
		return paradigm_name;
	}

	string CORBADataNetwork::getParadigmVersion() {
		return paradigm_version;
	}

	// Build up the DataNetwork, and send it to the server side
	long CORBADataNetwork::publishDocument(const char *docName, const char *docVersion, const char *remarks, const long keepAlive, const long previous_docId) {

		try {
			CORBA::ULong doc_length = _additional_files.size() + _attached_documents.size() + 1;

			OTIF::Document_var document = new OTIF::Document();
			OTIF::DocumentElement_S_var docelements = new OTIF::DocumentElement_S();
			docelements->length(doc_length);


			// structured element
			OTIF::DataNetwork_var udm_datanetwork = new OTIF::DataNetwork();
			udm_datanetwork->tId = __diagram_id;

			Udm::Object root_obj = testdn()->GetRootObject();
			udm_datanetwork->root.id = obtainID( root_obj.uniqueId() );

			Udm::ObjectImpl *so = root_obj.__impl();	

			// get the number of the objects in the datanetwork
			datanetwork_object_offset = 0;
			status_percent = 0;
			CORBA::ULong size = 1;

			_getNumberOfObjectsInDatanetwork(so, size);

			OTIF::Obj_S_var udm_objects = new OTIF::Obj_S();
			udm_objects->length(size);
		
			// build up the corba datanetwork structure
			ObjSeqPass(so, udm_objects, tooladaptor, size);
			
			udm_datanetwork->objects = udm_objects;

//			cout << "Objects wrapped in the datanetwork: " << udm_objects->length() << endl;

			OTIF::NameVersion_var document_name = new OTIF::NameVersion();
			document_name->name = CORBA::string_dup(docName);
			document_name->version = CORBA::string_dup(docVersion);

			OTIF::NameVersion_var paradigm = new OTIF::NameVersion();
			paradigm->name = CORBA::string_dup(paradigm_name.c_str());
			paradigm->version = CORBA::string_dup(paradigm_version.c_str());

			OTIF::StructuredElement_var se = new OTIF::StructuredElement();
			se->data = udm_datanetwork;
			se->paradigm = paradigm;

			docelements[(CORBA::ULong)0].sElem(se);


			// attached documents
			CORBA::ULong offset = 1;
			for(map<long, ATTACHED_DOCUMENT>::iterator ad=_attached_documents.begin(); ad!=_attached_documents.end(); ad++, offset++) {

				OTIF::Document_var document = new OTIF::Document();

				if (tooladaptor) {
					long session_id = tooladaptor->session_id;
					map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
					if (i!=tooladaptor_rep.end()) {
						OTIF::ToolAdaptorSession_var tool_adaptor = (*i).second;
						document = tool_adaptor->fetchDocument( (*ad).first );
					}
				} else if (translator) {
					// Obtain TranslatorFeed object reference

					CORBA::Object_var obj = __orb->string_to_object( 
						((string)"corbaloc::" + SERVER_IP_PORT + "/OTIF/" + translator->feed_servant).c_str() );

					// Narrow to TranslatorFeed object
					OTIF::TranslatorFeed_var feed = OTIF::TranslatorFeed::_narrow(obj.in());
					if (CORBA::is_nil(feed.in())) {
						throw udmcorba_exception("Failed to narrow to TranslatorFeed", 35);
					}
					document = feed->fetchDocument( (*ad).first );

				} else {
					throw udmcorba_exception("fetchDocument error: Could not get the document", 33);
				}

				int which = 0;
				for(CORBA::ULong q=0; q<document->documents.length(); q++) {
					if (document->documents[q]._d() == OTIF::structElem) {
						string par_name = document->documents[q].sElem().paradigm.name;
						string par_version = document->documents[q].sElem().paradigm.version;

						if (!par_name.compare((*ad).second.paradigm_name) && !par_version.compare((*ad).second.paradigm_version)) {
							if (which==(*ad).second.offset) {

								// structured elements
								OTIF::StructuredElement_var se = new OTIF::StructuredElement();
								se->data = document->documents[q].sElem().data;
								se->paradigm = document->documents[q].sElem().paradigm;

								docelements[offset].sElem(se);
							}
							which++;
						}
					}
				}
			}


			// unstructured elements
			for(map<string, PHYSICAL_FILE>::iterator i=_additional_files.begin(); i!=_additional_files.end(); i++, offset++) {
				
				OTIF::NameVersion_var filename = new OTIF::NameVersion();
				filename->name = CORBA::string_dup( (*i).first.c_str() );
				filename->version = CORBA::string_dup("");

				OTIF::DataBlob_var datablob = new OTIF::DataBlob();
				
				ifstream additional_file( ((*i).second.directory + (*i).second.name).c_str() );
				if (!additional_file.fail()) {

					long n = 0;
					for (;; ++n) {
						additional_file.get();
						if (additional_file.eof()) break;
					}
					datablob->data.length(n);
					datablob->size = n;
					additional_file.close();

					ifstream additional_file( ((*i).second.directory + (*i).second.name).c_str() );
					n = 0;
					for (;; ++n) {
						char c = additional_file.get();

						if (additional_file.eof()) break;
						datablob->data[n] = c;
					} 
					additional_file.close();

				} else {
					throw udmcorba_exception("Could not find the additional file!", 0);
				}

				OTIF::UnstructuredElement_var use = new OTIF::UnstructuredElement();
				use->name = filename;
				use->data = datablob;

				docelements[offset].usElem(use);
			}
			
			// document is ready!
			document->documents = docelements;

			long return_value;

			if (tooladaptor) {
				long session_id = tooladaptor->session_id;
				map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
				if (i!=tooladaptor_rep.end()) {
					OTIF::ToolAdaptorSession_var tool_adaptor = (*i).second;

					return_value = tool_adaptor->publishDocument(session_id, document_name, remarks, keepAlive, document, previous_docId);
				}

			} else {
				throw udmcorba_exception("publishDocument error: Could not publish the document. There is no active ToolAdaptor session.", 37);
			}

			return return_value;

		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is not found", 11);
		} catch (OTIF::InvalidSession&) {
			throw udmcorba_exception("Invalid session", 31);
		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in CORBADataNetwork::publishDocument function.", 0);
		}
	}


	// Get the Datanetwork from the server side, and build the rootobject
	void CORBADataNetwork::fetchDocument(const long docID, const long offset) {
		try {
			OTIF::Document_var document = new OTIF::Document();
			datanetwork_id_rep.clear();

			if (tooladaptor) {
				long session_id = tooladaptor->session_id;
				map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
				if (i!=tooladaptor_rep.end()) {
					OTIF::ToolAdaptorSession_var tool_adaptor = (*i).second;
					document = tool_adaptor->fetchDocument(docID);
				}

			} else if (translator) {
				// Obtain TranslatorFeed object reference

				CORBA::Object_var obj = __orb->string_to_object( 
					((string)"corbaloc::" + SERVER_IP_PORT + "/OTIF/" + translator->feed_servant).c_str() );

				// Narrow to TranslatorFeed object
				OTIF::TranslatorFeed_var feed = OTIF::TranslatorFeed::_narrow(obj.in());
				if (CORBA::is_nil(feed.in())) {
					throw udmcorba_exception("Failed to narrow to TranslatorFeed", 35);
				}

				document = feed->fetchDocument(docID);
				if (translator->logger) translator->logger->log(3, "Starts fetching a document:", docID);

			} else {
				throw udmcorba_exception("fetchDocument error: Could not get the document", 33);
			}

			int which = 0;
			for(CORBA::ULong q=0; q<document->documents.length(); q++) {
				if (document->documents[q]._d() == OTIF::structElem) {
					string par_name = document->documents[q].sElem().paradigm.name;
					string par_version = document->documents[q].sElem().paradigm.version;

					if (!par_name.compare(this->paradigm_name) && !par_version.compare(this->paradigm_version)) {
						if (which==offset) {

							// structured elements
							OTIF::DataNetwork udm_datanetwork = document->documents[q].sElem().data;

							Uml::Diagram root_meta = testdn()->GetRootMeta();
							set<Uml::Class> uml_classes = root_meta.classes();

							// Search the root object
							OTIF::Obj_var root = new OTIF::Obj();
							for(CORBA::ULong i=0; i<udm_datanetwork.objects.length(); i++) {
								if (udm_datanetwork.objects[i].id.id==udm_datanetwork.root.id) {
									*root=udm_datanetwork.objects[i];
									break;
								}
							}
						
							Udm::Object root_obj = this->GetRootObject();
							Udm::ObjectImpl *so = root_obj.__impl()->clone();

							objectimpl_rep[root->id.id] = so;
							_setDNObjectParameters(root_obj.uniqueId(), root->id);

							// Build up the attributes for the rootobject
							ObjectImplAttrsPass(so, root); 

							if (translator) {
								if (translator->logger) __printTime(translator->logger);
							}
							// Build up the children of the rootobject
							ObjectImplsPass(testdn(), root_meta, udm_datanetwork.objects, so, root->id.id, tooladaptor);
							if (translator) {
								if (translator->logger) __printTime(translator->logger);
							}

							if (translator) {
								if (translator->logger) translator->logger->log(3, "Fetched a document:", docID);
							}
						}
						which++;
					}
				}
			}


			// unstructured elements
			for(CORBA::ULong j=0; j<document->documents.length(); j++) {
				if (document->documents[j]._d() == OTIF::unstructElem) {

					OTIF::NameVersion filename = document->documents[j].usElem().name;
					OTIF::DataBlob datablob = document->documents[j].usElem().data;
					
					string temp_filename;
					srand( (unsigned)time( NULL ) );
					for(;;) {
						char t_chr[16];
						itoa(rand(), t_chr, 10);
						temp_filename = (string)"___tmp" + t_chr + ".tmp";

						FILE * t_file;
						if ((t_file = fopen(temp_filename.c_str(), "r")) == NULL) break;
						fclose(t_file);
					}

					PHYSICAL_FILE pf;
					pf.name = temp_filename;
					pf.directory = "";
					pf.fetched = true;

					this->_additional_files.insert(pair<string, PHYSICAL_FILE>((string)filename.name, pf));

					ofstream output_file( temp_filename.c_str() );
				
					for(CORBA::ULong i=0; i<datablob.data.length(); i++) output_file << datablob.data[i];

					output_file.close();

					if (translator) {
						if (translator->logger) translator->logger->log(3, "Fetched additional document:", (string)filename.name);
					}
				}
			}

		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is not found", 11);
		} catch (OTIF::InvalidSession&) {
			throw udmcorba_exception("Invalid session", 31);
		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in CORBADataNetwork::fetchDocument function.", 0);
		}
	}


	// release the datanetwork pointers and delete the temporary additional files
	void CORBADataNetwork::release() {

		for(map<long, ObjectImpl*>::iterator i=objectimpl_rep.begin(); i!=objectimpl_rep.end(); i++) {
			(*i).second->release();
		}
		
		objectimpl_rep.clear();
		
		datanetwork_id_rep.clear();
		datanetwork_host_rep.clear();
		datanetwork_process_rep.clear();
		datanetwork_thread_rep.clear();

		for(map<string, PHYSICAL_FILE>::iterator j=_additional_files.begin(); j!=_additional_files.end(); j++)
			if ((*j).second.fetched) DeleteFile((*j).second.name.c_str());
		_additional_files.clear();
	}


	// attach structured element to the publishable document
	void CORBADataNetwork::attachDocument(const long docId, const char *paradigm_name, const char *paradigm_version, const long offset) {

		ATTACHED_DOCUMENT ad;
		ad.paradigm_name = paradigm_name;
		ad.paradigm_version = paradigm_version;
		ad.offset = offset;
	
		_attached_documents.insert(pair<long, ATTACHED_DOCUMENT>(docId, ad));
	}

	// additional files for the structured datanetwork
	void CORBADataNetwork::addAdditionalFile(const char *filename, const char *directory) {

		string file = (string)directory + (string)filename;

		ifstream additional_file( file.c_str() );
		if (!additional_file.fail()) {
			PHYSICAL_FILE pf;
			pf.name = filename;
			pf.directory = directory;
			pf.fetched = false;
			_additional_files.insert(pair<string, PHYSICAL_FILE>(filename, pf));
		} else {
			throw udmcorba_exception("Could not find the additional file!", 0);
		}

		additional_file.close();
	}

	void CORBADataNetwork::getAdditionalFileList(multimap<string, string> &additional_files) {
		additional_files.clear();

		for(map<string, PHYSICAL_FILE>::iterator i=_additional_files.begin(); i!=_additional_files.end(); i++) {
			additional_files.insert(pair<string, string>((*i).first, (*i).second.name));
		}
	}

	string CORBADataNetwork::getPhysicalFilename(const char *filename) {
		string return_value;

		map<string, PHYSICAL_FILE>::iterator i = _additional_files.find( filename );
		if (i==_additional_files.end()) return "";

		return (*i).second.name;
	}

	void CORBADataNetwork::copyAdditionalFilesFrom(CORBADataNetwork *source) {
		for(map<string, PHYSICAL_FILE>::iterator i = source->_additional_files.begin(); i != source->_additional_files.end(); i++) {
			PHYSICAL_FILE pf;
			pf.name = (*i).second.name;
			pf.directory = "";	// directory is the working directory, where the fetch method saved the temporary files!
			pf.fetched = false; // nevermind that true or false cause the source will take care of deleting these files!
			this->_additional_files.insert(pair<string, PHYSICAL_FILE>((*i).first, pf));
		}
	}

// *******************************************************************************************************************************	
// ===============================================================================================================================
//  ToolAdaptor functions
// ===============================================================================================================================
// *******************************************************************************************************************************	
	ToolAdaptor::ToolAdaptor(int argc, char **argv) {
		try {
			this->callback_is_sent = false;
			this->dn = NULL;
			__UdmCORBA_argc = argc;
			__UdmCORBA_argv = argv;

			vector<string> null_vector;
			if (CORBA::is_nil(__backplane_obj)) corba_startup(null_vector);

			// Get the UML fetching tooladaptor type
			OTIF::NameVersion_var paradigm = new OTIF::NameVersion();
			paradigm->name = CORBA::string_dup("UML");
			paradigm->version = CORBA::string_dup("1.0");
			OTIF::taMode tam = OTIF::fetching;
			CORBA::String_var meta_uml_tooladaptor_type_var = __backplane_obj->getToolAdaptorType(paradigm, tam);
			string meta_uml_tooladaptor_type = meta_uml_tooladaptor_type_var;

			// Login as the corresponding type
			CORBA::Long session_id;
			OTIF::ToolAdaptorSession_var tooladaptor = __backplane_obj->loginToolAdaptor(meta_uml_tooladaptor_type.c_str(), session_id);
			tooladaptor_rep[session_id] = tooladaptor;
			this->session_id = session_id;
			this->tooladaptor_type = meta_uml_tooladaptor_type;
			
//			OTIF::NameVersion_S_var paradigms = __backplane_obj->getParadigmsOfToolAdaptorType(session_id);

			OTIF::ToolAdaptorInfo_var ta_info = __backplane_obj->getToolAdaptorInfo(session_id);

			working_paradigms.clear();
			for(CORBA::ULong j=0; j<ta_info->fetched.length(); j++) {
				for(CORBA::ULong i = 0; i<ta_info->fetched[j].paradigmName().length(); i++) {
					WORKING_PARADIGM wp;
					wp.name = ta_info->fetched[j].paradigmName()[i].name;
					wp.version = ta_info->fetched[j].paradigmName()[i].version;
					wp.dn = NULL;
					wp.subscribed = false;
					working_paradigms.push_front(wp);
				}
			}
			for(CORBA::ULong k=0; k<ta_info->published.length(); k++) {
				for(CORBA::ULong i = 0; i<ta_info->published[k].paradigmName().length(); i++) {
					WORKING_PARADIGM wp;
					wp.name = ta_info->published[k].paradigmName()[i].name;
					wp.version = ta_info->published[k].paradigmName()[i].version;
					wp.dn = NULL;
					wp.subscribed = false;
					
					bool exists = false;
					for(list<WORKING_PARADIGM>::iterator qq=working_paradigms.begin(); qq!=working_paradigms.end(); qq++) {
						if (!(*qq).name.compare(wp.name) && !(*qq).version.compare(wp.version)) exists = true;
					}

					if (!exists) working_paradigms.push_front(wp);
				}
			}

		} catch (OTIF::InvalidSession&) {
			throw udmcorba_tooladaptor_exception("ToolAdaptor type is invalid.", 81);
		} catch (OTIF::InvalidToolAdaptorType&) {
			throw udmcorba_tooladaptor_exception("ToolAdaptor type is invalid.", 81);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::constructor function.", 0);
		}

	};

	ToolAdaptor::ToolAdaptor(const char *tooladaptor_type, int argc, char **argv, Logger* logger)
    :logger(logger)
  {
		try {
			this->callback_is_sent = false;
			this->dn = NULL;
			__UdmCORBA_argc = argc;
			__UdmCORBA_argv = argv;

			vector<string> null_vector;
			if (CORBA::is_nil(__backplane_obj)) corba_startup(null_vector);
			
			CORBA::Long session_id;
			OTIF::ToolAdaptorSession_var tooladaptor = __backplane_obj->loginToolAdaptor(tooladaptor_type, session_id);
			tooladaptor_rep[session_id] = tooladaptor;
			this->session_id = session_id;
			this->tooladaptor_type = tooladaptor_type;

//			OTIF::NameVersion_S_var paradigms = __backplane_obj->getParadigmsOfToolAdaptorType(session_id);

			OTIF::ToolAdaptorInfo_var ta_info = __backplane_obj->getToolAdaptorInfo(session_id);

			working_paradigms.clear();
			for(CORBA::ULong j=0; j<ta_info->fetched.length(); j++) {
				for(CORBA::ULong i = 0; i<ta_info->fetched[j].paradigmName().length(); i++) {
					WORKING_PARADIGM wp;
					wp.name = ta_info->fetched[j].paradigmName()[i].name;
					wp.version = ta_info->fetched[j].paradigmName()[i].version;
					wp.dn = NULL;
					wp.subscribed = false;
					working_paradigms.push_front(wp);
				}
			}
			for(CORBA::ULong k=0; k<ta_info->published.length(); k++) {
				for(CORBA::ULong i = 0; i<ta_info->published[k].paradigmName().length(); i++) {
					WORKING_PARADIGM wp;
					wp.name = ta_info->published[k].paradigmName()[i].name;
					wp.version = ta_info->published[k].paradigmName()[i].version;
					wp.dn = NULL;
					wp.subscribed = false;

					bool exists = false;
					for(list<WORKING_PARADIGM>::iterator qq=working_paradigms.begin(); qq!=working_paradigms.end(); qq++) {
						if (!(*qq).name.compare(wp.name) && !(*qq).version.compare(wp.version)) exists = true;
					}

					if (!exists) working_paradigms.push_front(wp);
				}
			}

		} catch (OTIF::InvalidSession&) {
			throw udmcorba_tooladaptor_exception("ToolAdaptor type is invalid.", 81);
		} catch (OTIF::InvalidToolAdaptorType&) {
			throw udmcorba_tooladaptor_exception("ToolAdaptor type is invalid.", 81);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::constructor function.", 0);
		}
	};
		
	ToolAdaptor::~ToolAdaptor() {
		try {
			map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
			if (i==tooladaptor_rep.end()) return;
			tooladaptor_rep.erase(session_id);
			__backplane_obj->logoutToolAdaptor(session_id);

//			for(map<long, OTIF_ToolAdaptorNotifyCallback_i *>::iterator j=subscriptions_rep.begin(); j!=subscriptions_rep.end(); j++) delete (*j).second;
//			subscriptions_rep.erase(session_id);

			// All dn-s have to be deleted!!!!
			for(list<WORKING_PARADIGM>::iterator q=working_paradigms.begin(); q!=working_paradigms.end(); q++) {
				if ((*q).dn) {
					(*q).dn->release();
					delete (*q).dn;
				}
			}

//			for(map<long, CORBAObject*>::iterator w=servant_rep.begin(); w!=servant_rep.end(); w++) (*w).second->release();
		
		} catch (OTIF::InvalidSession&) {
//			cerr << "Invalid session!" << endl;
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::destructor function.", 0);
		}
	};

	string ToolAdaptor::getToolAdaptorType() {
		return this->tooladaptor_type;
	};		
	
	void ToolAdaptor::getWorkingParadigms(map<long, PARADIGM_STRUCT> &paradigms) {

		try {
			paradigms.clear();

			OTIF::ToolAdaptorInfo_var ta_info = __backplane_obj->getToolAdaptorInfo(session_id);

			for(CORBA::ULong j=0; j<ta_info->fetched.length(); j++) {
				for(CORBA::ULong i = 0; i<ta_info->fetched[j].paradigmName().length(); i++) {
					PARADIGM_STRUCT par;
					par.name = ta_info->fetched[j].paradigmName()[i].name;
					par.version = ta_info->fetched[j].paradigmName()[i].version;
					paradigms.insert(pair<const long, PARADIGM_STRUCT>(i, par));
				}
			}
			for(CORBA::ULong k=0; k<ta_info->published.length(); k++) {
				for(CORBA::ULong i = 0; i<ta_info->published[k].paradigmName().length(); i++) {
					PARADIGM_STRUCT par;
					par.name = ta_info->published[k].paradigmName()[i].name;
					par.version = ta_info->published[k].paradigmName()[i].version;

					bool exists = false;
					for(map<long, PARADIGM_STRUCT>::iterator qq=paradigms.begin(); qq!=paradigms.end(); qq++) {
						if (!(*qq).second.name.compare(par.name) && !(*qq).second.version.compare(par.version)) exists = true;
					}

					if (!exists) paradigms.insert(pair<const long, PARADIGM_STRUCT>(i, par));
				}
			}

		} catch (OTIF::InvalidSession&) {
			throw udmcorba_tooladaptor_exception("ToolAdaptor type is invalid.", 81);
		} catch (OTIF::InvalidToolAdaptorType&) {
			throw udmcorba_tooladaptor_exception("ToolAdaptor type is invalid.", 81);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::getWorkingParadigms function.", 0);
		}
	};
	
	void ToolAdaptor::getTAConfig(list<CONFIG> &config) {

		try {
			map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
			if (i!=tooladaptor_rep.end()) {
				OTIF::ToolAdaptorSession_var tooladaptor = (*i).second;

				OTIF::ToolAdaptorInfo_var ta_info = tooladaptor->getToolAdaptorInfo(session_id);

				for(CORBA::ULong k=0; k<ta_info->published.length(); k++) {
					list<PARADIGM_STRUCT> paradigms;
					
					for(CORBA::ULong i = 0; i<ta_info->published[k].paradigmName().length(); i++) {
						PARADIGM_STRUCT par;
						par.name = ta_info->published[k].paradigmName()[i].name;
						par.version = ta_info->published[k].paradigmName()[i].version;
						paradigms.push_back(par);
					}

					CONFIG cnf(paradigms);
					config.push_back(cnf);
				}
			}

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::getTAConfig function.", 0);
		}
	};

	void ToolAdaptor::getTASubscribable(list<CONFIG> &config) {

		try {
			map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
			if (i!=tooladaptor_rep.end()) {
				OTIF::ToolAdaptorSession_var tooladaptor = (*i).second;

				OTIF::ToolAdaptorInfo_var ta_info = tooladaptor->getToolAdaptorInfo(session_id);

				list<PARADIGM_STRUCT> paradigms;
				for(CORBA::ULong j=0; j<ta_info->fetched.length(); j++) {
					for(CORBA::ULong i = 0; i<ta_info->fetched[j].paradigmName().length(); i++) {
						PARADIGM_STRUCT par;
						par.name = ta_info->fetched[j].paradigmName()[i].name;
						par.version = ta_info->fetched[j].paradigmName()[i].version;
						paradigms.push_back(par);
					}
				}

				CONFIG cnf(paradigms);
				config.push_back(cnf);
			}

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::getTASubscribable function.", 0);
		}
	};

	vector<DOCUMENT_STRUCT> ToolAdaptor::getDocuments() {

		vector<DOCUMENT_STRUCT> result;
		try {
			map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
			if (i!=tooladaptor_rep.end()) {
				OTIF::ToolAdaptorSession_var tooladaptor = (*i).second;

				OTIF::DocumentInfo_S_var datanetworks = tooladaptor->getDocuments(session_id);

				for(CORBA::ULong i=0; i<datanetworks->length(); i++) {
					DOCUMENT_STRUCT ps;
					ps.docID = (*datanetworks)[i].docId; 
					ps.name = (*datanetworks)[i].documentName.name;
					ps.version = (*datanetworks)[i].documentName.version;
					ps.remarks = (*datanetworks)[i].remarks;
					ps.keepAlive = (*datanetworks)[i].keepAlive;

					OTIF::NameVersion_S_var paradigms = new OTIF::NameVersion_S((*datanetworks)[i].paradigms);
					list<PARADIGM_STRUCT> pds;
					for(CORBA::ULong j=0; j<paradigms->length(); j++) {
						PARADIGM_STRUCT pdms;
						pdms.name = (*paradigms)[j].name;
						pdms.version = (*paradigms)[j].version;
						pds.push_back(pdms);
					}
					ps.paradigms = pds;

					result.push_back(ps);
				}
			}

			return result;
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::getDocuments function.", 0);
		}
	};
	
	bool ToolAdaptor::_registerDN(WORKING_PARADIGM &wp, const Udm::UdmDiagram &metainfo, const char *name, const char *version) {
		if (!wp.dn) {
			__paradigm.name = CORBA::string_dup(name);
			__paradigm.version = CORBA::string_dup(version);
			this->dn = new CORBADataNetwork(metainfo, this);
			wp.dn = this->dn;
			return true;
		}
		return false;
	};
	
	CORBADataNetwork* ToolAdaptor::getDatanetwork(const Udm::UdmDiagram &metainfo) {
		try {
			if (working_paradigms.size()!=1) throw udmcorba_tooladaptor_exception("ToolAdaptor createDatanetwork function is ambiguous, because two or more paradigms are available.", 82);
		
			WORKING_PARADIGM wp = working_paradigms.front();
			if (_registerDN(wp, metainfo, wp.name.c_str(), wp.version.c_str())) {
				working_paradigms.clear();
				working_paradigms.push_front(wp);
			}
			return wp.dn;

		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::createDatanetwork function.", 0);
		}
	};

	CORBADataNetwork* ToolAdaptor::getDatanetwork(const Udm::UdmDiagram &metainfo, const char *paradigm_name, const char *paradigm_version) {
		try {
			bool found = false;
			WORKING_PARADIGM wp;

			for(list<WORKING_PARADIGM>::iterator q=working_paradigms.begin(); q!=working_paradigms.end(); q++) {
				if ( !((*q).name).compare(paradigm_name) && !((*q).version).compare(paradigm_version) ) {
					wp = *q;
					if (_registerDN(wp, metainfo, wp.name.c_str(), wp.version.c_str())) {
						working_paradigms.erase(q);
						working_paradigms.push_front(wp);
					}
					found = true;
					break;
				}
			}

			if (!found) throw udmcorba_tooladaptor_exception("The " + (string)paradigm_name + " v" + (string)paradigm_version + " paradigm is not allowed for this type of ToolAdaptor.", 86);

			return wp.dn;

		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::createDatanetwork function.", 0);
		}
	};
	
	bool ToolAdaptor::isRegisteredDatanetwork(const char *paradigm_name, const char *paradigm_version) {
		try {
			for(list<WORKING_PARADIGM>::iterator q=working_paradigms.begin(); q!=working_paradigms.end(); q++) {
				if ( !((*q).name).compare(paradigm_name) && !((*q).version).compare(paradigm_version) && (*q).dn) return true;
			}
			return false;

		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::isRegisteredDatanetwork function.", 0);
		}
	}

	// It sends the callback point to the server. 
	void ToolAdaptor::_subscribe() {
		map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
		if (i!=tooladaptor_rep.end()) {
			OTIF::ToolAdaptorSession_var tooladaptor = (*i).second;

			OTIF_ToolAdaptorNotifyCallback_i *ta_callback_obj = new OTIF_ToolAdaptorNotifyCallback_i();
			datanetwork_rep = (DatanetworkRepository *) ta_callback_obj;
			subscriptions_rep.insert(pair<long, OTIF_ToolAdaptorNotifyCallback_i *>(session_id, ta_callback_obj));

			PortableServer::ObjectId_var oid = __poa->activate_object(ta_callback_obj);
		    CORBA::Object_var callback_obj = __poa->id_to_reference(oid.in());
			OTIF::ToolAdaptorNotifyCallback_var ta_callback = OTIF::ToolAdaptorNotifyCallback::_narrow(callback_obj.in());

			tooladaptor->subscribe( session_id, ta_callback.in() );
			this->callback_is_sent = true;
		} else {
			throw udmcorba_exception("Subscribe failed: ToolAdaptor is not found", 34);
		}
	};
	
	void ToolAdaptor::subscribe(const Udm::UdmDiagram &metainfo) {
		try {
			if (working_paradigms.size()!=1) throw udmcorba_tooladaptor_exception("ToolAdaptor subscribe function is ambiguous, because two or more paradigms are available.", 85);

			WORKING_PARADIGM wp = working_paradigms.front();

			_registerDN(wp, metainfo, wp.name.c_str(), wp.version.c_str());
			wp.subscribed = true;
			working_paradigms.clear();
			working_paradigms.push_front(wp);

			if (!callback_is_sent) _subscribe();

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::subscribe function.", 0);
		}
	};

	void ToolAdaptor::subscribe(const Udm::UdmDiagram &metainfo, const char *paradigm_name, const char *paradigm_version) {
		try {
			bool found = false;
			WORKING_PARADIGM wp;

			for(list<WORKING_PARADIGM>::iterator q=working_paradigms.begin(); q!=working_paradigms.end(); q++) {
				if ( !((*q).name).compare(paradigm_name) && !((*q).version).compare(paradigm_version) ) {
					wp = *q;

					_registerDN(wp, metainfo, wp.name.c_str(), wp.version.c_str());

					wp.subscribed = true;
					working_paradigms.erase(q);
					working_paradigms.push_front(wp);

					found = true;
					break;
				}
			}

			if (!found) throw udmcorba_tooladaptor_exception("The " + (string)paradigm_name + " v" + (string)paradigm_version + " paradigm is not allowed for this type of ToolAdaptor.", 86);

			if (!callback_is_sent) _subscribe();

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::subscribe function.", 0);
		}
	};

	
	long ToolAdaptor::publishMultipleDocuments(list<CORBADataNetwork*> dns, const char *docName, const char *docVersion, const char *remarks, const long keepAlive, const long previous_docId) {

		try {
			CORBA::ULong doc_length = dns.size();
			{
				for(list<CORBADataNetwork*>::iterator i=dns.begin(); i!=dns.end(); i++) doc_length += (*i)->_additional_files.size();
			}
			
			OTIF::Document_var document = new OTIF::Document();
			OTIF::DocumentElement_S_var docelements = new OTIF::DocumentElement_S();
			docelements->length(doc_length);

			status_percent = 0;

			CORBA::ULong md_offset = 0;
			for(list<CORBADataNetwork*>::iterator cdn=dns.begin(); cdn!=dns.end(); cdn++) {
			
				// structured element
				OTIF::DataNetwork_var udm_datanetwork = new OTIF::DataNetwork();
				udm_datanetwork->tId = __diagram_id;

				Udm::Object root_obj = (*cdn)->GetRootObject();
				udm_datanetwork->root.id = obtainID( root_obj.uniqueId() );

				Udm::ObjectImpl *so = root_obj.__impl();	

				// get the number of the objects in the datanetwork
				datanetwork_object_offset = 0;
				CORBA::ULong size = 1;
				_getNumberOfObjectsInDatanetwork(so, size);
			
				OTIF::Obj_S_var udm_objects = new OTIF::Obj_S();
				udm_objects->length(size);

				// build up the corba datanetwork structure
				ObjSeqPass(so, udm_objects, this, size);
				udm_datanetwork->objects = udm_objects;

	//			cout << "Objects wrapped in the datanetwork: " << udm_objects->length() << endl;

				OTIF::NameVersion_var paradigm = new OTIF::NameVersion();
				paradigm->name = CORBA::string_dup((*cdn)->getParadigmName().c_str());
				paradigm->version = CORBA::string_dup((*cdn)->getParadigmVersion().c_str());

				OTIF::StructuredElement_var se = new OTIF::StructuredElement();
				se->data = udm_datanetwork;
				se->paradigm = paradigm;

				docelements[md_offset++].sElem(se);

				// unstructured elements
				for(map<string, CORBADataNetwork::PHYSICAL_FILE>::iterator i=(*cdn)->_additional_files.begin(); i!=(*cdn)->_additional_files.end(); i++) {
					
					OTIF::NameVersion_var filename = new OTIF::NameVersion();
					filename->name = CORBA::string_dup( (*i).first.c_str() );
					filename->version = CORBA::string_dup("");

					OTIF::DataBlob_var datablob = new OTIF::DataBlob();
					
					ifstream additional_file( ((*i).second.directory + (*i).second.name).c_str() );
					if (!additional_file.fail()) {

						long n = 0;
						for (;; ++n) {
							additional_file.get();
							if (additional_file.eof()) break;
						}
						datablob->data.length(n);
						datablob->size = n;
						additional_file.close();

						ifstream additional_file( ((*i).second.directory + (*i).second.name).c_str() );
						n = 0;
						for (;; ++n) {
							char c = additional_file.get();

							if (additional_file.eof()) break;
							datablob->data[n] = c;
						} 
						additional_file.close();

					} else {
						throw udmcorba_exception("Could not find the additional file!", 0);
					}

					OTIF::UnstructuredElement_var use = new OTIF::UnstructuredElement();
					use->name = filename;
					use->data = datablob;

					docelements[md_offset++].usElem(use);
				}
			}

			// document is ready!
			document->documents = docelements;

			
			OTIF::NameVersion_var document_name = new OTIF::NameVersion();
			document_name->name = CORBA::string_dup(docName);
			document_name->version = CORBA::string_dup(docVersion);

			long return_value;

			map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(this->session_id);
			if (i!=tooladaptor_rep.end()) {
				OTIF::ToolAdaptorSession_var tool_adaptor = (*i).second;
				return_value = tool_adaptor->publishDocument(this->session_id, document_name, remarks, keepAlive, document, previous_docId);
			}

			return return_value;

		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is not found", 11);
		} catch (OTIF::InvalidSession&) {
			throw udmcorba_exception("Invalid session", 31);
		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in ToolAdaptor::publishMultipleDocuments function.", 0);
		}
	};

	
	void ToolAdaptor::logout() {
		try {
			map<long, OTIF::ToolAdaptorSession_var>::iterator i = tooladaptor_rep.find(session_id);
			if (i==tooladaptor_rep.end()) return;
			tooladaptor_rep.erase(session_id);
			__backplane_obj->logoutToolAdaptor(session_id);
		} catch (OTIF::InvalidSession&) {
			throw udmcorba_exception("Invalid session", 31);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::logout function.", 0);
		}
	};
	
	void ToolAdaptor::run() {
		running = true;

    if (logger) logger->log(3, "Starts working");
		// keep running until it gets a terminate
		while (running) {
			process();
		}
	};

	void ToolAdaptor::process() {
		try {

			ACE_Time_Value tv(1);
			__orb->run(tv);

      if (logger) logger->log(3, "Starts working 1");
			while (!datanetwork_rep->changed_datanetwork_repository.empty())
      {
        if (logger) logger->log(3, "Starts working 2");
				map<long, DATANETWORK>::iterator i = datanetwork_rep->changed_datanetwork_repository.begin();
				DATANETWORK dn = (*i).second;
				changedExistingDocument(dn.docName, dn.docVersion, dn.remarks, (*i).first, dn.keepAlive);
				datanetwork_rep->changed_datanetwork_repository.erase( (*i).first );
			};

			while (!datanetwork_rep->datanetwork_repository.empty())
      {

        if (logger) logger->log(3, "Starts working 3");

				map<long, DATANETWORK>::iterator i = datanetwork_rep->datanetwork_repository.begin();
				DATANETWORK dn = (*i).second;

				long docId = (*i).first;
				
				map<long, OTIF::ToolAdaptorSession_var>::iterator q = tooladaptor_rep.find(session_id);
				if (q==tooladaptor_rep.end()) throw udmcorba_tooladaptor_exception("Session of the ToolAdaptor could not be found.", 87);
				
				OTIF::ToolAdaptorSession_var tooladaptor = (*q).second;
				OTIF::NameVersion_var paradigm = tooladaptor->getDocumentParadigmName(docId);
				string par_name = paradigm->name;
				string par_version = paradigm->version;

				for(list<WORKING_PARADIGM>::iterator wp=working_paradigms.begin(); wp!=working_paradigms.end(); wp++) {
					if (!par_name.compare((*wp).name) && !par_version.compare((*wp).version) && (*wp).subscribed) {
						notify(par_name.c_str(), par_version.c_str(), dn.docName, dn.docVersion, dn.remarks, docId, dn.keepAlive);
					}
				}

				datanetwork_rep->datanetwork_repository.erase( docId );
			};

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_corbalayer_exception("Invalid document", 32);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::process function.", 0);
		}
	};
	
	void ToolAdaptor::terminate() 
  {
    if (logger) logger->log(3, "Terminate");
		running = false;
	};

	string ToolAdaptor::getHostInfoString() {
		return IP_ADDRESS;
	};

	long ToolAdaptor::getHostInfoLong() {
		return inet_addr(IP_ADDRESS.c_str());
	};

	long ToolAdaptor::getProcessInfo() {
		return PROCESS_ID;
	};
	
	long ToolAdaptor::getThreadInfo() {
		return THREAD_ID;
	};

	// ------------------------------------------------------------------------------------------------------------
	// ------------------------------------------------------------------------------------------------------------
	// ------------------------------------------------------------------------------------------------------------

	void ToolAdaptor::getBackplaneToolAdaptorTypes(list<string> &types, int argc, char **argv) {

		if (CORBA::is_nil(__backplane_obj) && argv==NULL) throw udmcorba_corbalayer_exception("CORBA layer was not properly initialized.", 89);

		try {
			__UdmCORBA_argc = argc;
			__UdmCORBA_argv = argv;

			vector<string> null_vector;
			if (CORBA::is_nil(__backplane_obj)) corba_startup(null_vector);

			OTIF::Name_S_var ta_types = new OTIF::Name_S();
			ta_types = __backplane_obj->ToolAdaptorTypes();
			for(CORBA::ULong i=0; i<ta_types->length(); i++) {
				types.push_back((string)ta_types[i]);
			}

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::getBackplaneToolAdaptorTypes function.", 0);
		}
	}

	void ToolAdaptor::getBackplaneParadigms(map<long, PARADIGM_STRUCT> &paradigms, int argc, char **argv) {

		if (CORBA::is_nil(__backplane_obj) && argv==NULL) throw udmcorba_corbalayer_exception("CORBA layer was not properly initialized.", 89);

		try {
			__UdmCORBA_argc = argc;
			__UdmCORBA_argv = argv;

			vector<string> null_vector;
			if (CORBA::is_nil(__backplane_obj)) corba_startup(null_vector);
		
			OTIF::NameVersion_S_var pars = new OTIF::NameVersion_S();
			pars = __backplane_obj->Paradigms();
			for(CORBA::ULong i=0; i<pars->length(); i++) {
				PARADIGM_STRUCT par;
				par.name = pars[i].name;
				par.version = pars[i].version;
				paradigms.insert(pair<const long, PARADIGM_STRUCT>(i, par));
			}

		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::getBackplaneParadigms function.", 0);
		}

	};


	Uml::Class ToolAdaptor::getDocumentRootobjectMeta(long docId, const Uml::Diagram &meta_diagram) {
		CORBA::String_var rootobjectmeta;

		try {
			__backplane_obj->fetchRootTypeName(docId, rootobjectmeta);

			string rootobject = rootobjectmeta;
			set<Uml::Class> classes = meta_diagram.classes();
			for(set<Uml::Class>::iterator j=classes.begin(); j!=classes.end(); j++) {
				string name = (*j).name();
				if (!name.compare(rootobjectmeta)) return *j;
			}

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ToolAdaptor::getDocumentRootobjectMeta function.", 0);
		}

		return NULL;
	};


	Uml::Diagram ToolAdaptor::remoteFetchMyParadigm(const char *name, const char *version) {

		if (!dn) throw udmcorba_tooladaptor_exception("ToolAdaptor could not fetch a remote paradigm because datanetwork is NULL.", 88);
		
		OTIF::Diagram_var diagram;
		OTIF::NameVersion paradigm;
		paradigm.name = CORBA::string_dup(name);
		paradigm.version = CORBA::string_dup(version);

		try {
			diagram = __backplane_obj->fetchParadigm(paradigm);
		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is unknown", 18);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		}

		// Create & Init Diagram
		Uml::Diagram meta_diagram = Uml::CreateCORBADiagram();
		CORBAObject *d = (CORBAObject*) meta_diagram.__impl()->clone();
		d->m_servant._diagram = OTIF::Diagram::_duplicate(diagram);
		d->m_servant_id = d->m_servant._diagram->tId();

		servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );

		__diagram_id = d->m_servant_id;


		// Create & Init Classes & Attribs
		OTIF::Class_S_var classes = d->m_servant._diagram->classes();
		for(CORBA::ULong i=0; i<classes->length(); i++) {
			Uml::Class umlclass = Uml::CreateCORBAClass();
			CORBAObject *par = (CORBAObject*) meta_diagram.__impl();
			CORBAObject *d = (CORBAObject*) umlclass.__impl()->clone();
			d->m_servant._class = OTIF::Class::_duplicate(classes[i]);
			d->m_servant_id = d->m_servant._class->tId();
			d->m_parent = par;
			// Optimalization
			d->m_name = d->m_servant._class->name();;

			servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );

			OTIF::Attrib_S_var attribs = classes[i]->attributes();
			for(CORBA::ULong j=0; j<attribs->length(); j++) {
				Uml::Attribute umlattrib = Uml::CreateCORBAAttribute();
				CORBAObject *par = (CORBAObject*) umlclass.__impl();
				CORBAObject *d = (CORBAObject*) umlattrib.__impl()->clone();
				d->m_servant._attrib = OTIF::Attrib::_duplicate(attribs[j]);
				d->m_servant_id = d->m_servant._attrib->tId();
				d->m_parent = par;
				// Optimalization
				d->m_name = d->m_servant._attrib->name();
				d->m_attrib_type = d->m_servant._attrib->typeName();

				servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );
			}		
		}

		// Create & Init Compositions 
		OTIF::Composition_S_var comps = d->m_servant._diagram->compositions();
		for(i=0; i<comps->length(); i++) {
			Uml::Composition umlcomp = Uml::CreateCORBAComposition();
			
			CORBAObject *par = (CORBAObject*) meta_diagram.__impl();
			CORBAObject *d = (CORBAObject*) umlcomp.__impl()->clone();
			d->m_servant._composition = OTIF::Composition::_duplicate(comps[i]);
			d->m_servant_id = d->m_servant._composition->tId();
			servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );
			d->m_parent = par;

			Uml::CompositionParentRole umlcompparent = Uml::CreateCORBACompositionParentRole();
			CORBAObject *d2 = (CORBAObject*) umlcompparent.__impl()->clone();
			d2->m_servant._parentrole = OTIF::CompositionParentRole::_duplicate(comps[i]->parentRole());
			d2->m_servant_id = d2->m_servant._parentrole->tId();
			servant_rep.insert( pair<long const, CORBAObject*>(d2->m_servant_id, (CORBAObject*) d2) );
			d2->m_parent = d;

			Uml::CompositionChildRole umlcompchild = Uml::CreateCORBACompositionChildRole();
			CORBAObject *d3 = (CORBAObject*) umlcompchild.__impl()->clone();
			d3->m_servant._childrole = OTIF::CompositionChildRole::_duplicate(comps[i]->childRole());
			d3->m_servant_id = d3->m_servant._childrole->tId();
			servant_rep.insert( pair<long const, CORBAObject*>(d3->m_servant_id, (CORBAObject*) d3) );
			d3->m_parent = d;
		}

		// Create & Init Associations
		OTIF::Association_S_var assocs = d->m_servant._diagram->associations();
		for(i=0; i<assocs->length(); i++) {
			Uml::Association umlassoc = Uml::CreateCORBAAssociation();
			CORBAObject *par = (CORBAObject*) meta_diagram.__impl();
			CORBAObject *d = (CORBAObject*) umlassoc.__impl()->clone();
			d->m_servant._association = OTIF::Association::_duplicate(assocs[i]);
			d->m_servant_id = d->m_servant._association->tId();
			servant_rep.insert( pair<long const, CORBAObject*>(d->m_servant_id, (CORBAObject*) d) );
			d->m_parent = par;

			OTIF::AssociationRole_S_var ass_roles = assocs[i]->roles();
			if (ass_roles->length()!=2) throw udm_exception("UdmCORBA ERROR: An association without two endpoints is invalid!");
			
			Uml::AssociationRole umlassocrole1 = Uml::CreateCORBAAssociationRole();
			CORBAObject *d2 = (CORBAObject*) umlassocrole1.__impl()->clone();
			d2->m_servant._associationrole = OTIF::AssociationRole::_duplicate(ass_roles[(CORBA::ULong) 0]);
			d2->m_servant_id = d2->m_servant._associationrole->tId();
			d2->m_parent = d;
			// Optimalization
			d2->m_name = d2->m_servant._associationrole->name();

			servant_rep.insert( pair<long const, CORBAObject*>(d2->m_servant_id, (CORBAObject*) d2) );

			Uml::AssociationRole umlassocrole2 = Uml::CreateCORBAAssociationRole();
			CORBAObject *d3 = (CORBAObject*) umlassocrole2.__impl()->clone();
			d3->m_servant._associationrole = OTIF::AssociationRole::_duplicate(ass_roles[(CORBA::ULong) 1]);
			d3->m_servant_id = d3->m_servant._associationrole->tId();
			d3->m_parent = d;
			// Optimalization
			d3->m_name = d3->m_servant._associationrole->name();

			servant_rep.insert( pair<long const, CORBAObject*>(d3->m_servant_id, (CORBAObject*) d3) );
		}

		return meta_diagram;
	}

	void ToolAdaptor::changedExistingDocument(const char *name, const char *version, const char *remarks, long docId, long keepAlive) {};

	void ToolAdaptor::getPublishStatus(long offset, long length) {};
	void ToolAdaptor::getPublishStatus(long percent) {};
	void ToolAdaptor::getFetchStatus(long offset, long length) {};
	void ToolAdaptor::getFetchStatus(long percent) {};


// *******************************************************************************************************************************	
// ===============================================================================================================================
//  Translator functions
// ===============================================================================================================================
// *******************************************************************************************************************************	
	Translator::Translator(const Udm::UdmDiagram &from_metainfo, const char *from_name, const char *from_version, const Uml::Class &from_rootclass,
					const Udm::UdmDiagram &to_metainfo, const char *to_name, const char *to_version, const Uml::Class &to_rootclass,
					int argc, char **argv, Logger *logger) 
		: from_rootclass(from_rootclass), to_rootclass(to_rootclass) {

		if (argc<4) {
			if (logger) logger->log(1, "Invalid translator arguments");
			throw udmcorba_corbalayer_exception("Invalid translator arguments", 74);
		}

		try {	
			long log_level = atoi(argv[1]);
			feed_servant = argv[2];
			sink_servant = argv[3];
			this->logger = logger;
			if (logger) logger->setLevel(log_level);

			// create fromDN
			__UdmCORBA_argc = argc;
			__UdmCORBA_argv = argv;
			__paradigm.name = CORBA::string_dup(from_name);
			__paradigm.version = CORBA::string_dup(from_version);
			fromDN = new CORBADataNetwork(from_metainfo, this);

			// add this to fromDN map
			fromDN_map.clear();
			WORKING_PARADIGM par(from_name, from_version, from_rootclass);
			fromDN_map.insert(pair<CORBADataNetwork*, WORKING_PARADIGM>(fromDN, par));

			// create toDN
			__paradigm.name = CORBA::string_dup(to_name);
			__paradigm.version = CORBA::string_dup(to_version);
			toDN = new CORBADataNetwork(to_metainfo, this);

			// Obtain TranslatorFeed object reference
			CORBA::Object_var obj = __orb->string_to_object( 
				((string)"corbaloc::" + SERVER_IP_PORT + "/OTIF/" + feed_servant).c_str() );

			// Narrow to TranslatorFeed object
			OTIF::TranslatorFeed_var feed = OTIF::TranslatorFeed::_narrow(obj.in());
			if (CORBA::is_nil(feed.in())) {
				throw udmcorba_exception("Failed to narrow to TranslatorFeed", 35);
		    }

			// auto subscribe
			OTIF_TranslatorNotifyCallback_i *tr_callback_obj = new OTIF_TranslatorNotifyCallback_i();
			datanetwork_rep = (DatanetworkRepository *) tr_callback_obj;

			PortableServer::ObjectId_var oid = __poa->activate_object(tr_callback_obj);
		    CORBA::Object_var callback_obj = __poa->id_to_reference(oid.in());
		    OTIF::TranslatorNotifyCallback_var tr_callback = OTIF::TranslatorNotifyCallback::_narrow(callback_obj.in());

			feed->subscribe( tr_callback.in() );

			if (logger) logger->log(3, "Subscribed:", (long)tr_callback_obj);

		} catch (const CORBA::Exception &) {
			if (logger) logger->log(1, "Fatal error");
			throw udmcorba_corbalayer_exception("Fatal error.", 7);
		} catch (udm_exception &e) {
			if (logger) logger->log(1, "UDM exception: ", e.what());
			throw;
		} catch (udmcorba_exception &e) {
			if (logger) logger->log(1, "UDMCORBA exception: ", e.what());
			throw;
		} catch (...) {
			if (logger) logger->log(1, "Fatal exception in the Translator::constructor function");
			throw udmcorba_exception("Fatal exception in the Translator::constructor function.", 0);
		}
	}

	Translator::~Translator() {
		delete fromDN;
		delete toDN;

		for(map<CORBADataNetwork*, WORKING_PARADIGM>::iterator i=fromDN_map.begin(); i!=fromDN_map.end(); i++) {
			(*i).first->CloseNoUpdate();
			delete (*i).first;
		}

		fromDN_map.clear();

		if (logger) logger->log(3, "Destroyed");
	}


	void Translator::addFromDN(const Udm::UdmDiagram &from_metainfo, const char *from_name, const char *from_version, const Uml::Class &from_rootclass) {
		for(map<CORBADataNetwork*, WORKING_PARADIGM>::iterator i=fromDN_map.begin(); i!=fromDN_map.end(); i++) {
			WORKING_PARADIGM par = (*i).second;
			if (!par.name.compare(from_name) && !par.version.compare(from_version)) return;
		}
		
		__paradigm.name = CORBA::string_dup(from_name);
		__paradigm.version = CORBA::string_dup(from_version);
		CORBADataNetwork *dn = new CORBADataNetwork(from_metainfo, this);

		WORKING_PARADIGM par(from_name, from_version, from_rootclass);
		fromDN_map.insert(pair<CORBADataNetwork*, WORKING_PARADIGM>(dn, par));
	}
	
	CORBADataNetwork* Translator::get_fromDN(const char *name, const char *version, const long offset) {
		int which = 0;
		for(map<CORBADataNetwork*, WORKING_PARADIGM>::iterator q=fromDN_map.begin(); q!=fromDN_map.end(); q++) {
			WORKING_PARADIGM par = (*q).second;
			if (!par.name.compare(name) && !par.version.compare(version)) {
				if (which==offset) return (*q).first;
				which++;
			}
		}
		return NULL;
	}
	
	CORBADataNetwork* Translator::get_fromDN() {
		if (fromDN_map.size()>1) throw udmcorba_exception("Translator get_fromDN function is ambiguous, because two or more paradigms are available.", 75);
		return fromDN;
	}

	CORBADataNetwork* Translator::get_toDN() {
		return toDN;
	}

	void Translator::_publish(list<CORBADataNetwork*> dns, const char *doc_name, const char *doc_version, const char *doc_remarks) {

		try {
			// Obtain TranslatorSink object reference
			CORBA::Object_var obj = __orb->string_to_object( 
				((string)"corbaloc::" + SERVER_IP_PORT + "/OTIF/" + sink_servant).c_str() );

			// Narrow to TranslatorSink object
			OTIF::TranslatorSink_var sink = OTIF::TranslatorSink::_narrow(obj.in());
			if (CORBA::is_nil(sink.in())) {
				throw udmcorba_exception("Failed to narrow to TranslatorSink", 36);
			}

			CORBA::ULong doc_length = dns.size();
			{
				for(list<CORBADataNetwork*>::iterator i=dns.begin(); i!=dns.end(); i++) doc_length += (*i)->_additional_files.size();
			}

			OTIF::Document_var document = new OTIF::Document();
			OTIF::DocumentElement_S_var docelements = new OTIF::DocumentElement_S();
			docelements->length(doc_length);

			CORBA::ULong md_offset = 0;
			for(list<CORBADataNetwork*>::iterator cdn=dns.begin(); cdn!=dns.end(); cdn++) {

				// structured elements
				OTIF::DataNetwork_var udm_datanetwork = new OTIF::DataNetwork();
				udm_datanetwork->tId = __diagram_id;

				Udm::Object root_obj = (*cdn)->GetRootObject();
				udm_datanetwork->root.id = obtainID( root_obj.uniqueId() );

				Udm::ObjectImpl *so = root_obj.__impl();

				// get the number of the objects in the datanetwork
				datanetwork_object_offset = 0;
				CORBA::ULong size = 1;
				_getNumberOfObjectsInDatanetwork(so, size);

				OTIF::Obj_S_var udm_objects = new OTIF::Obj_S();
				udm_objects->length(size);

				ObjSeqPass(so, udm_objects, NULL, size);
				udm_datanetwork->objects = udm_objects;

				if (logger) logger->log(3, "Objects wrapped in the datanetwork:", (long)udm_objects->length());

				OTIF::NameVersion_var paradigm = new OTIF::NameVersion();
				paradigm->name = CORBA::string_dup((*cdn)->getParadigmName().c_str());
				paradigm->version = CORBA::string_dup((*cdn)->getParadigmVersion().c_str());

				OTIF::StructuredElement_var se = new OTIF::StructuredElement();
				se->data = udm_datanetwork;
				se->paradigm = paradigm;

				docelements[md_offset++].sElem(se);

				// unstructured elements
				for(map<string, CORBADataNetwork::PHYSICAL_FILE>::iterator i=(*cdn)->_additional_files.begin(); i!=(*cdn)->_additional_files.end(); i++) {
					
					OTIF::NameVersion_var filename = new OTIF::NameVersion();
					filename->name = CORBA::string_dup( (*i).first.c_str() );
					filename->version = CORBA::string_dup("");

					OTIF::DataBlob_var datablob = new OTIF::DataBlob();
					
					ifstream additional_file( ((*i).second.directory + (*i).second.name).c_str() );
					if (!additional_file.fail()) {

						long n = 0;
						for (;; ++n) {
							additional_file.get();
							if (additional_file.eof()) break;
						}
						datablob->data.length(n);
						datablob->size = n;
						additional_file.close();

						ifstream additional_file( ((*i).second.directory + (*i).second.name).c_str() );
						n = 0;
						for (;; ++n) {
							char c = additional_file.get();

							if (additional_file.eof()) break;
							datablob->data[n] = c;
						} 
						additional_file.close();

					} else {
						throw udmcorba_exception("Could not find the additional file!", 0);
					}

					OTIF::UnstructuredElement_var use = new OTIF::UnstructuredElement();
					use->name = filename;
					use->data = datablob;

					docelements[md_offset++].usElem(use);
				}
			}
			
			// document is ready!
			document->documents = docelements;

			
			OTIF::NameVersion_var document_name = new OTIF::NameVersion();
			document_name->name = CORBA::string_dup(doc_name);
			document_name->version = CORBA::string_dup(doc_version);

			sink->publishDocument(document_name, doc_remarks, keepAlive, document, this->working_docID);
			if (logger) logger->log(3, "Published a document:", (string)document_name->name + " v" + (string)document_name->version + " " + doc_remarks);

		} catch (const CORBA::Exception &) {
			if (logger) logger->log(1, "Fatal error");
			throw udmcorba_corbalayer_exception("Fatal error.", 7);
		} catch (udm_exception &e) {
			if (logger) logger->log(1, "UDM exception: ", e.what());
			throw;
		} catch (udmcorba_exception &e) {
			if (logger) logger->log(1, "UDMCORBA exception: ", e.what());
			throw;
		} catch (...) {
			if (logger) logger->log(1, "Fatal exception in the Translator::publish function");
			throw udmcorba_exception("Fatal exception in the Translator::publish function.", 0);
		}
	}
	
	void Translator::publish(const char *doc_name, const char *doc_version, const char *doc_remarks) {
		if (logger) logger->log(3, "Translation ended. Start publishing the result datanetwork");
		list<CORBADataNetwork*> dns;
		dns.push_back(toDN);
		_publish(dns, doc_name, doc_version, doc_remarks);
	}
	
	void Translator::publish(CORBADataNetwork *dn, const char *doc_name, const char *doc_version, const char *doc_remarks) {
		if (logger) logger->log(3, "Translation ended. Start publishing the result datanetwork");
		list<CORBADataNetwork*> dns;
		dns.push_back(dn);
		_publish(dns, doc_name, doc_version, doc_remarks);
	}
	
	void Translator::publishMultipleDocuments(list<CORBADataNetwork*> dns, const char *doc_name, const char *doc_version, const char *doc_remarks) {
		if (logger) logger->log(3, "Translation ended. Start publishing the result datanetwork");
		_publish(dns, doc_name, doc_version, doc_remarks);
	}

	void Translator::run() {
		if (logger) logger->log(3, "Starts working");
		while (true) {
			ACE_Time_Value tv(1);
			__orb->run(tv);
			process();
		}
	}

/*
	void Translator::fetchDocuments(const long docId, const unsigned long numberOfDocuments) {

		try {
			OTIF::Document_var document = new OTIF::Document();

			// Obtain TranslatorFeed object reference
			CORBA::Object_var obj = __orb->string_to_object( 
				((string)"corbaloc::" + SERVER_IP_PORT + "/OTIF/" + feed_servant).c_str() );

			// Narrow to TranslatorFeed object
			OTIF::TranslatorFeed_var feed = OTIF::TranslatorFeed::_narrow(obj.in());
			if (CORBA::is_nil(feed.in())) {
				throw udmcorba_exception("Failed to narrow to TranslatorFeed", 35);
			}

			document = feed->fetchDocument(docId);
			if (logger) logger->log(3, "Starts fetching a document:", docId);


			for(CORBA::ULong offset=0; offset<numberOfDocuments; offset++) {
				datanetwork_id_rep.clear();

				string par_name = document->documents[offset].sElem().paradigm.name;
				string par_version = document->documents[offset].sElem().paradigm.version;

				// structured elements
				fromDN = get_fromDN(par_name.c_str(), par_version.c_str());
				if (!fromDN) throw udmcorba_paradigm_exception("Paradigm is not found", 11);

				OTIF::DataNetwork udm_datanetwork = document->documents[offset].sElem().data;

				Uml::Diagram root_meta = fromDN->GetRootMeta();
				set<Uml::Class> uml_classes = root_meta.classes();

				// Search the root object
				OTIF::Obj_var root = new OTIF::Obj();
				for(CORBA::ULong i=0; i<udm_datanetwork.objects.length(); i++) {
					if (udm_datanetwork.objects[i].id.id==udm_datanetwork.root.id) {
						*root=udm_datanetwork.objects[i];
						break;
					}
				}
			
				Udm::Object root_obj = fromDN->GetRootObject();
				Udm::ObjectImpl *so = root_obj.__impl()->clone();

				objectimpl_rep[root->id.id] = so;
				_setDNObjectParameters(root_obj.uniqueId(), root->id);

				// Build up the attributes for the rootobject
				ObjectImplAttrsPass(so, root); 

				if (logger) __printTime(logger);
				// Build up the children of the rootobject
				ObjectImplsPass(fromDN, root_meta, udm_datanetwork.objects, so, root->id.id, NULL);
				if (logger) __printTime(logger);

			
				// unstructured elements
				for(CORBA::ULong j=0; j<document->documents.length(); j++) {
					if (document->documents[j]._d() == OTIF::unstructElem) {

						OTIF::NameVersion filename = document->documents[j].usElem().name;
						OTIF::DataBlob datablob = document->documents[j].usElem().data;
						
						string temp_filename;
						for(;;) {
							srand( (unsigned)time( NULL ) );
							char t_chr[16];
							itoa(rand(), t_chr, 10);
							temp_filename = (string)"___tmp" + t_chr + ".tmp";

							FILE * t_file;
							if ((t_file = fopen(temp_filename.c_str(), "r")) == NULL) break;
							fclose(t_file);
						}

						CORBADataNetwork::PHYSICAL_FILE pf;
						pf.name = temp_filename;
						pf.fetched = true;
						fromDN->_additional_files.insert(pair<string, CORBADataNetwork::PHYSICAL_FILE>((string)filename.name, pf));

						ofstream output_file( temp_filename.c_str() );
					
						for(CORBA::ULong i=0; i<datablob.data.length(); i++) output_file << datablob.data[i];

						output_file.close();
					}
				}
			}

			if (logger) logger->log(3, "Fetched a document:", docId);

		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is not found", 11);
		} catch (OTIF::InvalidSession&) {
			throw udmcorba_exception("Invalid session", 31);
		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (const CORBA::Exception &) {
			throw udmcorba_corbalayer_exception("Fatal exception: Caught a CORBA::Exception exception", 1);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in CORBADataNetwork::fetchMultipleDocument function.", 0);
		}
	}
*/
	
	void Translator::fetching_process(const char *name, const char *version, const char *remarks, const long docId, const unsigned long numberOfDocuments) {
		if (logger) logger->log(3, "Get a notification, starts fetching");
	
		// Create the DN for every fromDN
		for(map<CORBADataNetwork*, WORKING_PARADIGM>::iterator i=fromDN_map.begin(); i!=fromDN_map.end(); i++) {
			(*i).first->CreateNew(("___from_" + (string)(*i).second.name + "_" + (string)(*i).second.version + ".mem").c_str(), 
				"", (*i).second.rootclass, Udm::CHANGES_LOST_DEFAULT);//PERSIST_ALWAYS);//LOST_DEFAULT);
			(*i).first->fetchDocument(docId);
		}
		
//		fetchDocuments(docId, numberOfDocuments);

		toDN->CreateNew("___to.mem", "", to_rootclass, Udm::CHANGES_LOST_DEFAULT);//PERSIST_ALWAYS);//LOST_DEFAULT);

		this->working_docID = docId;
		notify(name, version, remarks, numberOfDocuments);

		for(i=fromDN_map.begin(); i!=fromDN_map.end(); i++) {
			(*i).first->CloseNoUpdate();
			(*i).first->release();
		}

		toDN->CloseNoUpdate();
		toDN->release();

		if (logger) logger->log(3, "Finished the translation process");
	}
	
	void Translator::process() {
		try {
			while (!datanetwork_rep->datanetwork_repository.empty()) {
				map<long, DATANETWORK>::iterator i = datanetwork_rep->datanetwork_repository.begin();
				DATANETWORK dn = (*i).second;

				this->keepAlive = dn.keepAlive;

				fetching_process(dn.docName, dn.docVersion, dn.remarks, (*i).first, dn.numberOfDocuments);

				datanetwork_rep->datanetwork_repository.erase((*i).first);

				ACE_Time_Value tv(1);
				__orb->run(tv);
				// translation ended if no more datanetwork exists, exit
				if (datanetwork_rep->datanetwork_repository.empty()) {
					if (logger) logger->log(3, "Exists");
					exit(0);
				}
			}

		} catch (const CORBA::Exception &) {
			if (logger) logger->log(1, "Fatal error");
			throw udmcorba_corbalayer_exception("Fatal error.", 7);
		} catch (udm_exception &e) {
			if (logger) logger->log(1, "UDM exception: ", e.what());
			throw;
		} catch (udmcorba_exception &e) {
			if (logger) logger->log(1, "UDMCORBA exception: ", e.what());
			throw;
		} catch (...) {
			if (logger) logger->log(1, "Fatal exception in the Translator::process function");
			throw udmcorba_exception("Fatal exception in the Translator::process function.", 0);
		}
	}

	Uml::Class Translator::getDocumentRootobjectMeta(long docId, const Uml::Diagram &meta_diagram) {
		CORBA::String_var rootobjectmeta;

		if (logger) logger->log(3, "Gets the rootobject of a document:", docId);
		try {
			__backplane_obj->fetchRootTypeName(docId, rootobjectmeta);

			string rootobject = rootobjectmeta;
			set<Uml::Class> classes = meta_diagram.classes();
			for(set<Uml::Class>::iterator j=classes.begin(); j!=classes.end(); j++) {
				string name = (*j).name();
				if (!name.compare(rootobjectmeta)) return *j;
			}

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (udm_exception &e) {
			if (logger) logger->log(1, "UDM exception: ", e.what());
			throw;
		} catch (udmcorba_exception &e) {
			if (logger) logger->log(1, "UDMCORBA exception: ", e.what());
			throw;
		} catch (const CORBA::Exception &) {
			if (logger) logger->log(1, "Fatal error. Maybe there is no Backplane server.");
			throw udmcorba_corbalayer_exception("Fatal error. Maybe there is no Backplane server.", 7);
		} catch (...) {
			if (logger) logger->log(1, "Fatal exception in the Translator::getDocumentRootobjectMeta function");
			throw udmcorba_exception("Fatal exception in the Translator::getDocumentRootobjectMeta function.", 0);
		}

		return NULL;
	};


// *******************************************************************************************************************************	
// ===============================================================================================================================
//  ManagerObject functions
// ===============================================================================================================================
// *******************************************************************************************************************************	
	OTIF::ManagerSession_var manager_obj;

	ManagerObject::ManagerObject(int argc, char **argv) {
		try {
			__UdmCORBA_argc = argc;
			__UdmCORBA_argv = argv;

			vector<string> null_vector;
			null_vector.push_back("");

			if (CORBA::is_nil(__backplane_obj)) corba_startup(null_vector);
			dn = NULL;
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::constructor function.", 0);
		}
	}
	
	ManagerObject::ManagerObject(const Udm::UdmDiagram &metainfo, const char *name, const char *version, int argc, char **argv) {
		try {
			__UdmCORBA_argc = argc;
			__UdmCORBA_argv = argv;

			vector<string> null_vector;
			null_vector.push_back("");

			if (CORBA::is_nil(__backplane_obj)) corba_startup(null_vector);

			__paradigm.name = CORBA::string_dup(name);
			__paradigm.version = CORBA::string_dup(version);
			dn = new CORBADataNetwork(metainfo);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::constructor function.", 0);
		}
	}
	
	ManagerObject::~ManagerObject() {
		if (dn) delete dn;
	}


	string ManagerObject::backplane_name() {
		string result;
		try {
			OTIF::NameVersion_var bname = __backplane_obj->name();
			result = bname->name;
			return result;

		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::backplane_name function.", 0);
		}
	}

	string ManagerObject::backplane_version() {
		string result;
		try {
			OTIF::NameVersion_var bname = __backplane_obj->name();
			result = bname->version;
			return result;

		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::backplane_version function.", 0);
		}
	}

	vector<PARADIGM_STRUCT> ManagerObject::backplane_paradigms() {
		vector<PARADIGM_STRUCT> result;
		try {
			OTIF::NameVersion_S_var metas = __backplane_obj->Paradigms();

			for(CORBA::ULong i=0; i<metas->length(); i++) {
				PARADIGM_STRUCT ps;
				ps.name = (*metas)[i].name;
				ps.version = (*metas)[i].version;
				result.push_back(ps);
			}

			return result;
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::backplane_paradigms function.", 0);
		}
	}

	vector<TRANSLATOR_STRUCT> ManagerObject::backplane_translators() {
		vector<TRANSLATOR_STRUCT> result;
		try {
			OTIF::TranslatorVersion_S_var translators = __backplane_obj->Translators();

			for(CORBA::ULong i=0; i<translators->length(); i++) {
				TRANSLATOR_STRUCT ps;
				ps.name = (*translators)[i].name.name;
				ps.version = (*translators)[i].name.version;
				ps.fromName = (*translators)[i].from.name;
				ps.fromVersion = (*translators)[i].from.version;
				ps.toName = (*translators)[i].to.name;
				ps.toVersion = (*translators)[i].to.version;
				result.push_back(ps);
			}

			return result;
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::backplane_translators function.", 0);
		}
	}

	vector<DOCUMENT_STRUCT> ManagerObject::backplane_documents() {
		vector<DOCUMENT_STRUCT> result;
		try {
			OTIF::DocumentInfo_S_var datanetworks = __backplane_obj->Documents();

			for(CORBA::ULong i=0; i<datanetworks->length(); i++) {
				DOCUMENT_STRUCT ps;
				ps.docID = (*datanetworks)[i].docId; 
				ps.name = (*datanetworks)[i].documentName.name;
				ps.version = (*datanetworks)[i].documentName.version;
				ps.remarks = (*datanetworks)[i].remarks;
				ps.keepAlive = (*datanetworks)[i].keepAlive;

				OTIF::NameVersion_S_var paradigms = new OTIF::NameVersion_S((*datanetworks)[i].paradigms);
				list<PARADIGM_STRUCT> pds;
				for(CORBA::ULong j=0; j<paradigms->length(); j++) {
					PARADIGM_STRUCT pdms;
					pdms.name = (*paradigms)[j].name;
					pdms.version = (*paradigms)[j].version;
					pds.push_back(pdms);
				}
				ps.paradigms = pds;

				result.push_back(ps);
			}

			return result;
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::backplane_documents function.", 0);
		}
	}

	void ManagerObject::lockParadigm(PARADIGM_STRUCT paradigm) {

		OTIF::NameVersion_var meta = new OTIF::NameVersion();
		meta->name = CORBA::string_dup(paradigm.name.c_str());
		meta->version = CORBA::string_dup(paradigm.version.c_str());

		try {
			__backplane_obj->lockParadigm(meta);
		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_exception("Paradigm is unknown", 18);
		} catch (OTIF::ParadigmLocked&) {
			throw udmcorba_exception("Paradigm is already locked", 15);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::lockParadigm function.", 0);
		}
	}

	void ManagerObject::unlockParadigm(PARADIGM_STRUCT paradigm) {

		OTIF::NameVersion_var meta = new OTIF::NameVersion();
		meta->name = CORBA::string_dup(paradigm.name.c_str());
		meta->version = CORBA::string_dup(paradigm.version.c_str());

		try {
			__backplane_obj->unlockParadigm(meta);
		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_exception("Paradigm is unknown", 18);
		} catch (OTIF::ParadigmNotLocked&) {
			throw udmcorba_exception("Paradigm is already unlocked", 20);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::unlockParadigm function.", 0);
		}
	}

	bool ManagerObject::isLocked(PARADIGM_STRUCT paradigm) {

		OTIF::NameVersion_var meta = new OTIF::NameVersion();
		meta->name = CORBA::string_dup(paradigm.name.c_str());
		meta->version = CORBA::string_dup(paradigm.version.c_str());

		try {
			__backplane_obj->lockParadigm(meta);
			__backplane_obj->unlockParadigm(meta);
			
			return false;
		} catch (OTIF::ParadigmLocked&) {
			return true;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::isLocked function.", 0);
		}
	}
	
	void ManagerObject::login(const char *backplane_password) {

		try {
			manager_obj = __backplane_obj->loginManager(backplane_password, manager_obj_session);
			if (CORBA::is_nil(manager_obj.in())) {
				throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
			}
		} catch (OTIF::InvalidPassword&) {
			throw udmcorba_manager_exception("The password is incorrect", 62);
		} catch (OTIF::OnlyOneManager&) {
			throw udmcorba_manager_exception("There is another active Manager application", 63);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::login function.", 0);
		}
	}
	
	void ManagerObject::logout() {
		
		try {
			__backplane_obj->logoutManager(manager_obj_session);
		} catch (OTIF::InvalidSession&) {
			throw udmcorba_exception("Invalid session", 31);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		} catch (udm_exception&) {
			throw;
		} catch (udmcorba_exception&) {
			throw;
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::logout function.", 0);
		}
	}

	void ManagerObject::registerMetaModel(PARADIGM_STRUCT paradigm, const char *temp_filename) {

		// Setup the parameters of the MetaModel
		OTIF::NameVersion_var meta = new OTIF::NameVersion();
		meta->name = CORBA::string_dup(paradigm.name.c_str());
		meta->version = CORBA::string_dup(paradigm.version.c_str());
		OTIF::ManagerSession::RegistrationAction action = OTIF::ManagerSession::RegistrationAction::add;

		dn->OpenExisting(temp_filename, "uml");

		OTIF::DataNetwork_var udm_datanetwork = new OTIF::DataNetwork();
		udm_datanetwork->tId = __diagram_id;

		Udm::Object root_obj = dn->GetRootObject();
		udm_datanetwork->root.id = obtainID( root_obj.uniqueId() );
		Udm::ObjectImpl *so = root_obj.__impl();	

		// get the number of the objects in the datanetwork
		datanetwork_object_offset = 0;
		CORBA::ULong size = 1;
		_getNumberOfObjectsInDatanetwork(so, size);
		
		OTIF::Obj_S_var udm_objects = new OTIF::Obj_S();
		udm_objects->length(size);
		
		// build up the corba datanetwork structure
		ObjSeqPass(so, udm_objects, NULL, size);
		udm_datanetwork->objects = udm_objects;

		dn->CloseNoUpdate();
		dn->release();

		// Register a paradigm
		try {
			manager_obj->registerMetaModel(meta, udm_datanetwork, action);
	    } catch (OTIF::ParadigmExists&) {
			throw udmcorba_paradigm_exception("Paradigm exists", 14);
	    } catch (OTIF::ParadigmLocked&) {
			throw udmcorba_paradigm_exception("Paradigm is locked", 15);
		} catch (OTIF::ParadigmInUse&) {
			throw udmcorba_paradigm_exception("Paradigm is in use", 16);
	    } catch (OTIF::InvalidFile&) {
			throw udmcorba_paradigm_exception("Paradigm file is invalid", 17);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}
/*
	void ManagerObject::registerMetaModelFile(PARADIGM_STRUCT paradigm, const char *filename) {

		// Setup the parameters of the MetaModel
		OTIF::NameVersion_var meta = new OTIF::NameVersion();
		meta->name = CORBA::string_dup(paradigm.name.c_str());
		meta->version = CORBA::string_dup(paradigm.version.c_str());
		OTIF::ManagerSession::RegistrationAction action = OTIF::ManagerSession::RegistrationAction::add;

		// Register a paradigm
		try {
			manager_obj->registerMetaModelFile(meta, filename, action);
	    } catch (OTIF::ParadigmExists&) {
			throw udmcorba_paradigm_exception("Paradigm exists", 14);
	    } catch (OTIF::ParadigmLocked&) {
			throw udmcorba_paradigm_exception("Paradigm is locked", 15);
		} catch (OTIF::ParadigmInUse&) {
			throw udmcorba_paradigm_exception("Paradigm is in use", 16);
	    } catch (OTIF::InvalidFile&) {
			throw udmcorba_paradigm_exception("Paradigm file is invalid", 17);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::unregisterMetaModel(PARADIGM_STRUCT paradigm) {

		// Setup the parameters of the MetaModel
		OTIF::NameVersion_var meta = new OTIF::NameVersion();
		meta->name = CORBA::string_dup(paradigm.name.c_str());
		meta->version = CORBA::string_dup(paradigm.version.c_str());
		OTIF::ManagerSession::RegistrationAction action = OTIF::ManagerSession::RegistrationAction::add;

		// Unregister a paradigm
		try {
			manager_obj->unregisterMetaModel(meta);
		} catch (OTIF::ParadigmInUse&) {
			throw udmcorba_paradigm_exception("Paradigm is in use", 16);
		} catch (OTIF::ParadigmLocked&) {
			throw udmcorba_paradigm_exception("Paradigm is locked", 15);
		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is unknown", 18);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::registerTranslator(TRANSLATOR translator, PARADIGM_STRUCT from_paradigm, PARADIGM_STRUCT to_paradigm, const char *filename) {
	
		// Setup the parameters of the Translator
		OTIF::ManagerSession::RegistrationAction action = OTIF::ManagerSession::RegistrationAction::add;
		OTIF::TranslatorVersion_var transl = new OTIF::TranslatorVersion();
		OTIF::NameVersion_var tr = new OTIF::NameVersion();
		tr->name = CORBA::string_dup(translator.name);
		tr->version = CORBA::string_dup(translator.version);
		transl->name_version = tr;
		transl->fromName = CORBA::string_dup(from_paradigm.name.c_str());
		transl->fromVersion = CORBA::string_dup(from_paradigm.version.c_str());
		transl->toName = CORBA::string_dup(to_paradigm.name.c_str());
		transl->toVersion = CORBA::string_dup(to_paradigm.version.c_str());

		try {
			manager_obj->registerTranslator(transl, filename, action);
		} catch (OTIF::TranslatorExists&) {
			throw udmcorba_translator_exception("Translator exists", 71);
		} catch (OTIF::TranslatorInUse&) {
			throw udmcorba_translator_exception("Translator is in use", 72);
		} catch (OTIF::InvalidTranslator&) {
			throw udmcorba_translator_exception("Invalid translator", 73);
		} catch (OTIF::InvalidParadigm&) {
			throw udmcorba_paradigm_exception("Paradigm is unknown", 18);
		}
	}

	void ManagerObject::unregisterTranslator(TRANSLATOR translator, PARADIGM_STRUCT from_paradigm, PARADIGM_STRUCT to_paradigm) {
	
		// Setup the parameters of the Translator
		OTIF::TranslatorVersion_var transl = new OTIF::TranslatorVersion();
		OTIF::NameVersion_var tr = new OTIF::NameVersion();
		tr->name = CORBA::string_dup(translator.name);
		tr->version = CORBA::string_dup(translator.version);
		transl->name_version = tr;
		transl->fromName = CORBA::string_dup(from_paradigm.name.c_str());
		transl->fromVersion = CORBA::string_dup(from_paradigm.version.c_str());
		transl->toName = CORBA::string_dup(to_paradigm.name.c_str());
		transl->toVersion = CORBA::string_dup(to_paradigm.version.c_str());

		try {
			manager_obj->unregisterTranslator(transl);
		} catch (OTIF::TranslatorInUse&) {
			throw udmcorba_translator_exception("Translator is in use", 72);
		} catch (OTIF::InvalidTranslator&) {
			throw udmcorba_translator_exception("Invalid translator", 73);
		}
	}
*/

	void ManagerObject::change_password(const char *old_password, const char *new_password) {

		try {
				manager_obj->changePassword( CORBA::string_dup(old_password), CORBA::string_dup(new_password) );

		} catch (OTIF::InvalidPassword&) {
			throw udmcorba_manager_exception("The password is incorrect! Password is not changed.", 62);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}


	void ManagerObject::backplane_shutdown() {

		try {
			manager_obj->shutdown();
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::destroyDocument(long docId) {
		try {
			manager_obj->destroyDocument(docId);

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::modifyDocumentName(long docId, const char* name) {
		try {
			manager_obj->modifyDocumentName(docId, name);

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::modifyDocumentVersion(long docId, const char* version) {
		try {
			manager_obj->modifyDocumentVersion(docId, version);

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::modifyDocumentRemarks(long docId, const char* remarks) {
		try {
			manager_obj->modifyDocumentRemarks(docId, remarks);

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::modifyDocumentKeepAlive(long docId, long keepAlive) {
		try {
			manager_obj->modifyDocumentKeepAlive(docId, keepAlive);

		} catch (OTIF::InvalidDocument&) {
			throw udmcorba_exception("Invalid document", 32);
		} catch (CORBA::Exception &) {
			throw udmcorba_manager_exception("Could not connect to the Backplane server", 61);
		}
	}

	void ManagerObject::uploadStartup(string &xml_file) {
		try {
			__backplane_obj->uploadStartup(xml_file.c_str());

		} catch (OTIF::InvalidStartupFile &) {
			throw udmcorba_exception("The configuration file is corrupt.", 0);
		} catch (const CORBA::Exception &) {
			throw udmcorba_exception("Could not send the configuration. Probably there is no active Backplane server.", 0);
		} catch (...) {
			throw udmcorba_exception("Fatal exception in the ManagerObject::uploadStartup function.", 0);
		}
	}
	
	
// Namespace UdmCORBA end
}



// *******************************************************************************************************************************	
// ===============================================================================================================================
//	Uml namespace
// ===============================================================================================================================
// *******************************************************************************************************************************	
namespace Uml
{
	Namespace CreateCORBANamespace() 
	{
		return Namespace::Cast(UdmCORBA::CreateObject(Namespace::meta)); 
	}


	Diagram CreateCORBADiagram() 
	{
		return Diagram::Cast(UdmCORBA::CreateObject(Diagram::meta)); 
	}

	Class CreateCORBAClass() 
	{
		return Class::Cast(UdmCORBA::CreateObject(Class::meta)); 
	}

	Attribute CreateCORBAAttribute() 
	{
		return Attribute::Cast(UdmCORBA::CreateObject(Attribute::meta));
	}

	Association CreateCORBAAssociation() 
	{
		return Association::Cast(UdmCORBA::CreateObject(Association::meta)); 
	}

	AssociationRole CreateCORBAAssociationRole()
	{ 
		return AssociationRole::Cast(UdmCORBA::CreateObject(AssociationRole::meta));
	}

	Composition CreateCORBAComposition() 
	{ 
		return Composition::Cast(UdmCORBA::CreateObject(Composition::meta)); 
	}

	CompositionParentRole CreateCORBACompositionParentRole()
	{ 
		return CompositionParentRole::Cast(UdmCORBA::CreateObject(CompositionParentRole::meta));
	}

	CompositionChildRole CreateCORBACompositionChildRole()
	{
		return CompositionChildRole::Cast(UdmCORBA::CreateObject(CompositionChildRole::meta));
	}

	void InitCORBADiagram(const Diagram &obj, const char *name, const char *version)
	{
		UdmCORBA::InitDiagram(obj);	
	}

	void InitCORBAClass(const Class &obj, const Diagram &parent, const char *name, bool isAbstract, const char *stereo)
	{
		UdmCORBA::InitClass(obj, parent, name);
	}

	void InitCORBAAttribute(const Attribute &obj, const Class &parent, const char *name, const char *type, bool nonpersistent, int min, int max, const bool ordered, const string& visibility, const vector<string> & defval)
	{

		UdmCORBA::InitAttribute(obj ,parent, name);
	}
	
// For InitAssociation
	Association s_obj;
	Diagram s_parent;
	char s_name[1024];
	AssociationRole s_ass1_role;
	char s_ass1_name[1024];
	Class s_ass1_target;
	
	void InitCORBAAssociation(const Association &obj, const Diagram &parent, const char *name) {
		s_obj = obj;
		s_parent = parent;
		strcpy(s_name, name);
	}

	void InitCORBAAssociationClass(const Association &aobj, const Class &cobj) {}
	void InitCORBAAssociationRole(const AssociationRole &obj, const Association &parent, 
		const char *name, bool navigable, bool primary, long min, long max, const Class &target) {
	
		if (!s_ass1_role) {
			s_ass1_role = obj;
			strcpy(s_ass1_name, name);
			s_ass1_target = target;
		} else {
			UdmCORBA::InitAssociation(s_obj, s_parent, s_name,
					s_ass1_role, s_ass1_name, s_ass1_target,
					obj, name, target);

			s_ass1_role = NULL;
		}
	}
	
// For InitComposition
	Composition s_comp;
	Diagram s_cparent;
	char s_cname[1024];
	CompositionParentRole s_cpr;
	char s_cpr_name[1024];
	Class s_cpr_target;

	void InitCORBAComposition(const Composition &obj, const Diagram &parent, const char *name) {
		s_comp = obj;
		s_cparent = parent;
		strcpy(s_cname, name);
	}
	void InitCORBACompositionParentRole(const CompositionParentRole &obj, 
		const Composition &parent, const char *name, bool navigable, const Class &target) {
	
		s_cpr = obj;
		strcpy(s_cpr_name, name);
		s_cpr_target = target;
	}
	void InitCORBACompositionChildRole(const CompositionChildRole &obj,
		const Composition &parent, const char *name, bool navigable, long min, long max, const Class &target) {
	
		UdmCORBA::InitComposition(parent, s_cparent, s_cname, 
			    s_cpr, s_cpr_name, s_cpr_target, 
				obj, name, target);
	}

	void AddCORBAInheritance(const Class &baseType, const Class &subType) {}


// Namespace Uml end
}
